/*
 * Copyright (C) 2002-5 Edscott Wilson Garcia
 * EMail: edscott@imp.mx
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

typedef struct preview_manager_t{
    gint population_serial;
    view_t *view_p;
} preview_manager_t;

#define CELLHSPACING (view_p->text_height+TEXTSPACING)
#define CELLWIDTH (view_p->column_width)
#define CELLHEIGHT (view_p->icon_size+(2*CELLHSPACING))
/* some of the code in this function contributed by Thomas Leonard: */
static void
save_thumbnail (const gchar * thumbnail_path, const gchar * original_path, GdkPixbuf * tgt) {
    int original_width,
      original_height;
    char *swidth,
     *sheight,
     *ssize,
     *smtime;
    struct stat st;
    GError *error = NULL;

    if(!thumbnail_path || !original_path)
        return;
    TRACE ("creating thumbnail for %s (%s)\n", original_path, thumbnail_path);

    original_width = gdk_pixbuf_get_width (tgt);
    original_height = gdk_pixbuf_get_height (tgt);

    if(stat (original_path, &st) != 0)
        return;

    swidth = g_strdup_printf ("%d", original_width);
    sheight = g_strdup_printf ("%d", original_height);
    ssize = g_strdup_printf ("%ld", (long int)st.st_size);
    smtime = g_strdup_printf ("%ld", st.st_mtime);

    gdk_pixbuf_save (tgt, thumbnail_path, "png", &error,
                     /*"tEXt::Thumb::Image::Width", swidth,
                        "tEXt::Thumb::Image::Height", sheight,
                        "tEXt::Thumb::Size", ssize, 
                        "tEXt::Thumb::MTime", smtime, */
                     "tEXt::Software", "Rodent", NULL);
    if(error)
        g_error_free (error);
    g_free (swidth);
    g_free (sheight);
    g_free (ssize);
    g_free (smtime);
}

static
    void
insert_pixbuf_tag (const GdkPixbuf * tag, GdkPixbuf * composite_pixbuf, const gchar * where, double scale_factor) {
    int width = gdk_pixbuf_get_width (composite_pixbuf);
    int height = gdk_pixbuf_get_height (composite_pixbuf);
    // "SW"
    int dest_x = 0;
    int dest_y = 0;
    int dest_width = width;
    int dest_height = height;
    double scale_x = 1.0 / scale_factor;
    double scale_y = 1.0 / scale_factor;
    int overall_alpha = 180;
    // default SW
    double offset_x = 0.0;
    double offset_y = height - ((double)height) / scale_factor;
    if(strcmp (where, "SE") == 0) {
        offset_x = width - ((double)width) / scale_factor;
    } else if(strcmp (where, "NW") == 0) {
        offset_y = 0.0;
    } else if(strcmp (where, "NE") == 0) {
        offset_x = width - ((double)width) / scale_factor;
        offset_y = 0.0;
    } else if(strcmp (where, "C") == 0) {
        offset_x = ((double)width) / 2 - ((double)width) / scale_factor / 2;
        offset_y = ((double)height) / 2 - ((double)height) / scale_factor / 2;
    } else if(strcmp (where, "SC") == 0) {
        offset_x = ((double)width) / 2 - ((double)width) / scale_factor / 2;
    }

    gdk_pixbuf_composite (tag, composite_pixbuf,
                          dest_x, dest_y,
                          dest_width, dest_height, offset_x, offset_y, scale_x, scale_y, GDK_INTERP_BILINEAR, overall_alpha);
    return;
}

static GdkPixbuf *
make_pixbuf_from_id (const gchar * id, int size) {
    GdkPixbuf *pixbuf;

    TRACE ("ICONx:make_pixbuf_from_id: %s\n", id);
    if(!id || !strlen (id))
        return NULL;

    if(strstr (id, "/composite")) {
        gchar *basic_type = g_strdup (id);
        TRACE ("composite type = %s\n", id);

        GdkPixbuf *basic_pixbuf,
         *composite_pixbuf;

        *(strstr (basic_type, "/composite")) = 0;

        basic_pixbuf = rfm_get_pixbuf (basic_type, size);
        g_free (basic_type);
        if(basic_pixbuf) {
            composite_pixbuf = gdk_pixbuf_copy (basic_pixbuf);
        } else {
            composite_pixbuf = NULL;
            g_warning ("cannot get basic pixbuf for %s", id);
            return NULL;
        }

        if(strstr (id, "/composite/") == NULL) {
            // incomplete composite specification.
            return composite_pixbuf;
        }

        const gchar *sub_id[] =
            { "/link", "/error", "/show-previews", "/show-hidden", "/stock_yes", "/stock_no", "/stock_insert-object", NULL };
        gchar *where[] = { "SW", "SE", "SW", "NW", "NW", "NW", "C" };
        double scale_factor[] = { 2.0, 2.0, 6.0, 3.0, 3.0, 3.0, 2.0 };
        int i;
        const GdkPixbuf *tag;

        for(i = 0; sub_id[i]; i++) {
            if(strstr (id, sub_id[i])) {
                TRACE ("composite link done for %s\n", id);
                gchar *c_id = g_strdup_printf ("xffm%s", sub_id[i]);
                tag = rfm_get_pixbuf (c_id, size);
                g_free (c_id);
                if(!tag) {
                    g_warning ("cannot get pixbuf for id: %s", sub_id[i]);
                    return NULL;
                }
                insert_pixbuf_tag (tag, composite_pixbuf, where[i], scale_factor[i]);
            }
        }
        return composite_pixbuf;
    }

    gchar *file = NULL;
    GError *error = NULL;
    double factor = 1.0;
    // hack factor: just looks better to my eyes...
    //if (view_p->icon_size > 48 && strcmp("xffm/stock_go-up",id)==0) factor = 0.75;
    file = ICON_get_filename_from_id (id);
    DBG ("ICON_get_filename_from_id(%s) ---> %s\n", id, file);
    if(!file) {
        file = ICON_get_filename_from_id ("xffm/stock_justify-fill");
    }
    pixbuf = gdk_pixbuf_new_from_file_at_size (file, factor * size, factor * size, &error);
    g_free(file);
    return pixbuf;
}

// this is used by the preview thread manager to determine if 
// population_pp under study is still valid.
static gboolean
invalid_population (view_t * view_p, gint population_serial) {
    gboolean result = FALSE;
    g_mutex_lock (view_p->population_serial_mutex);
    if(view_p->population_serial != population_serial) {
	DBG("invalid population serial %d != %d\n", 
		view_p->population_serial, population_serial);
        result = TRUE;
    }
    g_mutex_unlock (view_p->population_serial_mutex);
    return result;
}


static gpointer
thread_preview_manager (gpointer data) {
    preview_manager_t *preview_manager_p = data;
    view_t *view_p = preview_manager_p->view_p;
    int population_serial = preview_manager_p->population_serial;
    DBG ("PREVIEWS: thread_preview_manager\n");
    // should protect population p  from here!
    rfm_population_read_lock (view_p);
    if(invalid_population (view_p, population_serial)) {
        DBG ("PREVIEWS: invalid_population\n");
        g_free (preview_manager_p);
        rfm_population_read_unlock (view_p);
        return NULL;
    }
    // get thread's private copy of population_pp
    int grid_elements = view_p->xfdir_p->pathc;
    population_t **thread_population_pp=
	(population_t **)malloc((grid_elements + 1) * sizeof(population_t *));
    memcpy (thread_population_pp, view_p->population_pp, (grid_elements + 1) * sizeof (population_t *));
    view_p->thumbnailer_active=TRUE;
    rfm_population_read_unlock (view_p);

    /* schedule image previews... */
    int active_threads = 0;
    GThread *thread_id[PREVIEW_THREAD_LIMIT];

    int i;

    // this is to tell monitor thread to wait for mutex
    if(view_p->thumbnail_loop_mutex == NULL) {
        view_p->thumbnail_loop_mutex = g_mutex_new ();
    }
    g_mutex_lock (view_p->thumbnail_loop_mutex);
    for(i = 0; thread_population_pp[i]; i++) {
	if(invalid_population (view_p, population_serial)){
	    break;
	}
	
	if (view_p->thumbnailer_exit) {
	    break;
	}

    //  rfm_population_try_read_lock should not be used here.
    //  A blocking lock is better: after the lock is obtained,
    //  a check for valid population condition will determine if
    //  the thread should continue or exit.
	population_t *population_p = thread_population_pp[i];
	rfm_population_read_lock (view_p);
	if(invalid_population (view_p, population_serial)) {
	    DBG("invalid_population (population_serial=%d)\n", population_serial);
	    rfm_population_read_unlock (view_p);
	    break;
	}
        if(population_p->preview_done) {
            DBG ("SEMAPHORE: CONTINUE with %d active threads\n", active_threads);
	    rfm_population_read_unlock (view_p);
            continue;
        }
        if(!population_p->en || !population_p->en->path) {
            DBG ("SEMAPHORE: CONTINUE2 with %d active threads\n", active_threads);
	    rfm_population_read_unlock (view_p);
            continue;
        }
	// default is to show thumbnails.
        if(view_p->en && rfm_is_image (population_p->en)) {
            DBG ("SEMAPHORE: scheduling thread: %d\n", active_threads + 1);
            DBG ("THREAD: g_thread_create(rfm_preview_thread_f) at primary-icons.c\n");
	    preview_child_t *preview_child_p = 
		(preview_child_t *) malloc (sizeof (preview_child_t));
	    preview_child_p->population_p = population_p;
	    preview_child_p->population_serial = population_serial;
	    preview_child_p->view_p = view_p;
		
            thread_id[active_threads] = g_thread_create (rfm_preview_thread_f, (gpointer) preview_child_p, TRUE, NULL);
            active_threads++;

            //every PREVIEW_THREAD_LIMIT threads, don't spawn more until joined
            if(active_threads == PREVIEW_THREAD_LIMIT) {
                do {
                    DBG ("SEMAPHORE: thread_preview_manager: inner-join %d\n", active_threads);
                    g_thread_join (thread_id[active_threads - 1]);
                    active_threads--;
                }
                while(active_threads > 0);
            }
        }
	rfm_population_read_unlock (view_p);
    }

    while(active_threads > 0) {
        DBG ("SEMAPHORE: thread_preview_manager outer-join %d\n", active_threads);
        g_thread_join (thread_id[active_threads - 1]);
        active_threads--;
    }

    // this may fix the deskview expose problem, but definitely
    // causes thread segv crashes in gridview:
    // rodent_expose_all (view_p);

    view_p->thumbnailer_active=FALSE;
    view_p->thumbnailer_exit=FALSE;
    // this is to tell monitor thread to continue
    g_mutex_unlock (view_p->thumbnail_loop_mutex);
    g_free(thread_population_pp);
    g_free (preview_manager_p);

    return NULL;
}


static GdkPixbuf *
create_background_pixbuf (char *file, int width, int height) {
    GdkPixbuf *tgt;
    GError *error = NULL;
    double factor,
      x_factor,
      y_factor;
    int w,
      h;
    if(!file)
        return NULL;

    TRACE ("getting pixbuf %s", file);
        {
        GdkPixbuf *src = gdk_pixbuf_new_from_file (file, &error);
        if(error || !src) {
            if(error) {
                DBG ("file=%s (%s)", file, error->message);
                g_error_free (error);
            }
            error = NULL;
            return NULL;
        }

        if(!src)
            return NULL;

        w = gdk_pixbuf_get_width (src);
        h = gdk_pixbuf_get_height (src);
        TRACE (" pixbuf is %d,%d width,height is %d,%d", w, h, width, height);
        x_factor = (double)width / w;
        y_factor = (double)height / h;
        factor = (x_factor < y_factor) ? x_factor : y_factor;
        TRACE ("factors %lf,%lf --> %lf, distorsion=%lf", x_factor, y_factor, factor, fabs (x_factor - y_factor) / factor);
        if(fabs (x_factor - y_factor) / factor < 0.20) {
          TRACE ("scaling pixbuf %s to %d,%d", file, width, height);
          tgt = gdk_pixbuf_scale_simple (src, width, height, GDK_INTERP_BILINEAR);
        } else {
          TRACE ("factor=%lf scaling pixbuf %s to %lf,%lf", factor, file, factor * w, factor * h);
          tgt = gdk_pixbuf_scale_simple (src, factor * w, factor * h, GDK_INTERP_BILINEAR);
        }
        g_object_unref (G_OBJECT (src));
    }
    TRACE ("got pixbuf %s", file);
    return tgt;
}


static GdkPixbuf *
create_preview (const gchar * file, gint size) {
    GdkPixbuf *src;
    GError *error = NULL;
    gchar *thumbnail_path = rfm_get_thumbnail_path (file, size);
    if(!file)
        return NULL;

    if(thumbnail_path && g_file_test (thumbnail_path, G_FILE_TEST_EXISTS)) {
        struct stat st;
        struct stat file_st;
        stat (thumbnail_path, &st);
        stat (file, &file_st);
        if(st.st_mtime > file_st.st_mtime) {
            DBG ("PREVIEW thread: retrieving thumbnail for %s (%s)\n", file, thumbnail_path);
            src = gdk_pixbuf_new_from_file (thumbnail_path, NULL);
            if(src) {
                g_free (thumbnail_path);
                return src;
            }
        }
    }
    DBG ("PREVIEW thread: no thumbnail for %s (%s)\n", file, thumbnail_path);

    // first we try the width at specified size
    src = gdk_pixbuf_new_from_file_at_size (file, size, -1, &error);
    if(error) {
        DBG ("%s:%s", error->message, file);
        g_error_free (error);
        error = NULL;
        g_free (thumbnail_path);
        return NULL;
    }
    gint ph = gdk_pixbuf_get_height (src);
    gint pw = gdk_pixbuf_get_width (src);
    if (ph > pw) {
	// other way around
	g_object_unref(src);
	src = gdk_pixbuf_new_from_file_at_size (file, -1, size, &error);
	ph = gdk_pixbuf_get_height (src);
	pw = gdk_pixbuf_get_width (src);
    }


    if(pw < 10 || ph < 10 ) {
        if(src)
            g_object_unref (G_OBJECT (src));
        g_free (thumbnail_path);
        return NULL;
    }

    if(!src) {
        g_free (thumbnail_path);
        return NULL;
    }

    save_thumbnail (thumbnail_path, file, src);
    DBG ("save_thumbnail %s\n", thumbnail_path);
    g_free (thumbnail_path);
    return src;
}

// internal hash record structure
static GHashTable *pixbuf_hash = NULL;
static GMutex *pixbuf_hash_mutex = NULL;

typedef struct pixbuf_t {
    time_t mtime;
    gint size;
    GdkPixbuf *pixbuf;
    union {
        gchar *mime_id;
        gchar *path;
    };
} pixbuf_t;

static
gchar *
get_hash_key (const gchar * key, int size) {
    gint bucket = 0;
    pixbuf_t *pixbuf_p = NULL;
    gchar *hash_key = NULL;
    do {
        gchar *fullkey = g_strdup_printf ("%s-%d:%d", key, size, bucket);
        GString *gs = g_string_new (fullkey);
        g_free (fullkey);
        hash_key = g_strdup_printf ("%10u", g_string_hash (gs));
        g_string_free (gs, TRUE);
        g_strstrip (hash_key);
        g_mutex_lock (pixbuf_hash_mutex);
        pixbuf_p = g_hash_table_lookup (pixbuf_hash, hash_key);
        g_mutex_unlock (pixbuf_hash_mutex);
        if(!pixbuf_p)
            break;
        if(strcmp (pixbuf_p->path, key) == 0 && pixbuf_p->size == size)
            break;
        g_free (hash_key);
    } while(1);
    return hash_key;
}

static
    void
init_pixbuf_hash () {
    if(!pixbuf_hash) {
        pixbuf_hash = g_hash_table_new (g_str_hash, g_str_equal);
    }
    if(!pixbuf_hash_mutex) {
        pixbuf_hash_mutex = g_mutex_new ();
    }
    return;
}

static
pixbuf_t *
find_in_pixbuf_hash (const gchar * key, int size) {
    if (size > MAX_PIXBUF_HASH_SIZE) return NULL;
    init_pixbuf_hash ();
    gchar *hash_key = get_hash_key (key, size);
    //DBG("HASH: find in hash: %s --> %s\n",key, hash_key);

    g_mutex_lock (pixbuf_hash_mutex);
    pixbuf_t *pixbuf_p = g_hash_table_lookup (pixbuf_hash, hash_key);
    g_mutex_unlock (pixbuf_hash_mutex);
    // mime_id or NULL return immediately:
    if(!pixbuf_p)
        return pixbuf_p;
    //DBG("found  %s --> %s in hash!\n",key, hash_key);
    if(!g_path_is_absolute (key))
        return pixbuf_p;
    // check if item is out of date
    if(g_file_test (key, G_FILE_TEST_EXISTS)) {
        struct stat st;
        stat (key, &st);
        if(pixbuf_p->mtime != st.st_mtime) {
            g_mutex_lock (pixbuf_hash_mutex);
            g_hash_table_steal (pixbuf_hash, hash_key);
            g_mutex_unlock (pixbuf_hash_mutex);
            g_free (pixbuf_p->path);
            g_object_unref (pixbuf_p->pixbuf);
            g_free (pixbuf_p);
            g_free (hash_key);
            return NULL;
        }
    }
    g_free (hash_key);
    return pixbuf_p;
}


