/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@imp.mx
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/**  miscelaneous */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <regex.h>
#include <dirent.h>
#include <X11/Xlib.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>
#include <grp.h>
#include <pwd.h>

#ifdef HAVE_STDARG_H
# include <stdarg.h>
#elif HAVE_VARARGS_H
# include <varargs.h>
#endif

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#define EXPAND      TRUE
#define FILL        TRUE
#define NOEXPAND    FALSE
#define NOFILL      FALSE


#include "rfm_libs.h"
#include "rfm_modules.h"
#include "primary-dialogs.i"

typedef struct sequence_t {
    const gchar *id;
    const gchar *sequence;
} sequence_t;

static sequence_t sequence_p[] = {
        {"xffm_tag/black", "[30m"},
        {"xffm_tag/red", "[31m"},
        {"xffm_tag/green", "[32m"},
        {"xffm_tag/yellow", "[33m"},
        {"xffm_tag/blue", "[34m"},
        {"xffm_tag/magenta", "[35m"},
        {"xffm_tag/cyan", "[36m"},
        {"xffm_tag/white", "[37m"},
        {"xffm_tag/bold", "[1m"},
        {"xffm_tag/italic", "[4m"},
        {NULL, "[m"},
        {NULL, "[0m"},
        {NULL, "[22m"},
        {NULL, "[24m"},
        {NULL, NULL}            // this marks the end of sequences.
};

void
rfm_set_store_data_from_list(GtkListStore *list_store, GList **list){
    GtkTreeIter iter;
    GList *p;
    gtk_list_store_clear (list_store);
    for (p=g_list_first(*list); p && p->data; p=p->next) {
	gtk_list_store_append (list_store, &iter);
	gtk_list_store_set (list_store, &iter,
                          0, (gchar *)p->data,
                          -1);
      /* Note: The store will keep a copy of the string internally, 
       * so the list may be freed */
    }
}


char *
rfm_display_host_name (GtkWidget * widget){
    Window xid=GDK_WINDOW_XID(widget->window);
    char *name = NULL;
    unsigned char *property_data;
    unsigned long items, remaining;
    int actual_format;
    Atom atomo, actual_atom;

    atomo = XInternAtom(GDK_DISPLAY(), "WM_CLIENT_MACHINE", FALSE);
    if(XGetWindowProperty(GDK_DISPLAY(), xid,
		atomo, 0, 255, FALSE, XA_STRING, 
		&actual_atom, &actual_format, 
		&items, &remaining, &property_data) == Success)
    {
	TRACE("property_data=%s\n",((property_data)?property_data:(unsigned char *)"null")); 
	if (!property_data) name = g_strdup("localhost");
	else {
	    name = g_strdup((const gchar *)property_data);
	    XFree(property_data);
	}
    }
    else name = g_strdup(g_get_host_name());
    return name;
}

GtkWidget *
rfm_mk_little_button (widgets_t * widgets_p, const gchar * icon_id, void *callback, void *callback_data, const gchar * tooltip_text) {
    GtkWidget *button;
    if(!widgets_p)
        return NULL;
    button = gtk_button_new ();
    GTK_WIDGET_UNSET_FLAGS (button, GTK_CAN_FOCUS);
    gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
    if(icon_id) {
        GdkPixbuf *pb;
        GtkWidget *image;
        pb = rfm_get_pixbuf (icon_id, SIZE_BUTTON);
        image = gtk_image_new_from_pixbuf (pb);
        gtk_widget_show (image);
        gtk_container_add (GTK_CONTAINER (button), image);
    }
    if(tooltip_text && strlen (tooltip_text)) {
	gtk_widget_set_tooltip_text (button, tooltip_text);
    }
    if(callback) {
        DBG ("DIAGNOSTICS: run button connected to callback\n");
        g_signal_connect ((gpointer) button, "clicked", G_CALLBACK (callback), callback_data);
    }
    return button;
}

void
rfm_cursor_wait (GtkWidget * w) {
    static GdkCursor *cursor = NULL;
    if(!w)
        return;
    if(!cursor)
        cursor = gdk_cursor_new (GDK_WATCH);
    gdk_window_set_cursor (w->window, cursor);
    //gdk_flush();
}

void
rfm_cursor_reset (GtkWidget * w) {
    if(!w)
        return;
    gdk_window_set_cursor (w->window, NULL);

}

    
static void 
clear_diagnostics (widgets_t * widgets_p);

static void
clear_diagnostics_window (GtkButton * button, gpointer data) {
    widgets_t *widgets_p = data;
    DBG ("DIAGNOSTICS: hide_diagnostics_window: now ");
    if(!widgets_p || !widgets_p->diagnostics_window)
        return;
    gtk_widget_hide (widgets_p->diagnostics_window);
    clear_diagnostics ((widgets_t *)data);
    return;
}
    
static void
hide_diagnostics_window (GtkButton * button, gpointer data) {
    widgets_t *widgets_p = data;
    DBG ("DIAGNOSTICS: hide_diagnostics_window: now ");
    if(!widgets_p || !widgets_p->diagnostics_window)
        return;
    gtk_window_iconify((GtkWindow *)widgets_p->diagnostics_window);
    //gtk_widget_hide (widgets_p->diagnostics_window);
    return;
}
static gboolean
destroy_diagnostics (GtkWidget * widget, GdkEvent * event, gpointer data) {
    //widgets_t *widgets_p = data;
    TRACE ("destroy_diagnostics: now ");
    hide_diagnostics_window (NULL, data);
    // clear diagnostics window
    clear_diagnostics ((widgets_t *)data);
    return TRUE;
}

GtkWidget *
rfm_create_diagnostics_window (widgets_t * widgets_p) {
    GtkWidget *dialog,
     *button,
     *scrolledwindow,
     *hbox;
    dialog = gtk_dialog_new ();
    widgets_p->diagnostics_window = dialog;
    gchar *title=g_strdup_printf("%s (stdout/stderr)", _("Output"));
    gtk_window_set_title (GTK_WINDOW (dialog), title);
    g_free(title);
    gtk_window_set_resizable (GTK_WINDOW (dialog), TRUE);
    gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
    gtk_container_set_border_width (GTK_CONTAINER (dialog), 6);

    /* create diagnostics textview */
    widgets_p->diagnostics = gtk_text_view_new ();
    scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_set_size_request (scrolledwindow, 600, 400);

    gtk_container_add (GTK_CONTAINER (scrolledwindow), widgets_p->diagnostics);
    gtk_container_set_border_width (GTK_CONTAINER (widgets_p->diagnostics), 2);
    GTK_WIDGET_UNSET_FLAGS (widgets_p->diagnostics, GTK_CAN_FOCUS);
    gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (widgets_p->diagnostics), GTK_WRAP_WORD);
    gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (widgets_p->diagnostics), FALSE);
    gtk_box_pack_start (GTK_BOX ((GTK_DIALOG (dialog))->vbox), scrolledwindow, TRUE, TRUE, 0);
    gtk_widget_show_all (scrolledwindow);
    gtk_window_stick (GTK_WINDOW(dialog));

    hbox = gtk_hbox_new (FALSE, 0);

    gtk_box_pack_start (GTK_BOX ((GTK_DIALOG (dialog))->vbox), hbox, FALSE, FALSE, 0);

    button = gtk_button_new_with_label (_("Close"));
    GdkPixbuf *pixbuf=rfm_get_pixbuf("xffm/stock_close", SIZE_BUTTON);
    gtk_button_set_image (GTK_BUTTON(button),
	    gtk_image_new_from_pixbuf(pixbuf));
    
    gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
    g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (clear_diagnostics_window), widgets_p);


    button = gtk_button_new_with_label (_("Ok"));
    pixbuf=rfm_get_pixbuf("xffm/stock_yes", SIZE_BUTTON);
    gtk_button_set_image (GTK_BUTTON(button),
	    gtk_image_new_from_pixbuf(pixbuf));
    gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
    g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (hide_diagnostics_window), widgets_p);
    g_signal_connect (G_OBJECT (dialog), "destroy_event", G_CALLBACK (destroy_diagnostics), widgets_p);
    g_signal_connect (G_OBJECT (dialog), "delete_event", G_CALLBACK (destroy_diagnostics), widgets_p);

    widgets_p->button_space = gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (hbox), widgets_p->button_space, FALSE, FALSE, 0);
    /* in run.i: gtk_widget_show_all(widgets_p->button_space); */
    /*gtk_widget_realize(dialog); */
    gtk_widget_show_all (dialog);
    return dialog;

}

void
rfm_update_status_line (view_t * view_p) {
    if(!view_p->widgets.status)
        return;
    if(view_p->selection_list && g_list_length (view_p->selection_list) == 1) {
        record_entry_t *en = view_p->selection_list->data;
        if(en && en->path) {
            GError *error = NULL;
            const gchar *id = rfm_resolve_icon_id (en);
            gchar *g = g_path_get_basename (en->path);
            gchar *gg = g_locale_to_utf8 (g, -1, NULL, NULL, &error);
            if(gg)
                rfm_status (&(view_p->widgets), id, gg, NULL);
            else {
                g_warning ("g_locale_to_utf8(%s): %s", g, error->message);
                g_error_free (error);
            }
            g_free (g);
            g_free (gg);
        }
    } else if(view_p->selection_list && g_list_length (view_p->selection_list) >= 1) {
        gchar *g=
            g_strdup_printf (ngettext ("%'d item selected", "%'d items selected",
                g_list_length (view_p->selection_list)),
                    g_list_length (view_p->selection_list));
        rfm_status (&(view_p->widgets), "xffm/stock_info", g, NULL);
    } else {
        const gchar *icon = "xffm/default";
        if(view_p->en && view_p->en->path && g_file_test (view_p->en->path, G_FILE_TEST_IS_DIR)) {
            icon = "xffm/open_folder";
        } else if(view_p->en) {
            icon = rfm_resolve_icon_id (view_p->en);
        }
        const gchar *tag = "FIXME: rfm_update_status_line()";
        if(view_p->en && view_p->en->tag)
            tag = view_p->en->tag;
        rfm_status (&(view_p->widgets), icon, tag, NULL);
    }
}

typedef struct lpterm_colors_t {
    const gchar *id;
    GdkColor fore;
    GdkColor back;
} lpterm_colors_t;

static
GtkTextTag *
resolve_tag (GtkTextBuffer * buffer, const gchar * id) {

    GtkTextTag *tag = NULL;
    lpterm_colors_t lpterm_colors_p[] = {
        {"xffm_tag/command",
         {101, 0x5858, 0x3434, 0xcfcf}, {0, 0xffff, 0xffff, 0xffff}},
        {"xffm_tag/stderr",
         {102, 0xcccc, 0, 0}, {0, 0xffff, 0xffff, 0xffff}},
        {"xffm_tag/command_id",
         {103, 0x0000, 0x0000, 0xffff}, {201, 0xffff, 0xffff, 0xffff}},
        {"xffm_tag/green",
         {104, 0x0000, 0x8888, 0x0000}, {0, 0xffff, 0xffff, 0xffff}},
        {"xffm_tag/red",
         {105, 0xaaaa, 0x0000, 0x0000}, {0, 0xffff, 0xffff, 0xffff}},
        {"xffm_tag/blue",
         {106, 0x0000, 0x0000, 0xffff}, {0, 0xffff, 0xffff, 0xffff}},
        {"xffm_tag/yellow",
         {107, 0xffff, 0xffff, 0x0000}, {202, 0x0000, 0x0000, 0x0000}},
        {"xffm_tag/magenta",
         {108, 0x8888, 0x0000, 0x8888}, {0, 0xffff, 0xffff, 0xffff}},
        {"xffm_tag/cyan",
         {109, 0x0000, 0x8888, 0x8888}, {0, 0xffff, 0xffff, 0xffff}},
        {"xffm_tag/black",
         {1, 0x0000, 0x0000, 0x0000}, {0, 0xffff, 0xffff, 0xffff}},
        {"xffm_tag/white",
         {0, 0xffff, 0xffff, 0xffff}, {0, 0x0000, 0x0000, 0x0000}},
        {NULL,
         {0, 0, 0, 0}, {0, 0, 0, 0}}

    };

    tag = gtk_text_tag_table_lookup (gtk_text_buffer_get_tag_table (buffer), id);
    // bold an italic don't work...
    if(!tag) {                  // bold and italic tags
        if(strcmp (id, "xffm_tag/bold") == 0) {
            GdkColor fore = { 300, 0x0a0a, 0x0505, 0xc3c3 };
            tag = gtk_text_buffer_create_tag (buffer, id, "weight", PANGO_WEIGHT_BOLD, "foreground_gdk", &fore, NULL);
        } else if(strcmp (id, "xffm_tag/italic") == 0) {
            GdkColor fore = { 300, 0x2b2b, 0x6e6e, 0x3333 };
            tag = gtk_text_buffer_create_tag (buffer, id, "style", PANGO_STYLE_ITALIC, "foreground_gdk", &fore, NULL);
        }
    }
    if(!tag) {                  // color tags
        /*DBG("dbg:creating tag %s\n",tag_id); */
        lpterm_colors_t *p;
        for(p = lpterm_colors_p; p && p->id; p++) {
            if(strcmp (id, p->id) == 0) {
                tag = gtk_text_buffer_create_tag (buffer, p->id, "background_gdk", &(p->back), "foreground_gdk", &(p->fore), NULL);
                break;
            }
        }
    }

    return tag;
}

void
rfm_clear_sh_command_history (view_t * view_p, gboolean disk_too) {
    if(disk_too) {
        gchar *history = g_build_filename (g_get_user_cache_dir (), LP_TERMINAL_HISTORY, NULL);
        unlink (history);
    }
    GList *p;
    for(p = g_list_first (view_p->sh_command); p && p->data; p = p->next) {
        g_free (p->data);
    }
    g_list_free (view_p->sh_command);
    view_p->sh_command = NULL;
    view_p->sh_command_counter = 0;
}

// thread function
void
rfm_operate_stderr (void *user_data, void *stream, int childFD) {
    widgets_t *widgets_p = user_data;
    char *line;
    line = (char *)stream;

    if (!rfm_diagnostics_is_visible (widgets_p) ) {
GDK_THREADS_ENTER ();
        rfm_show_text (widgets_p);
GDK_THREADS_LEAVE ();
    }
    //if(line[0] != '\n' && rfm_diagnostics_is_visible (widgets_p)) {
    if(line[0] != '\n') {
GDK_THREADS_ENTER ();
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", line, NULL);
GDK_THREADS_LEAVE ();
    }
    return;
}


// thread function
void
rfm_operate_stdout (void *user_data, void *stream, int childFD) {
    widgets_t *widgets_p = user_data;
    char *line;
    line = (char *)stream;
    DBG ("FORK stdout: %s\n", line);

    if(line[0] == '\n')
        return;

/*    this is directly in tubo library
 *    if (!valid_ansi_sequence(line)) {
        DBG("invalid ansi sequence!\n");
        return;
    }*/


    // eliminate all ^G found (as in nano output)


    //int bell=0x07; // bell
    int bs=0x08;   // backspace
    //int ht=0x09;   // horizontal tab
    //int lf=0x0a;   // linefeed
    //int vt=0x0b;   // vertical tab
    //int ff=0x0c;   // formfeed
    //int cr=0x0d;   // carriage return

    // apply all ^H (bs) found (as in rar output)
    int i,
      j;
    gchar *outline = g_strdup (line);
    for(i = 0, j = 0; line[i]; i++) {
	/*if (line[i] == bell) {
	    gdk_window_beep (widgets_p->window->window);
	} else*/ 
	if(line[i] == bs && j > 0){
            j--;
	} else {
            outline[j] = line[i];
            j++;
        }
    }
    outline[j] = 0;

    //DBG ("%s\n",outline);

    if(strncmp (line, "Tubo-id exit:", strlen ("Tubo-id exit:")) == 0) {
            if(strchr (line, '\n')) *strchr (line, '\n') = 0;
GDK_THREADS_ENTER ();
            rfm_diagnostics (widgets_p, "xffm/stock_stop", NULL);
            rfm_diagnostics (widgets_p, "xffm_tag/command_id", strchr (line, ':') + 1, ".", NULL);
            rfm_diagnostics (widgets_p, NULL, "\n", NULL);
GDK_THREADS_LEAVE ();
    } else {
        GDK_THREADS_ENTER ();
        rfm_diagnostics (widgets_p, NULL, outline, NULL);
        GDK_THREADS_LEAVE ();
    }
    return;
}

#if 0
# define BOLD 0x01
# define ITALIC 0x02
static gchar *
att_text (const gchar * s, int *att_flag) {
    gchar *ne = g_strdup (s);
    gchar *e = strchr (ne, 0x1B);
    const gchar *span = "";
    *e = 0;
    e++;
    if(strncmp (e, "[1m", strlen ("[1m")) == 0) {
        (*att_flag) |= BOLD;
        e += strlen ("[1m");
    } else if(strncmp (e, "[0m", strlen ("[0m")) == 0) {
        if(*att_flag & BOLD && *att_flag & ITALIC) {
            *att_flag = 0;
        }
        if(*att_flag & BOLD) {
            (*att_flag) &= ITALIC;
        }
        if(*att_flag & ITALIC) {
            (*att_flag) &= BOLD;
        }
        e += strlen ("[0m");
    } else if(strncmp (e, "[22m", strlen ("[22m")) == 0) {
        e += strlen ("[22m");
    } else if(strncmp (e, "[4m", strlen ("[4m")) == 0) {
        (*att_flag) |= ITALIC;
        e += strlen ("[4m");
    } else if(strncmp (e, "[24m", strlen ("[24m")) == 0) {
        e += strlen ("[24m");
    }

    gchar *nee = g_strdup_printf ("%s%s%s", ne, span, e);
    g_free (ne);
    if(strchr (nee, 0x1B)) {    //vt escape sequence
        gchar *me = att_text (nee, att_flag);
        g_free (nee);
        return me;
    }
    return nee;
}
#endif


static void
insert_string (GtkTextBuffer * buffer, const gchar * s, GtkTextTag * tag) {
    if(!s)
        return;
    GtkTextIter start,
      end,
      real_end;
    gint line = gtk_text_buffer_get_line_count (buffer);
    gchar *a = NULL;
    if(strchr (s, 0x1B)) {      //vt escape sequence
        a = g_strdup (s);
        gchar *e = strchr (a, 0x1B);
        *e = 0;
        e++;
        TRACE ("VT: insert a=%s\n", a);
        if(strlen (a))
            insert_string (buffer, a, tag);
        // figure out the tag...
        sequence_t *p;
        for(p = sequence_p; p && p->sequence; p++) {
            TRACE ("VT: %s <=> %s\n", p->sequence, e);
            if(strncmp (e, p->sequence, strlen (p->sequence)) == 0) {
                TRACE ("VT: id=%s\n", p->id);
                e += strlen (p->sequence);
                if(p->id)
                    tag = resolve_tag (buffer, p->id);
                else
                    tag = NULL;
                break;
            }
        }
        TRACE ("VT: insert e=%s\n", e);
	if (!tag) {
	    // invalid tag?
	    //DBG("invalid tag\n");
	} else {
	    insert_string (buffer, e, tag);
	}
        // bug: this tag should be kept set on the buffer until 
        // turned off explicitly
        // instead, turn off at end of input string
        g_free (a);
        // we're done
        return;
    }

    GtkTextMark *mark = gtk_text_buffer_get_mark (buffer, "rfm-ow");
    if(strchr (s, 0x0D)) {      //CR
        gchar *aa = g_strdup (s);
        *strchr (aa, 0x0D) = 0;
        insert_string (buffer, aa, tag);
        g_free (aa);
        aa = strchr (s, 0x0D) + 1;
        if(mark) {
            gtk_text_buffer_get_iter_at_line (buffer, &start, line);
            gtk_text_buffer_move_mark (buffer, mark, &start);
        }
        insert_string (buffer, aa, tag);
        g_free (a);
        // we're done
        return;
    }

    gchar *q = rfm_utf_string (s);
    /// this should be mutex protected since this function is being called
    //  from threads all over the place.
    static GMutex *insert_mutex=NULL;
    if (insert_mutex==NULL) {
        insert_mutex=g_mutex_new();
    }
    g_mutex_lock(insert_mutex);
    if(mark == NULL) {
        gtk_text_buffer_get_end_iter (buffer, &end);
        gtk_text_buffer_create_mark (buffer, "rfm-ow", &end, FALSE);
    } else {
        gtk_text_buffer_get_iter_at_mark (buffer, &end, mark);
    }

    gtk_text_buffer_get_iter_at_line (buffer, &start, line);
    gtk_text_buffer_get_end_iter (buffer, &real_end);

    // overwrite section
    gchar *pretext = gtk_text_buffer_get_text (buffer, &start, &end, TRUE);
    gchar *posttext = gtk_text_buffer_get_text (buffer, &end, &real_end, TRUE);
    long prechars = g_utf8_strlen (pretext, -1);
    long postchars = g_utf8_strlen (posttext, -1);
    long qchars = g_utf8_strlen (q, -1);
    g_free (pretext);
    g_free (posttext);
    if(qchars < postchars) {
        gtk_text_buffer_get_iter_at_line_offset (buffer, &real_end, line, prechars + qchars);
    }
    /////
    gtk_text_buffer_delete (buffer, &end, &real_end);
    if(tag) {
        gtk_text_buffer_insert_with_tags (buffer, &end, q, -1, tag, NULL);
    } else {
        gtk_text_buffer_insert (buffer, &end, q, -1);
    }
    g_mutex_unlock(insert_mutex);
    g_free (q);
    g_free (a);
    return;
}

static void
set_monospace (GtkWidget * widget) {
    static PangoFontDescription *desc = NULL;
#if 10
    static int size = 8;
    int newsize = size;
    if(getenv ("RFM_DIAGNOSTICS_FONT_SIZE")
       && strlen (getenv ("RFM_DIAGNOSTICS_FONT_SIZE"))) {
        gboolean ok = TRUE;
        char *p = getenv ("RFM_DIAGNOSTICS_FONT_SIZE");
        for(; *p; p++) {
            if(!isdigit (*p)) {
                TRACE ("%c is not a digit\n", *p);
                ok = FALSE;
            }
        }
        if(ok) {
            newsize = atoi (getenv ("RFM_DIAGNOSTICS_FONT_SIZE"));
            TRACE ("newsize=%d\n", newsize);
        }
        if(newsize < 6 || newsize > 24)
            newsize = size;
    }
    if(!desc || newsize != size)
        size = newsize;
#endif
    if(!desc) {
        if(desc)
            pango_font_description_free (desc);
        desc = pango_font_description_new ();
        pango_font_description_set_family (desc, "monospace");
        pango_font_description_set_size (desc, size * PANGO_SCALE);
    }
    gtk_widget_modify_font (widget, desc);
}

static void
set_sans (GtkWidget * widget) {
    static PangoFontDescription *desc = NULL;
    if(!desc) {
        if(desc)
            pango_font_description_free (desc);
        desc = pango_font_description_new ();
        pango_font_description_set_family (desc, "sans");
    }
    gtk_widget_modify_font (widget, desc);
}

void
rfm_status (widgets_t * widgets_p, const gchar * id, ...
    ) {
    va_list var_args;
    char *t;
    GtkTextIter start,
      end;
    GtkTextBuffer *buffer;
    GdkPixbuf *icono;
    int icono_width = 0;

    if (!widgets_p) {
        return;
    }
//    if(!widgets_p) return;

    if(!widgets_p->status || !GTK_IS_TEXT_VIEW (widgets_p->status)) {
#ifdef DEBUG
        /*g_warning("rfm_status: status==NULL"); */
        va_start (var_args, id);
        do {
            t = va_arg (var_args, char *);
            if(t && strlen (t)) {
                printf ("%s\n", t);
            }
        }
        while(t);
        va_end (var_args);
#endif
        return;
    }
    set_sans (widgets_p->status);
    /*gtk_text_view_set_justification ((GtkTextView *) widgets_p->status,GTK_JUSTIFY_LEFT); */
    buffer = gtk_text_view_get_buffer ((GtkTextView *) widgets_p->status);
    gtk_text_buffer_set_text (buffer, " ", -1);
    gtk_text_buffer_get_bounds (buffer, &start, &end);

    if(!id)
        id = "xffm/default";

    icono = rfm_get_pixbuf (id, SIZE_BUTTON);

    if(icono) {
        gtk_text_buffer_insert_pixbuf (buffer, &start, icono);
        icono_width = gdk_pixbuf_get_width (icono);
    }

    {
        char *string = NULL;    /*,*tag; */
        gchar *f = NULL;
        va_start (var_args, id);
        do {
            t = va_arg (var_args, char *);
            if(t && strlen (t)) {
                if(!f)
                    f = g_strdup ("");
                string = g_strconcat (f, t, NULL);
                g_free (f);
                f = string;
            }
        }
        while(t);
        va_end (var_args);
        if(string) {
            insert_string (buffer, string, NULL);
            g_free (string);
        }
    }
    g_object_set_data (G_OBJECT (widgets_p->status), "clean", (gpointer) ((long)0x01));
/*  gtk_text_buffer_get_end_iter (buffer,&end);
  gtk_text_buffer_place_cursor (buffer,&end);
//  gtk_text_view_move_visually ( (GtkTextView *)(widgets_p->status),&end,10);
  gtk_text_view_scroll_to_iter  ((GtkTextView *)(widgets_p->status),&end,
				0.0, FALSE,
				1.0,0.0);
				  //gtk_text_buffer_place_cursor (buffer,&end);*/

    /*gdk_flush(); */
}

void
rfm_diagnostics (widgets_t * widgets_p, const gchar * id, ...
    ) {
    char *t;
    va_list var_args;
    GtkTextIter start,
      end;
    GtkTextMark *mark;
    GtkTextBuffer *buffer;
    GdkPixbuf *icono = NULL;

    if (!widgets_p) {
        return;
    }
    /*if(!widgets_p) {
        TRACE ("!widgets_p\n");
        return;
    }*/

    if(getenv ("RFM_DISABLE_DIAGNOSTICS")
       && strlen (getenv ("RFM_DISABLE_DIAGNOSTICS"))) {
        return;
    }

    if(!widgets_p->diagnostics || !GTK_IS_TEXT_VIEW (widgets_p->diagnostics)) {
#ifdef DEBUG
        /*g_warning("diagnostics==NULL"); */
        va_start (var_args, id);
        do {
            t = va_arg (var_args, char *);
            if(t && strlen (t)) {
                printf ("%s\n", t);
            }
        }
        while(t);
        va_end (var_args);
#endif
        return;
    }


    if(!rfm_diagnostics_is_visible (widgets_p)) {
        TRACE ("!rfm_diagnostics_is_visible(diagnostics)\n");
        return;
    }
    set_monospace (widgets_p->diagnostics);

    /****  */
    buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widgets_p->diagnostics));

    /*DBG("TRACE:rfm_diagnostics\n"); */
    gtk_text_buffer_get_bounds (buffer, &start, &end);

    GtkTextTag *tag = NULL;
    //DBG("DIAGNOSTICS: id= %s\n",id);
    if(id) {
        if(strncmp (id, "xffm_tag/", strlen ("xffm_tag/")) == 0) {
            tag = resolve_tag (buffer, id);
        } else {
            icono = rfm_get_pixbuf (id, SIZE_BUTTON);
        }
    }

    if(icono) {
        gtk_text_buffer_insert_pixbuf (buffer, &end, icono);
    }

    va_start (var_args, id);
    do {
        t = va_arg (var_args, char *);
        if(t && strlen (t)) {
            insert_string (buffer, t, tag);
        }
    }
    while(t);
    va_end (var_args);
    gtk_text_buffer_get_bounds (buffer, &start, &end);
    mark = gtk_text_buffer_create_mark (buffer, "scrollmark", &end, FALSE);
    gtk_text_view_scroll_to_mark (GTK_TEXT_VIEW (widgets_p->diagnostics), mark, 0.2,    /*gdouble within_margin, */
                                  FALSE, 0.0, 0.0);
    gtk_text_buffer_delete_mark (buffer, mark);
    //gdk_flush();
}

void
rfm_show_text (widgets_t * widgets_p) {

    if(!widgets_p) return;
    if(widgets_p->diagnostics == NULL) {
	if (widgets_p->diagnostics_window){
	    gtk_widget_show_all(widgets_p->diagnostics_window);
	    //gtk_window_deiconify(widgets_p->diagnostics_window);
	    return;
	}
        rfm_create_diagnostics_window (widgets_p);
    }    
    if(!rfm_diagnostics_is_visible (widgets_p)) {
        if(!widgets_p->vpane){
            return;
	}
        if(gtk_paned_get_position ((GtkPaned *) widgets_p->vpane) > widgets_p->vpane->allocation.height * 0.80)
            gtk_paned_set_position (GTK_PANED (widgets_p->vpane), widgets_p->vpane->allocation.height * 0.75);
    }
}

gboolean
rfm_diagnostics_is_visible (widgets_t * widgets_p) {
    if (!widgets_p) return FALSE;
    if (widgets_p->diagnostics_window){
       if (gtk_widget_get_visible(widgets_p->diagnostics_window)) {
	    return TRUE;
       } else {
	   return TRUE;
	   //return FALSE;
       }
    }

    if(!widgets_p->diagnostics){
	    return FALSE;
    }

    if(!widgets_p->vpane)
        return FALSE;
    if(gtk_paned_get_position (GTK_PANED(widgets_p->vpane)) > widgets_p->vpane->allocation.height * 0.95)
        return FALSE;
    else
        return TRUE;
}

void
rfm_hide_text (widgets_t * widgets_p) {
    GtkWidget *vpaned;
    if(!widgets_p->diagnostics)
        return;
    if(widgets_p->diagnostics_window) {
        gtk_widget_hide (widgets_p->diagnostics_window);
        return;
    }
    vpaned = rfm_get_widget_by_name (widgets_p->paper, "vpaned1");
    if(!vpaned)
        return;
    gtk_paned_set_position (GTK_PANED (vpaned), vpaned->allocation.height);
}

static void
clear_diagnostics (widgets_t * widgets_p) {
    GtkTextIter start,
      end;
    GtkTextBuffer *buffer;
    if(!widgets_p->diagnostics)
        return;
    buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (widgets_p->diagnostics));
    gtk_text_buffer_get_bounds (buffer, &start, &end);
    gtk_text_buffer_delete (buffer, &start, &end);
    rfm_hide_text (widgets_p);
}

void
rfm_clear_text_window (GtkButton * button, gpointer data) {

    widgets_t *widgets_p = (widgets_t *) data;
    if(!widgets_p->diagnostics)
        return;
    clear_diagnostics (widgets_p);
    gtk_widget_grab_focus (widgets_p->paper);
    rfm_update_status_line (widgets_p->view_p);
}

static GtkWidget *
dialog_button (const char *icon_id, const char *text) {
    GtkWidget *button = gtk_button_new_with_label (text);

    if(icon_id) {
        GdkPixbuf *pb = rfm_get_pixbuf (icon_id, SIZE_BUTTON);
        GtkWidget *image = gtk_image_new_from_pixbuf (pb);
	gtk_button_set_image(GTK_BUTTON(button), image);
    }
    gtk_widget_show_all (button);

    return button;

}

static
GtkWidget *
rfm_confirm_dialog (widgets_t * widgets_p, 
	GtkMessageType type,
	const gchar * text, 
	const gchar * action_false,
	const gchar * action_true 
	) {
    GtkWidget *dialog;
    GtkWidget *button;

    DBG ("DIAGNOSTICS: rfm_confirm_dialog\n");
    gchar *q = rfm_utf_string (text);
    dialog = gtk_message_dialog_new_with_markup (
	    NULL, GTK_DIALOG_MODAL,
	    type, GTK_BUTTONS_NONE, 
	    "%s", q);
    g_free (q);

    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
    if(action_false) {
        button = dialog_button ("xffm/stock_no", action_false);
        gtk_widget_show (button);
        gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, GTK_RESPONSE_NO);
        g_object_set_data (G_OBJECT (dialog), "action_false_button", button);
    }

    button = dialog_button ("xffm/stock_yes", (action_true) ? action_true : _("Ok"));
    gtk_widget_show (button);
    g_object_set_data (G_OBJECT (dialog), "action_true_button", button);
    gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, GTK_RESPONSE_YES);

    if(widgets_p && widgets_p->window)
        gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (widgets_p->window));
    gtk_widget_show (dialog);
    return dialog;
}

G_MODULE_EXPORT gboolean
rfm_confirm (widgets_t * widgets_p, 
	// type: GTK_MESSAGE_INFO, GTK_MESSAGE_WARNING, GTK_MESSAGE_QUESTION, GTK_MESSAGE_ERROR
	GtkMessageType type,
	const gchar * text, 
	const gchar * action_false,  // if NULL, button not shown
	const gchar * action_true   // if NULL, "Ok" button shown
	) {
    int response = GTK_RESPONSE_NONE;
    DBG ("DIAGNOSTICS: rfm_confirm\n");
    GtkWidget *dialog = rfm_confirm_dialog (widgets_p, type, text, action_false, action_true);
    if(!dialog){
        return FALSE;
    }
    DBG ("DIAGNOSTICS: rfm_confirm-- gtk_dialog_run\n");
    response = gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_hide (dialog);
    gtk_widget_destroy (dialog);

    if(response == GTK_RESPONSE_YES)
        return TRUE;
    else
        return FALSE;
}

const gchar *
rfm_get_password (widgets_t * widgets_p, const gchar * ptext) {
    const gchar *p;
    g_free (password);
    password = NULL;
    p = passwd_dialog (widgets_p, 11, NULL, ptext);
    if(p && strchr (p, '%'))
        p = strchr (p, '%') + 1;
    return p;
}

const gchar *
rfm_get_smbuserpass (widgets_t * widgets_p, record_entry_t * en) {
    const gchar *p;
    const gchar *u = NULL;
    /* browse password: */
    if(en && en->tag && strlen (en->tag)) {
        u = en->tag;
    } else {
        if(getenv ("SMB_USER") && strlen (getenv ("SMB_USER"))) {
            u = getenv ("SMB_USER");
        }
    }
    p = passwd_dialog (widgets_p, 1, u, NULL);
    return p;
}
