/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

// FIXME: this static variable is thread-unsafe.
static int relative_sort_column = TYPE_SORT;

#define MODULE_STAT(x,y) GPOINTER_TO_INT(rfm_rational("rfm", xfdir_p->view_p->module, x,y, "module_stat"))
#define MODULE_COUNT(x) GPOINTER_TO_INT(rfm_natural("rfm", xfdir_p->view_p->module, x, "module_count"))
#define MODULE_XFDIR(x) GPOINTER_TO_INT(rfm_natural("rfm", xfdir_p->view_p->module, x, "module_xfdir_get"))
#define MODULE_DATA(x) (rfm_natural("rfm", xfdir_p->view_p->module, x, "module_data"))
#define SUBMODULE_NAME (rfm_void("rfm", xfdir_p->view_p->module_name, "submodule_name"))
// check if monitor should wait:
/**
 * xfdir_monitor_skip:
 * @xfdir_p: pointer to an xfdir_t structure
 * Returns: TRUE if monitor should skip tests on current loop
 * or FALSE if monitor should go ahead with the update tests.
 * 
 * Checks whether the xfdir monitor should skip monitor tests on the current
 * loop or go ahead with monitor tests. It is an exported symbol so that modules
 * may use the same monitor algorithm as the xfdir monitor thread.
 **/
static
  gboolean
xfdir_monitor_skip (xfdir_t * xfdir_p) {
    static gboolean said = FALSE;
    gboolean skip = FALSE;
    if(xfdir_p == NULL) {
        g_error ("monitor_skip: xfdir_p ==NULL!");
    }
    view_t *view_p = xfdir_p->view_p;
    GtkWidget *w = view_p->widgets.popup;
    /*XXX view_p not set up yet */
    /*if (xfdir_p->active) {
       skip = TRUE;
       DBG("thread monitor: xfdir_p->active\n");
       } */
    /* update in progress */

    if(view_p->monitor_loop_mutex == NULL) {
        view_p->monitor_loop_mutex = g_mutex_new ();
    }
    g_mutex_lock (xfdir_p->view_p->monitor_loop_mutex);
    if(xfdir_p->view_p->refresh) {
        skip = TRUE;
        DBG ("thread monitor is skipping: xfdir_p->refresh\n");
    }
    g_mutex_unlock (xfdir_p->view_p->monitor_loop_mutex);

    if(view_p->thumbnail_loop_mutex == NULL) {
        view_p->thumbnail_loop_mutex = g_mutex_new ();
    }
    g_mutex_lock (view_p->thumbnail_loop_mutex);
    TRACE ("thread monitor: thumbnail_loop_mutex OK\n");
    g_mutex_unlock (view_p->thumbnail_loop_mutex);

    /* popup active? */
    if(w && gtk_widget_get_visible (w)) {
        skip = TRUE;
        DBG ("thread monitor: popup visible\n");
    }
    /* entry window is mapped? */
    else if(view_p->widgets.rename) {
        skip = TRUE;
        DBG ("thread monitor: entry window is mapped\n");
    }
    /* tip window is mapped? This could be disabled, but better be safe. */

    else if(view_p->tooltip_active) {
        // But is it a valid active state?
	// State should be valid here since
	// condition is taken care of in rodent_mouse.c
	// with the on_leave signal
        TRACE ("thread monitor: widgets.tooltip_active\n");
        skip = TRUE;
    }
    /* button down? */
    else if(view_p->boxX != -1 && view_p->boxY != -1) {
        skip = TRUE;
        DBG ("thread monitor: boxX != -1\n");
    }
    /* dragging? */
    else if(view_p->doing_drag_p)
        skip = TRUE;
    /* saving cache? */
    if(!skip && said) {
        said = FALSE;
        DBG ("thread monitor: active\n");
    }
    if(!skip && said) {
        said = FALSE;
        DBG ("thread monitor: active\n");
    }
    if(skip) {
        if(!said) {
            said = TRUE;
            DBG ("thread monitor: sleeping\n");
        }
        rfm_threadwait ();
    }
    return skip;
}

/**
 * xfdir_free_data:
 * @xfdir_p:  pointer to an xfdir_t structure
 *
 *  Frees allocated resources of a loaded xfdir structure
 *
 **/
static
    int
xfdir_free_data (xfdir_t * xfdir_p) {

    if(!xfdir_p) {
        g_warning ("xfdir_p==NULL in xfdir_free_data");
        return -1;
    }
    if(xfdir_p->gl != NULL && xfdir_p->pathc) {
        int i;
        for(i = 0; i < xfdir_p->pathc; i++) {
            TRACE ("thread monitor: will free xfdir_p->gl[i].pathv=0x%lx\n", (long unsigned)xfdir_p->gl[i].pathv);
            g_free (xfdir_p->gl[i].pathv);
            TRACE ("thread monitor: will destroy xfdir_p->gl[i].en=0x%lx\n", (long unsigned)xfdir_p->gl[i].en);
            rfm_destroy_entry (xfdir_p->gl[i].en);
        }
        TRACE ("thread monitor: will free xfdir_p->gl=0x%lx\n", (long unsigned)xfdir_p->gl);
        g_free (xfdir_p->gl);
        xfdir_p->gl = NULL;
        xfdir_p->pathc = 0;
    }
    TRACE ("thread monitor: will free xfdir_p->data=0x%lx\n", (long unsigned)xfdir_p->data);
    g_free (xfdir_p->data);
    return 0;
}

static
  gboolean
exit_monitor (xfdir_t * xfdir_p) {
    gboolean result = FALSE;
    g_mutex_lock (xfdir_p->view_p->monitor_id_mutex);
    if(xfdir_p->view_p->monitor_id != xfdir_p->monitor_id)
        result = TRUE;
    g_mutex_unlock (xfdir_p->view_p->monitor_id_mutex);
    return result;
}

/**
 * xfdir_monitor_continue:
 * @xfdir_p:  pointer to an xfdir_t structure
 * Returns: TRUE if monitor should continue alive or FALSE if monitor should 
 * finish.
 *
 * This function is a generic check whether the local file monitor thread
 * should continue alive or exit. It is an exported symbol so that modules
 * may use the same monitor algorithm as the xfdir monitor thread.
 **/
static
  gboolean
xfdir_monitor_continue (xfdir_t * xfdir_p) {
    if(!xfdir_p) {
        DBG ("thread monitor: monitor_continue !xfdir_p\n");
        return FALSE;
    }
    /* signaled exit: */
    if(exit_monitor (xfdir_p)) {
        DBG ("thread monitor: exit_monitor(xfdir_p) = 1\n");
        return FALSE;
    }
    /* invalid view: */
    if(!xfdir_p->view_p)
        return FALSE;
    /* root window */
    if(!xfdir_p->view_p->en) {
        DBG ("thread monitor: exit on !xfdir_p->view_p->en\n");
        return FALSE;
    }
    /* directory has vanished */
    if(xfdir_p->view_p->module) {
        // module defined test.
        if(!MODULE_STAT (NULL, NULL))
            return FALSE;

    } else if(xfdir_p->path == NULL || !g_file_test (xfdir_p->path, G_FILE_TEST_EXISTS)) {
        // this is the test for local files.
        DBG ("monitor: !g_file_test(view_p->xfdir_p->path,G_FILE_TEST_EXISTS)\n");
        return FALSE;
    }
    return TRUE;
}

static const gchar *
get_local_shm_name (gchar * directory) {
    static gchar key[11];
    GString *gs = g_string_new (directory);
    sprintf (key, "%010u", g_string_hash (gs));
    g_string_free (gs, TRUE);
    return (const gchar *)key;
}

static off_t
st_sum (struct stat *st) {
    off_t sum;
    if(!st)
        sum = 0;
    else {
        sum = st->st_mtime + st->st_size + st->st_mode + st->st_nlink + st->st_uid + st->st_gid;
    }
    return sum;
}

static off_t
st_sum_dir (struct stat *st) {
    off_t sum;
    int build=atoi(BUILD);
    if(!st){
        sum = 0;
    } else {
        sum = st->st_mtime + st->st_ctime + st->st_size + st->st_mode + st->st_nlink + st->st_uid + st->st_gid + build;
    }
    return sum;
}

static
    void
set_sort_column (int column) {
    relative_sort_column = column;
}

static
int update_xfdir (xfdir_t * xfdir_p);

static
    void
set_deepest_dir (view_t * view_p) {
    if(view_p->en && view_p->en->path && g_file_test (view_p->en->path, G_FILE_TEST_IS_DIR)) {
        if(!view_p->deepest_dir || !strstr (view_p->deepest_dir, view_p->en->path)) {
            g_free (view_p->deepest_dir);
            view_p->deepest_dir = g_strdup_printf ("%s/.", view_p->en->path);
        }
    } else {
        //g_warning("cannot set deepest dir");
    }
}

// Two possible races with other processes:
//  1. There are more files present than those counted
//  2. There are less files present than those counted
static
  gint
read_files_local (xfdir_t * xfdir, gboolean fullread) {
    DIR *directory;
    gint count = 0;
//    gint smallcount = 0;
    struct dirent *d;
    struct stat st;

    directory = opendir (xfdir->path);
    if(!directory)
        return count;

    while((d = readdir (directory)) != NULL) {
        gchar *fullpath = NULL;
        gboolean broken = FALSE;
        if(strcmp (d->d_name, ".") == 0)
            continue;
        if(!SHOWS_HIDDEN (xfdir->type) && d->d_name[0] == '.' && strcmp (d->d_name, "..") != 0)
            continue;

        if(fullread && xfdir != NULL) {
            if(!fullpath) {
                if(strcmp (d->d_name, "..") == 0)
                    fullpath = g_path_get_dirname (xfdir->path);
                else
                    fullpath = g_build_filename (xfdir->path, d->d_name, NULL);
                if(stat (fullpath, &st) < 0) {
                    if(lstat (fullpath, &st) < 0) {
                        g_free (fullpath);
                        fullpath = NULL;
                        continue;
                    }
                    broken = TRUE;
                }
            }
            if(strcmp (d->d_name, "..") == 0) {
                if(strcmp (xfdir->path, "/") == 0) {
                    xfdir->gl[count].en = NULL;
                    xfdir->gl[count].pathv = g_strdup (g_get_host_name ());
                } else {
                    xfdir->gl[count].en = rfm_stat_entry (fullpath, 0);
                    xfdir->gl[count].pathv = g_path_get_basename (fullpath);
                    xfdir->gl[count].en->path = fullpath;
                    SET_DUMMY_TYPE (xfdir->gl[count].en->type);
                    SET_UP_TYPE (xfdir->gl[count].en->type);

                }

            } else {
                xfdir->gl[count].en = rfm_stat_entry (fullpath, xfdir->type);
                xfdir->gl[count].pathv = g_strdup (d->d_name);
                TRACE ("POP_SEM:   %s st sum is %ld\n", xfdir->gl[count].en->path, (long)st_sum (&st));
            }
            count++;
            // ensure that we only read files for which we have enough malloc'd memory:
            if(count >= xfdir->pathc) {
                break;
            }
        } else {                // only counting files...
            g_free (fullpath);
            count++;
        }
    }
    closedir (directory);
    // ignore malloc'd space for files which have disappeared between count and read steps:
    if(fullread && xfdir != NULL && count < xfdir->pathc) {
        xfdir->pathc = count;
    }
    return (count);
}

static
    void
save_xfdir_cache (xfdir_t * xfdir_p, int checksum) {
    int fd;
    gchar *cache_path;

    if(!xfdir_p || !xfdir_p->path || !strlen (xfdir_p->path))
        return;
    if((cache_path = rfm_xfdir_get_local_cache_path (xfdir_p->path)) == NULL) {
        return;
    }
    DBG ("writing local cache: %s for %s\n", cache_path, xfdir_p->path);

    fd = creat (cache_path, 0660);
    if(fd < 0) {
        g_warning ("open(%s): %s", cache_path, strerror (errno));
        g_free (cache_path);
        return;
    }
    // get the file lock, wait if necesary:
    struct flock lock;

    lock.l_type = F_WRLCK;
    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;             // to eof   
    if(fcntl (fd, F_SETLKW, &lock) < 0) {
        g_warning ("fcntl(F_WRLCK,F_SETLKW)(%s): %s", cache_path, strerror (errno));
    }

    if(write (fd, &checksum, sizeof (int)) < 0) {
        g_warning ("write: %s", strerror (errno));
    }
    if(write (fd, &(xfdir_p->pathc), sizeof (int)) < 0) {
        g_warning ("write: %s", strerror (errno));
    }

    int j;
    for(j = 0; j < xfdir_p->pathc; j++) {
        int path_length = 0;
        if(!xfdir_p->gl[j].pathv || !xfdir_p->gl[j].en) {
            continue;
        }
        path_length = strlen (xfdir_p->gl[j].en->path);
        if(write (fd, &path_length, sizeof (int)) < 0) {
            g_warning ("write: %s", strerror (errno));
        }
        if(write (fd, xfdir_p->gl[j].en->path, path_length) < 0) {
            g_warning ("write: %s", strerror (errno));
        }
        if(write (fd, &(xfdir_p->gl[j].en->type), sizeof (int)) < 0) {
            g_warning ("write: %s", strerror (errno));
        }
        if(write (fd, xfdir_p->gl[j].en->st, sizeof (struct stat)) < 0) {
            g_warning ("write: %s", strerror (errno));
        }
    }

    // release the lock:
    lock.l_type = F_UNLCK;
    if(fcntl (fd, F_SETLKW, &lock) < 0) {
        g_warning ("fcntl(F_UNLCK)(%s): %s", cache_path, strerror (errno));
    }
    close (fd);
    g_free (cache_path);
    return;
}

#define MAX_CACHE_LINE 255

static
  gboolean
read_xfdir_cache (xfdir_t * xfdir_p) {
    int fd;
    gchar *cache_path;
    int j,
      k;
    char *line = NULL;

    if(!xfdir_p || !xfdir_p->path || !strlen (xfdir_p->path)) {
        DBG ("read_xfdir_cache: !xfdir_p || !xfdir_p->path || !strlen(xfdir_p->path\n");
        return FALSE;
    }
    if((cache_path = rfm_xfdir_get_local_cache_path (xfdir_p->path)) == NULL) {
        DBG ("read_xfdir_cache: rfm_xfdir_get_local_cache_path(xfdir_p->path)==NULL\n");
        return FALSE;
    }
    // it is already stated... stat(xfdir_p->path,&xfdir_p.st);

    DBG ("reading local cache: %s for %s\n", cache_path, xfdir_p->path);
    fd = open (cache_path, 0);

    if(fd < 0) {
        //g_warning("open(%s): %s",cache_path,strerror(errno));
        g_free (cache_path);
        return FALSE;
    }
    // get the file lock, wait if necesary:
    struct flock lock;

    lock.l_type = F_RDLCK;
    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;             // to eof   
    if(fcntl (fd, F_SETLKW, &lock) < 0) {
        g_warning ("fcntl(F_RDLCK,F_SETLKW)(%s): %s", cache_path, strerror (errno));
    }

    xfdir_p->gl = NULL;
    if(read (fd, &j, sizeof (int)) != sizeof (int)) {
      short_read:
        DBG ("cache Short read in %s. Deleting\n", cache_path);
        g_free (line);
        g_free (xfdir_p->gl);
        // release the lock:
        lock.l_type = F_UNLCK;
        if(fcntl (fd, F_SETLKW, &lock) < 0) {
            g_warning ("fcntl(F_UNLCK)(%s): %s", cache_path, strerror (errno));
        }
        close (fd);
        unlink (cache_path);
        g_free (cache_path);
        return FALSE;
    }
    k = st_sum_dir (&(xfdir_p->st)) + SHOWS_HIDDEN (xfdir_p->type);

    if(j != k) {
        TRACE ("cache outofdate: %s (%d != %d)\n", xfdir_p->path, j, k);
        goto short_read;
    }

    if(read (fd, &(xfdir_p->pathc), sizeof (int)) != sizeof (int))
        goto short_read;
    // last test, file count is the same in cache as on disk
    if(xfdir_p->pathc != read_files_local (xfdir_p, FALSE)) {
        DBG ("cache out of date: xfdir_p->pathc != read_files_local()\n");
        goto short_read;
    }
    if(!xfdir_p->pathc) {
        DBG ("cache xfdir_p->pathc == 0\n");
        goto short_read;
    }

    xfdir_p->gl = (dir_t *) malloc (xfdir_p->pathc * sizeof (dir_t));
    if(!xfdir_p->gl) {
        DBG ("cache cannot malloc xfdir_p->gl\n");
        goto short_read;
    }
    /* xfdir_p->gl is allocated... */

    line = (gchar *) malloc (MAX_CACHE_LINE + 1);
    for(j = 0; j < xfdir_p->pathc; j++) {
        int path_length = 0;
        if(read (fd, &path_length, sizeof (int)) != sizeof (int))
            goto short_read;

        if(path_length > MAX_CACHE_LINE) {
            g_free (line);
            line = (gchar *) malloc (path_length + 1);
            if(!line)
                goto short_read;
        }

        if(read (fd, line, path_length) != path_length)
            goto short_read;

        line[path_length] = 0;
        xfdir_p->gl[j].en = rfm_mk_entry (xfdir_p->type);
        if(!xfdir_p->gl[j].en)
            goto short_read;

        if(read (fd, &(xfdir_p->gl[j].en->type), sizeof (int)) != sizeof (int))
            goto short_read;

        xfdir_p->gl[j].en->path = g_strdup (line);
        xfdir_p->gl[j].pathv = g_path_get_basename (line);
        xfdir_p->gl[j].en->st = (struct stat *)malloc (sizeof (struct stat));
        if(read (fd, xfdir_p->gl[j].en->st, sizeof (struct stat)) != sizeof (struct stat))
            goto short_read;
    }
    g_free (line);
    // release the lock:
    lock.l_type = F_UNLCK;
    if(fcntl (fd, F_SETLKW, &lock) < 0) {
        g_warning ("fcntl(F_UNLCK)(%s): %s", cache_path, strerror (errno));
    }
    close (fd);
    DBG ("cache read ok\n");
    g_free (cache_path);
    return TRUE;
}

// this function checks if a directory (or module) currently in 
// xfdir structure needs to be updated in the view (in other words,
// is a new xfdir structure necessary?)
static
  gboolean
reload_condition (xfdir_t * xfdir_p, gint * current_count, struct stat *current_st) {
    gboolean instant_return = FALSE;

    // Test 1. Short circuit on instruction from another thread.
    //
    // put a lock on xfdir structure and use the reload variable
    // to shortcircuit test on condition set by other threads
    g_mutex_lock (xfdir_p->view_p->monitor_loop_mutex);
    instant_return = xfdir_p->reload;
    xfdir_p->reload = FALSE;
    g_mutex_unlock (xfdir_p->view_p->monitor_loop_mutex);
    if(instant_return){
        return TRUE;
    }

    // Test 2. inode count condition.
    // This will count the number of icons in the view.
    gint instant_count;
    if(xfdir_p->view_p->module) {
        // module defined count test.
        instant_count = MODULE_COUNT (xfdir_p);
    } else {
        // local directory test 
        instant_count = read_files_local (xfdir_p, FALSE);
    }
    DBG ("MONITOR current count= %d, instant=%d\n", *current_count, instant_count);
    if(*current_count != instant_count) {
        DBG ("reload condition true on COUNT condition %d != %d\n", 
		*current_count, instant_count);
	// We update the current count, otherwise the reload condition
	// will not cease, and it must cease to trigger the actual reload.
	*current_count=instant_count;
        return TRUE;
    }

    // Test 3. Check stat information of current view.
    struct stat instant_st;
    if(xfdir_p->view_p->module) {
        // module defined test.
        if(!MODULE_STAT (NULL, &instant_st)) {
            DBG ("MONITOR: module_stat not defined for %s\n", xfdir_p->view_p->module);
        }
    } else {
        // local file test
        stat (xfdir_p->path, &instant_st);
    }
    if(st_sum_dir (current_st) != st_sum_dir (&instant_st)) {
 	// We update the current stat, otherwise the reload condition
	// will not cease, and it must cease to trigger the actual reload.
        memcpy (current_st, &instant_st, sizeof (struct stat));
        DBG ("MONITOR: reload condition true on STAT condition\n");
        return TRUE;
    }

    // Test 4. Module specific test: comparison of character data 
    // 	       in xfdir structure.
    if(xfdir_p->view_p->module && xfdir_p->data) {
        void *data = MODULE_DATA (xfdir_p);
        int different = strcmp ((char *)(xfdir_p->data), (char *)data);
        if(different) {
            g_free (xfdir_p->data);
            xfdir_p->data = data;
            DBG ("MONITOR: reload condition true on DATA condition\n");
            return TRUE;
        }
        g_free (data);
    }

    return FALSE;
}

// If a full reload is not necessary, this function just updates
// stat information and thumbnails.
static gboolean
update_stat_info (xfdir_t * xfdir_p) {
    int i;
    view_t *view_p = xfdir_p->view_p;

    // We skip first element (which is the dummy "up" icon).
    for(i = 1; i < xfdir_p->pathc; i++) {
	// Is the work still necessary or has the thread met an exit condition?
        if(exit_monitor (xfdir_p)){
            return FALSE;
	}
	// This is a safeguard, it really should not happen though.
        if(xfdir_p->gl[i].en == NULL){
            continue;
	}

        struct stat st;
        if(xfdir_p->view_p->module) {
            // module defined test.
            if(!MODULE_STAT (xfdir_p->gl[i].en->path, &st)) {
                DBG ("module_stat not defined for %s\n", xfdir_p->view_p->module);
            }
        } else {
            // local files test
            if (stat (xfdir_p->gl[i].en->path, &st) < 0) {
		// Failed stat instruction means that the file count test
		// (which passed by now) is incorrect and does not reflect
		// the actual directory condition.
		return TRUE;
	    }
        }
	
	// This is the stat comparison test.
        if(st_sum (xfdir_p->gl[i].en->st) != st_sum (&st)) {
	    // Stat information update is necessary.
            DBG ("POP_SEM:  stat differs for %s (%ld != %ld\n",
                 xfdir_p->gl[i].en->path, 
		 (long)st_sum (xfdir_p->gl[i].en->st), 
		 (long)st_sum (&st));
	    // Step 1. update stat information in xfdir structure.
            memcpy (xfdir_p->gl[i].en->st, &st, sizeof (struct stat));
            DBG ("POP_SEM:  update_stat_info lock..\n");
	    // Step 2. update stat information in corresponding view population item.
            // First get the population semaphore. A simple read lock is OK since
	    // the view_p->population_pp is not affected.
            if(!rfm_population_try_read_lock (view_p)){
                continue;
	    }
            if(exit_monitor (xfdir_p)) {
		// Now, is the thread doing required work or not?
		// Condition may have changed after the read lock wait.
                rfm_population_read_unlock (view_p);
                return FALSE;
            }

            // Do the update on current view population.
            population_t **tmp;
            for(tmp = view_p->population_pp; *tmp; tmp++) {
                population_t *population_p = *tmp;
		gboolean found =  population_p->en
			       && population_p->en->path
			       && strcmp (population_p->en->path, xfdir_p->gl[i].en->path) == 0;
                if(found) {
                    DBG ("POP_SEM: found population_p for %s\n", xfdir_p->gl[i].en->path);
		    // Update current population item stat information.
                    memcpy (population_p->en->st, &st, sizeof (struct stat));
		    // Does the referred item have a thumbnail?
                    if(population_p->preview_done) {
                        // It does have a thumbnail. This should be redone.
                        DBG ("THREAD: thread(rfm_preview_thread_f) at primary-xfdir.i\n");
                        // First we invalidate preview popup so that it will be
			// regenerated the next time the user triggers a tooltip preview.
                        population_p->preview_pixbuf_timestamp = 0;
			// Now assign memory block which is to be used by the thumbnail
			// preview thread.
			preview_child_t *preview_child_p = 
			    (preview_child_t *) malloc (sizeof (preview_child_t));
			// Initialize preview thread memory block.
			preview_child_p->population_p = population_p;
			preview_child_p->population_serial = view_p->population_serial;
			preview_child_p->view_p = view_p;
			// Fire up preview thread and forget.
                        g_thread_create (rfm_preview_thread_f, (gpointer) preview_child_p, FALSE, NULL);
			// Exit the item in population search.
			break; 
                    } else { 
			// update the icon pixmap, for directories so that
			// greenball dissappears if unmounted or appears
			// if mounted.
			if (IS_SDIR(population_p->en->type)) {
			  gchar *icon_id = rfm_get_composite_icon_id (population_p->en, NULL); 
			  GdkPixbuf *pixbuf = 
			    rfm_get_pixbuf (icon_id, view_p->icon_size);
			  g_free (icon_id);
			  population_p->pixbuf = 
			    population_p->base_pixbuf = 
			    population_p->normal_pixbuf = pixbuf;
			  if (population_p->saturated_pixbuf) {
			      g_object_unref(population_p->saturated_pixbuf);
			      population_p->saturated_pixbuf=NULL;
			  }
			  if (population_p->cut_pixbuf) {
			      g_object_unref(population_p->cut_pixbuf);
			      population_p->cut_pixbuf=NULL;
			  }
			  if (population_p->selected_pixbuf) {
			      g_object_unref(population_p->selected_pixbuf);
			      population_p->selected_pixbuf=NULL;
			  }

			}
		    }
                }
            } // This is the end of the item in population search.
            // Release the population read lock.
            rfm_population_read_unlock (view_p);
        } // This is the end of the stat comparison test.
    } // This is the end of the stat comparison loop on all items of xfdir structure
    return FALSE;
}

static gint
pathv_compare_up (const gchar * a, const gchar * b) {
    if(!a && !b)
        return 0;
    if(!a)
        return -1;
    if(!b)
        return 1;
    if(strchr (a, G_DIR_SEPARATOR))
        a = strrchr (a, G_DIR_SEPARATOR) + 1;
    if(strchr (b, G_DIR_SEPARATOR))
        b = strrchr (b, G_DIR_SEPARATOR) + 1;
    return strcmp (a, b);
}

static gint
pathv_compare_down (const gchar * a, const gchar * b) {
    if(!a && !b)
        return 0;
    if(!a)
        return 1;
    if(!b)
        return -1;
    if(strchr (a, G_DIR_SEPARATOR))
        a = strrchr (a, G_DIR_SEPARATOR) + 1;
    if(strchr (b, G_DIR_SEPARATOR))
        b = strrchr (b, G_DIR_SEPARATOR) + 1;
    return strcmp (a, b);
}

static
  gint
entry_compare_up (int caso, record_entry_t * en_a, record_entry_t * en_b) {
    gboolean d1d,
      d2d;                      /* d1, d2 is directory flag */
    if(!en_a && !en_b)
        return 0;
    if(!en_a)
        return 1;
    if(!en_b)
        return -1;
    if(IS_DUMMY_TYPE (en_a->type))
        return 1;
    if(IS_DUMMY_TYPE (en_b->type))
        return -1;

    if(!en_a->path && !en_b->path)
        return 0;
    if(!en_a->path)
        return -1;
    if(!en_b->path)
        return 1;

    d1d = IS_SDIR (en_a->type);
    d2d = IS_SDIR (en_b->type);

    if(d1d && !d2d)
        return 1;
    if(!d1d && d2d)
        return -1;

    /* subsorting... */
    if(caso == TYPE_SORT && IS_FILE (en_a->type) && IS_FILE (en_b->type)) {
        char *a,
         *b;
        a = strrchr (en_a->path, '.');
        b = strrchr (en_b->path, '.');
        if(a || b) {
            if(!a)
                return -1;
            if(!b)
                return 1;
            if(strcmp (a, b))
                return strcmp (a, b);
        }
    }

    switch (caso) {
    case TYPE_SORT:
    case NAME_SORT:
        {
            gchar *a = en_a->path,
                *b = en_b->path;
            if(strchr (a, G_DIR_SEPARATOR))
                a = strrchr (a, G_DIR_SEPARATOR) + 1;
            if(strchr (b, G_DIR_SEPARATOR))
                b = strrchr (b, G_DIR_SEPARATOR) + 1;
            return strcmp (a, b);
        }
    default:
        if(!en_a->st && !en_b->st)
            return 0;
        if(!en_a->st)
            return -1;
        if(!en_b->st)
            return 1;
        switch (caso) {
        case SIZE_SORT:
            return en_a->st->st_size - en_b->st->st_size;
        case DATE_SORT:
            return en_a->st->st_mtime - en_b->st->st_mtime;
        case OWNER_SORT:
            return en_a->st->st_uid - en_b->st->st_uid;
        case GROUP_SORT:
            return en_a->st->st_gid - en_b->st->st_gid;
        case MODE_SORT:
            return en_a->st->st_mode - en_b->st->st_mode;
        }
    }
    return 0;
}

static
  gint
entry_compare_down (int caso, record_entry_t * en_a, record_entry_t * en_b) {
    gboolean d1d,
      d2d;                      /* d1, d2 is directory flag */
    if(!en_a && !en_b)
        return 0;
    if(!en_a)
        return -1;
    if(!en_b)
        return 1;
    if(IS_DUMMY_TYPE (en_a->type))
        return -1;
    if(IS_DUMMY_TYPE (en_b->type))
        return 1;

    if(!en_a->path && !en_b->path)
        return 0;
    if(!en_a->path)
        return 1;
    if(!en_b->path)
        return -1;

    d1d = IS_SDIR (en_a->type);
    d2d = IS_SDIR (en_b->type);

    if(d1d && !d2d)
        return -1;
    if(!d1d && d2d)
        return 1;

    /* subsorting... */
    if(caso == TYPE_SORT && IS_FILE (en_a->type)
       && IS_FILE (en_b->type)) {
        char *a,
         *b;
        a = strrchr (en_a->path, '.');
        b = strrchr (en_b->path, '.');
        if(a || b) {
            if(!a)
                return -1;
            if(!b)
                return 1;
            if(strcmp (a, b))
                return strcmp (a, b);
        }
    }

    switch (caso) {
    case TYPE_SORT:
    case NAME_SORT:
        {
            gchar *a = en_a->path,
                *b = en_b->path;
            if(strchr (a, G_DIR_SEPARATOR))
                a = strrchr (a, G_DIR_SEPARATOR) + 1;
            if(strchr (b, G_DIR_SEPARATOR))
                b = strrchr (b, G_DIR_SEPARATOR) + 1;
            return strcmp (a, b);
        }
    default:
        if(!en_a->st && !en_b->st)
            return 0;
        if(!en_a->st)
            return 1;
        if(!en_b->st)
            return -1;
        switch (caso) {
        case SIZE_SORT:
            return en_a->st->st_size - en_b->st->st_size;
        case DATE_SORT:
            return en_a->st->st_mtime - en_b->st->st_mtime;
        case OWNER_SORT:
            return en_a->st->st_uid - en_b->st->st_uid;
        case GROUP_SORT:
            return en_a->st->st_gid - en_b->st->st_gid;
        case MODE_SORT:
            return en_a->st->st_mode - en_b->st->st_mode;
        }
    }
    return 0;
}

static
    int
xfdir_compare_up (const void *a, const void *b) {
    int result;
    const dir_t *d1 = (const dir_t *)a;
    const dir_t *d2 = (const dir_t *)b;
    if(!d1 && !d2)
        return 0;
    if(!d1)
        return -1;
    if(!d2)
        return 1;

    /* we sort here by displayed name, not by the path's basename,
     * which is what entry compare will do... */

    if(relative_sort_column == NAME_SORT) {
        if(!d1->en || IS_DUMMY_TYPE (d1->en->type))
            return 1;
        if(!d2->en || IS_DUMMY_TYPE (d2->en->type))
            return -1;
        if(d1->en && d2->en) {  /* directories on top */
            gboolean d1d = IS_SDIR (d1->en->type);
            gboolean d2d = IS_SDIR (d2->en->type);
            if(d1d && !d2d)
                return -1;
            if(!d1d && d2d)
                return 1;
        }

        result = pathv_compare_up (d1->pathv, d2->pathv);
        if(result)
            return result;
        /* break name ties with date: */
        return (entry_compare_up (DATE_SORT, d1->en, d2->en));
    }
    result = entry_compare_up (relative_sort_column, d1->en, d2->en);
    if(result)
        return result;
    /* break other ties with name */
    return pathv_compare_up (d1->pathv, d2->pathv);

    /*return (entry_compare(relative_sort_column,d1->en,d2->en)); */
}

static
    int
xfdir_compare_down (const void *a, const void *b) {
    int result;
    const dir_t *d1 = (const dir_t *)a;
    const dir_t *d2 = (const dir_t *)b;
    if(!d1 && !d2)
        return 0;
    if(!d1)
        return -1;
    if(!d2)
        return 1;

    /* we sort here by displayed name, not by the path's basename,
     * which is what entry compare will do... */

    if(relative_sort_column == NAME_SORT) {
        if(!d1->en || IS_DUMMY_TYPE (d1->en->type))
            return -1;
        if(!d2->en || IS_DUMMY_TYPE (d2->en->type))
            return 1;
        if(d1->en && d2->en) {  /* directories on top */
            gboolean d1d = IS_SDIR (d1->en->type);
            gboolean d2d = IS_SDIR (d2->en->type);
            if(d1d && !d2d)
                return -1;
            if(!d1d && d2d)
                return 1;
        }

        result = pathv_compare_down (d1->pathv, d2->pathv);
        if(result)
            return result;
        /* break name ties with date: */
        return (entry_compare_down (DATE_SORT, d1->en, d2->en));
    }
    result = entry_compare_down (relative_sort_column, d1->en, d2->en);
    if(result)
        return result;
    /* break other ties with name */
    return pathv_compare_down (d1->pathv, d2->pathv);

    /*return (entry_compare(relative_sort_column,d1->en,d2->en)); */
}

static
    int
xfdir_compare_default (const void *a, const void *b) {
    int result;
    const dir_t *d1 = (const dir_t *)a;
    const dir_t *d2 = (const dir_t *)b;
    if(!d1 && !d2)
        return 0;
    if(!d1)
        return -1;
    if(!d2)
        return 1;

    {
        /* equivalent to no sort (treeview uses 3 states:
         * no sort is ascending by name and
         * subsorted by filetype).*/
        d2 = (const dir_t *)a;
        d1 = (const dir_t *)b;
    }
    /* we sort here by displayed name, not by the path's basename,
     * which is what entry compare will do... */

    if(relative_sort_column == NAME_SORT) {
        if(!d1->en || IS_DUMMY_TYPE (d1->en->type))
            return 1;
        if(!d2->en || IS_DUMMY_TYPE (d2->en->type))
            return -1;
        if(d1->en && d2->en) {  /* directories on top */
            gboolean d1d = IS_SDIR (d1->en->type);
            gboolean d2d = IS_SDIR (d2->en->type);
            if(d1d && !d2d)
                return 1;
            if(!d1d && d2d)
                return -1;
        }

        result = pathv_compare_up (d1->pathv, d2->pathv);
        if(result)
            return result;
        /* break name ties with date: */
        return (entry_compare_up (DATE_SORT, d1->en, d2->en));
    }
    result = entry_compare_up (relative_sort_column, d1->en, d2->en);
    if(result)
        return result;
    /* break other ties with name */
    return pathv_compare_up (d1->pathv, d2->pathv);

    /*return (entry_compare(relative_sort_column,d1->en,d2->en)); */
}

/**
 * xfdir_sort:
 * @xfdir_p: pointer to an xfdir_t structure
 *
 * Sorts an xfdir_t structure according to the settings in xfdir_p->ascending 
 * and xfdir_p->sort_column. Ascending is TRUE or FALSE and sort_column may be: 
 * 
 * TYPE_SORT, NAME_SORT, SIZE_SORT,
 * DATE_SORT, OWNER_SORT, GROUP_SORT, MODE_SORT.
 * <para>
 * qsort library routine is used as a sorting method.
 * </para>
 * 
 **/
static void
xfdir_sort (xfdir_t * xfdir_p) {
    TRACE ("reload_iconview: do quicksort here: sortcolumn is %d, ascending=%d\n", xfdir_p->sort_column, xfdir_p->ascending);
    TRACE ("xfdir_p=0x%lx,pathc is %ld\n", (unsigned long)xfdir_p, (long)xfdir_p->pathc);

    set_sort_column (xfdir_p->sort_column);

    switch (xfdir_p->ascending) {
    case 0:
        qsort ((void *)xfdir_p->gl, xfdir_p->pathc, sizeof (dir_t),
#ifdef __COMPAR_FN_T
               (__compar_fn_t)
#endif
               xfdir_compare_down);
        break;
    case 1:
        qsort ((void *)xfdir_p->gl, xfdir_p->pathc, sizeof (dir_t),
#ifdef __COMPAR_FN_T
               (__compar_fn_t)
#endif
               xfdir_compare_up);
        break;
    default:
        qsort ((void *)xfdir_p->gl, xfdir_p->pathc, sizeof (dir_t),
#ifdef __COMPAR_FN_T
               (__compar_fn_t)
#endif
               xfdir_compare_default);
        break;
    }
}


// function to do a background reload and set things up for the update
static xfdir_t *
immediate_reload_action (xfdir_t *xfdir_p){
    // allocate memory for reloaded xfdir structure
    xfdir_t *new_xfdir_p = 
	new_xfdir_p = (xfdir_t *) malloc (sizeof (xfdir_t));

    DBG ("monitor new_xfdir_p=0x%lx\n", (long unsigned)new_xfdir_p);

    // initialize new xfdir structure with current xfdir structure information
    memcpy (new_xfdir_p, xfdir_p, sizeof (xfdir_t));

    // not necesary: new_xfdir_p->view_p = xfdir_p->view_p;
    //
    // new xfdir structure specific initializations:
    new_xfdir_p->pathc = 0;
    new_xfdir_p->gl = NULL;
    new_xfdir_p->xfdir_p = NULL;

    // reload directory (or module) into new xfdir structure
    update_xfdir (new_xfdir_p);

    // sort directory (or module) using current user settings
    xfdir_sort (new_xfdir_p);

    // safeguard check: if updated xfdir structure is no longer
    // required, discard updated directory 
    if(exit_monitor (xfdir_p)) {
	DBG ("monitor now exiting and dumping new_xfdir_p!\n");
	// rodent no longer needs this.
	xfdir_free_data (new_xfdir_p);
	g_free (new_xfdir_p);
	DBG ("monitor: id= %d is now defunct (updated xfdir discarded)\n",
		xfdir_p->monitor_id);
	return NULL;
    }
    
    // this mutex is for working on the current xfdir structure
    // without confusing other threads which may require access
    // to the xfdir structure. multiple view_p in a single rodent instance
    // (XXX actually, I wonder if it is really necessary...)
    g_mutex_lock (xfdir_p->view_p->monitor_loop_mutex);
    // clean up xfdir playground
    if(xfdir_p->xfdir_p) {
	TRACE ("monitor freeing xfdir_p->xfdir_p=0x%lx\n", (long unsigned)xfdir_p->xfdir_p);
	xfdir_free_data (xfdir_p->xfdir_p);
	g_free (xfdir_p->xfdir_p);
	xfdir_p->xfdir_p = NULL;
    }

    // save a pointer to current xfdir structure
    new_xfdir_p->xfdir_p = xfdir_p;
    // mutex end.
    g_mutex_unlock (xfdir_p->view_p->monitor_loop_mutex);

    return new_xfdir_p;
}

// thread_monitor_f is in charge of xfdir_p cleanup!
// AND view_p cleanup!
static gpointer
thread_monitor_f (gpointer data) {
    gboolean disabled = FALSE;

    // memory structure should be received from parent thread
    xfdir_t *xfdir_p = (xfdir_t *) data;
    if(xfdir_p == NULL || !xfdir_p->path) {
        return NULL;
    }

    // This is the main monitor loop. It is the event loop
    // which controls the monitor thread (outer loop)
    do { // outer loop (1)                 
        gboolean refresh_pending = FALSE;
        gboolean refresh_action = FALSE;

	// count items present in directory (or module)
        gint current_count;
        if(xfdir_p->view_p->module) {
            current_count = MODULE_COUNT (xfdir_p);
        } else {
	    // simple file count may prove buggy
	     current_count = read_files_local (xfdir_p, FALSE);
	    // maybe we do an inode count, which is more complete
	    // (sum the inodes of all files)
	    //current_count = inode_count();
        }

	// check whether stat condition has changed on directory (or module)
        struct stat current_st;
        if(xfdir_p->view_p->module) {
            if(!MODULE_STAT (NULL, &current_st)) {
                // if there is no module stat function, then enter suspend mode
		// until event loop is broken on a reload condition
                disabled = TRUE;
            }
        } else {
            stat (xfdir_p->path, &current_st);
        }

	// Setup to enter inner loop is complete. For this we have a count of 
	// items which should be present (inode count) and a stat structure 
	// with current information.
        DBG ("monitor %d setup for %s\n", xfdir_p->monitor_id, xfdir_p->path);
        do { // inner loop (2)
	    // check if the inner event loop must be broken
	    // this happens if monitor thread will exit or a reload condition
	    // has been detected
            if(!xfdir_monitor_continue (xfdir_p)) {
                break;
            }
            // check if the event loop must be placed in suspend mode
	    if(disabled || xfdir_monitor_skip (xfdir_p)) {
                rfm_threadwait ();
                continue;
            }
            // check for an immediate reload condition which requires
	    // immediate reload action (gboolean refresh_action)
            if(refresh_action) {
                DBG ("monitor now setting up refresh!\n");
		// obtain new xfdir structure
		xfdir_p = immediate_reload_action(xfdir_p);
		// assign new xfdir structure to view
		xfdir_p->view_p->xfdir_p = xfdir_p;
	        // tell main thread that refresh information is ready
		xfdir_p->view_p->refresh = 1;
		// put current monitor thread in suspend state until 
		// reload has completed.
		refresh_action = FALSE;
		// take a breath
		rfm_threadwait ();
                break;
            }
            // if a refresh is not pending, check for a full reload condition: 
            else if(!refresh_pending && 
		    reload_condition (xfdir_p, &current_count, &current_st)) {
		// reload condition is true, so we queue a refresh
		// refresh is queued so that refresh will not actually
		// occur until reload condition stops occurring. 
		// (this is to avoid a cascade of unnecessary reloads
		// on a rapidly changing directory)
                DBG ("MONITOR: reload_condition is TRUE\n");
                refresh_pending = TRUE;
		// here we used to have a status line condition
		// message, but that will now be annoying to lpterm users (me).
            }
            // check if reload pending and changes have ceased 
            else if(refresh_pending && 
		    !reload_condition (xfdir_p, &current_count, &current_st)) {
                DBG ("MONITOR: reload_condition has ceased. Now scheduling refresh\n");
		// Since reload condition has ceased, we can schedule an 
		// immediate reload action on the next loop iteration
                refresh_action = TRUE;
            } else {
		// reload condition has not occurred, but stat information
		// or thumbnail image may have changed for an existing item
		// without affecting directory stat information. In this case
		// we update stat information for individual xfdir structure items. 
                TRACE ("MONITOR: now checking update_stat_info...\n");
                if (update_stat_info (xfdir_p)){
		    // A TRUE return value indicates that the directory
		    // count does not correspond to the actual files 
		    // which currently exist and the file count is wrong.
		    // In this case we queue a refresh.
		    refresh_pending = TRUE;
		} 
            }
	    // take a breath
            rfm_threadwait ();
        } while (2);   // end inner loop 2
        TRACE ("monitor %d loop broken for %s\n", xfdir_p->monitor_id, xfdir_p->path);
	// inner loop has exited. This means one of two things:
	// 	1. monitor thread has met an exit condition and is no longer necessary
	// 	2. a reload condition has occurred and monitor thread must pause
	// 	   until reload has been accomplished.
	//
        // Here we consider case 1, monitor thread had met an exit condition:
        if(exit_monitor (xfdir_p)) {
            TRACE ("MONITOR: all is cool to release xfdir memory\n");
            TRACE ("monitor: id= %d ending now\n", xfdir_p->monitor_id);
            TRACE ("monitor: will free xfdir_p=0x%lx\n", (long unsigned)xfdir_p);
            DBG ("monitor: id= %d is now defunct\n", xfdir_p->monitor_id);

            // monitor thread will now verify if a view cleanup is
	    // necessary (as when the view has been destroyed)
	    // This will occur if view_p->exit_status == TRUE.
	    // (set in schedule_destroy() at rodent_population.c)
	    // 
            rfm_xfdir_view_cleanup ((gpointer) xfdir_p->view_p);
	    // current xfdir structure (no longer needed) cleanup
            if(xfdir_p->xfdir_p) {
                xfdir_free_data (xfdir_p->xfdir_p);
                g_free (xfdir_p->xfdir_p);
            }
            xfdir_free_data (xfdir_p);
            g_free (xfdir_p);
            xfdir_p = NULL;
	    // monitor thread end here: 
            return NULL;
        }
        // main thread must free view_p->xfdir_p upon condition where
        // xfdir_p->xfdir_p replaces it on upload...
        DBG ("monitor: id= %d continues with xfdir_p->xfdir_p\n", xfdir_p->monitor_id);

    } while(1);
    // monitor event loop ends here (end outer loop)
}

/**
 * xfdir_start_monitor:
 * @view_p: A Rodent view_t pointer
 * @en: An icon entry structure.
 *
 * Starts a monitor thread for the loaded @xfdir_p
 * structure. This is done automatically for local files
 * but modules should define a FIXME function to 
 * test for updates and a FIXME function to reload.
 * 
 **/
static
    void
xfdir_start_monitor (view_t * view_p, xfdir_t * xfdir_p) {

    // (mutex lock might be overkill here)
    g_mutex_lock (view_p->monitor_id_mutex);
    xfdir_p->monitor_id = ++view_p->monitor_id;
    view_p->monitor_enabled=TRUE;
    g_mutex_unlock (view_p->monitor_id_mutex);

    // start the background monitor thread
    xfdir_p->thread = g_thread_create (thread_monitor_f, (gpointer) xfdir_p, FALSE, NULL);
    if (xfdir_p->thread == NULL) {
        g_warning("cannot create monitor thread at primary_xfdir.i");
        view_p->monitor_enabled=FALSE;
    }

}

static long long 
get_tama(xfdir_t *xfdir_p){
    long long tama=0;
    if (!xfdir_p) return tama;
    int j;
    for(j = 0; j < xfdir_p->pathc; j++)
            if(xfdir_p->gl[j].en && xfdir_p->gl[j].en->st) {
                tama += xfdir_p->gl[j].en->st->st_size;
    }
    return tama;
}

// update xfdir structure for directory (or module)
static int
update_xfdir (xfdir_t * xfdir_p) {
    if(xfdir_p->view_p->module) {
        DBG ("obtaining view_p->xfdir_p from module %s\n", xfdir_p->view_p->module);
        if(!MODULE_STAT (NULL, &xfdir_p->st)) {
            DBG ("module_stat not defined for %s\n", xfdir_p->view_p->module);
        }
        xfdir_p->pathc = MODULE_COUNT (xfdir_p);
        xfdir_p->data = MODULE_DATA (xfdir_p);
        xfdir_p->view_p->submodule_name = SUBMODULE_NAME;
        return MODULE_XFDIR (xfdir_p);
    } else {
        xfdir_p->view_p->submodule_name = NULL;
    }
    int checksum;

    xfdir_p->tama = 0;

    stat (xfdir_p->path, &xfdir_p->st);

    DBG ("update_xfdir...cache first\n");
#if 10
    /* cache short circuit (only if not using filter) */
    if(read_xfdir_cache (xfdir_p)) {
        /* calculate checksums for the directory (for later use by monitor) 
         * */
        DBG ("---------cache read: pathc=%zd\n", xfdir_p->pathc);
        if(!xfdir_p->pathc) {
            DBG ("---------cache read: !xfdir_p->pathc\n");
            return 0;
        }
        xfdir_p->tama = get_tama(xfdir_p);
        return 1;
//        return xfdir_p->tama;
    }
#endif
    /* count step */
    xfdir_p->pathc = read_files_local (xfdir_p, FALSE);
    xfdir_p->tama = 0;
    if(xfdir_p->pathc <= 0) {
        g_warning ("read_files_local returned %zd", xfdir_p->pathc);
        return -1;
    }
    DBG ("monitor file count=%zd xfdir_p=0x%lx xfdir_p->xfdir_p=0x%lx\n", xfdir_p->pathc, (long unsigned)xfdir_p,
         (long unsigned)xfdir_p->xfdir_p);
    /* read part */
    if(xfdir_p->pathc) {
        xfdir_p->gl = (dir_t *) malloc (xfdir_p->pathc * sizeof (dir_t));
        if(!xfdir_p->gl) {
            g_warning ("unable to malloc struct dir_t");
            return -1;
        } else {
            memset (xfdir_p->gl, 0, xfdir_p->pathc * sizeof (dir_t));
        }

        checksum = st_sum_dir (&(xfdir_p->st)) + SHOWS_HIDDEN (xfdir_p->type);

        /* read step (read <= count step) */
        if(read_files_local (xfdir_p, TRUE) < 0) {
            g_warning ("read_files_local returned < 0");
            return -1;
        }
        xfdir_p->tama = get_tama(xfdir_p);
        
        // a write lock is placed on cache
        save_xfdir_cache (xfdir_p, checksum);
        TRACE ("monitor: cache saved...\n");

    }

    return 1;
}

