/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define RFM_PRIMARY_C

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <dirent.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "rfm_libs.h"
#include "rfm_modules.h"

// static functions:
#include "primary.i"
#include "primary-winops-x11.i"
#include "primary-ls.i"

gboolean
rfm_write_ok (struct stat *st){
        if (!st) return FALSE;
        if (geteuid() == 0){
                return TRUE; // user is root
        } else
        if (st->st_uid == geteuid() && (st->st_mode & S_IWUSR)){
                return TRUE; // user has write permission
        } else
        if (st->st_gid == getegid() && (st->st_mode & S_IWGRP)){
                return TRUE; // group has write permission
        } else 
        if (st->st_mode & S_IWOTH) {
                return TRUE; // others have write permission
        }
        return FALSE;
}

GtkWidget *
rfm_get_widget_by_name (GtkWidget * parent, const gchar * name) {
    if(!parent)
        g_warning ("rfm_get_widget_by_name: !parent");
    if(!name)
        g_warning ("rfm_get_widget_by_name: !name");
    GtkWidget *widget = GTK_WIDGET (g_object_get_data (G_OBJECT (parent), name));
    if(!widget)
        g_warning ("Cannot find widget associated to \"%s\"", name);
    return widget;

}

void
rfm_set_widget_by_name (GtkWidget * parent, const gchar * name, GtkWidget * widget) {
    if(!parent)
        g_warning ("rfm_set_widget_by_name: !parent");
    g_object_set_data (G_OBJECT (parent), name, widget);
}

void
rfm_lock_fd_read (int fd) {
    //  return;
    // get the file lock, wait if necesary:
    struct flock lock;

    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;             // to eof  

    lock.l_type = F_RDLCK;

    if(fcntl (fd, F_SETLKW, &lock) < 0) {
        fprintf (stderr, "fcntl(F_SETLKW)(%d): %s\n", fd, strerror (errno));
    }
}

void
rfm_lock_fd_write (int fd) {
    //  return;
    // get the file lock, wait if necesary:
    struct flock lock;

    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;             // to eof  

    lock.l_type = F_WRLCK;

    if(fcntl (fd, F_SETLKW, &lock) < 0) {
        fprintf (stderr, "fcntl(F_SETLKW)(%d): %s\n", fd, strerror (errno));
    }
}

void
rfm_unlock_fd (int fd) {
    //return;
    // release the lock:
    struct flock lock;

    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;             // to eof  

    lock.l_type = F_UNLCK;
    if(fcntl (fd, F_SETLKW, &lock) < 0) {
        fprintf (stderr, "fcntl(F_SETLKW): %s\n", strerror (errno));
    }
}

gchar *
rfm_esc_string (const gchar * string) {
    int i,
      j,
      k;
    char *charset = "\\\"\' ()`|<>";

    for(j = 0, i = 0; i < strlen (string); i++) {
        for(k = 0; k < strlen (charset); k++) {
            if(string[i] == charset[k])
                j++;
        }
    }
    gchar *estring = (gchar *) malloc (strlen (string) + j + 1);
    memset (estring, 0, strlen (string) + j + 1);
    for(j = 0, i = 0; i < strlen (string); i++, j++) {
        for(k = 0; k < strlen (charset); k++) {
            if(string[i] == charset[k])
                estring[j++] = '\\';
        }
        estring[j] = string[i];
    }
    DBG ("ESC:estring=%s\n", estring);
    return estring;
}

#define MAX_PATH_LABEL 40
#define MAX_PATH_START_LABEL 18
const gchar *
rfm_chop_excess (gchar * b) {
    // chop excess length...
    int len = strlen (b);
    if(len > MAX_PATH_LABEL) {
        b[MAX_PATH_START_LABEL - 1] = '~';
        strcpy (b + MAX_PATH_START_LABEL, b + (len - (MAX_PATH_LABEL - MAX_PATH_START_LABEL)));
    }
    return b;
}

/*
 * This function converts a time value specified in seconds since 1970-01-01
 * to a string representation. It shows the date and the time if the point
 * if less then six month in the past and only the date otherwise.
 * The exact format must be provided by a translation string. 
 *
 * The function should be thread-save since there are not used static
 * (or even global) variables if the system provided a localtime_r() function.
 *
 * Arguments:     when:    time in seconds since 1970-01-01
 *                string:  the result char-array in which the string is placed
 *                length:  the length of the string-array
 *                
 * Return value:  string on success, NULL on failure
 */
gchar *
rfm_time_to_string (time_t when) {
    time_t now = time (NULL);
#ifdef HAVE_LOCALTIME_R
    struct tm timestruct;
#endif
    struct tm *timestruct_ptr;
    char *formatstring;
    gchar *s = NULL;
    gchar string[64];

#ifdef HAVE_MEMSET
    memset (string, 0, 64);
#else
    string[0] = 0;
#endif

    formatstring = difftime (now, when) > 24 * 60 * 60 * 30 * 6
        /* strftime format for non-recent files (older than 6 months)  */
        ? _("%b %e  %Y")
        /* strftime format for recent files */
        : _("%b %e %H:%M");

#ifdef HAVE_LOCALTIME_R
    timestruct_ptr = &timestruct;
    localtime_r (&when, timestruct_ptr);
#else
    timestruct_ptr = localtime (&when);
#endif

    if(strftime (string, 64, formatstring, localtime (&when)) == 0) {
        return NULL;
    }
    s = rfm_utf_string (string);
    return s;
}

char *
rfm_mode_string (mode_t mode) {
    return mode_string (mode);
}

gchar *
rfm_utf_string (const gchar * t) {
    if(!t) {
        return g_strdup ("FIXME: rfm_utf_string(NULL) not valid!");
    }

    if(g_utf8_validate (t, -1, NULL)) {
        return g_strdup (t);
    }
    /* so we got a non-UTF-8 */
    /* but is it a valid locale string? */
    gchar *actual_tag;
    actual_tag = g_locale_to_utf8 (t, -1, NULL, NULL, NULL);
    if(actual_tag)
        return actual_tag;
    /* So it is not even a valid locale string... 
     * Let us get valid utf-8 caracters then: */
    const gchar *p;
    actual_tag = g_strdup ("");
    for(p = t; *p; p++) {
        // short circuit end of string:
        gchar *r = g_locale_to_utf8 (p, -1, NULL, NULL, NULL);
        if(g_utf8_validate (p, -1, NULL)) {
            gchar *qq = g_strconcat (actual_tag, p, NULL);
            g_free (actual_tag);
            actual_tag = qq;
            break;
        } else if(r) {
            gchar *qq = g_strconcat (actual_tag, r, NULL);
            g_free (r);
            g_free (actual_tag);
            actual_tag = qq;
            break;
        }
        // convert caracter to utf-8 valid.
        gunichar gu = g_utf8_get_char_validated (p, 2);
        if(gu == (gunichar) - 1) {
            //g_warning("gu=%d",(int)gu);
            gu = g_utf8_get_char_validated ("?", -1);
        }
        gchar outbuf[8];
        memset (outbuf, 0, 8);
        gint outbuf_len = g_unichar_to_utf8 (gu, outbuf);
        if(outbuf_len < 0) {
            g_warning ("unichar=%d char =%c outbuf_len=%d", gu, p[0], outbuf_len);
        }
        gchar *qq = g_strconcat (actual_tag, outbuf, NULL);
        g_free (actual_tag);
        actual_tag = qq;
    }
    return actual_tag;
}

gchar *
rfm_sizetag (off_t tama, gint count) {
    gchar *tag = _("bytes");
    gchar *buf = NULL;
    double utama = tama;

    buf = NULL;
    if(utama > 0) {
        if(utama >= (long long)1024 * 1024 * 1024) {
            utama /= ((long long)1024 * 1024 * 1024);
            tag = _("Gigabytes");
        } else if(utama >= 1024 * 1024) {
            utama /= (1024 * 1024);
            tag = _("Megabytes");
        } else if(utama >= 1024) {
            utama /= 1024;
            tag = _("Kilobytes");
        }
        if(count <= 0) {
            /* format for size column of regular files */
            buf = g_strdup_printf ("%.2lf %s", utama, tag);
        } else {
            gchar *plural_text=
                g_strdup_printf (ngettext ("%'u item", "%'u items", 
                            count),count);
            buf = g_strdup_printf ("%s: %.2lf %s.", plural_text,
                    utama, tag);
            g_free(plural_text);
    
        }
    } else {
        if(count <=0) {
            buf = g_strdup_printf (_("The location is empty."));
        } else {
            buf=
                g_strdup_printf (ngettext ("%'u item", "%'u items", count),
                        count);
        }
    }
    return buf;
}

/**
 * rfm_host_name:
 * @xid: Window X id
 *
 * This gets the hostname of the window client. This is different from
 * #g_get_host_name() which gets the name of the window server. This may be 
 * different when the window is displayed on a remote server.
 *
 **/
gchar *
rfm_host_name (Window xid) {
    char *name = NULL;
    unsigned char *property_data;
    unsigned long items,
      remaining;
    int actual_format;
    Atom atomo,
      actual_atom;

    atomo = XInternAtom (GDK_DISPLAY (), "WM_CLIENT_MACHINE", FALSE);
    if(XGetWindowProperty (GDK_DISPLAY (), xid,
                           atomo, 0, 255, FALSE, XA_STRING,
                           &actual_atom, &actual_format, &items, &remaining, &property_data) == Success) {
        TRACE ("property_data=%s", ((property_data) ? property_data : (unsigned char *)"null"));
        if(!property_data)
            name = g_strdup (g_get_host_name ());
        else {
            name = g_strdup ((const gchar *)property_data);
            XFree (property_data);
        }
    } else
        name = g_strdup (g_get_host_name ());
    return name;
}

void
rfm_rm_local_cache (view_t * view_p) {
    if(!view_p->en || !view_p->en->path || !g_file_test (view_p->en->path, G_FILE_TEST_IS_DIR)) {
        return;
    }
    gchar *cache_path = rfm_xfdir_get_local_cache_path (view_p->en->path);
    if(cache_path) {
        unlink (cache_path);
    }
    g_free (cache_path);
}

int
rfm_mkdir (const gchar * dir) {
    if(g_file_test (dir, G_FILE_TEST_EXISTS)) {
        gchar *message;
        if(!g_file_test (dir, G_FILE_TEST_IS_DIR)) {
            message = g_strdup_printf ("%s: %s (ENOTDIR: %s)", dir, strerror (EEXIST), strerror (ENOTDIR));
        } else {
            message = g_strdup_printf ("%s: %s", dir, strerror (EEXIST));
        }
        rfm_confirm (NULL, GTK_MESSAGE_ERROR, message, NULL, NULL);
        g_free (message);
        return 0;
    }
    if(g_mkdir_with_parents (dir, 0700) < 0) {
        //g_warning("cannot create %s: %s", dir, strerror(errno));
    }
    if(g_file_test (dir, G_FILE_TEST_EXISTS)) {
        if(g_file_test (dir, G_FILE_TEST_IS_DIR)) {
            return 0;
        }
    }
    DBG ("!g_file_test(%s, G_FILE_TEST_IS_DIR\n", dir);
    return -1;
}

void
rfm_save_to_go_history (char *p) {
    gchar *f = g_build_filename (g_get_user_data_dir (), GOTO_DBH_FILE, NULL);
    COMBOBOX_save_to_history (f, p);
    g_free (f);
}

void
rfm_init_env (void
    ) {
    int i;
    for(i = 0; environ_v[i].env_var; i++) {
        // copy default values from static memory to heap
        if(environ_v[i].env_string) {
            if(strcmp (environ_v[i].env_var, "SUDO_ASKPASS") == 0) {
                environ_v[i].env_string = g_find_program_in_path (environ_v[i].env_string);
                DBG ("ENV: ***** %s %s\n", environ_v[i].env_var, environ_v[i].env_string);

            } else {
                environ_v[i].env_string = g_strdup (environ_v[i].env_string);
                DBG ("ENV: %s %s\n", environ_v[i].env_var, environ_v[i].env_string);
            }
        }
    }
    return;
}

void
rfm_setenv (const char *name, char *value, gboolean verbose) {
    int which;
    gboolean go_ahead = FALSE;
    for(which = 0; environ_v[which].env_var; which++)
        if(strcmp (name, environ_v[which].env_var) == 0)
            break;
    if(!environ_v[which].env_var)
        return;
    if(!value || !strlen (value)) {
        /*DBG("TRACE: erasing %s\n",name); */
        g_free (environ_v[which].env_string);
        environ_v[which].env_string = NULL;
/* environment handling is a blast:
 * freebsd only clears environment with unsetenv, not putenv
 * */

#ifdef HAVE_SETENV
        unsetenv (name);
        environ_v[which].env_string = NULL;
#else
        environ_v[which].env_string = g_strconcat (name, "=", NULL);
        putenv (environ_v[which].env_string);
#endif
#ifdef DEBUG
        if(verbose) {
            if(strcmp (name, "SMB_USER") == 0) {
                DBG ("Mcs manager changed rfm environment: %s\n", name);
            } else {
                DBG ("Mcs manager changed rfm environment: %s=%s\n", name, ((value) ? value : " "));
            }
        }
#endif
        return;
    }
    if(strcmp (name, "RFM_MAX_PREVIEW_SIZE") == 0) {
        if(is_number (value))
            go_ahead = TRUE;
    } else if(strcmp (name, "TERMCMD") == 0) {
        /*if (getenv(name) && access(getenv(name),X_OK)==0) */
        /*DBG("TRACE:TERMCMD=%s!\n",value); */

        if(value && strlen (value)) {
            gchar *c,
             *t = g_strdup (value);
            t = g_strstrip (t);
            if(strchr (t, ' '))
                t = strtok (t, " ");
            c = g_find_program_in_path (t);
            if(c && access (c, X_OK) == 0) {
                go_ahead = TRUE;
            }
            g_free (c);
            c = NULL;
            g_free (t);
            t = NULL;
        }
    } else
        go_ahead = TRUE;
    if(go_ahead) {
        g_free (environ_v[which].env_string);
        environ_v[which].env_string = NULL;
        if(strcmp (name, "SMB_USER") == 0 && !strchr (value, '%')) {
            environ_v[which].env_string = g_strconcat (name, "=", value, "%", NULL);
        } else {
            environ_v[which].env_string = g_strconcat (name, "=", value, NULL);
        }
        putenv (environ_v[which].env_string);
        /*DBG("TRACE: mapping %s=%s\n",name,value); */
#ifdef DEBUG
        if(verbose) {
            if(strcmp (name, "SMB_USER") == 0) {
                DBG ("changed rfm environment: %s\n", name);
            } else {
                DBG ("changed rfm environment: %s=%s\n", name, ((value) ? value : " "));
            }
        }
#endif
    } else {                    /* not go_ahead */
        g_warning ("failed to change rfm environment: %s", name);
    }
    return;
}

void
rfm_threadwait (void) {
    struct timespec thread_wait = {
        0, 100000000
    };
    nanosleep (&thread_wait, NULL);
}

#ifdef DEBUG
# include <sys/times.h>
static struct tms *last_clock = NULL;

const gchar *
profile (void) {
    struct tms this_clock;
    static gchar *p = NULL;
    g_free (p);
    if(!last_clock) {
        last_clock = (struct tms *)malloc (sizeof (struct tms));
        times (last_clock);
    }
    times (&this_clock);
    p = g_strdup_printf ("\n\tPROFILE*** user=%ld, system=%ld",
                         (long)(this_clock.tms_utime - last_clock->tms_utime),
                         (long)(this_clock.tms_stime - last_clock->tms_stime));
    memcpy (last_clock, &this_clock, sizeof (struct tms));

    return (const gchar *)p;
}
#endif

gboolean
rfm_population_write_lock (view_t * view_p) {
    gboolean unlocked = TRUE;
    DBG ("population_write_lock (0x%lx)...\n", (long unsigned)g_thread_self ());

    if(view_p->population_lock_mutex == NULL) {
        view_p->population_lock_mutex = g_mutex_new ();
    }
    if(view_p->population_semaphore == NULL) {
        view_p->population_semaphore = (sem_t *) malloc (sizeof (sem_t));
        memset (view_p->population_semaphore, 0, sizeof (sem_t));
        sem_init (view_p->population_semaphore, 0, 1);
    }
    while(unlocked) {
        g_mutex_lock (view_p->population_lock_mutex);
        gint lock_count = view_p->population_lock_count;
        g_mutex_unlock (view_p->population_lock_mutex);
        if(lock_count == 0) {
            sem_wait (view_p->population_semaphore);
            unlocked = FALSE;
        }
        if(unlocked) {
            rfm_threadwait ();
        }
    }
    return TRUE;
}

static
  gboolean
population_read_lock (view_t * view_p, gboolean trywait) {
    DBG ("population_read_lock-%d (0x%lx)...\n", trywait, (long unsigned)g_thread_self ());
    if(view_p->population_lock_mutex == NULL) {
        view_p->population_lock_mutex = g_mutex_new ();
    }
    g_mutex_lock (view_p->population_lock_mutex);
    if(view_p->population_semaphore == NULL) {
        view_p->population_semaphore = (sem_t *) malloc (sizeof (sem_t));
        memset (view_p->population_semaphore, 0, sizeof (sem_t));
        TRACE ("population_read_lock (0x%lx)...sem_init\n", (long unsigned)g_thread_self ());
        sem_init (view_p->population_semaphore, 0, 1);
    }
    g_mutex_unlock (view_p->population_lock_mutex);
    if(trywait) {
        if(sem_trywait (view_p->population_semaphore) < 0) {
            DBG ("population_read_lock (0x%lx): sem_trywait failed\n", (long unsigned)g_thread_self ());
            return FALSE;
        }
    } else {
        int sval;
        sem_getvalue (view_p->population_semaphore, &sval);
        TRACE ("population_read_lock (0x%lx) ...sem_getvalue=%d\n", (long unsigned)g_thread_self (), sval);

        sem_wait (view_p->population_semaphore);
    }
    TRACE ("population_read_lock...g_mutex_lock (0x%lx)\n", (long unsigned)g_thread_self ());
    g_mutex_lock (view_p->population_lock_mutex);
    view_p->population_lock_count++;
    g_mutex_unlock (view_p->population_lock_mutex);
    // allow multiple read locks on population:
    DBG ("population_read_lock (0x%lx) = %d\n", (long unsigned)g_thread_self (), view_p->population_lock_count);
    sem_post (view_p->population_semaphore);
    return TRUE;
}

gboolean
rfm_population_read_lock (view_t * view_p) {
    return population_read_lock (view_p, FALSE);
}

gboolean
rfm_population_try_read_lock (view_t * view_p) {
    return population_read_lock (view_p, TRUE);
}

void
rfm_population_write_unlock (view_t * view_p) {
    DBG ("rfm_population_write_unlock (0x%lx)...\n", (long unsigned)g_thread_self ());
    TRACE ("POP_SEM:  sem_post(view_p->population_semaphore)\n");
    DBG ("rfm_population_write_unlock...sem_post\n");
    sem_post (view_p->population_semaphore);
}

void
rfm_population_read_unlock (view_t * view_p) {
    g_mutex_lock (view_p->population_lock_mutex);
    view_p->population_lock_count--;
    DBG ("rfm_population_read_unlock (0x%lx)...%d\n", (long unsigned)g_thread_self (), view_p->population_lock_count);
    g_mutex_unlock (view_p->population_lock_mutex);
}

// This function is to be run in its own thread, without any read locks
// enabled, otherwise a deadlock may occur in main thread.

typedef struct expose_rect_t {
    GdkWindow *window;
    GdkRectangle rect;
} expose_rect_t;

static
  gpointer
rfm_expose_rect_thread_f (gpointer data) {
    expose_rect_t *expose_rect_p = data;
    GDK_THREADS_ENTER ();
    gdk_window_invalidate_rect (expose_rect_p->window, 
	    &(expose_rect_p->rect), TRUE);
    GDK_THREADS_LEAVE ();
    g_free (expose_rect_p);
    return NULL;
}

void
rfm_expose_rect (GdkWindow * window, GdkRectangle * rect) {
    // this should be a separate thread because if the main
    // thread wins the race for the gdk_mutex, and then
    // requests a population write lock, there is no way 
    // this thread will release the read lock.
    // function is in primary.c
    expose_rect_t *expose_rect_p = (expose_rect_t *) malloc (sizeof (expose_rect_t));
    expose_rect_p->window = window;
    memcpy (&(expose_rect_p->rect), rect, sizeof (GdkRectangle));

    DBG ("THREAD: g_thread_create(rfm_expose_rect_thread_f) at primary.c\n");
    g_thread_create (rfm_expose_rect_thread_f, expose_rect_p, FALSE, NULL);
    return;
}

static
gchar *
get_paste_buffer (void
    ) {
    int len;
    DBG ("XFetchBuffer...\n");
    char *b = XFetchBuffer (GDK_DISPLAY (), &len, 0);
    if(!b)
        return NULL;
    gchar *pasteboard = g_strdup (b);
    XFree (b);
    return pasteboard;

}

int
rfm_valid_pasteboard (view_t * view_p) {
    if(!view_p)
        return 0;
    gchar *b = view_p->xbuffer;
    if(!b || !strlen (b))
        return 0;
    const gchar *cut = "#xfvalid_buffer:cut";
    const gchar *copy = "#xfvalid_buffer:copy";
    if(strncmp (b, copy, strlen (copy)) == 0)
        return 1;
    if(strncmp (b, cut, strlen (cut)) == 0)
        return 2;
    return 0;
}

/* returns 0 if not in pasteboard, 1 if in copy pasteboard or 2 if
 * in cut pasteboard */
int
rfm_in_pasteboard (view_t * view_p, record_entry_t * en) {
    if(!en || !en->path)
        return FALSE;
    if(IS_ROOT_TYPE (en->type) && !g_file_test (en->path, G_FILE_TEST_EXISTS))
        return FALSE;
    //if (__XF_MASK & en->subtype) return FALSE;
    int paste_type = rfm_valid_pasteboard (view_p);
    if(!paste_type)
        return FALSE;
    gchar *b = g_strdup (view_p->xbuffer);

    // this is the valid buffer line:
    gchar *search = strtok (b, "\n");
    if(!search) {
        g_free (b);
        return 0;
    }
    // now we check line per line for the path.
    search = strtok (NULL, "\n");
    while(search) {
        if(strcmp (search, en->path) == 0) {
            DBG ("PASTE type =%d %s\n", paste_type, en->path);
            g_free (b);
            return paste_type;
        }
        search = strtok (NULL, "\n");
    }
    g_free (b);
    return 0;
}

gboolean
rfm_update_pasteboard (view_t * view_p) {
    if(!view_p->xbuffer)
        view_p->xbuffer = get_paste_buffer ();
    gchar *current_xbuffer = get_paste_buffer ();
    if(!view_p->xbuffer || strcmp (current_xbuffer, view_p->xbuffer)) {
        DBG ("XBUFFER: xbuffer has changed! %s\n", current_xbuffer);
        g_free (view_p->xbuffer);
        view_p->xbuffer = current_xbuffer;
        return TRUE;
    } else {
        TRACE ("XBUFFER: xbuffer OK!\n");
        g_free (current_xbuffer);
        return FALSE;
    }

}

void
rfm_soft_reload (view_t * view_p) {
    if(!view_p || !view_p->xfdir_p)
        return;
    if(view_p->monitor_loop_mutex == NULL) {
        view_p->monitor_loop_mutex = g_mutex_new ();
    }
    // tell monitor to do a soft reload
    g_mutex_lock (view_p->monitor_loop_mutex);
    //view_p->xfdir_p->invalidate_icons=TRUE;
    view_p->xfdir_p->reload = TRUE;
    g_mutex_unlock (view_p->monitor_loop_mutex);
}
