/*
 * 
 * copywrite 2003-2010 under GNU/GPL which means what it means.
 * Edscott Wilson Garcia 
 *
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <limits.h>

#include <dirent.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <utime.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>


#define MAX_LINE_SIZE (_POSIX_PATH_MAX*3)
#define CHILD_FILE_LENGTH 64

#include "rodent_libs.h"
#include "rodent_cpy.i"
/*****************************************************************************/
/**********************     exported symbols     *****************************/
/*****************************************************************************/
// this exported variable is necesary for scp...
//char *src_host = NULL;

gboolean
rodent_cp (int mode, widgets_t * widgets_p, GList * in_list, const gchar * target_path) {
    view_t *view_p=widgets_p->view_p;
    GList *verified_list = NULL;
    DBG ("CPY: rodent_cp\n");
    verify_list (widgets_p, in_list, &verified_list, target_path);
    if(verified_list == NULL) {
        DBG ("CPY: verified_list == NULL\n");
        return FALSE;
    }
    gchar *sources[MAX_COMMAND_ARGS];
    int i = 0;
// command:
    switch (mode) {
    case (TR_MOVE):
        sources[i++] = "mv";
        break;
    case (TR_LINK):
        sources[i++] = "ln";
        break;
    default:
        sources[i++] = "cp";
        break;
    }
// common options:
    switch (mode) {
    case (TR_COPY):
        if(getenv ("RFM_CP_A") && strlen (getenv ("RFM_CP_A"))) {
            sources[i++] = "-a";
        }
        if(getenv ("RFM_CP_RR") && strlen (getenv ("RFM_CP_RR"))) {
            sources[i++] = "-R";
        }
    case (TR_MOVE):
#ifdef GNU_CP
        if(getenv ("RFM_CP_B") && strlen (getenv ("RFM_CP_B"))) {
            sources[i++] = "-b";
        }
        if(getenv ("RFM_CP_U") && strlen (getenv ("RFM_CP_U"))) {
            sources[i++] = "-u";
        }
#endif
    case (TR_LINK):
        if(getenv ("RFM_CP_VERBOSE") && strlen (getenv ("RFM_CP_VERBOSE"))) {
            sources[i++] = "-v";
        }
        sources[i++] = "-f";
    }
// exclusive options:
    switch (mode) {
    case (TR_COPY):
#ifdef GNU_CP
        if(getenv ("RFM_CP_D") && strlen (getenv ("RFM_CP_D"))) {
            sources[i++] = "-d";
        }
        if(getenv ("RFM_CP_S") && strlen (getenv ("RFM_CP_S"))) {
            sources[i++] = "-s";
        }
#endif
        if(getenv ("RFM_CP_L") && strlen (getenv ("RFM_CP_L"))) {
            sources[i++] = "-l";
        }
        if(getenv ("RFM_CP_LL") && strlen (getenv ("RFM_CP_LL"))) {
            sources[i++] = "-L";
        }
        if(getenv ("RFM_CP_N") && strlen (getenv ("RFM_CP_N"))) {
            sources[i++] = "-n";
        }
        if(getenv ("RFM_CP_PP") && strlen (getenv ("RFM_CP_PP"))) {
            sources[i++] = "-P";
        }
        if(getenv ("RFM_CP_P") && strlen (getenv ("RFM_CP_P"))) {
            sources[i++] = "-p";
        }
        if(getenv ("RFM_CP_X") && strlen (getenv ("RFM_CP_X"))) {
            sources[i++] = "-x";
        }
        if(getenv ("RFM_CP_H") && strlen (getenv ("RFM_CP_HH"))) {
            sources[i++] = "-H";
        }
        break;
    case (TR_MOVE):
        break;
    case (TR_LINK):
        if(getenv ("RFM_LN_S") && strlen (getenv ("RFM_LN_S"))) {
            sources[i++] = "-s";
        }
        break;
    }

    GList *list;
    GList *ln_list=NULL;
    gboolean ok = TRUE;
    if (mode==TR_LINK) {
	gchar *tgtdir = g_strdup (target_path);
	record_entry_t *ln_en=verified_list->data;

	
        gchar *base_dir = g_path_get_dirname (ln_en->path);
        int up = 0;
        while(base_dir && tgtdir && !strstr (base_dir, tgtdir)
              && strchr (tgtdir, '/')) {
            up++;
            *(strrchr (tgtdir, '/')) = 0;
            DBG ("CPY: tgtdir=%s\n", tgtdir);
        }
        g_free (base_dir);



	for(list = g_list_first (verified_list); 
		list && list->data; list = list->next) {
	    record_entry_t *orig_en=list->data;
	    const gchar *source_substring = orig_en->path + strlen (tgtdir) + 1;
	    int newsource_length = up * strlen ("../") + strlen (source_substring) + 1;
	    gchar *newsource = (gchar *)malloc (newsource_length);
	    memset (newsource, 0, newsource_length);
	    int k;
	    for(k = 0; k < up; k++) {
		strcat (newsource, "../");
	    }
	    strcat (newsource, source_substring);
	    sources[i++] = newsource;
	    ln_list=g_list_append(ln_list, newsource);
	    if(i == MAX_COMMAND_ARGS - 3) {
		rfm_diagnostics (widgets_p, "xffm/warning", NULL);
                gchar *max=g_strdup_printf("%d",MAX_COMMAND_ARGS);
                rfm_diagnostics (widgets_p, "xffm_tag/stderr", sources[0], ": ", strerror(E2BIG)," (> ",max,")","\n", NULL);
                g_free(max);
		ok = FALSE;
		break;
	    }
	}
	g_free (tgtdir);
        if(g_file_test (target_path, G_FILE_TEST_IS_DIR)) {
            tgtdir = g_strdup (target_path);
        } else {
            tgtdir = g_path_get_dirname (target_path);
        }
        sources[i++] = tgtdir;
        if(!g_file_test(tgtdir, G_FILE_TEST_IS_DIR)) {
            g_warning ("Cannot chdir to %s\n", tgtdir);
            g_free (tgtdir);
	    ok = FALSE;
        } else {
	    view_t *view_p=widgets_p->view_p;
            g_free (view_p->workdir);
            view_p->workdir = tgtdir;
        }

    } else {
	for(list = g_list_first (verified_list); 
		list && list->data; list = list->next) {
	    record_entry_t *src_en = list->data;
	    sources[i++] = src_en->path;
	    if(i == MAX_COMMAND_ARGS - 3) {
		rfm_diagnostics (widgets_p, "xffm/warning", NULL);
                gchar *max=g_strdup_printf("%d",MAX_COMMAND_ARGS);
                rfm_diagnostics (widgets_p, "xffm_tag/stderr", sources[0], ": ", strerror(E2BIG)," (> ",max,")","\n", NULL);
                g_free(max);
		ok = FALSE;
		break;
	    }
	}
        DBG ("CPY: normal copy mode=%d ?= %d, list length=%d\n", mode, TR_LINK, g_list_length (verified_list));
        sources[i++] = (gchar *) target_path;
    }

    sources[i++] = NULL;

    if (view_p->type != DESKVIEW_TYPE) {
	rfm_show_text(widgets_p);
    }
    // note: with run_argv we do not need to escape paths.
    if(ok) {
        rfm_thread_run_argv (widgets_p, sources, FALSE);
    } else {
        DBG ("PASTE CP: ok == FALSE\n");
    }

    for(list = g_list_first (verified_list); list && list->data; list = list->next) {
        record_entry_t *src_en = list->data;
        rfm_destroy_entry (src_en);
    }
    g_list_free (verified_list);
    for(list = g_list_first (ln_list); list && list->data; list = list->next) {
        g_free(list->data);
    }
    g_list_free (ln_list);

    return TRUE;
}

/**********************************************************************/
