
/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <gdk/gdkx.h>

#include "rodent_libs.h"
#include "rodent_deskview.i"


#define NUM_COLORS 8

void
rodent_select_pen (view_t * view_p, int p) {
    static gchar *desktop_color = NULL;
    static GdkColor *user_bg_color = NULL;
    static int offset = NUM_COLORS;

    static GdkColor pen_colors[] = {
/* light background */
        {0, 65535, 65535, 65535},       /*white */
        {1, 0, 0, 0},           /*black */
        {2, 65535, 0, 0},       /*red */
        {3, 0, 30000, 0},       /*green */
        {4, 0, 0, 65535},       /*blue */
        {5, 32000, 32000, 0},   /*brown */
        {6, 65535, 0, 65535},   /*magenta */
        {7, 48000, 48000, 48000},       /*grey */
/* dark background */
        {1, 0, 0, 0},           /*black */
        {0, 65535, 65535, 65535},       /*white */
        {12, 65535, 20000, 20000},      /*red */
        {13, 20000, 60000, 20000},      /*green */
        {14, 40000, 40000, 65535},      /*blue */
        {15, 32000, 32000, 0},  /*brown */
        {16, 65535, 0, 65535},  /*magenta */
        {17, 48000, 48000, 48000}       /*grey */
    };
    if(p < 0 || p >= NUM_COLORS) {
        g_warning ("rodent_select_pen: pen %d out of range.", p);
        return;
    }

    gchar *bkg_var;
    if(view_p->type == DESKVIEW_TYPE) {
        bkg_var = "RFM_DESKTOP_COLOR";
    } else {
        bkg_var = "RFM_ICONVIEW_COLOR";
    }
    if(!view_p->cmap) {
        int i;
        view_p->cmap = gdk_drawable_get_colormap (view_p->widgets.paper->window);
        if(getenv (bkg_var) && strlen (getenv (bkg_var))) {
            XColor db_def,
              hw_def;
            desktop_color = g_strdup (getenv (bkg_var));
            user_bg_color = NULL;
            if(!XAllocNamedColor (GDK_DISPLAY (), GDK_COLORMAP_XCOLORMAP (view_p->cmap), getenv (bkg_var), &db_def, &hw_def)) {
                g_warning ("Cannot allocate color %s", getenv (bkg_var));
                offset = 0;
            } else {
                user_bg_color = (GdkColor *) malloc (sizeof (GdkColor));
                user_bg_color->red = db_def.red * 0xff - 1;
                user_bg_color->green = db_def.green * 0xff - 1;
                user_bg_color->blue = db_def.blue * 0xff - 1;
                user_bg_color->pixel = db_def.pixel;
            }
        }
        for(i = 0; i < NUM_COLORS * 2; i++) {
            if(!gdk_colormap_alloc_color (view_p->cmap, pen_colors + i, TRUE, TRUE))
                g_error ("couldn't allocate color");
        }

        view_p->penGC = gdk_gc_new (view_p->widgets.paper->window);
        gdk_gc_set_colormap (view_p->penGC, view_p->cmap);
    }

    if(getenv (bkg_var) && strlen (getenv (bkg_var))) {
        if(!desktop_color || strcmp (desktop_color, getenv (bkg_var))) {
            XColor db_def,
              hw_def;
            g_free (desktop_color);
            g_free (user_bg_color);
            user_bg_color = NULL;
            desktop_color = g_strdup (getenv (bkg_var));
            if(!XAllocNamedColor (GDK_DISPLAY (), GDK_COLORMAP_XCOLORMAP (view_p->cmap), getenv (bkg_var), &db_def, &hw_def)) {
                offset = 0;
                g_warning ("Cannot allocate color %s", getenv (bkg_var));
            } else {
                user_bg_color = (GdkColor *) malloc (sizeof (GdkColor));
                user_bg_color->red = db_def.red * 0xff - 1;
                user_bg_color->green = db_def.green * 0xff - 1;
                user_bg_color->blue = db_def.blue * 0xff - 1;
                user_bg_color->pixel = db_def.pixel;
            }
        }

    }
    if(user_bg_color) {
        int sum = (65535 * 3) - user_bg_color->red - (int)user_bg_color->green - (int)user_bg_color->blue;
        if(sum >= 128 * 3)
            offset = 0;
        else
            offset = NUM_COLORS;
        if(65535 - (int)user_bg_color->green > 200)
            offset = 0;
        TRACE ("DESK: offset now is %d (%d,%d,%d)\n", offset,
               65535 - user_bg_color->red, 65535 - user_bg_color->green, 65535 - user_bg_color->blue);
    }

    if(p == BACKGROUND_COLOR && user_bg_color) {
        TRACE ("DESK: selecting BACKGROUND_COLOR %s\n", desktop_color);
        gdk_gc_set_foreground (view_p->penGC, user_bg_color);
    } else
        gdk_gc_set_foreground (view_p->penGC, pen_colors + p + offset);
}


void
rodent_clean_paper (view_t * view_p) {
    if(!view_p->widgets.paper)
        g_error ("view_p->widgets.paper==NULL");
    if(view_p->pixmap) {
        g_object_unref (G_OBJECT (view_p->pixmap));
    }
    view_p->pixmap = gdk_pixmap_new (view_p->widgets.paper->window,
                                             view_p->widgets.paper->allocation.width + 2 * view_p->icon_size,
                                             view_p->widgets.paper->allocation.height + 2 * view_p->icon_size, -1);

    if(!view_p->pixmap) {
        g_warning ("pixmap is null!");
        return;
    }
    rodent_select_pen (view_p, BACKGROUND_COLOR);

    DBG ("CLEAN: %d,%d,%d,%d\n", 0, 0, view_p->widgets.paper->allocation.width, view_p->widgets.paper->allocation.height);

    rodent_clean_rectangle (view_p, 0, 0, view_p->widgets.paper->allocation.width, view_p->widgets.paper->allocation.height);
    /*if (view_p->root_pxm == None) {
       gint root_w, root_h, root_d;     
       gdk_window_get_geometry  (gdk_get_default_root_window(),
       NULL,NULL,&root_w,&root_h,&root_d);      
       view_p->root_pxm = XCreatePixmap(GDK_DISPLAY(), GDK_ROOT_WINDOW(),
       root_w,root_h,root_d);
       XCopyArea(       GDK_DISPLAY(), 
       GDK_PIXMAP_XID(view_p->pixmap), 
       view_p->root_pxm, 
       GDK_GC_XGC(view_p->penGC), 
       0,0,root_w,root_h,0,0);
       TRACE("setting pixmap to 0x%x\n", view_p->root_pxm);
       mb_util_set_root_pixmap(GDK_DISPLAY(),  view_p->root_pxm, True);

       } */

    gtk_widget_queue_draw_area (view_p->widgets.paper, 0, 0,
                                view_p->widgets.paper->allocation.width, view_p->widgets.paper->allocation.height);
}

void
rodent_expose (GtkWidget * widget, GdkEventExpose * event, gpointer data) {
    DBG ("on_expose x,y=%d,%d w,h=%d,%d\n", event->area.x, event->area.y, event->area.width, event->area.height);
    view_t *view_p = (view_t *) data;
    if(!view_p || !view_p->pixmap || !view_p->widgets.paper)
        return;
    if(view_p->redraw_pixmap && view_p->population_pp) {
	DBG("rodent_expose: redraw_pixmap\n");
        redraw_pixmap (view_p, event->area.x, event->area.y, event->area.width, event->area.height);
    }
    XCopyArea (GDK_DISPLAY (),
               GDK_PIXMAP_XID (view_p->pixmap),
               GDK_DRAWABLE_XID (view_p->widgets.paper->window),
               GDK_GC_XGC (view_p->penGC),
               event->area.x, event->area.y, event->area.width, event->area.height, event->area.x, event->area.y);
}

void
rodent_insert_all_population (view_t * view_p) {
    int j;
    for(j = 0; j < view_p->xfdir_p->pathc; j++) {
        DBG ("TRACE: inserting population %s\n", view_p->xfdir_p->gl[j].pathv);
        insert_population_item (view_p, NULL, j);
        view_p->count++;
    }

}

void 
rodent_erase_deskview_geometries (void) {
    gchar *f = g_build_filename (g_get_user_config_dir (), DESKVIEW_GEOMETRY_DBH_FILE, NULL);
    if(g_file_test (f, G_FILE_TEST_EXISTS)) {
	unlink(f);
    }
    g_free (f);
    return;
}

void
rodent_save_deskview_geometry (const gchar * fullpath, int x, int y) {
    DBHashTable *preferences;
    GString *gs;
    gchar *f;
    deskview_geometry_t deskview_geometry;

    if(!fullpath || !strlen (fullpath))
        return;

    f = g_build_filename (g_get_user_config_dir (), DESKVIEW_GEOMETRY_DBH_FILE, NULL);
    if(!g_file_test (f, G_FILE_TEST_EXISTS)) {
        DBG ("Creating preferences file: %s", f);
        preferences = dbh_create (f, 11);
    } else {
        preferences = dbh_open (f);
        if(!preferences) {
            unlink (f);
            preferences = dbh_create (f, 11);
        }
    }
    if(!preferences) {
        g_warning ("cannot open %s", f);
        g_free (f);
        return;
    }
    g_free (f);

    gs = g_string_new (fullpath);
    sprintf ((char *)DBH_KEY (preferences), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);

    deskview_geometry.x = x;
    deskview_geometry.y = y;

    memcpy (DBH_DATA (preferences), &deskview_geometry, sizeof (deskview_geometry_t));
    dbh_set_recordsize (preferences, sizeof (deskview_geometry_t));

    if(!dbh_update (preferences)) {
        g_warning ("!dbh_update(deskview_geometry preferences)");
    }
    dbh_close (preferences);
    TRACE ("saved deskview_geometry preferences with key=%s\n", fullpath);
    return;
}

void
rodent_recalc_population_geometry (view_t * view_p) {
    int root_w, root_h;
    gdk_window_get_geometry (gdk_get_default_root_window(),
			    NULL,NULL,
			    &root_w,&root_h,
			    NULL);
    if (root_w != view_p->root_w || root_h != view_p->root_h) {
	// XXX I haven't tested this resize thingy yet...
	gtk_window_set_default_size ((GtkWindow *)view_p->widgets.window, 	     root_w,root_h);
	view_p->root_w=root_w;
	view_p->root_h=root_h;
    }


    if (view_p->grid_flags == NULL){
	int columns = root_w / CELLWIDTH;
	int rows = root_h / CELLHEIGHT;
	view_p->grid_extent = rows * columns;
        view_p->grid_flags = (char *)malloc(view_p->grid_extent*sizeof(char));
    } 
    memset(view_p->grid_flags, 0, view_p->grid_extent*sizeof(char));

    population_t **population_pp;
    for(population_pp = view_p->population_pp; *population_pp; population_pp++) {
        recalc_population_geometry_item (view_p, *population_pp, TRUE);
    }
    for(population_pp = view_p->population_pp; *population_pp; population_pp++) {
        recalc_population_geometry_item (view_p, *population_pp, FALSE);
    }
}
