
/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

typedef struct deskview_geometry_t {
    int x;
    int y;
} deskview_geometry_t;

static void
rodent_clean_rectangle (view_t * view_p, int x, int y, int w, int h) {
    rodent_select_pen (view_p, BACKGROUND_COLOR);

    /*if (view_p->root_pxm == None) {
       view_p->root_pxm=mb_util_get_root_pixmap(GDK_DISPLAY());
       } */

    TRACE ("DESK: clean_rectangle %d %d\n", w, h);
    gdk_draw_rectangle (view_p->pixmap, view_p->penGC, TRUE, x, y, w, h);

    if(view_p->background != None) {
        TRACE ("copy area %d %d\n", w, h);

        XCopyArea (GDK_DISPLAY (),
                   view_p->background, GDK_PIXMAP_XID (view_p->pixmap), GDK_GC_XGC (view_p->penGC), x, y, w, h, x, y);

    }
    gtk_widget_queue_draw_area (view_p->widgets.paper, x, y, w, h);

}

static void
erase_text (view_t * view_p, population_t * population_p) {
    int lowX,
      lowY;
    lowX = ((population_p->x - 1 < 0) ? 0 : population_p->x - 1);
    lowY = ((population_p->y - 1 < 0) ? 0 : population_p->y - 1);

    rodent_clean_rectangle (view_p, 
	    lowX, lowY + population_p->pixbufH, 
	    CELLWIDTH, 2 * (population_p->logical_rect.height));
}

static void
redraw_text (view_t * view_p, population_t * population_p) {
    population_p->labelX = population_p->x;
    population_p->labelY = population_p->y + population_p->pixbufH;
    erase_text (view_p, population_p);
    // redraw_text_outline
    rodent_select_pen (view_p, BACKGROUND_COLOR);
    {
        int x,
          y,
          lowX,
          lowY,
          i,
          j;
        x = population_p->labelX;
        y = population_p->labelY;
        lowX = ((x - 1 < 0) ? 0 : x - 1);
        lowY = ((y - 1 < 0) ? 0 : y - 1);
        for(i = lowX; i <= x + 1; i++)
            for(j = lowY; j <= y + 1; j++) {
                if(i == x && j == y)
                    continue;
                gdk_draw_layout (view_p->pixmap, view_p->penGC, i, j, population_p->layout);
                if(population_p->layout2) {
                    gdk_draw_layout (view_p->pixmap, view_p->penGC,
                                     i, j + population_p->logical_rect.height, population_p->layout2);
                }
            }
    }

    // redraw_text;
    int text_color = rodent_text_color(population_p);
    rodent_select_pen (view_p, text_color);
    gdk_draw_layout (view_p->pixmap, view_p->penGC, population_p->labelX, population_p->labelY, population_p->layout);
    if(population_p->layout2) {
        gdk_draw_layout (view_p->pixmap,
                         view_p->penGC,
                         population_p->labelX, population_p->labelY + population_p->logical_rect.height, population_p->layout2);
    }
}

static void
insert_layout (view_t * view_p, population_t * population_p) {
    if(!view_p || !population_p) {
        g_warning ("function parameter is null");
        return;
    }
    redraw_text (view_p, population_p);
}

static void
insert_pixbuf (view_t * view_p, population_t * population_p) {

    gdk_draw_pixbuf ((GdkDrawable *) view_p->pixmap, 
	    view_p->penGC, population_p->pixbuf, 
	    0, 0, /* src */
	    population_p->x, population_p->y, 
	    population_p->pixbufW, population_p->pixbufH, 
	    GDK_RGB_DITHER_NONE, 0, 0);
    /*gtk_widget_queue_draw_area (view_p->widgets.paper,
       population_p->x, 
       population_p->y, 
       population_p->w,
       population_p->h); */
    return;
}

static void
redraw_pixmap (view_t * view_p, int x, int y, int w, int h) {
    //rodent_clean_paper (view_p);
    rodent_select_pen (view_p, BACKGROUND_COLOR);
    DBG ("CLEAN: %d,%d,%d,%d\n", x, y, w, h);
    rodent_clean_rectangle (view_p, x, y, w, h);

    population_t **tmp;
    // this draws into pixbuf
    for(tmp = view_p->population_pp; *tmp; tmp++) {
        population_t *population_p = *tmp;
        if(!population_p->en) {
            continue;
        }
	gboolean visible=FALSE;
	DBG("x: %d, %d; exposex: %d, %d\n", 
		population_p->x, population_p->x + CELLWIDTH, x, x+w);
	DBG("x: %d, %d; exposey: %d, %d\n", 
		population_p->y, population_p->y + CELLHEIGHT, y, y+h);
	if (population_p->x >= x && population_p->x + CELLWIDTH <= x+w &&
	    population_p->y >= y && population_p->y + CELLHEIGHT <= y+h) {
	    visible=TRUE;
	}

	/*if (population_p->x < 0) {
	    TRACE ("%s bad: x=%d  y=%d \n", population_p->en->path, population_p->x, population_p->y);
	    visible=FALSE;
	}*/
/*	if(view_p->type == DESKVIEW_TYPE) {
	    gint width, height;
	    gdk_window_get_geometry ((GTK_WIDGET(view_p->widgets.window))->window,
			NULL, NULL, &width, &height, NULL);
	    if (population_p->x > width || population_p->y > height) {
		visible=FALSE;
	    } else {
		DBG ("%s ok: x=%d < %d y=%d < %d\n", population_p->en->path, population_p->x, width, population_p->y, height);
	    }
	    
	}*/

        if (visible) {
        //if (1) {
	    insert_pixbuf (view_p, population_p);
	    insert_layout (view_p, population_p);
	}

    }
    view_p->redraw_pixmap = FALSE;
}


static int
insert_population_item (view_t * view_p, gchar * icon_id, int element) {
    if(!view_p->xfdir_p->gl[element].pathv) {
        g_warning ("!view_p->gl->pathv");
        return 0;
    }

    view_p->population_pp[element] = rodent_create_population_t (view_p,
                                                          view_p->xfdir_p->gl[element].en, element, icon_id,
                                                          view_p->xfdir_p->gl[element].pathv);

    return 1;
}


static
deskview_geometry_t *
get_deskview_geometry_p (const gchar * fullpath) {
    DBHashTable *geometry;
    GString *gs;
    static deskview_geometry_t deskview_geometry;
    gchar *f;

    if(!fullpath || !strlen (fullpath))
        return NULL;

    TRACE ("looking for geometry with key=%s\n", fullpath);

    f = g_build_filename (g_get_user_config_dir (), DESKVIEW_GEOMETRY_DBH_FILE, NULL);
    if(!g_file_test (f, G_FILE_TEST_EXISTS)) {
        TRACE ("no deskview geometry file: %s\n", f);
        g_free (f);
        return NULL;
    }

    geometry = dbh_open (f);
    g_free (f);

    gs = g_string_new (fullpath);
    sprintf ((char *)DBH_KEY (geometry), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);

    if(!dbh_load (geometry)) {
        TRACE ("no deskview_geometry record for %s (%s)\n", fullpath, (char *)DBH_KEY (geometry));
        dbh_close (geometry);
        return NULL;
    }
    memcpy (&deskview_geometry, DBH_DATA (geometry), sizeof (deskview_geometry_t));
    TRACE ("got geometry with key=%s  position=%d,%d\n", fullpath, deskview_geometry.x, deskview_geometry.y);
    dbh_close (geometry);
    return &deskview_geometry;
}

/*************************************************************************/
/*       rodent_deskview.c      ******************************************/
/* these functions are only used by type==DESKVIEW_TYPE ******************/
/*************************************************************************/
static void
grid_mark (view_t * view_p, const gchar * key, int x, int y) {
    x -= X_OFFSET;
    y -= Y_OFFSET;
    int column = x / CELLWIDTH;
    int row = y / CELLHEIGHT;
    int columns_per_row = view_p->root_w / CELLWIDTH;

    DBG("x=%d y=%d, row=%d column=%d, cellwidth=%d cellheight=%d modX=%d modY=%d\n", x,y,row,column,CELLWIDTH,CELLHEIGHT,x%CELLWIDTH,y%CELLHEIGHT);
    int item= columns_per_row * row + column;
    if (view_p->grid_flags[item]==0) {
	DBG ("DESK: grid_mark: %s, item=%d\n", key, item);
	view_p->grid_flags[item] = 1;
    }

    if (x % CELLWIDTH) {
	item= columns_per_row * row + (column+1);
	if (view_p->grid_flags[item]==0) {
	    DBG ("DESK: grid_mark: %s, item=%d\n", key, item);
	    view_p->grid_flags[item] = 1;
	}
    }
    if (y % CELLHEIGHT) {
	item= (columns_per_row * (row+1)) + column;
	if (view_p->grid_flags[item]==0) {
	    DBG ("DESK: grid_mark: %s, item=%d\n", key, item);
	    view_p->grid_flags[item] = 1;
	}
    }
    if (x % CELLWIDTH && y % CELLHEIGHT) {
	item= (columns_per_row * (row+1)) + (column+1);
	if (view_p->grid_flags[item]==0) {
	    DBG ("DESK: grid_mark: %s, item=%d\n", key, item);
	    view_p->grid_flags[item] = 1;
	}
    }
}

static gboolean
grid_occupied (view_t * view_p, int x, int y) {
    x -= X_OFFSET;
    y -= Y_OFFSET;
    int column = x / CELLWIDTH;
    int row = y / CELLHEIGHT;
    int columns_per_row = view_p->root_w / CELLWIDTH;

    int item= columns_per_row * row + column;
    if (view_p->grid_flags[item]) return TRUE;

    if (x % CELLWIDTH) {
	item= columns_per_row * row + (column+1);
	if (view_p->grid_flags[item]) return TRUE;
    }
    if (y % CELLHEIGHT) {
	item= columns_per_row * (row+1) + column;
	if (view_p->grid_flags[item]) return TRUE;
    }
    if (x % CELLWIDTH && y % CELLHEIGHT) {
	item= columns_per_row * (row+1) + (column+1);
	if (view_p->grid_flags[item]) return TRUE;
    }
    return FALSE;
}


static void
recalc_population_geometry_item (view_t * view_p, population_t * population_p, gboolean user_defined) {
    gboolean grid_geometry = TRUE;
    static gint root_w=-1, root_h=-1;
    gchar *key;
    deskview_geometry_t *deskview_geometry_p;
    if (root_w < 0 ||root_h < 0) {
	gdk_window_get_geometry (gdk_get_default_root_window (), 
		NULL, NULL, &root_w, &root_h, NULL);
    }

    view_p->text_height = population_p->logical_rect.height;
    if(!population_p->en)
        key = "ROOT_RFM";
    else
        key = population_p->en->path;
    if((deskview_geometry_p = get_deskview_geometry_p (key)) != NULL) {
        // is the geometry saved by user?
	DBG ("geometry found for %s\n", population_p->en->path);
        if(deskview_geometry_p->x < root_w && deskview_geometry_p->y < root_h) {
	    // is saved geometry within display coordinates?
            if (user_defined) {
		// are we doing initial markup of saved geometries?
		population_p->x = deskview_geometry_p->x;
		population_p->y = deskview_geometry_p->y;
		grid_mark (view_p, key, population_p->x, population_p->y);
		return;
	    } else {
		grid_geometry=FALSE;
	    }
	    
	}
    }
    if(!user_defined && grid_geometry) {
	// here we are doing secondary placement of icons not placed by user
        int x = X_OFFSET;
        int y = Y_OFFSET;
        if(grid_occupied (view_p, x, y)) {
            do {
                TRACE ("DESK: grid location %d,%d is occupied.......\n", x, y);
                y += (CELLHEIGHT);
                /* not too close to the bottom, panel rules here */
                if(y + (CELLHEIGHT * 2) >= view_p->root_h) {
                    y = Y_OFFSET;
                    x += (CELLWIDTH);
                    if(x >= view_p->root_w) {
                        DBG ("Too many items on desktop directory!\n");
                        x = -1;
			break;
                    }
                }
            }
            while(x >= 0 && grid_occupied (view_p, x, y));
        }
        population_p->x = x;
        population_p->y = y;
        DBG ("DESK: grid_geometry is %d,%d\n", x, y);
        if(x >= 0)
            grid_mark (view_p, key, x, y);
    }
}


