/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <pwd.h>

#include <ctype.h>
#include <glob.h>
#include <memory.h>
#include <signal.h>
#include <dirent.h>

#include <X11/Xlib.h>
#ifdef HAVE_LIBSM
# include <X11/SM/SMlib.h>
#endif

#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gmodule.h>

#include "rodent_libs.h"

#include "rodent_entry.i"

/* entry function: */
void
rodent_mk_text_entry (view_t * view_p, population_t * population_p, int caso) {
    GtkWidget *entry;
    GtkWidget *hbox;

    widgets_t *widgets_p=&(view_p->widgets);
    
    GdkRectangle frame_extent;
    GdkRectangle window_extent;

    gint delta_y;
    gint delta_h;
    gint y_offset;
    gdk_window_get_position ((GTK_WIDGET(widgets_p->window))->window, 
    	&window_extent.x, &window_extent.y);
    gdk_window_get_geometry((GTK_WIDGET(widgets_p->window))->window, NULL, NULL, &window_extent.width, &window_extent.height, NULL);
    TRACE("window_geometry:  x=%d,  y=%d, w=%d, h=%d", window_extent.x, window_extent.y, window_extent.width, window_extent.height); 

    gdk_window_get_frame_extents ((GTK_WIDGET(widgets_p->window))->window,&frame_extent);
    TRACE("window_frame_extents: x=%d, y=%d, w=%d, h=%d\n", 
    	frame_extent.x, frame_extent.y, frame_extent.width, frame_extent.height); 


    delta_y = window_extent.y - frame_extent.y;
    delta_h = frame_extent.height - window_extent.height;
    y_offset = delta_y + delta_h;
    TRACE("window: delta_y=%d, delta_h=%d, y_offset=%d\n", 
    	delta_y, delta_h, y_offset); 


    TRACE ("rodent_mk_text_entry...\n");
    if(!population_p || !population_p->en || !population_p->en->path)
        return;

    /* caso=0, rename; caso=1, duplicate; caso=2, symlink */

    entry = gtk_entry_new ();
    hbox = gtk_hbox_new (FALSE, 0);
    /*view_p->widgets.rename = gtk_window_new (GTK_WINDOW_TOPLEVEL); */
    view_p->widgets.rename = gtk_window_new (GTK_WINDOW_POPUP);
    /*gtk_window_set_position (GTK_WINDOW (view_p->widgets.rename), GTK_WIN_POS_MOUSE); */

    gint x_coordinate;
    gint y_coordinate;
    /* change relative coordinates to absolute coordinates and place entry window */
    {
        gint y;
        double sh=0;
        if(GTK_IS_SCROLLED_WINDOW (view_p->widgets.scrolled_window)) {
            sh = gtk_adjustment_get_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
        }
        gtk_window_get_position ((GtkWindow *) view_p->widgets.window, &x_coordinate, &y);
        x_coordinate += ((frame_extent.width - window_extent.width)/2);
        y_coordinate=population_p->labelY + frame_extent.y - sh 
                    + y_offset + CELLHSPACING;
         TRACE("ENTRY: gtk_adjustment_get_value(sh)=%lf, population_p->labelX=%d, population_p->labelY=%d, y=%d population_p->y + population_p->h =%d\n", sh, population_p->labelX, population_p->labelY, y, population_p->y + population_p->h);
        gtk_window_move ((GtkWindow *) view_p->widgets.rename,
                         population_p->labelX + x_coordinate, 
			 y_coordinate);
        TRACE("y=%d\n", y_coordinate );
    }

    gtk_window_set_resizable (GTK_WINDOW (view_p->widgets.rename), FALSE);
    gtk_container_set_border_width (GTK_CONTAINER (view_p->widgets.rename), 0);
    gtk_window_set_modal (GTK_WINDOW (view_p->widgets.rename), FALSE);

    gchar *g = NULL;
    gchar *b = g_path_get_basename (population_p->en->path);
    gchar *path = g_strdup (population_p->en->path);
    g_object_set_data (G_OBJECT (view_p->widgets.rename), "path", path);

    if(caso == 0) {
        g = g_strdup (b);
        rfm_status (&(view_p->widgets), "xffm/warning", _("Rename"), " ", b, NULL);
        g_signal_connect (G_OBJECT (entry), "activate", G_CALLBACK (entry_activate_rename), view_p);
    } else if(caso == 1) {
        gchar *dir = g_path_get_dirname (population_p->en->path);
        g = g_strdup_printf (_("%s (copy)%s"), b, "");
        //g = g_strdup_printf (_("Copy of %s"), b);
        g_free (dir);
        rfm_status (&(view_p->widgets), "xffm/warning", _("Duplicate each selected item"), " ", b, NULL);
        g_signal_connect (G_OBJECT (entry), "activate", G_CALLBACK (entry_activate_duplicate), view_p);
    } else if(caso == 2) {
        g = g_strdup_printf (_("link to %s"), b);
        rfm_status (&(view_p->widgets), "xffm/warning", _("Create a symbolic link for each selected item"), " ", b, NULL);
        g_signal_connect (G_OBJECT (entry), "activate", G_CALLBACK (entry_activate_symlink), view_p);
    }
    gchar *q = rfm_utf_string (g);
    gchar *p=q;
    double extra=0;

    while (p && *p) {  //hack: gtk bug workaround
    gunichar up=g_utf8_get_char_validated (p, -1);
    if (up > 0) {
    if (g_unichar_isupper(up)) extra += 0.3;
    //if (g_unichar_islower(up)) extra -= 0.1;
    if (g_unichar_iswide(up)) extra += 1.0;
    if (g_unichar_ispunct(up)) extra -= 0.2;
    if (g_unichar_iszerowidth(up)) extra -= 1.0;
    }
    p++;
    }
    extra += 0.5;
    int chars = extra + g_utf8_strlen (q, -1);
    if (chars < 7) chars=7;

    //gtk_entry_set_width_chars ((GtkEntry *)entry, g_utf8_strlen (q, -1));
    gtk_entry_set_width_chars ((GtkEntry *)entry, chars);
    gtk_entry_set_text ((GtkEntry *) entry, q);
    g_free (q);
    g_free (b);
    g_free (g);

    gtk_editable_set_editable ((GtkEditable *) entry, TRUE);

    gtk_container_set_border_width (GTK_CONTAINER (hbox), 0);
    gtk_container_add (GTK_CONTAINER (view_p->widgets.rename), hbox);

    gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, FALSE, 0);

    g_signal_connect (G_OBJECT (view_p->widgets.rename), "destroy-event", G_CALLBACK (destroy_dialog), view_p);
    g_signal_connect (G_OBJECT (view_p->widgets.rename), "key_press_event", G_CALLBACK (on_key_press), view_p);
    g_signal_connect (G_OBJECT (view_p->widgets.rename), "delete-event", G_CALLBACK (destroy_dialog), view_p);
    g_signal_connect (G_OBJECT (entry), "enter-notify-event", G_CALLBACK (grab_focus), view_p);
    /* this no longer works here, gtk insists on preselecting all... */
    /*gtk_editable_select_region((GtkEditable *) entry,0,endpos); */

    /*gtk_widget_realize(view_p->widgets.rename);
    gtk_widget_realize(hbox);
    gtk_widget_realize(entry);
    gdk_flush ();*/
    TRACE("entry box width=%d, window width=%d\n",
        view_p->widgets.rename->allocation.width,
        view_p->widgets.window->allocation.width);
    

    DBG ("now showing view_p->widgets.rename \n");
    gtk_widget_show_all (view_p->widgets.rename);
    /* this sucks: gtk_widget_grab_focus (view_p->widgets.rename); */
    XSetInputFocus (GDK_DISPLAY (), GDK_WINDOW_XID (gtk_widget_get_parent_window (entry)), RevertToParent, CurrentTime);

    gtk_window_set_transient_for (GTK_WINDOW (view_p->widgets.rename), GTK_WINDOW (view_p->widgets.window));
    
    GdkWindow *root=gdk_get_default_root_window();
    gdk_window_get_geometry(root, NULL, NULL, &window_extent.width, &window_extent.height, NULL);

    TRACE("***entry box width=%d, window width=%d root width=%d\n",
        view_p->widgets.rename->allocation.width,
        view_p->widgets.window->allocation.width,
        window_extent.width
        );

    //test 1, is the box larger than the root window?
    if (view_p->widgets.rename->allocation.width > window_extent.width) {
        gtk_widget_set_size_request(entry, window_extent.width-6, -1);
        gtk_widget_set_size_request(view_p->widgets.rename, window_extent.width, -1);
        x_coordinate=0;
        // this does not do what I think it should do (show start of truncated text):
        //gtk_entry_set_alignment((GtkEntry *)entry, 0);
    } else { // it fits, try to place it.
        int center = x_coordinate + population_p->labelX;
        // overspill?
        if (center + view_p->widgets.rename->allocation.width > window_extent.width) x_coordinate = window_extent.width - view_p->widgets.rename->allocation.width;
        // center is OK
        else x_coordinate=center;
    }


    gtk_window_move ((GtkWindow *) view_p->widgets.rename,
                x_coordinate, //    population_p->labelX + x, 
		y_coordinate);
    XGrabPointer (GDK_DISPLAY (),
                  GDK_WINDOW_XID (gtk_widget_get_parent_window (entry)),
                  TRUE, 0, GrabModeSync, GrabModeAsync,
                  GDK_WINDOW_XID (gtk_widget_get_parent_window (entry)),
                  None, CurrentTime);
    XUngrabPointer(GDK_DISPLAY(),CurrentTime);

    return;
}

void
rodent_done_with_rename (gpointer data) {
    view_t *view_p = (view_t *) data;
    gchar *path;

    TRACE ("rodent_done_with_rename\n");
    if(!view_p->widgets.rename) {
        TRACE ("!view_p->widgets.rename\n");
        return;
    }
    path = g_object_get_data (G_OBJECT (view_p->widgets.rename), "path");
    g_free (path);

    TRACE ("gtk_widget_destroy(GTK_WIDGET(view_p->widgets.rename));\n");
    gtk_widget_destroy (GTK_WIDGET (view_p->widgets.rename));
    view_p->widgets.rename = NULL;
    return;
}
