/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

static int
entry_rename (widgets_t * widgets_p, const gchar * nfile, const gchar * ofile) {

    struct stat st;
    gboolean overwrite = FALSE;

    TRACE ("at rename\n");
    if(!widgets_p || !nfile || !ofile || !strlen (nfile) || !strlen (ofile))
        return FALSE;

    if(lstat (nfile, &st) >= 0) {
        /* dir over dir. file over dir, file over file, dir over file:
         * keep retrying if cannot unlink... */
        /*if (!query_rm(nfile, ofile,overwrite)){ */
        if(!rodent_query_rm (widgets_p, nfile, ofile, TRUE)) {
            return FALSE;
        }
        overwrite = TRUE;
    }

    GList *in_list = NULL;
    in_list = g_list_append (in_list, (void *)ofile);
    rodent_cp (TR_MOVE, widgets_p, in_list, nfile);
    g_list_free (in_list);

 /*   rfm_diagnostics (widgets_p, NULL, "$mv ", ofile, " ", nfile, "\n", NULL);
    if(rename (ofile, nfile) == -1) {
        return FALSE;
    }*/
    return TRUE;
}

static int
entry_duplicate (widgets_t * widgets_p, const gchar * nfile, const gchar * ofile) {

    struct stat n_st;
    struct stat o_st;
    if(!widgets_p || !nfile || !ofile || !strlen (nfile) || !strlen (ofile))
        return FALSE;

    if(lstat (nfile, &n_st) >= 0) {
        /*DBG ("TRACE:query_rm\n"); */
        gboolean overwrite;
        if(lstat (ofile, &o_st) != 0) {
            g_warning ("lstat(ofile, &o_st) != 0");
            return FALSE;
        }
        /* dir over dir, do not retry if cannot unlink... */
        if(S_ISDIR (o_st.st_mode) && S_ISDIR (n_st.st_mode))
            overwrite = TRUE;
        else
            overwrite = FALSE;
        /* file over dir, file over file, dir over file:
         * keep retrying if cannot unlink... */
        if(!rodent_query_rm (widgets_p, nfile, ofile, overwrite)) {
            return FALSE;
        }

    }
    GList *in_list = NULL;
    in_list = g_list_append (in_list, (void *)ofile);
    rodent_cp (TR_COPY, widgets_p, in_list, nfile);
    g_list_free (in_list);

    return TRUE;
}

static int
entry_symlink (widgets_t * widgets_p, const gchar * nfile, const gchar * ofile) {

    struct stat st;

    if(!widgets_p || !nfile || !ofile || !strlen (nfile) || !strlen (ofile))
        return FALSE;
    if(lstat (nfile, &st) >= 0) {
        if(!rodent_query_rm (widgets_p, nfile, ofile, FALSE))
            return FALSE;
    }

    gchar *command = g_strdup_printf ("ln -s \"%s\" \"%s\"", ofile, nfile);
    rfm_thread_run2argv (widgets_p, command, FALSE);
    g_free (command);

    return TRUE;
}

static void
entry_activate (GtkEntry * entry, view_t * view_p, int caso) {
    gchar *actual_tag;
    gchar *tag;
    gchar *b;
    int result = FALSE;
    gchar *path;
    gtk_widget_hide (GTK_WIDGET (view_p->widgets.rename));

    path = g_object_get_data (G_OBJECT (view_p->widgets.rename), "path");

    TRACE ("entry_activate\n");

    if(!path) {
        return;
    }
    actual_tag = gtk_editable_get_chars ((GtkEditable *) entry, 0, -1);
    tag = g_locale_from_utf8 (actual_tag, -1, NULL, NULL, NULL);
    g_free (actual_tag);
    actual_tag = tag;
    b = g_path_get_basename (path);

    gchar *d = g_path_get_dirname (path);
    gchar *p = g_build_filename (d, actual_tag, NULL);
    g_free (d);
    switch (caso) {
    case 0:
        result = entry_rename (&(view_p->widgets), p, path);
        if(result) {
            TRACE ("renaming %s to %s\n", path, p);
        }
        break;
    case 1:
        result = entry_duplicate (&(view_p->widgets), p, path);
        if(result) {
            TRACE ("duplication %s to %s\n", path, p);
        }
        break;
    case 2:
        result = entry_symlink (&(view_p->widgets), p, path);
        if(result) {
            TRACE ("symlinking %s to %s\n", path, p);
        }
        break;
    }
    g_free (p);

    g_free (b);
    g_free (actual_tag);
    rodent_done_with_rename ((gpointer) view_p);
}

static void
entry_activate_rename (GtkEntry * entry, gpointer data) {
    entry_activate (entry, (view_t *) data, 0);
}

static void
entry_activate_duplicate (GtkEntry * entry, gpointer data) {
    entry_activate (entry, (view_t *) data, 1);
}

static void
entry_activate_symlink (GtkEntry * entry, gpointer data) {
    entry_activate (entry, (view_t *) data, 2);
}

static void
destroy_dialog (GtkWidget * widget, gpointer data) {
    rodent_done_with_rename (data);
}

static gint
on_key_press (GtkWidget * entry, GdkEventKey * event, gpointer data) {
    if(event->keyval == GDK_Escape) {
        rodent_done_with_rename (data);
        //rfm_status (&(view_p->widgets), NULL, _("Omitting"), NULL);
        return TRUE;
    }
    return FALSE;
}

static gboolean
grab_focus (GtkWidget * widget, GdkEventCrossing * event, gpointer data) {
    static GtkWidget *w = NULL;
    XSetInputFocus (GDK_DISPLAY (), GDK_WINDOW_XID (gtk_widget_get_parent_window (widget)), RevertToParent, CurrentTime);
    XUngrabPointer (GDK_DISPLAY (), CurrentTime);
/*    XGrabPointer(GDK_DISPLAY(),
	  GDK_WINDOW_XID(gtk_widget_get_parent_window(widget)),
          TRUE,
          0,
          GrabModeAsync,
          GrabModeAsync,
          GDK_WINDOW_XID(gtk_widget_get_parent_window(widget)),
          None, CurrentTime);*/
    /* this is here so that we can override gtk's selection of everything */
    if(w != widget) {
        int endpos;
        char *g = gtk_editable_get_chars ((GtkEditable *) widget, 0, -1);
        if(strchr (g, '.')) {
            gtk_editable_select_region ((GtkEditable *) widget, 0, 0);
            for(endpos = strlen (g) - 1; g[endpos] >= 0; endpos--) {
                if(g[endpos] == '.')
                    break;
            }
            DBG ("entry box: endpos=%d\n", endpos);
            gtk_editable_select_region ((GtkEditable *) widget, 0, endpos);
        }
        g_free (g);
        w = widget;
    }

    return TRUE;
}
