
/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include <string.h>
#include <stdlib.h>

#include "rodent_libs.h"
#include "rodent_gridview.i"

/*************************************************************************/
/*       rodent_gridview.c      ******************************************/
/* these functions are only used by type==ICONVIEW_TYPE ******************/
/*************************************************************************/

view_t *
rodent_get_current_view ( GtkWidget * window) {
    GList *window_view_list = g_object_get_data (G_OBJECT (window), "window_view_list");
    view_t *view_p = window_view_list->data;
    if (view_p->type == DESKVIEW_TYPE) return view_p;
    // we get the current page ...
    gint npage = gtk_notebook_get_current_page (GTK_NOTEBOOK (view_p->widgets.notebook));
    if (npage < 0) return NULL;
    GtkWidget *child =gtk_notebook_get_nth_page(GTK_NOTEBOOK (view_p->widgets.notebook), npage);
    view_p = g_object_get_data (G_OBJECT (child), "view_p");
    return view_p;
}

int 
rodent_text_color(population_t *population_p){
    int text_color = TEXT_COLOR;
    view_t *view_p=population_p->view_p;
    if(population_p && population_p->en && population_p->en->path && !g_utf8_validate (population_p->en->path, -1, NULL)){
        text_color = INVALID_UTF8_TEXT_COLOR;
    }
    if(population_p == view_p->label_p){
        text_color = SATURATE_LABEL_COLOR;
    }
    if(population_p->selected){
        text_color = SELECT_TEXT_COLOR;
    }
    else if(population_p == view_p->saturated_p){
        text_color = SATURATE_TEXT_COLOR;
    }
    return text_color;
}

void
rodent_init_grid (view_t * view_p, int pathc) {
    int width,
      height;

    view_p->max_elements = pathc;
    TRACE ("init_grid: view_p->max_elements=%d\n", pathc);

    view_p->column_width = view_p->icon_size * 2;       // this is CELLWIDTH !!!

    if(!view_p->widgets.window) {
        g_warning ("init_grid: view_p->widgets.window==NULL\n");
        view_p->grid_columns = MINIMUM_ICON_COLUMNS;
        return;
    }

    height = view_p->widgets.window->allocation.height - (CELLWIDTH * 1.5);
    // minus margin
    width = view_p->widgets.window->allocation.width - (CELLWIDTH / 2);
    // minus margin
    view_p->grid_columns = (double)width / CELLWIDTH + 0.5;
    view_p->grid_rows = (view_p->max_elements) / view_p->grid_columns;
    if(view_p->max_elements % view_p->grid_columns)
        view_p->grid_rows++;

    TRACE ("init_grid: rows=%d columns=%d\n", view_p->grid_rows, view_p->grid_columns);
    view_p->grid_area = view_p->grid_rows * view_p->grid_columns;
    view_p->paperX = CELLWIDTH;
    view_p->paperY = view_p->grid_rows * CELLHEIGHT;
}

/* restore window size   */
void
rodent_restore_view_size (view_t * view_p) {
    /* watchout, this is for paper resize */
    DBG (">> geometry: rodent_restore_view_size\n");
    view_geometry_t *view_geometry_p;
    view_geometry_p = rodent_get_view_geometry_p (view_p);
    if(!view_geometry_p) {
        DBG ("rodent_restore_view_size: no view_geometry_p!\n");
        return;
    }
    DBG ("rodent_restore_view_size: apply_window_size(%d,%d)\n", view_geometry_p->w, view_geometry_p->h);

    apply_window_size (view_p, view_geometry_p->w, view_geometry_p->h);

    gdk_flush ();
    DBG ("PAPER rodent_restore_view_size: paper size=(%d,%d), allocation=(%d,%d)\n", view_p->paperX, view_p->paperY,
         view_p->widgets.paper->allocation.width, view_p->widgets.paper->allocation.height);

    /*DBG ("PAPER rodent_restore_view_size: resize to %d,%d\n", view_geometry_p->w, view_geometry_p->h);
    rodent_set_scroll_position (view_p);*/
}

/**
 * rodent_on_size:
 *
 * figures out the number of columns and rows in the paper
 * from resize event data *
 * **/

gboolean
rodent_on_size (view_t * view_p, int width, int height) {
    DBG (">> rodent_on_size\n");
    int w,
      h;
    GtkAdjustment *adjustment;
    TRACE ("rodent_on_size:: w+h=area... %d x %d = %d\n", width, height, width * height);
    if(!view_p || !view_p->widgets.paper) {
        g_warning ("!view_p || !view_p->widgets.paper");
        return TRUE;
    }
    if(!CELLWIDTH) {
        TRACE ("!CELLWIDTH\n");
        return TRUE;
    } else {
        TRACE ("CELLWIDTH=%d\n", CELLWIDTH);
    }

    w = view_p->widgets.paper->allocation.width;
    h = view_p->widgets.paper->allocation.height;
    TRACE ("paper... w*h=area...%d x %d = %d\n", w, h, w * h);
    TRACE ("window... w*h=area...%d x %d = %d\n",
           view_p->widgets.window->allocation.width,
           view_p->widgets.window->allocation.height,
           view_p->widgets.window->allocation.width * view_p->widgets.window->allocation.height);
    set_rows_columns (view_p, width);

    DBG ("rodent_on_size:rows and columns: %d,%d\n", view_p->grid_rows, view_p->grid_columns);
    if(GTK_IS_SCROLLED_WINDOW (view_p->widgets.scrolled_window)) {
        adjustment = gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window);
        TRACE ("adjustment upper=%lf, adjustment pagesize=%lf\nf", adjustment->upper, adjustment->page_size);
        adjustment->upper = view_p->grid_rows * CELLHEIGHT;
        /* XXX this lies: */
        TRACE ("hscroll is visible=%d\n", GTK_WIDGET_VISIBLE (view_p->widgets.scrolled_window->hscrollbar));
        TRACE ("rodent_on_size:: done\n");
    }
    return TRUE;
}

// this arbitrarily set scroll position
void
rodent_set_scroll (view_t * view_p, double scroll_position) {
    DBG ("SCROLL: scroll_position = %lf\n", scroll_position);
    if(!GTK_IS_SCROLLED_WINDOW (view_p->widgets.scrolled_window))
        return;

    double upper = gtk_adjustment_get_upper (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
    double page = gtk_adjustment_get_page_size (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
    DBG ("SCROLLV: geometryZ: scroll_position = %lf, view_p->grid_rows * CELLHEIGHT= %d, view_p->widgets.paper->allocation.height = %d, upper value=%lf, page size=%lf\n", scroll_position, view_p->grid_rows * CELLHEIGHT, view_p->widgets.paper->allocation.height, upper, page);
    if(scroll_position < upper - page) {
        gtk_adjustment_set_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window), scroll_position);

    } else {
        DBG ("SCROLL: geometryZ3: setting scroll position to 0 \n");
        gtk_adjustment_set_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window), upper - page);
    }
    gtk_adjustment_changed (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
}

// this sets scroll position from saved geometry
void
rodent_set_scroll_position (view_t * view_p) {
    if(!GTK_IS_SCROLLED_WINDOW (view_p->widgets.scrolled_window))
        return;
    int window_area = view_p->widgets.window->allocation.height * view_p->widgets.window->allocation.width;
    int icon_area = view_p->widgets.window->allocation.width * view_p->grid_rows * CELLHEIGHT;

    DBG ("SCROLL: icon_area=%d, window_area=%d\n", icon_area, window_area);
    DBG ("SCROLL: geometry: setting scroll position grid_rows= %d grid_columns=%d\n", view_p->grid_rows, view_p->grid_columns);
    view_geometry_t *view_geometry_p = rodent_get_view_geometry_p (view_p);
    if(icon_area < window_area) {
        DBG ("SCROLL: geometryZ1: setting scroll position to 0 \n");
        gtk_adjustment_set_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window), 0.0);
        gtk_adjustment_changed (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
        return;
    }
    double scroll_position =
        view_geometry_p->scrollX * (gtk_adjustment_get_upper (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window))
                                    -
                                    gtk_adjustment_get_page_size (gtk_scrolled_window_get_vadjustment
                                                                  (view_p->widgets.scrolled_window)));
    DBG ("SCROLL: geometryZ2: setting scroll position to %lf geometry height=%d\n", scroll_position, view_geometry_p->h);
    rodent_set_scroll (view_p, scroll_position);

}

void
rodent_save_view_preferences (view_t * view_p) {
    view_preferences_t iconview_preferences;
    DBHashTable *preferences;
    GString *gs;
    gchar *f;
    const gchar *key;
    if(!view_p)
        return;

    if(!view_p->en || !view_p->en->path)
        key = "ROOT_ICONVIEW";
    else
        key = view_p->en->path;

    f = g_build_filename (g_get_user_config_dir (), GRIDVIEW_PREFERENCES_DBH_FILE, NULL);
    if(!g_file_test (f, G_FILE_TEST_EXISTS)) {
        DBG ("Creating preferences file: %s", f);
        preferences = dbh_create (f, 11);
    } else {
        preferences = dbh_open (f);
        if(!preferences) {
            if(unlink (f) < 0)
                g_warning ("cannot replace %s: %s", f, strerror (errno));
            else
                preferences = dbh_create (f, 11);
        }
    }
    if(!preferences) {
        g_warning ("cannot open file: %s", f);
        g_free (f);
        return;
    }

    g_free (f);

    gs = g_string_new (key);
    sprintf ((char *)DBH_KEY (preferences), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);

    iconview_preferences.preferences = view_p->preferences;
    iconview_preferences.sortcolumn = view_p->sortcolumn;

    memcpy (DBH_DATA (preferences), &iconview_preferences, sizeof (view_preferences_t));
    dbh_set_recordsize (preferences, sizeof (view_preferences_t));

    if(!dbh_update (preferences)) {
        g_warning ("!dbh_update(preferences)");
    }
    dbh_close (preferences);
    TRACE ("saved preferences with key=%s\n", key);
    return;
}

view_preferences_t *
rodent_get_view_preferences (const gchar * path) {
    DBG (">> rodent_get_view_preferences\n");
    DBHashTable *preferences;
    const gchar *key;
    GString *gs;
    static view_preferences_t view_preferences;
    //default is to show images but not hidden files (constants.h)
    view_preferences.preferences = DEFAULT_VIEW_PREFERENCES;
    view_preferences.sortcolumn = DEFAULT_SORT_COLUMN;
    gchar *f = g_build_filename (g_get_user_config_dir (), GRIDVIEW_PREFERENCES_DBH_FILE,
                                 NULL);
    if(!g_file_test (f, G_FILE_TEST_EXISTS)) {
        DBG ("Creating preferences file: %s", f);
        preferences = dbh_create (f, 11);
    } else {
        preferences = dbh_open (f);
        if(!preferences) {
            if(unlink (f) < 0)
                g_warning ("cannot replace %s: %s", f, strerror (errno));
            else
                preferences = dbh_create (f, 11);
        }
    }

    if(!preferences) {
        g_warning ("cannot open file: %s", f);
        g_free (f);
        return NULL;
    }
    g_free (f);

    if(path)
        key = path;
    else
        key = "ROOT_ICONVIEW";
    TRACE ("looking for preferences with key=%s\n", key);

    gs = g_string_new (key);
    sprintf ((char *)DBH_KEY (preferences), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);

    if(!dbh_load (preferences)) {
        TRACE ("no preferences record for %s (%s)\n", key, (char *)DBH_KEY (preferences));
        dbh_close (preferences);

        return &view_preferences;
    }
    memcpy (&view_preferences, DBH_DATA (preferences), sizeof (view_preferences_t));

    TRACE ("got preferences with key=%s  (preferences=0x%x, sort_column%d)\n",
           key, view_preferences.preferences, view_preferences.sortcolumn);
    dbh_close (preferences);
    return &view_preferences;
}

view_geometry_t *
rodent_get_view_geometry_p (view_t * view_p) {
    TRACE (">> rodent_get_view_geometry_p\n");
    DBHashTable *geometry;
    const gchar *key;
    GString *gs;
    static view_geometry_t view_geometry;
    view_geometry_t path_view_geometry;
    gchar *f;

    if(!view_p)
        return NULL;
    /* defaults */
    view_geometry.w = DEFAULT_WIDTH;
    view_geometry.h = DEFAULT_HEIGHT;
    view_geometry.x = view_geometry.y = -1;
    view_geometry.scrollX = 0;

    if(view_p->en && view_p->en->path && strlen (view_p->en->path))
        key = view_p->en->path;
    else
        key = "ROOT_ICONVIEW";
    TRACE ("looking for geometry file %s with key=%s\n", f, key);

    f = g_build_filename (g_get_user_config_dir (), GRIDVIEW_GEOMETRY_DBH_FILE, NULL);
    if(!g_file_test (f, G_FILE_TEST_EXISTS)) {
        DBG ("no geometry file: %s\n", f);
        g_free (f);
        return &view_geometry;
    }

    geometry = dbh_open (f);
    g_free (f);

    /* load root record geometry */
    gs = g_string_new ("ROOT_ICONVIEW");
    sprintf ((char *)DBH_KEY (geometry), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);
    if(!dbh_load (geometry)) {
        TRACE ("no geometry record for %s (%s)", "ROOT_ICONVIEW", (char *)DBH_KEY (geometry));
        dbh_close (geometry);
        return &view_geometry;
    }
    memcpy (&view_geometry, DBH_DATA (geometry), sizeof (view_geometry_t));

    gs = g_string_new (key);
    sprintf ((char *)DBH_KEY (geometry), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);

    if(!dbh_load (geometry)) {
        TRACE ("no geometry record for %s (%s)\n", key, (char *)DBH_KEY (geometry));
    } else {
        memcpy (&path_view_geometry, DBH_DATA (geometry), sizeof (view_geometry_t));
        view_geometry.scrollX = path_view_geometry.scrollX;
#ifdef PER_PATH_GEOMETRY
        /* this is disabled by default */
        view_geometry.x = path_view_geometry.x;
        view_geometry.y = path_view_geometry.y;
        view_geometry.w = path_view_geometry.w;
        view_geometry.h = path_view_geometry.h;
#endif
    }
    //view_geometry.scrollX *= view_p->grid_rows;
    DBG ("(graphics_on_size) got geometry with key=%s  (scroll %lf,x=%d y=%d, w=%d h=%d)\n", key, view_geometry.scrollX,
         view_geometry.x, view_geometry.y, view_geometry.w, view_geometry.h);
    dbh_close (geometry);
    if(view_geometry.w <= 0 || view_geometry.h <= 0) {
        view_geometry.w = DEFAULT_WIDTH;
        view_geometry.h = DEFAULT_HEIGHT;
    }
    return &view_geometry;
}

void
rodent_save_view_geometry_p (view_t * view_p, const char *inkey) {
    DBG (">> rodent_save_view_geometry_p\n");
    view_geometry_t view_geometry;
    DBHashTable *geometry;
    GString *gs;
    gchar *f;
    const gchar *key;
    if(!view_p)
        return;

    if(!view_p->en || !view_p->en->path)
        key = "ROOT_ICONVIEW";
    else if(inkey)
        key = inkey;
    else
        key = view_p->en->path;

    f = g_build_filename (g_get_user_config_dir (), GRIDVIEW_GEOMETRY_DBH_FILE, NULL);
    if(!g_file_test (f, G_FILE_TEST_EXISTS)) {
        DBG ("Creating geometry file: %s", f);
        geometry = dbh_create (f, 11);
    }

    geometry = dbh_open (f);

    if(!geometry) {
        g_warning ("cannot open geometry file %s", f);
        g_free (f);
        return;
    }
    g_free (f);

    gs = g_string_new (key);
    sprintf ((char *)DBH_KEY (geometry), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);

    view_geometry.w = view_p->widgets.window->allocation.width;
    view_geometry.h = view_p->widgets.window->allocation.height;

    if(strcmp (key, "ROOT_ICONVIEW") == 0)
        view_geometry.scrollX = 0;
    else {
        view_geometry.scrollX = gtk_adjustment_get_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
//        view_geometry.scrollX /= view_p->grid_rows;
        view_geometry.scrollX /=
            (gtk_adjustment_get_upper
             (gtk_scrolled_window_get_vadjustment
              (view_p->widgets.scrolled_window)) -
             gtk_adjustment_get_page_size (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window)));
    }

    DBG ("geometry graphics: saving ajustment at %lf (%s)\n", view_geometry.scrollX, key);
    gtk_window_get_position ((GtkWindow *) view_p->widgets.window, &(view_geometry.x), &(view_geometry.y));

    DBG ("(graphics_on_size)  saving geometry %d,%d size %d,%d\n",
         view_geometry.x, view_geometry.y, view_geometry.w, view_geometry.h);

    /* here we save the per path geometry */
    memcpy (DBH_DATA (geometry), &view_geometry, sizeof (view_geometry_t));
    dbh_set_recordsize (geometry, sizeof (view_geometry_t));
    if(!dbh_update (geometry))
        g_warning ("!dbh_update(geometry)");
    DBG ("1.saved geometry with key=%s (%lf,%d,%d)\n", key, view_geometry.scrollX, view_geometry.w, view_geometry.h);

    dbh_close (geometry);
    return;
}

void
rodent_save_local_view_geometry_p (view_t * view_p) {
    rodent_save_view_geometry_p (view_p, NULL);
}

void
rodent_save_root_view_geometry_p (view_t * view_p) {
    rodent_save_view_geometry_p (view_p, "ROOT_ICONVIEW");
}
