
/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <grp.h>
#include <sys/mman.h>
#include <gdk/gdkx.h>
#include <time.h>

#include "rodent_libs.h"

#define CONTROL_MODE (event->state & GDK_CONTROL_MASK)
#define NO_CLICK 0
#define LABEL_CLICK 1
#define ICON_CLICK 2

#include "rodent_mouse.h"

#include "rodent_mouse-magic.i"
#include "rodent_mouse.i"
/*******************************************************************/
/*           GTK   signal callbacks */
/******************************************************************/

gboolean
rodent_signal_on_button_press (GtkWidget * widget, GdkEventButton * event, gpointer data) {
    DBG (">>  rodent_signal_on_button_press\n");
    view_t *view_p = (view_t *) data;
    gboolean label_click = FALSE;

    /* cancel any pending tip : */
    view_p->tip_timer = 0;
    if(view_p->widgets.rename) {
        rodent_done_with_rename ((gpointer) view_p);
        return TRUE;
    }

    if(event->button == 1) {
        view_p->mouseX = event->x;
        view_p->mouseY = event->y;
    } else {
        view_p->mouseX = -1;
        view_p->mouseY = -1;
    }
    TRACE ("on_button_press\n");
    /* ignore size events until button released... */
    view_p->eventtime = event->time;
    if(event->button == 2) {
        g_object_set_data (G_OBJECT (view_p->widgets.paper), "normal_tip", GINT_TO_POINTER (1));
    }
    // get population_p into stack memory (population_v) or return

    rfm_population_read_lock (view_p);

    record_entry_t en_v;
    population_t population_v;

    population_t *population_p = rodent_find_in_population (view_p, event->x, event->y);
    if(!population_p) {
        population_p = rodent_find_in_labels (view_p, event->x, event->y);
        label_click = TRUE;
    }
    if(population_p) {
        memcpy (&population_v, population_p, sizeof (population_t));
        if(population_p->en) {
            memcpy (&en_v, population_p->en, sizeof (record_entry_t));
            population_v.en = &en_v;
        } else {
            population_v.en = NULL;
        }
    }

    DBG ("OPEN: rodent_find_in_population=0x%lx\n", (unsigned long)population_p);
    if(event->button == 3) {
        TRACE ("uh-huh. a popup now.\n");
        if(!population_p) {
            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_pixbuf (view_p);
            view_p->selected_p = NULL;
        }
        view_p->doing_drag_p = NULL;
        view_p->dragstate = FALSE;
        view_p->old_X = view_p->old_Y = view_p->boxX = view_p->boxY = -1;
        hide_tip (view_p);
        // button_popup locks the monitor loop
        // population_p may or may not be valid, but population_v is 
        // no good here...
        rfm_population_read_unlock (view_p);
        if(population_p) {
            button_popup (event, view_p, population_p);
        } else {
            button_popup (event, view_p, NULL);
        }
        return TRUE;
    }
    // this will start the box selection:
    if(!population_p && event->button == 1) {
        view_p->selected_p = NULL;
        TRACE ("nothing found at %f %f\n", event->x, event->y);
        if(!(event->state & GDK_CONTROL_MASK)) {
            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_pixbuf (view_p);
        }
        view_p->boxX = event->x;
        view_p->boxY = event->y;
        rfm_population_read_unlock (view_p);
        return TRUE;
    }

    if(event->button == 2) {
        if((event->type == GDK_2BUTTON_PRESS)) {
            view_p->doing_drag_p = NULL;
            view_p->dragstate = FALSE;
            view_p->old_X = view_p->old_Y = view_p->boxX = view_p->boxY = -1;

            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_pixbuf (view_p);
            rodent_select_pixbuf (view_p, population_p);
            rfm_population_read_unlock (view_p);
            DBG ("unlock for doubleclick\n");
            rodent_double_click (view_p, &population_v, TRUE);
        } else {
            rfm_population_read_unlock (view_p);
        }
        return TRUE;
    }

    view_p->dragstate = FALSE;

    if(event->state & GDK_SHIFT_MASK && !label_click) {
	if (population_p && view_p->population_pp){
	    population_t **pp, **to=NULL, **from=NULL;
	    for (pp=view_p->population_pp; pp && *pp; pp++){
		population_t *p=*pp;
		gboolean from_c=
		    !from && 
		    (g_list_find (view_p->selection_list, p->en) || p==population_p);
		if (from_c) from=pp;
		gboolean to_c=
		    from &&  
		    (g_list_find (view_p->selection_list, p->en) || p==population_p);
		if (to_c && pp!=from) to=pp;
	    }
	    if (*from != population_p && *to  != population_p) to=NULL;
	    if (!to) to=from;
	    rodent_unselect_all_pixbuf (view_p);
	    for (pp=from; pp && *pp; pp++){
		rodent_select_pixbuf (view_p, *pp);
		if (pp==to) break;
	    }
	}
        rfm_population_read_unlock (view_p);
        return TRUE;
    }
    if(!CONTROL_MODE) {
        rodent_unselect_all_pixbuf (view_p);
        rodent_select_pixbuf (view_p, population_p);
    }

    /* set up drag status */
    view_p->doing_drag_p = NULL;
    view_p->dragstate = FALSE;
    view_p->old_X = view_p->old_Y = view_p->boxX = view_p->boxY = -1;
    int x_spacing;
    x_spacing = (CELLWIDTH - population_v.pixbufW) / 2;
    if(x_spacing < 0) {
        x_spacing = 0;
    }
//  view_p->doing_drag_p=population_p;
    DBG ("doing_drag_p=0x%lx \n", (unsigned long)view_p->doing_drag_p);
    if(event->button == 1) {
	setup_drag_state (view_p, event);
    }

    if(SINGLE_CLICK_MODE) {
        if(label_click) {
            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_pixbuf (view_p);
            view_p->single_click_mode_state = LABEL_CLICK;
        } else {
            view_p->single_click_mode_state = ICON_CLICK;
        }
        rfm_population_read_unlock (view_p);
        /* double click status gets processed on button release... */
        return TRUE;
    }

    /* now we process doubleclick, if any */
    view_p->single_click_mode_state = NO_CLICK;

    if(event->type == GDK_2BUTTON_PRESS) {
        if(!label_click) {
	    if (event->state & GDK_SHIFT_MASK) {
		rfm_population_read_unlock (view_p);
                return TRUE;
	    }
            // since we modify population_p we must
            // be sure population_serial is correct
            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_pixbuf (view_p);
            rodent_select_pixbuf (view_p, population_p);
            rfm_population_read_unlock (view_p);
            // even if population_serial is not correct, we can 
            // proceed with population_v...
            rodent_double_click (view_p, &population_v, FALSE);
            return TRUE;
        } else {
            int caso = 0;
            if(event->state & GDK_CONTROL_MASK)
                caso = 1;
            if(event->state & GDK_MOD1_MASK)
                caso = 2;
            if(event->state & GDK_MOD5_MASK)
                caso = 2;
            if(event->state & GDK_SHIFT_MASK && event->state & GDK_CONTROL_MASK)
                caso = 2;
            rodent_unselect_all_pixbuf (view_p);
            rodent_unsaturate_pixbuf (view_p);
            rodent_select_pixbuf (view_p, population_p);

            if(population_v.en && !IS_ROOT_TYPE (population_v.en->type)
               && !IS_DUMMY_TYPE (population_v.en->type)
               && population_v.en->path) {
                //view_p->selected_p = population_p;
                if(g_file_test (population_p->en->path, G_FILE_TEST_EXISTS)) {
                    // here we respecify labelX and labelY to avoid
                    // race with reload after a finished rename 
                    //population_p->labelX = CELLWIDTH * population_p->column;
                    //population_p->labelY =
                    //    CELLHEIGHT * population_p->row + view_p->icon_size;
                    rodent_mk_text_entry (view_p, &population_v, caso);
                } else if(view_p->module_name) {
                    // if it's not a valid path, and loaded by a module
                    // then the module should know what to do here...
                    g_warning ("not a valid path loaded by module... \nmodule should know what to do here!\n");
                }
            }
        }
    }
    DBG ("rodent_signal_on_button_press() done!\n");
    rfm_population_read_unlock (view_p);

    return TRUE;
}

gboolean
rodent_signal_on_button_release (GtkWidget * widget, GdkEventButton * event, gpointer data) {
    DBG ("RECT>> rodent_signal_on_button_release event x=%lf y=%lf\n", event->x, event->y);
    view_t *view_p = (view_t *) data;
    population_t *population_p;
    int rectangle_count;
    /*view_p->mouseX=event->x;
       view_p->mouseY=event->y; */
    TRACE ("button release...\n");
    if(event->x < 0)
        event->x = 0;
    if(event->y < 0)
        event->y = 0;
    DBG ("RECT>> rodent_signal_on_button_release event x=%lf y=%lf\n", event->x, event->y);

    view_p->mouseX = -1;
    view_p->mouseY = -1;
    if(event->button == 2) {
        g_object_set_data (G_OBJECT (view_p->widgets.paper), "normal_tip", NULL);
    }
    if (view_p->doing_drag_p) return TRUE;
    // get population_p into stack memory (population_v) or return
    if(!rfm_population_try_read_lock (view_p)) {
        return TRUE;
    }
    record_entry_t en_v;
    population_t population_v;

    DBG ("population_sem: rodent_signal_on_button_release() obtained...\n");
    population_p = rodent_find_in_population (view_p, (gdouble) event->x, (gdouble) event->y);
    if(!population_p) {
        population_p = rodent_find_in_labels (view_p, event->x, event->y);
    }
    if(population_p) {
        memcpy (&population_v, population_p, sizeof (population_t));
        if(population_p->en) {
            memcpy (&en_v, population_p->en, sizeof (record_entry_t));
            population_v.en = &en_v;
        } else {
            population_v.en = NULL;
        }
    }

    if(population_p && (event->state & GDK_SHIFT_MASK) && ! (event->state & GDK_CONTROL_MASK)) {
            rfm_population_read_unlock (view_p);
            return TRUE;
    }
    if(population_p && CONTROL_MODE) {
        if(!population_p->selected) {
            rodent_select_pixbuf (view_p, population_p);
        } else {
            rodent_unselect_pixbuf (view_p, population_p);
        }
    }


    if(population_p && SINGLE_CLICK_MODE && (event->button == 1)
       && view_p->single_click_mode_state) {
        if(CONTROL_MODE && view_p->single_click_mode_state != LABEL_CLICK) {
            view_p->single_click_mode_state = NO_CLICK;
            view_p->doing_drag_p = NULL;
            view_p->old_X = view_p->old_Y = view_p->boxX = view_p->boxY = -1;
            rfm_update_status_line (view_p);
            rfm_population_read_unlock (view_p);
            return TRUE;
        }

        if(view_p->single_click_mode_state == ICON_CLICK) {
            rodent_unselect_all_pixbuf (view_p);
            rodent_select_pixbuf (view_p, population_p);
            rfm_population_read_unlock (view_p);

            rodent_double_click (view_p, &population_v, FALSE);
            return TRUE;
        } else {
            if(population_v.en && !IS_ROOT_TYPE (population_v.en->type)
               && !IS_DUMMY_TYPE (population_v.en->type)
               && population_v.en->path) {
                int caso = 0;
                if(event->state & GDK_CONTROL_MASK)
                    caso = 1;
                if(event->state & GDK_MOD1_MASK)
                    caso = 2;
                if(event->state & GDK_MOD5_MASK)
                    caso = 2;
                if(event->state & GDK_SHIFT_MASK && event->state & GDK_CONTROL_MASK)
                    caso = 2;
                //view_p->selected_p = population_p;
                if(g_file_test (population_v.en->path, G_FILE_TEST_EXISTS)) {
                    /* here we respecify labelX and labelY to avoid
                     * race with reload after a finished rename */
                    //population_p->labelX = CELLWIDTH * population_p->column;
                    //population_p->labelY =
                    //    CELLHEIGHT * population_p->row + view_p->icon_size;
                    rodent_mk_text_entry (view_p, &population_v, caso);
                } else if(view_p->module_name) {
                    TRACE ("FIXME: ask module\n");
                }
            }
        }
        view_p->single_click_mode_state = NO_CLICK;
    }

    view_p->doing_drag_p = NULL;
    draw_window (view_p, event->x, event->y, FALSE);

    rectangle_count = select_population_area (view_p);

    view_p->old_X = view_p->old_Y = view_p->boxX = view_p->boxY = -1;
/*    if (view_p->selection_count > 1) {
#ifdef HAVE_DCGETTEXT
	gchar *g=g_strdup_printf(dngettext(GETTEXT_PACKAGE,"%d item selected","%d items selected",view_p->selection_count),view_p->selection_count);
#else
	gchar *g=g_strdup_printf("%d item(s) selected",view_p->selection_count);
#endif
	rfm_status(&(view_p->widgets),"xffm/warning",g,NULL);
	g_free(g);
    }*/
    rfm_update_status_line (view_p);
    rfm_population_read_unlock (view_p);

    DBG ("rodent_signal_on_button_release() done!\n");
    return TRUE;
}

gboolean
rodent_signal_on_motion (GtkWidget * widget, GdkEventMotion * event, gpointer data) {

    TRACE (">> rodent_signal_on_motion\n");
    view_t *view_p = (view_t *) data;
    if(event->x < 0) event->x = 0;
    if(event->y < 0)  event->y = 0;
    view_p->current_mouseX = event->x;
    view_p->current_mouseY = event->y;
    //
    // if we are window selecting, don't do individual saturation 
    // (at least not yet)
    // no read lock is necessary...
    if(view_p->boxX != -1 && view_p->boxY != -1) {
        draw_window (view_p, event->x, event->y, TRUE);
        int X,
          Y;
        X = event->x - view_p->boxX;
        Y = event->y - view_p->boxY;
        if(X * X + Y * Y > 2)
            view_p->single_click_mode_state = NO_CLICK;
        /* reselect */
        /* here you could select_population_area() on the fly.
         * the only difficulty is unselecting on the fly
         * as the user changes her mind */
        /*select_population_area(view_p); */
	// nothing else to do here...
        return TRUE;
    }  

    // get semaphore or return:
    if(!rfm_population_try_read_lock (view_p)) {
        return TRUE;
    }
    TRACE ("population_sem: rodent_signal_on_motion() obtained...\n");

    /* turn drag on? */
    if(view_p->current_mouseX >= 0 &&
       (abs (view_p->current_mouseX - view_p->mouseX) >
        view_p->icon_size / 4 || abs (view_p->current_mouseY - view_p->mouseY) > view_p->icon_size / 4)) {
        view_p->doing_drag_p = rodent_find_in_population (view_p, view_p->mouseX, view_p->mouseY);
	if (view_p->doing_drag_p && view_p->doing_drag_p->en) {
	    DBG ("turning drag on... %s\n", view_p->doing_drag_p->en->path);
	}
    }

    population_t *population_p = rodent_find_in_population (view_p, event->x, event->y);
     
    
    TRACE ("on motion\n");

    {
        graphics_saturation_event (view_p, population_p, event->x, event->y);

	/*TRACE("x=%f,y=%f (%f,%f)\n",
           event->x,event->y,view_p->mouseX,view_p->mouseY); */
	if(getenv ("RFM_DISABLE_TIPS") && strlen (getenv ("RFM_DISABLE_TIPS"))) {
	    // tips disabled condition 
	    TRACE ("tips disabled by environment variable\n");
	} 
	else if(population_p){
	    gtk_widget_set_tooltip_text (view_p->widgets.window, "Rodent");
	    activate_tip_t *activate_tip_p=
		(activate_tip_t *)malloc(sizeof(activate_tip_t));
	    activate_tip_p->view_p=view_p;
	    activate_tip_p->population_serial=view_p->population_serial;
	    activate_tip_p->population_p=population_p;
	    g_thread_create(activate_tip_thread, (gpointer) activate_tip_p, FALSE, NULL);
	    gtk_widget_trigger_tooltip_query(view_p->widgets.window);
	}
        population_p = rodent_find_in_labels (view_p, event->x, event->y);
        rodent_label_event (view_p, population_p);


    }
    TRACE ("population_sem: rodent_signal_on_motion() released!\n");
    rfm_population_read_unlock (view_p);
    return TRUE;
}

gboolean
rodent_signal_on_enter (GtkWidget * widget, GdkEventCrossing * event, gpointer data) {
    TRACE (">> rodent_signal_on_enter\n");
    //view_t *view_p = (view_t *)data;   
    return TRUE;
}

gboolean
rodent_signal_on_leave (GtkWidget * widget, GdkEventCrossing * event, gpointer data) {
    DBG ("TIP: rodent_signal_on_leave\n");
    view_t *view_p = (view_t *) data;
    widgets_t *widgets_p=&(view_p->widgets);
    // this does not always work since there are ways to get 
    // around this signal. like exiting the window by cruising
    // on top of a tooltip preview beyond the parent window's 
    // border...
    view_p->tooltip_active=FALSE;
    
#if 10
    if (widgets_p->rename) {
       // get pointer position
       GdkRectangle pointer;
       gdk_window_get_pointer(
          (GTK_WIDGET(widgets_p->window))->window, &pointer.x, &pointer.y, NULL);
       // get window position
       gdk_window_get_position ((GTK_WIDGET(widgets_p->window))->window, 
    	        &pointer.width, &pointer.height);
        pointer.x += pointer.width;
        pointer.y += pointer.height;
        TRACE("pointer: x=%d, y=%d; window: x=%d y=%d\n",
                pointer.x, pointer.y, pointer.width, pointer.height);
        GdkRectangle entry_extent;
        //gdk_window_get_geometry((GTK_WIDGET(widgets_p->window))->window, NULL, NULL, &entry_extent.width, &entry_extent.height, NULL);
        gdk_window_get_position ((GTK_WIDGET(widgets_p->rename))->window, 
    	        &entry_extent.x, &entry_extent.y);
        gdk_window_get_geometry((GTK_WIDGET(widgets_p->rename))->window, 
                NULL, NULL, &entry_extent.width, &entry_extent.height, NULL);
        entry_extent.width += entry_extent.x;
        entry_extent.height += entry_extent.y;
        TRACE("rename: x=%d-%d, y=%d-%d; \n", 
                entry_extent.x, entry_extent.x+entry_extent.width, 
                entry_extent.y, entry_extent.y+entry_extent.height);
        gboolean x_condition=pointer.x >= entry_extent.x && pointer.x < entry_extent.width;
        gboolean y_condition=pointer.y >= entry_extent.y && pointer.y < entry_extent.height;
        if (x_condition && y_condition) {
                DBG("OK");
        } else {
                rodent_done_with_rename(data);
                DBG("zap");
        }
    }
#endif
    // check if pointer is really out of window
    DBG ("population_sem: rodent_signal_on_leave() obtained...\n");
    if(view_p->saturated_p) {
        rodent_unsaturate_pixbuf (view_p);
    }
    rodent_label_event(view_p, NULL);
    DBG ("population_sem: rodent_signal_on_leave() released!\n");
    return TRUE;
}

/**  drag events **********************************************/
void
rodent_signal_drag_data (GtkWidget * widget,
                  GdkDragContext * context,
                  gint x, gint y, 
		  GtkSelectionData * selection_data, 
		  guint info, 
		  guint time, 
		  gpointer data) {
    DBG ("DND>> rodent_signal_drag_data\n");
    population_t *population_p;
    view_t *view_p = (view_t *) data;
    record_entry_t *target_en;
    DBG ("drag_data: drag_data...\n");
    /*view_p = (view_t *)g_object_get_data(G_OBJECT(widget),"view_p"); */
    if(!view_p) {
        DBG ("--DND>>rodent_signal_drag_data !view_p\n");
        gtk_drag_finish (context, FALSE, FALSE, time);
        return;
    }
    rfm_cursor_wait (view_p->widgets.window);
    target_en = view_p->en;
    /*if(abs (x - view_p->mouseX) <= CELLWIDTH / 2 && abs (y - view_p->mouseY)) {
        target_en = NULL;
    }*/

    if(!target_en || !target_en->path) {
        TRACE ("drag_data: !target_en || !target_en->path\n");
        rfm_cursor_reset (view_p->widgets.window);
        DBG ("--DND>>rodent_signal_drag_data !target_en || !target_en->path\n");
        gtk_drag_finish (context, FALSE, FALSE, time);
        return;
    }

    rfm_population_read_lock (view_p);
    //gint population_serial=rodent_get_population_serial(view_p);
    record_entry_t en_v;
    population_t population_v;

    DBG ("population_sem: rodent_signal_drag_data() obtained...\n");
    population_p = rodent_find_in_population (view_p, x, y);
    if(!population_p) {
        population_p = rodent_find_in_labels (view_p, x, y);
    }
    if(population_p) {
        memcpy (&population_v, population_p, sizeof (population_t));
        if(population_p->en) {
            memcpy (&en_v, population_p->en, sizeof (record_entry_t));
            population_v.en = &en_v;
        } else {
            population_v.en = NULL;
        }
    }

    if(population_p && population_v.en && population_v.en->path) {
        if(g_file_test (population_v.en->path, G_FILE_TEST_IS_DIR))
            target_en = population_v.en;
        /*else if (IS_NETDIR(population_p->en->subtype) ||
           IS_XF_NETSHARE(population_p->en->subtype)) {
           target_en = population_p->en;
           } */
    }

    DBG ("DND>>rodent_signal_drag_data: target entry is %s, source is %s\n", target_en->path, view_p->drag_src);
    if(drag_view_p==view_p && view_p->drag_src && strcmp (target_en->path, view_p->drag_src) == 0) {
        if (view_p->type == DESKVIEW_TYPE) {
	  GList *list = NULL;
	  int nitems = rfm_uri_parse_list ((const char *)selection_data->data, &list);
	  if (nitems == 1) {
	    gchar *url = (gchar *)list->data;
	    /*population_t *pop_p=rodent_find_in_population();
	    if (pop_p)*/ {
		rodent_save_deskview_geometry(url, x, y);
		DBG("rodent_save_deskview_geometry %s -> %d,%d, \n",url, x, y);
	    }
	  }
	  rfm_uri_free_list (list);
        } else {
	  DBG("ignoring drop command! drag_src=%s target=%s\n", view_p->drag_src ,target_en->path);
          /*rfm_diagnostics(&(view_p->widgets),"xffm/warning",NULL);
          rfm_diagnostics(&(view_p->widgets),"xffm_tag/stderr"," ",strerror(EADDRINUSE),"\n",NULL);*/
	  DBG("drag status=%d (%d, %d, %d)\n",
		context->actions, GDK_ACTION_MOVE, GDK_ACTION_COPY, GDK_ACTION_LINK);
	}
        rfm_cursor_reset (view_p->widgets.window);
        view_p->drag_src = NULL;
        drag_view_p = NULL;
        gtk_drag_finish (context, FALSE, FALSE, time);
        rfm_population_read_unlock (view_p);
	if (view_p->type == DESKVIEW_TYPE) {
	    record_entry_t *en = rfm_copy_entry (view_p->en);
	    rodent_refresh (&(view_p->widgets), en);
	}
        return;
    }

    TRACE ("drag_data...gui_drag_data\n");
    if(gui_drag_data (&(view_p->widgets), target_en, context, x, y, selection_data, info, time))
        TRACE ("drag_data...reload 0x%lx->0x%lx\n", (unsigned long)view_p, (unsigned long)view_p->en);
    /* monitor job otherwise (should ask if module will monitor): */
    if(view_p->module_name && !rfm_natural ("rfm", view_p->module_name, view_p->en, "module_monitor")) {
        DBG ("FIXME\n");
        //rodent_refresh(view_p,view_p->en);  
    } else {
        /* local fm job */
        /* not necesary any more:
	 * if(view_p->en && g_file_test (target_en->path, G_FILE_TEST_EXISTS)) {
            stat (target_en->path, view_p->en->st);
            SET_CHANGED (target_en->type);
        }*/
    }

    TRACE ("drag_data...all done\n");
    rfm_cursor_reset (view_p->widgets.window);
    rfm_population_read_unlock (view_p);
    DBG ("population_sem: rodent_signal_drag_data() released!\n");
    return;
}
//#define DBG TRACE

void
rodent_signal_drag_leave (GtkWidget * widget, GdkDragContext * drag_context, guint time, gpointer data) {
    DBG ("DND>> rodent_signal_drag_leave\n");

}

void
rodent_signal_drag_delete (GtkWidget * widget, GdkDragContext * context, gpointer data) {
#ifdef DEBUG
    view_t *view_p = (view_t *) data;
    DBG ("DND>> rodent_signal_drag_delete: %s\n", view_p->drag_src);
#endif
}

void
rodent_signal_drag_end (GtkWidget * widget, GdkDragContext * context, gpointer data) {
    view_t *view_p = (view_t *) data;
    DBG ("DND>> rodent_signal_drag_end: %s\n", view_p->drag_src);
    gui_drag_end (widget, context, data);
    drag_view_p = NULL;
    view_p->drag_src = NULL;
    rfm_cursor_wait (view_p->widgets.window);
    rfm_cursor_reset (view_p->widgets.window);
    DBG ("drag_end... alldone\n");
}

gboolean
rodent_signal_drag_motion (GtkWidget * widget, GdkDragContext * dc, gint x, gint y, guint t, gpointer data) {
    TRACE ("DND>> rodent_signal_drag_motion\n");
    gboolean target_ok = FALSE;
    view_t *view_p = (view_t *) data;
    rfm_population_read_lock (view_p);
    //    if(!rfm_population_try_read_lock (view_p)) { return TRUE; }
        

    TRACE ("population_sem: rodent_signal_drag_motion() obtained...\n");

    population_t *population_p = rodent_find_in_population (view_p, x, y);
    record_entry_t en_v;
    population_t population_v;
    if(population_p) {
        memcpy (&population_v, population_p, sizeof (population_t));
        if(population_p->en) {
            memcpy (&en_v, population_p->en, sizeof (record_entry_t));
            population_v.en = &en_v;
        } else {
            population_v.en = NULL;
        }
    }

    if(getenv ("RFM_DRAG_DOES_COPY") && strlen (getenv ("RFM_DRAG_DOES_COPY")))
        view_p->drag_action = GDK_ACTION_COPY;
    else
        view_p->drag_action = GDK_ACTION_MOVE;

    if(dc->actions == GDK_ACTION_MOVE)
        gdk_drag_status (dc, GDK_ACTION_MOVE, t);
    else if(dc->actions == GDK_ACTION_COPY)
        gdk_drag_status (dc, GDK_ACTION_COPY, t);
    else if(dc->actions == GDK_ACTION_LINK)
        gdk_drag_status (dc, GDK_ACTION_LINK, t);
    else if(dc->actions & view_p->drag_action)
        gdk_drag_status (dc, view_p->drag_action, t);
    else
        gdk_drag_status (dc, 0, t);
    TRACE ("on_drag_motion...x=%d, y= %d, population_p=0x%lx\n", x, y, (unsigned long)population_p);
    hide_tip (view_p);

    if(population_p) {
        /* if not valid drop target, return */
        if(view_p->module_name) {
            if(rfm_natural ("rfm", view_p->module_name, population_v.en, "valid_drop_site"))
                target_ok = TRUE;
        } else {                /* local */
	    target_ok =
		population_v.en && 
		population_v.en->path && 
		g_file_test(population_v.en->path, G_FILE_TEST_IS_DIR);    
	}
    }
    if(target_ok || !population_p) {
        graphics_saturation_event (view_p, population_p, -1, -1);
    }

    if(view_p->doing_drag_p) {
        TRACE ("widget ok\n");
    }
    rfm_population_read_unlock (view_p);
    TRACE ("population_sem: rodent_signal_drag_motion() released!\n");
    return (TRUE);
}

void
rodent_signal_drag_data_get (GtkWidget * widget,
                      GdkDragContext * context, GtkSelectionData * selection_data, guint info, guint time, gpointer data) {
    DBG ("DND>> rodent_signal_drag_data_get\n");
    view_t *view_p = (view_t *) data;
    hide_tip (view_p);
    gui_drag_data_get (&(view_p->widgets), view_p->selection_list, context, selection_data, info, time);
    TRACE ("drag_data_get: all done\n");
}

void
rodent_signal_drag_begin (GtkWidget * widget, GdkDragContext * drag_context, gpointer data) {
    view_t *view_p = (view_t *) data;
    drag_view_p = view_p;
    DBG ("DND>> rodent_signal_drag_begin: %s\n", view_p->drag_src);
    TRACE ("-------------------------on_drag_begin...\n");
    view_p->drag_event.context = drag_context;
    hide_tip (view_p);
}

/* end of drag signals ********************************/

void
rodent_create_target_list (view_t * view_p) {
    if(view_p->target_list)
        return;
    view_p->target_list = gtk_target_list_new (target_table, NUM_TARGETS);
/*    this does not seem to be necessary
 *    gtk_drag_source_set ((GtkWidget *) view_p->widgets.paper,
                         GDK_BUTTON1_MASK | GDK_BUTTON2_MASK, target_table,
                         NUM_TARGETS, GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK);*/
    gtk_drag_dest_set ((GtkWidget *) view_p->widgets.paper,
                       GTK_DEST_DEFAULT_DROP, target_table, NUM_TARGETS, GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK);
    return;
}

void
rodent_clear_paper (view_t * view_p) {
    hide_tip (view_p);
    rodent_select_pen (view_p, BACKGROUND_COLOR);
    gdk_draw_rectangle (view_p->widgets.paper->window,
                        view_p->penGC, TRUE,
                        0, 0, view_p->widgets.paper->allocation.width, view_p->widgets.paper->allocation.height);

}

GdkPixbuf *
rodent_make_selected_pixbuf (view_t * view_p, population_t * population_p, gboolean saturated) {
    // make pixmap
    GdkPixbuf *pixbuf;
    GdkPixmap *pixmap = gdk_pixmap_new (view_p->widgets.paper->window,
                                        gdk_pixbuf_get_width (population_p->normal_pixbuf),
                                        gdk_pixbuf_get_height (population_p->normal_pixbuf), -1);

    // fill in rectangle (or get background)
    if(view_p->background != None) {
        TRACE ("copy area %d %d\n",
		gdk_pixbuf_get_width (population_p->normal_pixbuf), 
		gdk_pixbuf_get_height (population_p->normal_pixbuf));

        XCopyArea (GDK_DISPLAY (),
                   view_p->background,
                   GDK_PIXMAP_XID (pixmap),
                   GDK_GC_XGC (view_p->penGC),
                   population_p->x, population_p->y,
                   gdk_pixbuf_get_width (population_p->normal_pixbuf), gdk_pixbuf_get_height (population_p->normal_pixbuf), 0, 0);

    } else {
        rodent_select_pen (view_p, BACKGROUND_COLOR);
        gdk_draw_rectangle (pixmap,
                            view_p->penGC, TRUE, 0, 0,
                            gdk_pixbuf_get_width (population_p->normal_pixbuf),
                            gdk_pixbuf_get_height (population_p->normal_pixbuf));
    }
    if (saturated) {
	rodent_select_pen(view_p, 2);
	int w=gdk_pixbuf_get_width(population_p->normal_pixbuf);
	int h=gdk_pixbuf_get_height(population_p->normal_pixbuf);
	gdk_draw_rectangle (pixmap, view_p->penGC, FALSE, 0, 0, w-1,h-2);
	//rodent_select_pen(view_p, BACKGROUND_COLOR);
    }
     // copy pixmap to pixbuf
    pixbuf = gdk_pixbuf_get_from_drawable (NULL, GDK_DRAWABLE (pixmap),
	    view_p->cmap, 0, 0, 0, 0,
	    gdk_pixbuf_get_width (population_p->normal_pixbuf),
            gdk_pixbuf_get_height (population_p->normal_pixbuf));

    // get bg color or bg pixmap... if pixbuf, do a pixbuf compose...
    // create a pixbuf of the size with appropriate bg color
    gdk_pixbuf_composite (population_p->base_pixbuf, pixbuf, 0, 0,       // destination x,y
                          gdk_pixbuf_get_width (population_p->normal_pixbuf), 
			  gdk_pixbuf_get_height (population_p->normal_pixbuf), 
			  0, 0,        // offset x,y
                          1.0, 1.0,     //scale x,y
                          GDK_INTERP_BILINEAR, 125);    // opacity
    g_object_ref (pixbuf);
    // destroy pixmap
    g_object_unref (pixmap);
    return pixbuf;
}

void
rodent_select_pixbuf (view_t * view_p, population_t * population_p) {
    DBG (">> rodent_select_pixbuf\n");
    if(!population_p) {
        TRACE ("!population_p\n");
        return;
    }
    if(!population_p->en)
        return;

    if(IS_DUMMY_TYPE (population_p->en->type))
        return;
    if(view_p->module_name) {
        /* ask the module whether the element is selectable 
         * (by default they will not be)*/
        if(!rfm_natural ("rfm", view_p->module_name, population_p->en, "is_selectable")) {
            return;
        }
    }
    if(IS_ROOT_TYPE (population_p->en->type) && population_p->en->module) {
        /*if (!population_p->en->module) return; */
        if(!rfm_natural ("rfm", population_p->en->module, population_p->en, "get_dnd_path"))
            return;
    }

    if(!population_p->selected_pixbuf && population_p->base_pixbuf) {
        population_p->selected_pixbuf = rodent_make_selected_pixbuf (view_p, population_p, FALSE);
    }
    population_p->pixbuf = population_p->selected_pixbuf;
    population_p->selected = TRUE;
    if(!g_list_find (view_p->selection_list, population_p->en))
        view_p->selection_list = g_list_append (view_p->selection_list, population_p->en);

    view_p->selected_p = population_p;
    rodent_expose_item (view_p, population_p);
    //XXX hack:
    if(view_p->type != DESKVIEW_TYPE) {
        rodent_expose_all (view_p);
    }
}

void
rodent_unselect_all_pixbuf (view_t * view_p) {
    DBG (">> rodent_unselect_all_pixbuf\n");
    population_t **tmp;
    for(tmp = view_p->population_pp; tmp && *tmp; tmp++) {
        population_t *population_p = *tmp;
        if(!population_p)
            continue;
        if(population_p == view_p->doing_drag_p)
            continue;
        //XXX FIXME: test for population_p->selected is broken
        //if (population_p->selected) {
        rodent_unselect_pixbuf (view_p, population_p);
        //}
    }
    if(view_p->selection_list)
        g_list_free (view_p->selection_list);
    view_p->selection_list = NULL;
}

void
rodent_saturate_pixbuf (view_t * view_p, population_t * population_p) {
    DBG (">> rodent_saturate_pixbuf\n");
    if(!population_p) {
        g_warning ("!population_p");
        return;
    }
    if(population_p->normal_pixbuf) {
	view_p->dnd_pixbuf=population_p->normal_pixbuf;
    }

    // why? if(population_p->selected) return;

    if(!population_p->saturated_pixbuf && population_p->base_pixbuf) {
        population_p->saturated_pixbuf =
	    rodent_make_selected_pixbuf (view_p, population_p, TRUE);
    }
    population_p->pixbuf = population_p->saturated_pixbuf;
    view_p->saturated_p = population_p;
    TRACE ("saturate row, column=(%d,%d)\n", population_p->row, population_p->column);
    rodent_expose_item (view_p, population_p);
    
}

void
rodent_unsaturate_pixbuf (view_t * view_p) {
    DBG (">> rodent_unsaturate_pixbuf\n");
    if(view_p->tip_timer) {

    }
    if(!view_p->saturated_p) {
        TRACE ("!view_p->saturated_p\n");
        return;
    }
    hide_tip (view_p);
    population_t *population_p = view_p->saturated_p;
    if(view_p->saturated_p->en && g_list_find (view_p->selection_list, view_p->saturated_p->en)) {
	view_p->saturated_p->pixbuf = view_p->saturated_p->selected_pixbuf;
        view_p->saturated_p = NULL;
        TRACE ("population is selected...\n");
    } else {
	view_p->saturated_p->pixbuf = view_p->saturated_p->base_pixbuf;
    }
    view_p->saturated_p = NULL;
    rodent_expose_item (view_p, population_p);
}

void
rodent_label_event (view_t * view_p, population_t * population_p) {
    TRACE (">> rodent_label_event\n");
    if(!view_p)
        return;
    if(!population_p && view_p->label_p) {
        TRACE ("unsaturating label\n");
	hide_tip (view_p);
        rodent_expose_item (view_p, view_p->label_p);
        view_p->label_p = NULL;
        return;
    }
    if(!population_p || population_p->selected)
        return;

    if(population_p && !view_p->label_p && population_p->en && !IS_ROOT_TYPE (population_p->en->type)
       && !IS_DUMMY_TYPE (population_p->en->type)) {
        if(view_p->module_name) {
            TRACE ("FIXME: ask module\n");
        } else if(g_file_test (population_p->en->path, G_FILE_TEST_EXISTS)) {
            DBG ("saturating label\n");//fflush(NULL);
            view_p->label_p = population_p;
            gtk_widget_set_tooltip_text (view_p->widgets.window, "Rodent");
	    normal_tip(population_p, FALSE);
            
	    rodent_expose_item (view_p, population_p);
            TRACE("tip=%s\n",population_p->en->path);
        }
        //hide_tip (view_p);
    }
}

///////////////////////
void
rodent_double_click (view_t * view_p, population_t * population_p, gboolean control) {
    view_t *view_target = view_p;
    // semaphore need not be set, for this should use a stack location
    // for a copy of selected population_p with copy of entry record.
    widgets_t *widgets_p = &(view_p->widgets);
    if(!view_p || !population_p)
        return;
    if(view_p->status) {
        switch (view_p->status) {
        case STATUS_CLOSING:
            /* ignore doubleclick if closing in progress */
            return;
        case STATUS_IDLE:
        default:
            break;
        }
    }
    hide_tip (view_p);

    //rfm_cursor_wait(view_p->widgets.window);

    if(population_p->en) {
        TRACE ("submodule_name=%s\n", population_p->en->module);
        if(IS_NOACCESS_TYPE (population_p->en->type)) {
            rfm_show_text(widgets_p);
            rfm_diagnostics (widgets_p, "xffm/error", NULL);
            rfm_diagnostics (widgets_p, "xffm_tag/stderr", population_p->en->path, ": ", strerror (EACCES), "\n", NULL);

            return;
        }
    }

    /* does the module have something else in mind for the
       double click on the particular entry ? */
    if(view_p->module_name && population_p->en) {
        DBG ("MOUNT: looking for alternate module double click here\n");
        if(rfm_rational ("rfm", view_p->module_name, population_p->en, &(view_p->widgets), "double_click")) {
            rfm_cursor_reset (view_p->widgets.window);
            return;
        }
        DBG ("MOUNT: no alternate found, going for default...\n");
    }

    if(population_p->en && !population_p->en->module && !g_file_test (population_p->en->path, G_FILE_TEST_IS_DIR)) {
        record_entry_t *en = population_p->en;
        DBG ("OPEN: \n");
        if(g_file_test (en->path, G_FILE_TEST_EXISTS)) {
            DBG ("OPEN: G_FILE_TEST_EXISTS %s\n", en->path);
            record_entry_t *en = population_p->en;
            rfm_cursor_reset (view_p->widgets.window);
            if(!en->mimetype)
                en->mimetype = MIME_magic (en->path);
            gchar *prg = NULL;
            if(en->mimetype)
                prg = MIME_command (en->mimetype);
            if(IS_EXE_TYPE (en->type)) {     /* run it (if not registered) */
                if(prg)
                    rodent_open_with (&(view_p->widgets), en);
                else {
		    g_free(view_p->workdir);
		    view_p->workdir=g_strdup(view_p->en->path);

                    /*  assume in_term to be safe */
                    rfm_thread_run2argv (widgets_p, en->path, TRUE);
                    //rfm_thread_run2argv (widgets_p, en->path, FALSE);
                }
            } else {            /* open with */
                DBG ("OPEN: calling rodent_open_with()\n");
                rodent_open_with (&(view_p->widgets), en);
            }
            g_free (prg);
            return;
        }

        g_warning ("nothing programed here with double click\n");
        rfm_cursor_reset (view_p->widgets.window);
        return;
    } else {
        record_entry_t *en = rfm_copy_entry (population_p->en);
        const gchar *submodule_name = NULL;
        if(population_p->en)
            submodule_name = population_p->en->module;

        TRACE ("RELOADING HERE... module_name=%s\n", view_target->module_name);

        if(en && en->path && g_file_test (en->path, G_FILE_TEST_IS_DIR)) {
            rfm_save_to_go_history (en->path);
        }

        if(control && (view_p->constructor || view_p->popup_constructor)) {
            DBG ("create_iconview \n");
            if(view_p->constructor) {
                DBG ("create iconview with %s\n", (en) ? en->path : "NULL entry");
                view_target = (*view_p->constructor) (en);
                view_target->preferences = view_p->preferences;
                view_target->sortcolumn = view_p->sortcolumn;
                view_target->module_name = submodule_name;
                rodent_refresh (&(view_target->widgets), en);

            } else if(en && en->path && view_p->popup_constructor) {
                (*view_p->popup_constructor) (widgets_p, en->path);
            }
            rfm_cursor_reset (view_target->widgets.window);
            return;
        }

        if(en && view_p->child_constructor) {
            DBG ("DOUBLE_CLICK: en && view_p->child_constructor\n");
            (*view_p->child_constructor) (widgets_p, population_p->en->path);
            rfm_cursor_reset (view_target->widgets.window);
            return;
        }

        rodent_push_view_go_history (view_p);

        record_entry_t *out_en = rfm_copy_entry (en);
        if(out_en) {
            DBG ("RELOADING: default goto action: %s !\n", out_en->path);
        }
        if(rodent_refresh (&(view_target->widgets), out_en)) {
            if(out_en && out_en->path && g_file_test (out_en->path, G_FILE_TEST_IS_DIR))
                rfm_save_to_go_history (out_en->path);
            rfm_cursor_reset (view_target->widgets.window);
        }
        return;
    }
    DBG ("population_sem: released!\n");
    return;
}


// this function takes normal population_p (with a read-lock if shared)
// and returns the GdkPixbuf (with an additional reference)
// if it is already loaded and referenced
// by population_p. This is called by rodent_preview
// to shortcircuit preview generation...
static GdkPixbuf *
rodent_preview_if_loaded(population_t *population_p){
    GdkPixbuf *pixbuf;
    // check if preview is already loaded
    if(population_p->preview_pixbuf) {
        // check if loaded preview is up to date: 
        if((population_p->preview_pixbuf_timestamp >= population_p->en->st->st_mtime)) {
            //fast exit here
            DBG ("SHOW_TIPx: preview is already loaded\n");
	    pixbuf=population_p->preview_pixbuf;
	    // pixbuf has ref==2 
	    g_object_ref(pixbuf);
            return pixbuf;
        }
    }
}

// this function should take a private population_p (thread-safe)
// and returns a newly reffed pixbuf

GdkPixbuf *
rodent_preview (population_t * population_p) {
    if(!population_p->en || !population_p->en->st) {
        DBG ("SHOW_TIPx: !population_p->en || !population_p->en->st\n");
        return NULL;
    }
    if(population_p->en->st->st_size == 0) {
        return NULL;
    }

    /*if(population_p->preview_busy) {
        DBG ("SHOW_TIPx: !population_p->preview_pixbuf && population_p->preview_busy=%d\n", population_p->preview_busy);
        return NULL;
    }*/
    if(!thumbnail_mutex) {
        thumbnail_mutex = g_mutex_new ();
    }

    // So it is not already loaded (rodent_preview_if_loaded()
    // should be called beforehand.
    // Is the pixbuf in the thumbnail cache?
    gchar *thumbnail = rfm_get_thumbnail_path (population_p->en->path, PREVIEW_IMAGE_SIZE);
    GdkPixbuf *pixbuf = rodent_preview_if_loaded(population_p);
    if (pixbuf) {
	return pixbuf;
    }


    if(g_file_test (thumbnail, G_FILE_TEST_EXISTS)) {
        struct stat st;
        if(stat (thumbnail, &st) < 0)
            g_error ("stat(%s): %s", thumbnail, strerror (errno));
        if(st.st_mtime >= population_p->en->st->st_mtime) {
            DBG ("SHOW_TIPx: %s: thumbnail is up to date\n", population_p->en->path);
            // pixbuf is reffed in routine:
            pixbuf = load_preview_pixbuf_from_disk (population_p->en->path, thumbnail);
            if(pixbuf) {
		// pixbuf has ref==1 
                DBG ("SHOW_TIPx: preview loaded from thumbnail file\n");
                g_free (thumbnail);
                return pixbuf;
            }
        }
    }

    // So it is not in thumbnail cache. So it will have to be regenerated
    // First we shall try internal gtk image manipulation:
    if(rfm_is_image (population_p->en)) {
        pixbuf = rfm_get_pixbuf (population_p->en->path, PREVIEW_IMAGE_SIZE);
		// pixbuf has ref==1 but should not be in the pixbuf hash if
		// PREVIEW_IMAGE_SIZE  >= MAX_PIXBUF_HASH_SIZE 
		// (400 > 128 last time I looked)
        DBG ("SHOW_TIPx: preview created by gtk\n");
        g_free (thumbnail);
        if(pixbuf) {
            // test is pixbuf is in pixbuf hash, if it is, then
	    // we should have an additional ref to that the pixbuf
	    // referenced in the hash remains valid when this
	    // pixbuf is unreffed
            if (PREVIEW_IMAGE_SIZE  <= MAX_PIXBUF_HASH_SIZE){
		g_object_ref (pixbuf);
	    }
            return pixbuf;
        } else {
            return NULL;
        }
    }

    // So it is not an image type
#ifdef THIS_IS_LINUX
    // XXX this use of shm is not FreeBSD compatible.
    //
    // Is it an OpenDocument file? get the embeded preview
    // Is it a zip file in disguise?
    if (!population_p->en->filetype) 
        population_p->en->filetype = MIME_file(population_p->en->path);
    gboolean OpenDocument = (strstr (population_p->en->filetype, "OpenDocument") != NULL);
    if(OpenDocument) {
        DBG (">>>>> mime file (%s) = %s %s\n", 
		population_p->en->path, population_p->en->filetype, 
		(strstr (population_p->en->filetype, "OpenDocument")) ? "YES" : "NO");
        int errorv;
        void *ptr = NULL;
        struct zip *Z = zip_open (population_p->en->path, 0, &errorv);
        struct zip_stat sb;
        if(zip_stat (Z, "Thumbnails/thumbnail.png", 0, &sb) == 0) {
            gchar *shm_thumbnail = mouse_shm_open (population_p->en->path, sb.size, &ptr);
            DBG ("shm_thumbnail=%s\n", shm_thumbnail);
            if(shm_thumbnail) {
                struct zip_file *ZF = zip_fopen (Z, "Thumbnails/thumbnail.png", 0);
                if(ZF) {
                    zip_fread (ZF, ptr, sb.size);
                    zip_fclose (ZF);
                    gchar *fs_path = g_strdup_printf ("/dev/shm%s", shm_thumbnail);
                    pixbuf = gdk_pixbuf_new_from_file (fs_path, NULL);
		    pixbuf=fix_pixbuf_scale(pixbuf);	    
		    // pixbuf: ref==1
                    if(GDK_IS_PIXBUF(pixbuf)) {
                        g_mutex_lock (thumbnail_mutex);
                        {
                            gdk_pixbuf_save (pixbuf, thumbnail,
				    "png", NULL, "tEXt::Software", "Rodent",
                                             NULL);
                        }
                        g_mutex_unlock (thumbnail_mutex);
                    }

                }
                munmap (ptr, sb.size);
                shm_unlink (shm_thumbnail);
            }
        } else {
            g_warning ("zip_stat(Z, Thumbnails/thumbnail.png, 0, &sb)==0");
        }
        zip_close (Z);
        g_free (thumbnail);
        if(pixbuf) {
            return pixbuf;
        } else {
            return NULL;
        }
    }
#endif

    // Ok, that didn't work either. Is it ghostscript (ps or pdf) or text?
    // this will construct the thumbnail in disk to load on next mousemove.
    const gchar *convert_type = want_imagemagick_preview (population_p->en);

    if(!convert_type) {
        DBG ("SHOW_TIPx: convert type=%s\n",convert_type);
        g_free (thumbnail);
        return NULL;
    }

    // pdf forks to ghostscript to create thumbnail file
    if(strcmp (convert_type, "PDF") == 0 || strcmp (convert_type, "PDF") == 0) {
        pixbuf = image_magic_preview_forked (population_p, thumbnail, convert_type);
    }
    // text uses pango cairo to create thumbnail file
    else if(strcmp (convert_type, "TXT") == 0) {
        view_t *view_p = population_p->view_p;
        pixbuf = text_preview (population_p, thumbnail, view_p);
    }
    return pixbuf;
}


