static GList *global_view_list = NULL;

static void
mk_logical_rect (PangoLayout * layout, PangoRectangle * logical_rect) {
    TRACE (">> mk_logical_rect\n");
    if(layout) {
        pango_layout_get_pixel_extents (layout, NULL, logical_rect);
        /* ++ because zero is actually unity, not nonexistance */
        logical_rect->width++;
        logical_rect->height++;
    }
}

//FIXME: use pango internals to do the cutting
static void
mk_layouts (view_t * view_p, gchar * tag, PangoLayout ** layout, PangoLayout ** layout2) {
    TRACE (">> rfm_mk_layouts\n");
    int i;
    gchar *optimal_chars = " .-_";
    static gchar *optimal_set[4] = { NULL, NULL, NULL, NULL };
    PangoRectangle logical_rect;
    gchar *tag_copy = NULL,
        *the_tag = NULL;

    if(!layout || !layout2)
        return;
    /* cleanup first */
    if(*layout)
        g_object_unref (*layout);
    if(*layout2)
        g_object_unref (*layout2);
    *layout = *layout2 = NULL;

    if(!tag || !strlen (tag))
        return;
    if(!optimal_set[0]) {
        for(i = 0; i < strlen (optimal_chars); i++) {
            gchar c[2] = { 0, 0 };
            c[0] = optimal_chars[i];
            optimal_set[i] = rfm_utf_string (c);
        }
    }

    /* new method... */
    /* get width of string
     * string MUST be a validated utf string here. */
    the_tag = g_strdup (tag);
    *layout = gtk_widget_create_pango_layout (view_p->widgets.paper, the_tag);
    pango_layout_get_pixel_extents (*layout, NULL, &(logical_rect));
    if(logical_rect.width + 1 <= CELLWIDTH - 2 * TEXTSPACING) {
        /* if it fits, were all done */
        TRACE ("no cut necessary for %s\n", tag);
        g_free (the_tag);
        return;
    }
    /* shucks, it doesn't fit. Let's work with a copy of the tag */
    TRACE ("shucks, it doesn't fit: %d > %d\n", logical_rect.width + 1, CELLWIDTH - 2 * TEXTSPACING);
    tag_copy = g_strdup (the_tag);
    while(strlen (tag_copy) > 1 && logical_rect.width + 1 > CELLWIDTH - 2 * TEXTSPACING) {
        gchar *rear;
        int length = strlen (tag_copy);
        gchar *p = tag_copy + length - 1;
        rear = g_utf8_find_prev_char (tag_copy, p);
        length = strlen (rear);
        memset (rear, 0, length);
        g_object_unref (*layout);
        *layout = gtk_widget_create_pango_layout (view_p->widgets.paper, tag_copy);
        pango_layout_get_pixel_extents (*layout, NULL, &(logical_rect));
    }
    TRACE ("\"%s\" cut to \"%s\"\n", the_tag, tag_copy);
    /* Now the first part fits, but the split may not be optimum
     * we must now look for one in " .-_" to do the split, but the
     * character search must be done in utf */
    /* is the cut optimal? */
    {
        int cut_location = strlen (tag_copy);
        gchar *optimal_ch = NULL;
        gchar *p = the_tag + (cut_location - 1);
        gchar *optimal_cut_p = NULL;
        p = g_utf8_find_prev_char (the_tag, p); /* previous validated location */

        p = g_utf8_find_next_char (p, NULL);    /* validated location */

        for(i = 0; i < strlen (optimal_chars); i++) {
            if(strncmp (p, optimal_set[i], strlen (optimal_set[i])) == 0) {
                optimal_ch = optimal_set[i];
                optimal_cut_p = p;
                TRACE ("optimal cut found for %s\n", tag);
                break;
            }
        }
        if(optimal_cut_p)
            goto go_on;

        for(i = 0; i < strlen (optimal_chars); i++) {
            gchar *g = g_utf8_strrchr (tag_copy, -1, g_utf8_get_char (optimal_set[i]));

            if(g) {
                int diff = strlen (tag_copy) - strlen (g);
                g = the_tag + diff + 1;
                *layout2 = gtk_widget_create_pango_layout (view_p->widgets.paper, g);
                pango_layout_get_pixel_extents (*layout2, NULL, &(logical_rect));
                if(logical_rect.width + 1 <= CELLWIDTH - 2 * TEXTSPACING) {
                    TRACE ("might improve cut for %s\n", tag);
                    g_object_unref (*layout);
                    memset (tag_copy + diff + 1, 0, strlen (tag_copy + diff + 1));
                    *layout = gtk_widget_create_pango_layout (view_p->widgets.paper, tag_copy);

                    goto done;
                } else {
                    TRACE ("NOT improveable: %s (could look for another cut char)\n", tag);
                    g_object_unref (*layout2);
                }
                break;
            }
        }
    }

    /* So we don't have an optimal split. Let's construct the rest. */
  go_on:
    if(strlen (tag_copy) && strlen (the_tag) > strlen (tag_copy)) {
        gchar *p = the_tag + (strlen (tag_copy));
        /* XXX should do this for all spaces, not just the first  */
        if(strncmp (p, optimal_set[0], strlen (optimal_set[0])) == 0)
            p = g_utf8_find_next_char (p, NULL);

        *layout2 = gtk_widget_create_pango_layout (view_p->widgets.paper, p);
        /* final question, does it fit? */
        pango_layout_get_pixel_extents (*layout2, NULL, &(logical_rect));
        while(logical_rect.width + 1 > CELLWIDTH - 2 * TEXTSPACING) {
            gchar *g;
            TRACE ("second line does not fit (%s)\n", p);
            g_object_unref (*layout2);
            p = g_utf8_find_next_char (p, NULL);
            g = g_strconcat ("~", p, NULL);
            *layout2 = gtk_widget_create_pango_layout (view_p->widgets.paper, g);
            pango_layout_get_pixel_extents (*layout2, NULL, &(logical_rect));
            g_free (g);
        }

    }
    /* finally we free our copies of the tag */
  done:
    g_free (tag_copy);
    g_free (the_tag);
    return;
}

#if 0
/* this is to see which icon gets placed on top
   and is incompatible with sorting icons by grid*/
static int
rfm_population_compare (const void *a, const void *b) {
    population_t *p_a = (population_t *) a;
    population_t *p_b = (population_t *) b;
    if(!a || !b)
        return 0;
    return (p_a->serial - p_b->serial);
}

// this sort function is just to see which icons get placed on
// top of each other in the desktop view.
static
void
sort_population_list (view_t * view_p, population_t * population_p) {
    if(!view_p->xfdir_p) {
        g_warning ("sort_population_list: !view_p->xfdir_p");
        return;
    }
    if(view_p->xfdir_p->pathc < 2) {
        return;
    }
    // place referenced population_p at the top of the list
    population_p->serial = view_p->serial++;
    // sort the list 
    qsort ((void *)view_p->population_pp, view_p->xfdir_p->pathc, sizeof (population_t),
#ifdef __COMPAR_FN_T
           (__compar_fn_t)
#endif
           rfm_population_compare);
}
#endif

void
destroy_population_item (population_t * population_p) {
    if(!population_p)
        return;
    /* objects that need to be unreffed: 
     * (tagged as referrence in types.h) */

    if(population_p->preview_pixbuf)
        g_object_unref (population_p->preview_pixbuf);
    if(population_p->initial_preview_pixbuf)
        g_object_unref (population_p->initial_preview_pixbuf);
    /* these are hash optimised: 
     * if (population_p->normal_pixbuf)
     g_object_unref(population_p->normal_pixbuf);*/
    if(population_p->saturated_pixbuf)
        g_object_unref (population_p->saturated_pixbuf);
    if(population_p->cut_pixbuf)
        g_object_unref (population_p->cut_pixbuf);
    // XXX: selected_pixbuf is tagged as reference not clear in types.h
    // releasing it here is wrong. Must find out where it is unreffed
    // and put those comments in the code for clarity.
     /*if (population_p->selected_pixbuf)
       g_object_unref(population_p->selected_pixbuf); */

    if(population_p->layout)
        g_object_unref (population_p->layout);
    if(population_p->layout2)
        g_object_unref (population_p->layout2);
    rfm_destroy_entry (population_p->en);
    g_free (population_p);
}

// a bit complicated, and a bit dificult to debug...
// but it avoids remaking thumnail pixmaps
static void
transfer_population (view_t * view_p) {
    int i;
    population_t **new_population_pp,
    **old_population_pp;
    new_population_pp = rodent_mk_grid_elements (view_p, view_p->xfdir_p->pathc);
    old_population_pp = view_p->population_pp;

    //DBG(" gl is 0x%lx for fstab\n", (unsigned long)((long)view_p->xfdir_p->gl));
    //DBG("monitor transfer_population xfdir_p 0x%lx != xfdir_p->xfdir_p 0x%lx\n", (unsigned long)((long)view_p->xfdir_p), (unsigned long)((long)view_p->xfdir_p->xfdir_p));

    DBG ("MONITOR TRANSFER, pathc is %ld\n", (long)view_p->xfdir_p->pathc);
    // find stuff in old_population which is in new_population
    for(i = 0; i < view_p->xfdir_p->pathc; i++) {
        char *newpath;
        if(view_p->xfdir_p->gl[i].en)
            newpath = view_p->xfdir_p->gl[i].en->path;
        else
            newpath = "root entry";
        population_t **tmp;
        gboolean hit = FALSE;
        population_t *population_p;
        for(tmp = old_population_pp; *tmp; tmp++) {
            population_p = *tmp;
            //if (!population_p) break;
            char *oldpath;
            if(population_p->en) {
                oldpath = population_p->en->path;
                if(view_p->xfdir_p->invalidate_icons) {
                    // if the invalidate_icons flag is set
                    // then all icons are updated,
                    // that means, nothing is found
                    // in the old population.
                    // This is practically a hard reload.
                    continue;
                }
            } else
                oldpath = "root entry";
            // if stat information is different, icon may need to be
            // updated, so the easy way is to test this here too.
            gboolean teskOK = population_p->en && population_p->en->st;
            if(teskOK && strcmp (newpath, oldpath) == 0 &&
               memcmp (population_p->en->st, view_p->xfdir_p->gl[i].en->st, sizeof (struct stat)) == 0) {
                hit = TRUE;
                break;
            }
        }
        if(hit) {
            // found in old_population
            // and the stat information is the same
            // image preview need not be updated. 
            new_population_pp[i] = population_p;
            /*  update the stat record */
            if(view_p->xfdir_p->gl[i].en) {
                if(!new_population_pp[i]->en->st) {
                    g_warning ("!new_population_pp[i]->en->st");
                }
                if(!view_p->xfdir_p->gl[i].en->st) {
                    g_warning ("!view_p->xfdir_p->gl[i].en->st");
                }
                memcpy (new_population_pp[i]->en->st, view_p->xfdir_p->gl[i].en->st, sizeof (struct stat));

            }

            DBG ("newmonitor: hit for %s\n", newpath);
            // mv population element

        } else {
            //create new population item
            DBG ("newmonitor: no-hit for %s\n", newpath);
            new_population_pp[i] = rodent_create_population_t (view_p, view_p->xfdir_p->gl[i].en, i, NULL, view_p->xfdir_p->gl[i].pathv);
            //  new x,y position
            //  image preview thread
        }
    }
    // reset the state variable
    view_p->xfdir_p->invalidate_icons = FALSE;
/* That done, update the population selection list. */

    GList *tmp;
    for(tmp = view_p->selection_list; tmp; tmp = tmp->next) {
        record_entry_t *en = (record_entry_t *) tmp->data;
        if(en) {
            int i;
            gboolean found = FALSE;
            for(i = 0; new_population_pp[i]; i++) {
                if(en == new_population_pp[i]->en) {
                    found = TRUE;
                    break;
                }
            }
            if(!found) {
                DBG ("newmonitor: removing  %s from selection list\n", en->path);
                view_p->selection_list = g_list_remove (view_p->selection_list, en);
            }
        }
    }
/*
 * Next, destroy population elements of
 * old_population_pp which are not in the new_population_pp
 *
 * 
 * */
#if 10
    int j;
    for(j = 0; old_population_pp[j]; j++) {
        gboolean found = FALSE;
        if(!old_population_pp[j])
            continue;
        int i;
        for(i = 0; new_population_pp[i]; i++) {
            if(old_population_pp[j] == new_population_pp[i]) {
                found = TRUE;
                break;
            }
        }
        if(!found) {
            DBG ("newmonitor: removing  %s population_p\n",
                 (old_population_pp[j]->en == NULL) ? "root entry" : old_population_pp[j]->en->path);
            // FIXME: we have a race here. we have a valid population
            // item with an invalid entry. entry seems ok, but entry
            // contents are full of junk.
            destroy_population_item (old_population_pp[j]);
            old_population_pp[j] = NULL;
        }
    }

#endif

    /* destroy old population pp and assign new population pp */
    view_p->population_pp = new_population_pp;
    g_free (old_population_pp);
    // for deskview expose:
    view_p->redraw_pixmap = TRUE;

    return;

}

static
GdkPixbuf *
cut_icon (view_t * view_p, population_t * population_p, int pasteboard_status) {
    GdkPixbuf *pixbuf = gdk_pixbuf_copy (population_p->normal_pixbuf);
    gchar *tag_id = "xffm/stock_copy";
    if(pasteboard_status == 2) {        //iow, cut
        tag_id = "xffm/stock_cut";
    }
    GdkPixbuf *scissors = rfm_get_pixbuf (tag_id, view_p->icon_size);
    int w = gdk_pixbuf_get_width (population_p->normal_pixbuf);
    int h = gdk_pixbuf_get_height (population_p->normal_pixbuf);

    gdk_pixbuf_composite (scissors, pixbuf, 0, 0,       // destination x,y
                          w, h, 0, 0,   // offset x,y
                          0.33, 0.33,   //scale x,y
                          GDK_INTERP_BILINEAR, 254);    // opacity

    g_object_ref (pixbuf);

    return pixbuf;
}


static void
increase_population_serial (view_t * view_p) {
    if(view_p->population_serial_mutex == NULL) {
        view_p->population_serial_mutex = g_mutex_new ();
    }
    g_mutex_lock (view_p->population_serial_mutex);
    view_p->population_serial++;
    g_mutex_unlock (view_p->population_serial_mutex);

}


static void
schedule_destroy(view_t *view_p){
    view_p->widgets.diagnostics=NULL;
    view_p->widgets.diagnostics_window=NULL;
    view_p->widgets.status=NULL;
    view_p->widgets.paper=NULL;

    view_p->exit_status = TRUE;
    if (view_p->widgets.notebook) {
        // this destroys the child widget!
	gint page_num = gtk_notebook_page_num (GTK_NOTEBOOK (view_p->widgets.notebook), view_p->widgets.page_child);
	gtk_notebook_remove_page (GTK_NOTEBOOK (view_p->widgets.notebook), page_num);
    }
    if (!view_p->monitor_enabled) {
	g_thread_create (rfm_xfdir_view_cleanup, view_p, FALSE, NULL);
    } else {
	rfm_xfdir_exit_monitor (view_p);
    }
}

///***////****///****/

//*** progress bar stuff ***/
#define SET_PROGRESS_FULL(w,x,y) set_progress_generic(w,x,y,1)

static const gchar *
plain_sizetag (int tama, int tamal) {
    char *tag = "KB";
    char *tagl = "KB";
    static gchar *buf = NULL;

    g_free (buf);
    buf = NULL;
    if(tama >= 1024 * 1024) {
        tama /= (1024 * 1024);
        tag = "GB";
    } else if(tama >= 1024) {
        tama /= 1024;
        tag = "MB";
    }

    if(tamal >= 1024 * 1024) {
        tamal /= (1024 * 1024);
        tagl = "GB";
    } else if(tamal >= 1024) {
        tamal /= 1024;
        tagl = "MB";
    }

    buf = g_strdup_printf ("%d %s / %d %s", tama, tag, tamal, tagl);

    return (const gchar *)buf;
}

static void
set_progress_generic (widgets_t * widgets_p, int count, int total, int caso) {
    static int smallcount = 1;
    GtkWidget *progress;
    gfloat fraction;

    if(!widgets_p)
        return;
    progress = widgets_p->progress;
    if(progress == NULL)
        return;
    if(caso == -1) {
        if(progress)
            gtk_widget_hide (progress);
        if(widgets_p->status)
            gtk_widget_show (widgets_p->status);
        return;
    }

    if(!progress) {
        DBG ("progress==NULL");
        return;
    }
    if(!gtk_widget_get_visible (progress)) {
        if(widgets_p->status) {
            GtkTextBuffer *buffer = gtk_text_view_get_buffer ((GtkTextView *) widgets_p->status);
            gtk_text_buffer_set_text (buffer, " ", -1);
        }
        gtk_widget_show (progress);
    }
    /* this does not solve the indented status, neither does look good:
       if (widgets_p->status)gtk_widget_hide(widgets_p->status);
     */
    if(count == -1) {
        /*DBG("TRACE:updating pulse\n"); */
        if(total < 0 || (smallcount++ & (1 << 7))) {
            gtk_progress_bar_pulse ((GtkProgressBar *) progress);
            //gdk_flush();
            smallcount = 1;
        }
    } else {
        char texto[_POSIX_PATH_MAX];
        if(!total)
            fraction = 0.0;
        else
            fraction = (float)count / (float)total;
        if(fraction < 0 || fraction > 1)
            return;
        gtk_progress_bar_set_fraction ((GtkProgressBar *) progress, fraction);
        if(count >= total)
            texto[0] = 0;
        else
            switch (caso) {
            case 1:
                sprintf (texto, "%d / %d", count, total);
                break;
            case 0:
            default:
                sprintf (texto, "%s", plain_sizetag (count, total));
                break;
            }
        gtk_progress_bar_set_text ((GtkProgressBar *) progress, texto);
    }
}

static
void
reset_view_preferences (view_t * view_p) {
    if(!view_p->en)
        return;
    view_preferences_t *view_preferences_p = rodent_get_view_preferences (view_p->en->path);
    if(view_preferences_p) {
        view_p->preferences = view_preferences_p->preferences;
        view_p->sortcolumn = view_preferences_p->sortcolumn;
        if(view_p->preferences & __SHOW_HIDDEN)
            SET_SHOWS_HIDDEN (view_p->en->type);
        else
            UNSET_SHOWS_HIDDEN (view_p->en->type);
        if(view_p->preferences & __SHOW_IMAGES)
            SET_SHOWS_IMAGES (view_p->en->type);
        else
            UNSET_SHOWS_IMAGES (view_p->en->type);
    } else {
    	// default settings
        view_p->preferences = DEFAULT_VIEW_PREFERENCES;
        view_p->preferences = DEFAULT_SORT_COLUMN;    
    }
    //XXX Is the following instruction really necesary?
    if(view_p->en->path && g_file_test (view_p->en->path, G_FILE_TEST_IS_DIR))
        SET_SDIR (view_p->en->type);
}

static
    void
set_application_icon (widgets_t * widgets_p, const gchar * icon_id) {
    GdkPixbuf *icon_pixbuf = rfm_get_pixbuf (icon_id, SIZE_ICON);
    if(icon_pixbuf) {
        gtk_window_set_icon (GTK_WINDOW (widgets_p->window), icon_pixbuf);
    }
}

static void
set_view_title (view_t * view_p) {
    if(view_p->type == DESKVIEW_TYPE)
        return;
    DBG (">> set_iconview_title\n");
    record_entry_t *en = view_p->en;
    const gchar *path;

    if(en && en->path && strlen (en->path)) {
        path = en->path;
    } else {
        path = g_get_host_name ();
    }

    if(view_p->module_name) {
        if(view_p->module_name && rfm_natural ("rfm", view_p->module_name, view_p->en, "window_title")) {
            path = rfm_natural ("rfm", view_p->module_name, view_p->en, "window_title");
        }
    }
    /*if (view_p->module_name && rfm_void("rfm",view_p->module_name,"exec_name")) path=rfm_void("rfm",view_p->module_name,"exec_name"); */
    rodent_set_icon_name (view_p->widgets.window, path);
    const gchar *icon_id = rfm_resolve_icon_id (en);

    set_application_icon (&(view_p->widgets), icon_id);

    gchar *basename = g_path_get_basename (path);
    gchar *dirname = g_path_get_dirname (path);
    gchar *parentname = g_path_get_basename (dirname);
    gchar *tabname;
    if(strcmp (parentname, ".") == 0 || strcmp (basename, "/") == 0) {
        tabname = g_strdup (basename);
    } else {
        tabname = g_build_filename (parentname, basename, NULL);
    }
    gtk_label_set_text (GTK_LABEL (view_p->widgets.page_label), tabname);
    gtk_label_set_text (GTK_LABEL (view_p->widgets.page_menu_label), path);
    g_free (basename);
    g_free (dirname);
    g_free (parentname);
    g_free (tabname);
    GdkPixbuf *pb = rfm_get_pixbuf (icon_id, SIZE_BUTTON);
    GtkWidget *image = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (image);
    GList *children = gtk_container_get_children (GTK_CONTAINER (view_p->widgets.page_icon));
    if(children) {
        gtk_container_remove (GTK_CONTAINER (view_p->widgets.page_icon), GTK_WIDGET (children->data));
    }
    gtk_container_add (GTK_CONTAINER (view_p->widgets.page_icon), image);
    g_list_free (children);

    /*gtk_notebook_set_tab_label_text ((GtkNotebook *)view_p->widgets.notebook,
       view_p->widgets.page,
       basename); */

    /* session management: this is for dummies */
    /*set_iconview_restart(view_p); */
    return;
}

static void
get_best_iconsize (view_t * view_p) {
    /* determine iconsize if images in xfdir structure */
    view_p->icon_size = SIZE_ICON;
    if(view_p->type == DESKVIEW_TYPE || !rfm_svg_supported ())
        return;
    return;
#if 0
    // deprecated code: too nerdy
    // FIXME: should get the directory preference.
    if(!view_p->en || !view_p->en->path ||
       !g_file_test (view_p->en->path, G_FILE_TEST_IS_DIR) || !(view_p->preferences & __SHOW_IMAGES))
        return;

    int j;
    DBG ("iconsize: looking at %s\n", view_p->en->path);
    if(chdir (view_p->en->path) < 0) {
        g_warning ("chdir(%s): %s", view_p->en->path, strerror (errno));
    }
    // get mimetype for icons (no magic yet)
    for(j = 0; j < view_p->xfdir_p->pathc; j++) {
        if(view_p->xfdir_p->gl[j].en == NULL)
            continue;           //skip root
        if(!view_p->xfdir_p->gl[j].en->mimetype) {
            view_p->xfdir_p->gl[j].en->mimetype = MIME_type (view_p->xfdir_p->gl[j].en->path, view_p->xfdir_p->gl[j].en->st);
        }
        if(view_p->xfdir_p->gl[j].en->mimetype && strstr (view_p->xfdir_p->gl[j].en->mimetype, "image")) {
            view_p->icon_size = BIG_ICON_SIZE;
            break;
        }
    }
    DBG ("-- pathc=%d iconsize=%d\n", (int)(view_p->xfdir_p->pathc), view_p->icon_size);
#endif
}

/////////   icon roots stuff //////
//
//

static population_t *
create_local_root (view_t * view_p, int i, const gchar * path) {
    record_entry_t *en;
    population_t *population_p;
    const gchar *icon_id = NULL;
    const gchar *label = NULL;

    if(!path) {
        path = "/";
        label = "/";
        icon_id = "inode/mount-point";
    }
    if((en = rfm_stat_entry ((char *)path, 0)) == NULL) {
        g_warning (" rfm_stat_entry((char *)path, 0) != NULL");
        return NULL;
    }
    if(strcmp (path, g_get_home_dir ()) == 0) {
        icon_id = "xffm/stock_home";
        label = _("Home");
    }
    SET_LOCAL_TYPE (en->type);
    SET_ROOT_TYPE (en->type);
    en->tag = g_strdup (label);
    population_p = rodent_create_population_t (view_p, en, i, icon_id, label);

    return population_p;
}

static population_t *
create_icon_root_element (view_t * view_p, int i, const gchar * module_name) {
    record_entry_t *en;
    population_t *population_p;
    const gchar *icon_id = NULL;
    const gchar *submodule_name = NULL;
    const gchar *label = NULL;

    if(!module_name) {
        DBG ("ROOTS: !module_name\n");
        return NULL;
    }
    DBG ("ROOTS: module create_icon_root_element: %s\n", module_name);

    if(!rfm_void ("rfm", module_name, "is_root_module")) {
        DBG (" ROOTS: %s is not root module\n", module_name);
        return NULL;
    }
    module_name = rfm_void ("rfm", module_name, "module_name");
    icon_id = rfm_natural ("rfm", module_name, NULL, "module_icon_id");
    label = rfm_natural ("rfm", module_name, NULL, "module_label");
    submodule_name = rfm_void ("rfm", module_name, "module_name");
    DBG (" ROOTS: module_name --> %s \n", module_name);
    DBG (" ROOTS: icon_id --> %s \n", icon_id);
    DBG (" ROOTS: label --> %s \n", label);
    DBG (" ROOTS: submodule_name --> %s \n", submodule_name);
    en = rfm_mk_entry (0);
    {
        const gchar *g = (const gchar *)rfm_natural ("rfm", module_name, NULL,
                                                          "module_label");
        const gchar *gg = (const gchar *)rfm_natural ("rfm", module_name, NULL,
                                                           "module_path");
        if(!g)
            g = "no module_label";
        if(!gg)
            gg = g;
        en->path = g_strdup (gg);
        en->tag = g_strdup (g);
    }
    if(!icon_id || !label) {
        g_warning ("(!icon_id || !label)");
        return NULL;
    }
    // what the hell was this doing here: view_p->module_name = NULL;
    population_p = rodent_create_population_t (view_p, en, i, icon_id, label);
    if(population_p->en)
        population_p->en->module = submodule_name;
    SET_ROOT_TYPE (en->type);
    return population_p;
}

    
typedef struct thread_wait_t {
    GThread *thread;
    gboolean wait_done;
} thread_wait_t;


static void *
thread_wait(gpointer data){
    thread_wait_t *thread_wait_p=data;
    g_thread_join(thread_wait_p->thread);
    thread_wait_p->wait_done=TRUE;
    return NULL;
}

static
void
thread_tryjoin (gpointer (*function) (gpointer), gpointer data) {
    view_t *view_p=data;
    thread_wait_t *thread_wait_p=(thread_wait_t *)malloc(sizeof(thread_wait_t));
    memset(thread_wait_p, 0, sizeof(thread_wait_t));

    thread_wait_p->thread=g_thread_create(function, data, TRUE, NULL);
    g_thread_create(thread_wait, (void *)thread_wait_p, FALSE, NULL);

    int progress_count = view_p->progress_count;
    while (!thread_wait_p->wait_done){
        gboolean progress = view_p->type != DESKVIEW_TYPE && view_p->widgets.progress
            && gtk_widget_get_visible (view_p->widgets.progress);
        if(progress && progress_count != view_p->progress_count) {
            progress_count = view_p->progress_count;
            SET_PROGRESS_FULL (&(view_p->widgets), progress_count, view_p->progress_total);
        }
        /* this borks things: (conflicts with monitor function)  */
	while(gtk_events_pending ()){
            gtk_main_iteration ();
	}
	gdk_flush();
	rfm_threadwait();
    }
    g_free(thread_wait_p);
    return;
}

static
gpointer
load_xfdir_structure (gpointer data) {
    view_t *view_p = data;
    /* Generalized xfdir get (includes modules) */
    view_p->xfdir_p = rfm_xfdir_get (view_p, view_p->en);

    if(!view_p->xfdir_p) {
        g_error ("reload_iconview: view_p->xfdir_p is NULL!");
    }

    DBG ("figuring out best_iconsize(view_p)\n");
    get_best_iconsize (view_p);

    return NULL;
}

static
gpointer
load_xfdir_icons (gpointer data) {
    view_t *view_p = data;
    unsigned int smallcount = 0;
    int j;
    /* add entry for each population_t */
    DBG ("Loading... load_xfdir_icons()\n");
    DBG ("pathc=%ld\n", (long)view_p->xfdir_p->pathc);

    /* previous value is freed in init_iconview() call: */
    if(view_p->type == DESKVIEW_TYPE) {
        view_p->population_pp = rodent_mk_grid_elements (view_p, view_p->xfdir_p->pathc);
        view_p->max_elements = view_p->xfdir_p->pathc;
    } else {
        view_p->population_pp = rodent_mk_grid_elements (view_p, view_p->grid_rows * view_p->grid_columns);
    }

    // we could mutex lock this, 
    // but that should not be necesary since there is only one writer (?)
    view_p->progress_count = 0;
    int maxY = 0;
    int lastY = 0;
    if(view_p->xfdir_p == NULL)
        return NULL;
    if(view_p->population_pp == NULL)
        return NULL;
    for(j = 0; j < view_p->xfdir_p->pathc; j++) {
        if(j >= view_p->max_elements) {
            g_warning ("sanity check: j < view_p->max_elements not met");
            continue;
        }
        /* this is OK, it may be an icon to the rodent root level
         * if (!view_p->xfdir_p->gl[j].en) {
         g_warning("sanity check: view_p->xfdir_p->gl[j].en != NULL not met");
         continue;
         }*/
        if(!view_p->xfdir_p->gl[j].pathv) {
            g_warning ("sanity check: view_p->gl[j]->pathv != NULL not met");
            continue;
        }
        if(view_p->xfdir_p->gl[j].en) {
            view_p->xfdir_p->gl[j].en->module = view_p->submodule_name;
        }
        view_p->population_pp[j] = rodent_create_population_t (view_p, view_p->xfdir_p->gl[j].en, j, NULL, view_p->xfdir_p->gl[j].pathv);
        DBG ("%d. created %s\n", j + 1, view_p->xfdir_p->gl[j].pathv);
        view_p->count++;

        DBG ("load_xfdir_icons: new view_p->population_pp[j](%d,%d)=0x%lx\n",
             view_p->population_pp[j]->row, view_p->population_pp[j]->column, (unsigned long)view_p->population_pp[j]);

        if(view_p->type != DESKVIEW_TYPE) {
            if(view_p->population_pp[j]->rect.y - CELLHEIGHT > maxY) {
                maxY = view_p->population_pp[j]->rect.y - CELLHEIGHT;
            }
            if(view_p->type != DESKVIEW_TYPE && (smallcount++) % 52 == 0) {
                view_p->progress_count = j;
                GdkRectangle rect;
                rect.x = 0;
                rect.y = lastY;
                rect.width = view_p->widgets.window->allocation.width;
                rect.height = maxY;
                rfm_expose_rect (view_p->widgets.paper->window, &rect);
                lastY = maxY;
            }
        }
        // if destroy event called from above, then:
        if(view_p->status == STATUS_CLOSING)
            return NULL;

    }
    if(view_p->type == DESKVIEW_TYPE)
        rodent_recalc_population_geometry (view_p);
    DBG ("view_p->xfdir_p->pathc=%ld\n", (long)view_p->xfdir_p->pathc);

    view_p->progress_count = view_p->xfdir_p->pathc;
    DBG ("memory in population structures=%ld\n", (long)sizeof (population_t) * view_p->xfdir_p->pathc);

    // for iconview, this will expose remaining area,
    // for deskview it will expose all.
    GdkRectangle rect;
    rect.x = 0;
    rect.y = lastY;
    rect.width = view_p->widgets.window->allocation.width;
    rect.height = view_p->widgets.window->allocation.height;
    rfm_expose_rect (view_p->widgets.paper->window, &rect);

    return NULL;
}


