/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
typedef struct keybind_t {
    guint key;
    guint mask;
    gpointer callback;
} keybind_t;

typedef struct filechooser_t {
    GtkWidget *parent;
    void *combo_info;
    int filechooser_action;
    const gchar *folder;
    GtkEntry *entry;
    void (*activate_func) (GtkEntry * entry, gpointer activate_user_data);
    gpointer activate_user_data;

} filechooser_t;


/**************************************************************************/
/**************************  pasteboard stuff  ****************************/
/**************************************************************************/

//extern char *src_host;

/*  */
static
int gui_pasteboard_list (GList ** list_p);

static
int gui_pasteboard_transfer (widgets_t * widgets_p,
                             record_entry_t * t_en, 
			     GList * list, 
			     gboolean cut, 
			     gboolean symlink);

static
    void
gui_pasteboard_copy_cut (widgets_t * widgets_p, gboolean cut, GList ** paste_list) {
    int len;
    char *buffer,
     *files;
    GList *tmp;

    if(*paste_list == NULL)
        return;

    if(cut)
        rfm_status (widgets_p, "xffm/info", _("Cut"), NULL);
    else
        rfm_status (widgets_p, "xffm/info", _("Copy"), NULL);

    gdk_flush ();

    XStoreBuffer (GDK_DISPLAY (), "", 1, CUT_BUFFER);   /* store a null string */

    len = 1 + strlen ("#xfvalid_buffer:copy:%%:\n");
    len += strlen (g_get_host_name ());
    for(tmp = *paste_list; tmp; tmp = tmp->next) {
        int addlen;
        record_entry_t *en = (record_entry_t *) tmp->data;
        /*if (IS_NETWORK_TYPE(en->type)) {
           addlen = strlen("smb://@://\n");
           addlen += ((en->tag) ? strlen(en->tag) : strlen("GUEST%%"));
           } else */
        addlen = 0;
        len += (1 + strlen (en->path) + addlen);
    }
    buffer = (char *)malloc (len * sizeof (char) + 1);
    if(!buffer) {
#ifdef DEBUG
        g_warning ("rfm: unable to allocate paste buffer\n");
#endif
        return;
    }
    sprintf (buffer, "#xfvalid_buffer:%s:%s:\n", (cut) ? "cut" : "copy", g_get_host_name ());
    files = buffer + strlen (buffer);

    for(tmp = *paste_list; tmp; tmp = tmp->next) {
        record_entry_t *en = (record_entry_t *) tmp->data;
        /*if (IS_NETWORK_TYPE(en->type)) {
           char *server, *remote_file;
           server = g_strdup(en->path + 2);
           strtok(server, "/");
           if (IS_XF_NETWS(en->subtype)) {
           sprintf(files, "%s://%s@%s/",
           IS_SAMBA_SERVER(en->subtype) ? "SMB" : "smb",
           (en->tag) ? en->tag : "GUEST%%", server);
           } else {
           remote_file = server + strlen(server) + 1;
           sprintf(files, "%s://%s@%s/%s%s",
           IS_SAMBA_SERVER(en->subtype) ? "SMB" : "smb",
           (en->tag) ? en->tag : "GUEST%%",
           server, remote_file,
           (IS_NETDIR(en->subtype)) ? "/\n" : "\n");
           }
           g_free(server);
           server = NULL;
           files = files + strlen(files);

           } 
           else */
        {
            strcat (buffer, en->path);
            strcat (buffer, "\n");
        }
    }
    /*TRACE("dbg:len=%d,strlen=%d,data=%s\n",len,strlen(buffer),buffer); */
    XStoreBuffer (GDK_DISPLAY (), buffer, len, CUT_BUFFER);
    g_free (buffer);
    buffer = NULL;

    if(cut)
        rfm_status (widgets_p, "xffm/info", _("Cut"), NULL);
    else
        rfm_status (widgets_p, "xffm/info", _("Copy"), NULL);
    view_t *view_p = widgets_p->view_p;
    //view_p->pasteboard_serial++;
    DBG ("PASTE: view_p->pasteboard_serial=%ld\n", view_p->pasteboard_serial);
    gchar *value = g_strdup_printf ("%ld", view_p->pasteboard_serial + 1);
    if (rfm_rational ("rfm", "settings", (void *)"RFM_PASTEBOARD_SERIAL", (void *)value, "mcs_set_var") == NULL){
        g_warning("cannot set RFM_PASTEBOARD_SERIAL");
    }
    g_free (value);
}


static
    int
gui_pasteboard_list (GList ** list_p) {
    gboolean cut;
    char *b = NULL,
        *word;
    int i,
      len = -1;

    b = XFetchBuffer (GDK_DISPLAY (), &len, 0);
    /*TRACE("dbg:bytes=%d,buffer0=%s\n",len,b); */
    if((!b) || (!strlen (b))) {
      no_pasteboard:
        if(b)
            XFree (b);
        return 0;
    }

    if((word = strtok (b, ":")) == NULL)
        goto no_pasteboard;
    if(!strstr (word, "#xfvalid_buffer"))
        goto no_pasteboard;
    if((word = strtok (NULL, ":")) == NULL)
        goto no_pasteboard;
    if(strstr (word, "cut"))
        cut = TRUE;
    else
        cut = FALSE;
    if((word = strtok (NULL, ":")) == NULL)
        goto no_pasteboard;
    if(!word) {
        goto no_pasteboard;
    }
    //src_host = g_strdup(word);

    word = word + strlen (word) + 1;
    if(word[0] == '\n') {
        word++;
        if(word[0] == 0)
            goto no_pasteboard;
    } else {
        if((word = strtok (NULL, "\n")) == NULL)
            goto no_pasteboard;
        word = word + strlen (word) + 1;
    }

    /* create list to send to CreateTmpList */
    i = rfm_uri_parse_list (word, list_p);
    if(!i)
        goto no_pasteboard;
    XFree (b);
    if(cut)
        return 1;
    else
        return 2;
}

static
    int
gui_pasteboard_transfer (widgets_t * widgets_p,
                         record_entry_t * t_en, 
			 GList * list, 
			 gboolean cut,
			 gboolean symlink) {
    gchar *url = (gchar *) list->data;
    if(!url)
        return 0;
    /*DBG ("PASTE: pastepath is %s\n", rfm_details->pastepath);
    if(strncmp (rfm_details->pastepath, "//", 2) == 0) {
        DBG ("TRACE:pastepath=%s,en->path=%s\n", rfm_details->pastepath, t_en ? t_en->path : "null");

    } else */
    {
        if(strncmp (url, "smb://", strlen ("smb://")) == 0 || strncmp (url, "SMB://", strlen ("SMB://")) == 0) {        /* src_host may be remote or local  
                                                                                                                           but target only local  or bookmark */
            if(g_file_test (t_en->path, G_FILE_TEST_EXISTS)) {
                /* this cannot live in module without excessive complication */
                rfm_natural ("rfm", "rfm_smb_list", t_en, "set_drop_entry");
                rfm_rational ("rfm", "rfm_smb_list", list, widgets_p, "SMBGetFile");
                rfm_void ("rfm", "rfm_smb_list", "clear_drop_entry");
            }
            return 1;
        }
    }

    if(t_en->module) {
        DBG ("PASTE: en->module=%s\n", t_en->module);
        if(rfm_natural ("rfm", t_en->module, t_en, "valid_drop_site")) {
            TRACE ("module: valid_drop_site for %s\n", t_en->module);
            rfm_natural ("rfm", t_en->module, t_en, "set_drop_entry");
            if(rfm_rational ("rfm", t_en->module, list, widgets_p, "process_drop")) {
                TRACE ("module: process_drop ok\n");
                /*result=TRUE; */
            }
            rfm_void ("rfm", t_en->module, "clear_drop_entry");
            return 1;
        }
    }
#if 0
    DBG ("PASTE: scp check\n");
    /* single command line for rsync not possible, only for scp! */
    if(strcmp (src_host, g_get_host_name ()) != 0) {
        int l = 0;
        char **srcs;
        GList *tmp;
        for(tmp = list; tmp != NULL; tmp = tmp->next) {
            l++;
        }
        srcs = (char **)malloc ((l + 1) * sizeof (char *));
        if(!srcs) {
            g_warning ("critical: %s at malloc\n", strerror (errno));
            return 0;
        }
        srcs[l] = NULL;
        for(l = 0, tmp = list; tmp != NULL; tmp = tmp->next, l++) {
            srcs[l] = tmp->data;
        }
        rodent_scp (widgets_p, (const gchar **)srcs, t_en->path);
        return 1;
    }
#endif
    DBG ("PASTE: must be local then\n");
    int mode;
    if(cut) {
        mode = TR_MOVE;
    } else {
        mode = TR_COPY;
    }
    if(symlink) {
        mode = TR_LINK;
    }
    rodent_cp (mode, widgets_p, list, t_en->path);
    return 1;
}

/**************************************************************************/
/*************************   end of pasteboard stuff **********************/
/**************************************************************************/
static
void rodent_save_workdir_history (char *p);

static
int rodent_diff (widgets_t * widgets_p, char *file1, char *file2);

static
    void



gui_add_menu_item (widgets_t * widgets_p,
                   GtkWidget * parent,
                   const gchar * label,
                   const gchar * icon_id,
                   const gchar * name,
                   gpointer callback, gpointer callback_data, GtkAccelGroup * accel_group, guint accel_key, guint accel_mask);

static
    void
  gui_add_to_keylist (guint in_key, guint mask, gpointer callback);

static
GtkWidget *gui_mk_radio_button (widgets_t * widgets_p,
                                const gchar * label, 
				const gchar * name, 
				GtkWidget * parent, 
				GSList ** radiogroup);

static void
gui_add_autotype_C_widgets (widgets_t * widgets_p,
                            GtkWidget * popup, 
			    gpointer callback, 
			    gpointer on_mount, 
			    gpointer on_unmount);

static
GtkWidget *gui_add_menu_separator (GtkWidget * parent);

/**************************************************************************/

///// popup callbacks....

/* this file is included by rodent_popup.c */
static void
execute (GtkMenuItem * menuitem, gpointer user_data, gboolean selection) {
    widgets_t *widgets_p;
    widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;

    /* set the working directory */
    g_free (view_p->workdir);
    if(view_p->en && view_p->en->path)
        view_p->workdir = g_strdup (view_p->en->path);
    else
        view_p->workdir = g_strdup (g_get_home_dir ());

    DBG ("on_open_with_activate...\n");
    gchar *files = g_strdup (" ");
    gchar *files_txt;

    gchar *first_path=NULL;
    files_txt = g_strdup_printf (_("Open with %s"),":   \n\n");
    if(selection) {
        if(g_list_length (view_p->selection_list)) {
            gchar *tt,
             *ttt;
            files = g_strdup (" ");
            GList *tmp;
            for(tmp = view_p->selection_list; tmp; tmp = tmp->next) {
                record_entry_t *en = (record_entry_t *) tmp->data;
                char *b = g_path_get_basename (en->path);
                gchar *q = rfm_utf_string (rfm_chop_excess (b));
                tt = g_strconcat (files_txt, q, "\n", NULL);
                gchar *esc_path = rfm_esc_string (en->path);
		if (!first_path){
		    first_path=g_strdup(esc_path);
		}
                ttt = g_strconcat (files, esc_path, " ", NULL);
                g_free (esc_path);
                g_free (files_txt);
                g_free (q);
                g_free (b);
                g_free (files);
                files_txt = tt;
                files = ttt;
            }
        }
    }
    DBG ("OPEN: files=%s\n", files);
    DBG ("OPEN: first_path=%s\n", first_path);
    gboolean interm = FALSE;
    const gchar *g;
    {
        gchar *f = g_build_filename (g_get_user_data_dir (), RUN_DBH_FILE, NULL);
        gchar *ff = g_build_filename (g_get_user_data_dir (), RUN_FLAG_FILE, NULL);
        DBG ("RUN_DBH_FILE=%s RUN_FLAG_FILE=%s\n", f, ff);

        g = rodent_get_response_history (NULL,
		_("Execute:"), 
		files_txt,
		f, 
		first_path,
		NULL,
		ff,
		_("Run in _Terminal"),
		GTK_FILE_CHOOSER_ACTION_OPEN,
		"/usr/bin");
        g_free (f);
        g_free (ff);
    }
    g_free (first_path);

    g_free (files_txt);
    if(!g) {
        DBG ("on_open_with_activate... !g\n");
        return;
    }
    if(g[strlen (g) + 1])
        interm = TRUE;

    if(selection) {
        /* if only one file selected, associate to mimetype... */
        if(g_list_length (view_p->selection_list) == 1 && view_p->selected_p) {
            DBG ("OPEN: MIME_add(%s,%s)\n", view_p->selected_p->en->mimetype, g);
            if(!view_p->selected_p->en->mimetype) {
                view_p->selected_p->en->mimetype = MIME_magic (view_p->selected_p->en->path);
            }
            if(view_p->selected_p->en->mimetype) {
                gchar *command = MIME_mk_command_line (g, NULL);
                if(interm) {
                    gchar *term_command = MIME_mk_terminal_line (command);
                    g_free (command);
                    command = term_command;
                }
                DBG ("OPEN: adding %s --> %s (from %s)\n", view_p->selected_p->en->mimetype, command, g);
                MIME_add (view_p->selected_p->en->mimetype, command);
                g_free (command);
            }
        }
    }
    gchar *command;
    if(strstr (g, "%s")) {
        command = g_strdup_printf (g, files);
    } else {
        command = g_strdup_printf ("%s %s", g, files);
    }
    DBG ("OPEN: command = %s\n", command);
    g_free (files);

    // do the call with argv so that command not saved in lpterm history
    rfm_thread_run2argv (widgets_p, command, interm);

    g_free (command);
}

/*   callback  */
static void
on_open_with_activate (GtkMenuItem * menuitem, gpointer user_data) {
    /*widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }*/
    execute (menuitem, user_data, TRUE);
    //rfm_population_read_unlock (view_p);
}

static void
on_open_exec (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p;
    widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;

    /* set the working directory */
    g_free (view_p->workdir);
    if(view_p->en && view_p->en->path)
        view_p->workdir = g_strdup (view_p->en->path);
    else
        view_p->workdir = g_strdup (g_get_home_dir ());

    if(g_list_length (view_p->selection_list)) {
          record_entry_t *en = (record_entry_t *) view_p->selection_list->data;
          if (en && en->path) {
              rfm_thread_run2argv (widgets_p, en->path, TRUE);
          }
    }
}

static void
gui_on_about_activate (GtkMenuItem * menuitem, gpointer data) {
    gchar *argv[3];
    argv[0] = "rodent";
    argv[1] = "--version";
    argv[2] = NULL;
    widgets_t *widgets_p = (widgets_t *) data;
    if(widgets_p->diagnostics == NULL) {
        rfm_create_diagnostics_window (widgets_p);
    }
    rfm_show_text (widgets_p);
    rfm_thread_run_argv (widgets_p, argv, FALSE);
}

static void
on_autotype_R (GtkMenuItem * menuitem, gpointer user_data) {
    record_entry_t *en;
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    en = view_p->en;
    rodent_open_with (&(view_p->widgets), en);
    return;
}

static void
on_view_activate (GtkMenuItem * w, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    view_t *view_target_p;
    if(view_p->constructor) {
        record_entry_t *en = rfm_copy_entry (view_p->en);
        view_target_p = (*view_p->constructor) (en);
    } else if(view_p->popup_constructor) {
        (*view_p->popup_constructor) (widgets_p, view_p->en->path);
        return;
    } else
        return;

    widgets_p = &(view_target_p->widgets);
    view_target_p->preferences = view_p->preferences;
    view_target_p->sortcolumn = view_p->sortcolumn;
    view_target_p->module_name = view_p->module_name;

    record_entry_t *en = rfm_copy_entry (view_p->en);
    rodent_refresh (widgets_p, en);
    return;
}

static void
rodent_open_in_terminal_activate (GtkMenuItem * menuitem, gpointer user_data) {
    static gchar *shell = NULL;
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    g_free (view_p->workdir);
    if(!view_p->en || !view_p->en->path)
        view_p->workdir = g_strdup (g_get_home_dir ());
    else
        view_p->workdir = g_strdup (view_p->en->path);
    g_free (shell);
    if(getenv ("SHELL") && strlen (getenv ("SHELL"))) {
        shell = g_find_program_in_path (getenv ("SHELL"));
    }
    if(!shell && getenv ("XTERM_SHELL") && strlen (getenv ("XTERM_SHELL"))) {
        shell = g_find_program_in_path (getenv ("XTERM_SHELL"));
    }
    if(!shell)
        shell = g_find_program_in_path ("bash");
    if(!shell)
        shell = g_find_program_in_path ("sh");
    if(!shell)
        shell = g_find_program_in_path ("csh");
    if(!shell)
        shell = g_find_program_in_path ("ksh");
    if(!shell)
        shell = g_find_program_in_path ("tcsh");
    if(!shell)
        shell = g_find_program_in_path ("zsh");
    if(!shell)
        shell = g_find_program_in_path ("ash");
    if(!shell)
        return;
    gchar *argv[] = { shell, NULL };

    // use argv version so that it won't go into history file...
    rfm_thread_run_argv (widgets_p, argv, TRUE);

    return;
}

//XXX not thread safe
static void
rodent_differences_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    gchar *file1 = NULL,
        *file2 = NULL;

    if(view_p->selection_list) {
        file1 = ((record_entry_t *) (view_p->selection_list->data))->path;
        if(g_list_length (view_p->selection_list) > 1)
            file2 = ((record_entry_t *) (view_p->selection_list->next->data))->path;

    }
    TRACE ("rodent-diff %s %s\n", (file1 ? file1 : "null"), (file2 ? file2 : "null"));
    if(view_p->en && view_p->en->path)
        view_p->workdir = g_strdup (view_p->en->path);
    else
        view_p->workdir = (char *)g_strdup (g_get_home_dir ());
    rodent_diff (widgets_p, file1, file2);
}

static void
on_settings_show (GtkMenuItem * w, gpointer data) {
    void *p;
    GtkDialog *d;

    if((p = rfm_void ("rfm", "settings", "module_init")) != NULL) {
        TRACE ("module_init=0x%lx\n", (unsigned long)p);
        d = (GtkDialog *) rfm_natural ("rfm", "settings", p, "run_rfm_settings_dialog");
        if(d) {
            gtk_dialog_run (d);
        }
    } else {
        g_warning ("cannot load libsettings");
    }
    return;

}

static void
on_newdir_activate (GtkMenuItem * menuitem, gpointer user_data) {
    record_entry_t *en = NULL;
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(view_p)
        en = view_p->en;
    if(!en || !en->path)
        return;

    gchar *label = g_strdup_printf ("%s/", en->path);
    gchar *q = rfm_utf_string (rfm_chop_excess (label));
    const gchar *g = rodent_get_response (NULL, _("Create New Folder"), q,
                                          _("Name"));
    g_free (label);
    g_free (q);

    if(g && strlen (g)) {
      gchar *p = g_build_filename (en->path, g, NULL);
      if(!g_file_test (p, G_FILE_TEST_EXISTS)) {
        gchar *argv[3];
        int j=0;
        argv[j++]="mkdir";
        argv[j++]=p;
        argv[j++]=NULL;

        gboolean authorized=rfm_write_ok(en->st);
        gchar *sudo = NULL;
        if (!authorized) {
                sudo = g_find_program_in_path ("sudo");
        } 
        if (sudo) {
           // shell processing if using sudo...
           gchar *pp = rfm_esc_string(p);
           gchar *command = 
                 g_strdup_printf("sudo -A mkdir %s && sudo -A chown %d:%d %s",
                        pp, (int)geteuid(), (int)getegid(), pp);
           g_free(pp);
           rfm_thread_run (widgets_p, command, FALSE);
           g_free(command);
        } else {
                rfm_thread_run_argv (widgets_p, argv, FALSE);
        }
        g_free(sudo);
      }
      else {
            rfm_confirm (NULL, GTK_MESSAGE_ERROR, strerror (EEXIST), NULL, NULL);
      }
      g_free(p);
    }
}

static void
on_newfile_activate (GtkMenuItem * menuitem, gpointer user_data) {
    record_entry_t *en = NULL;
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(view_p)
        en = view_p->en;
    if(!en || !en->path)
        return;
    gchar *label = g_strdup_printf ("%s/", en->path);
    gchar *q = rfm_utf_string (rfm_chop_excess (label));
    const gchar *g = rodent_get_response (NULL, _("Create a new empty file inside this folder"), q, _("Name"));
    g_free (q);
    g_free (label);
    if(g && strlen (g)) {
        gchar *p = g_build_filename (en->path, g, NULL);
        if(!g_file_test (p, G_FILE_TEST_EXISTS)) {
          gchar *argv[3];
          int j=0;
          argv[j++]="touch";
          argv[j++]=p;
          argv[j++]=NULL;

          gboolean authorized=rfm_write_ok(en->st);
          gchar *sudo = NULL;
          if (!authorized) {
                sudo = g_find_program_in_path ("sudo");
          } 
          if (sudo) {
           // shell processing if using sudo...
           gchar *pp = rfm_esc_string(p);
           gchar *command = 
                 g_strdup_printf("sudo -A touch %s && sudo -A chown %d:%d %s",
                        pp, (int)geteuid(), (int)getegid(), pp);
           g_free(pp);
           rfm_thread_run (widgets_p, command, FALSE);
           g_free(command);
          } else {
                rfm_thread_run_argv (widgets_p, argv, FALSE);
          }
          g_free(sudo);
        } else {
            rfm_confirm (NULL, GTK_MESSAGE_ERROR, strerror (EEXIST), NULL, NULL);
        }
        g_free (p);
    }
}

#if 0
static const gchar *
get_selection_list_path (view_t * view_p) {
    GList *tmp,
     *selection_list;
    int list_path_length = 0;
    gchar *list_path = NULL;

    selection_list = view_p->selection_list;

    for(tmp = selection_list; tmp; tmp = tmp->next) {
        record_entry_t *en;
        en = (record_entry_t *) tmp->data;
        gchar *g = g_path_get_basename (en->path);
        list_path_length += (strlen (g) + 3);
        g_free (g);
    }

    g_free (list_path);
    list_path = (gchar *) malloc (list_path_length + 1);
    memset (list_path, 0, list_path_length + 1);

    if(!list_path)
        return NULL;
    strcpy (list_path, "");
    for(tmp = selection_list; tmp; tmp = tmp->next) {
        record_entry_t *en;
        en = (record_entry_t *) tmp->data;
        strcat (list_path, "\"");
        gchar *g = g_path_get_basename (en->path);
        strcat (list_path, g);
        g_free (g);
        strcat (list_path, "\"");
        if(tmp->next)
            strcat (list_path, " ");
    }
    return (const gchar *)list_path;
}
#endif

// Determines the work directory of the command to be executed
// to do things like extract a tar file to a selected diretory
static gchar * 
autofunction_workdir (view_t * view_p, record_entry_t * en, const gchar * querypath) {
    widgets_t *widgets_p;
    widgets_p = &(view_p->widgets);
    static gchar *last_dir = NULL;
    const gchar *g;

    if(querypath) {
        gchar *f = g_build_filename (g_get_user_data_dir (), WORKDIR_DBH_FILE, NULL);
        const gchar *folder;
        if(view_p->en && view_p->en->path && g_file_test (view_p->en->path, G_FILE_TEST_IS_DIR)) {
            folder = view_p->en->path;
        } else {
            folder = g_get_home_dir ();
        }
        DBG ("COMBO: last_dir is %s\n", last_dir);
	const gchar *default_dir=last_dir;
	if (!default_dir) default_dir=en->path;
	if (!g_file_test(default_dir, G_FILE_TEST_IS_DIR) || 
		access(en->path, R_OK|W_OK|X_OK) < 0) {
	    default_dir=g_get_home_dir();
	}
        g = rodent_get_response_history (NULL, _(querypath), 
			    _("Path"), f, NULL,
                            default_dir,
                            NULL, NULL,
                            GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                            folder);
        g_free (f);

        if(!g)
            return NULL;
        if(g_file_test (g, G_FILE_TEST_EXISTS)
           && !g_file_test (g, G_FILE_TEST_IS_DIR)) {
            rfm_status (widgets_p, "xffm/warning", strerror (ENOTDIR), NULL);
            return NULL;
        }
        if(!g_file_test (g, G_FILE_TEST_EXISTS)) {
            gchar *b = g_strdup_printf ("%s: %s\n%s...", g,
                                        strerror (ENOENT),_("Create New Folder"));
            if(!rfm_confirm (widgets_p, GTK_MESSAGE_QUESTION, b, _("Cancel"), NULL)) {
                g_free (b);
                return NULL;
            }
            g_free (b);
            if(g_mkdir_with_parents (g, 0750) < 0) {
                rfm_confirm (widgets_p, GTK_MESSAGE_ERROR, strerror (errno), NULL, NULL);
                return NULL;
            }
        }

    
        g_free (last_dir);
        last_dir = g_strdup (g);
        DBG ("COMBO: setting last_dir to %s\n", last_dir);
        if(g_file_test(g,G_FILE_TEST_IS_DIR))
            rodent_save_workdir_history (view_p->workdir);
        else {
            rfm_show_text(widgets_p);
            rfm_diagnostics (widgets_p, "xffm/error", NULL);
            rfm_status (widgets_p, "xffm_tag/stderr", g, ": ", strerror (errno), "\n", NULL);
            return NULL;
        }
        return g_strdup(g);
    }

    return NULL;
}

static void
module_goto (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    const gchar *module_name = g_object_get_data (G_OBJECT (menuitem), "module_name");
    if(module_name) {
      if (view_p->child_constructor) {
	  gchar *m=g_strdup_printf("exec:%s",
		  (char *)rfm_natural ("rfm", module_name, NULL, "exec_name"));
            (*view_p->child_constructor) (widgets_p, (char *)m);
	  g_free(m);
      } else {
        rodent_push_view_go_history (view_p);
        record_entry_t *new_en = rfm_mk_entry (0);
        SET_ROOT_TYPE (new_en->type);
        new_en->module = module_name;
        new_en->path = rfm_natural ("rfm", module_name, NULL, "module_label");
        rodent_refresh (widgets_p, new_en);
      }
    } else
        g_warning ("cannot get path from module-goto menu");
    return;
}

static void
level_goto (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    const gchar *path;
    path = g_object_get_data (G_OBJECT (menuitem), "path");
    if(path) {
        if(view_p->child_constructor) {
            (*view_p->child_constructor) (widgets_p, (char *)path);
        } else {
            rodent_push_view_go_history (view_p);
            record_entry_t *new_en = rfm_stat_entry (path, __LOCAL_TYPE);
            view_p->module_name = NULL;
            rfm_save_to_go_history ((gchar *) path);
            rodent_refresh (widgets_p, new_en);
        }
    } else
        g_warning ("cannot get path from level-goto menu");
    return;
}
//ACO9003147HB
// autotype C is "open with" operator which is determined by
// means of the command associated data item of the widget
static void
execute_autotype (GtkMenuItem * menuitem, gpointer data) {
    const gchar *command;
    const gchar *dirname;
    gchar *new_command=NULL;
    const gchar *querypath,
     *output_arg,
     *output_ext;
    record_entry_t *en;
    view_t *view_p = g_object_get_data (G_OBJECT (menuitem), "view_p");
    widgets_t *widgets_p = &(view_p->widgets);

    en = view_p->en;

    output_arg = g_object_get_data (G_OBJECT (menuitem), "output_arg");
    command = g_object_get_data (G_OBJECT (menuitem), "command");
    dirname = g_object_get_data (G_OBJECT (menuitem), "workdir");
    querypath = g_object_get_data (G_OBJECT (menuitem), "querypath");
    output_ext = g_object_get_data (G_OBJECT (menuitem), "output_ext");

    TRACE("execute_autotype: %s\n", command);
    gint argc;
    gchar **argv;
    GError *error=NULL;

    if(!g_shell_parse_argv (command, &argc, &argv, &error)) {
        g_warning("execute_autotype(%s): %s", command, error->message);
        g_error_free (error);
        g_strfreev (argv);
        TRACE ("!g_shell_parse_argv at gui_autostuff\n");
        return;
    }

// internal commands:
    if (strcmp(argv[0], "rodent-newtab")==0){
        if(view_p->tab_constructor) {
            (*(view_p->tab_constructor))(widgets_p, argv[1]);
        }
        return;
    }
    // We don't use this any more:
    g_strfreev (argv);

    gchar *workdir=NULL;
    if(querypath) {
        workdir=autofunction_workdir (view_p, en, querypath);
        if (!workdir) return;
    }

    if (dirname) {// XXX I wonder what this is for?
        g_free (view_p->workdir);
        view_p->workdir = g_strdup (dirname);
    }
    if(output_ext) {
        gchar *basename = g_path_get_basename (output_arg);
        gchar *output_path=g_strconcat("\"",workdir, G_DIR_SEPARATOR_S, basename, output_ext,"\"",NULL);
        gchar *esc_path = g_strconcat("\"",basename,"\"",NULL);
        g_free(basename);
        basename=esc_path;

        gchar *s=strstr(command,"%s");
        if (s) {
        // XXX this should have at least two %s tokens...
            new_command = g_strdup_printf(command, output_path, basename);
        } else {
                DBG("no %%s in command");
            new_command = g_strconcat (command, " ", output_path, " ", basename, NULL);
        }
        command = (const gchar *)new_command;
        g_free (output_path);
        g_free (basename);
    } else if (workdir) {
        g_free (view_p->workdir);
        view_p->workdir = g_strdup (workdir); 
    }
    DBG ("output_arg=%s, command=%s, querypath(string)=%s, output_ext=%s, dirname=%s workdir=%s\n", output_arg, command, querypath,
           output_ext, dirname, view_p->workdir);

    rfm_thread_run2argv (widgets_p, command, FALSE);
        
    g_free (new_command);
    g_free (workdir);


}

static void
rodent_up_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!view_p) {
        g_warning ("!view_p");
    } else {
        if(view_p->population_pp && view_p->population_pp[0]) {
            if(view_p->population_pp[0]->en && !IS_DUMMY_TYPE (view_p->population_pp[0]->en->type))
                return;
            rodent_push_view_go_history (view_p);
            if(view_p->population_pp[0]->en)
                view_p->module_name = view_p->population_pp[0]->en->module;
            else
                view_p->module_name = NULL;
            record_entry_t *en = rfm_copy_entry (view_p->population_pp[0]->en);
            rodent_refresh (widgets_p, en);
        }
    }
}

#if 0
static const gchar *
fix_relative_path (view_t * view_p, const gchar * choice) {
    static gchar *fixed_choice = NULL;
    /* invalid stuff */
    if(!choice || !view_p)
        return NULL;
    g_free (fixed_choice);
    fixed_choice = NULL;
    if(!strlen (choice)) {
        fixed_choice = g_strdup (choice);
    } else if(strncmp (choice, "//", 2) == 0 || strncmp (choice, "smb://", 6) == 0) {
        /* network stuff */
        fixed_choice = g_strdup (choice);
    } else if(g_path_is_absolute (choice)
              && g_file_test (choice, G_FILE_TEST_IS_DIR)) {
        /* absolute path */
        fixed_choice = g_strdup (choice);
    } else if(view_p->en && view_p->en->path) {
        fixed_choice = g_build_filename (view_p->en->path, choice, NULL);
        if(!g_file_test (fixed_choice, G_FILE_TEST_IS_DIR)) {
            /* relative to homedir, maybe */
            g_free (fixed_choice);
            fixed_choice = g_build_filename (g_get_home_dir (), choice, NULL);
            if(!g_file_test (fixed_choice, G_FILE_TEST_IS_DIR)) {
                g_free (fixed_choice);
                fixed_choice = NULL;
            }
        }
    }
    return (const gchar *)fixed_choice;
}
#endif

/*********************** private callbacks ************************************/
#define RENAME_CASO 	0
#define DUPLICATE_CASO 	1
#define SYMLINK_CASO 	2

static gboolean
easy_data_test (view_t * view_p) {
    if(!view_p->en || !g_file_test (view_p->en->path, G_FILE_TEST_EXISTS) || !view_p->selected_p || !view_p->selected_p->en) {
        rfm_show_text(&(view_p->widgets));
        rfm_diagnostics (&(view_p->widgets), "xffm/error", NULL);
        rfm_diagnostics (&(view_p->widgets), "xffm_tag/stderr", strerror (EINVAL), "\n", NULL);
        return FALSE;
    }
    return TRUE;
}

static void
rodent_rename_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!easy_data_test (view_p)) {
        g_warning ("!easy_data_test");
        return;
    }
    if(!rfm_population_try_read_lock (view_p)) {
        g_warning ("cannot get readlock!");
        return;
    }

    rodent_mk_text_entry (view_p, view_p->selected_p, RENAME_CASO);
    rfm_population_read_unlock (view_p);
}

static void
rodent_duplicate_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!easy_data_test (view_p))
        return;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }

    rodent_mk_text_entry (view_p, view_p->selected_p, DUPLICATE_CASO);
    rfm_population_read_unlock (view_p);
}

static void
rodent_symlink_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!easy_data_test (view_p))
        return;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }
    rodent_mk_text_entry (view_p, view_p->selected_p, SYMLINK_CASO);
    rfm_population_read_unlock (view_p);
}

static void
rodent_unselect_all_view (view_t * view_p) {
    rfm_cursor_wait (view_p->widgets.window);
    rodent_unselect_all_pixbuf (view_p);
    rodent_unsaturate_pixbuf (view_p);
    rodent_expose_all (view_p);
    rfm_cursor_reset (view_p->widgets.window);
}

static int
rodent_select_all_view (view_t * view_p) {
    population_t **population_pp;
    int items = 0;
    if(!view_p || !view_p->en)
        return items;
    rfm_cursor_wait (view_p->widgets.window);
    population_pp = view_p->population_pp;
    for(; population_pp && *population_pp; population_pp++) {
        if((*population_pp)->en == NULL)
            continue;
        if(IS_DUMMY_TYPE ((*population_pp)->en->type))
            continue;
        items++;
        TRACE ("selected item %d", items);
        rodent_saturate_pixbuf (view_p, *population_pp);
        rodent_select_pixbuf (view_p, *population_pp);
    }
    view_p->saturated_p = NULL;
    rodent_expose_all (view_p);
    rfm_cursor_reset (view_p->widgets.window);
    return items;
}

enum {
    END_MATCH,
    START_MATCH,
    TOMATO_IN_SANDWICH_MATCH,
    BREAD_ON_SANDWICH_MATCH,
    EXACT_MATCH
};

static gint
rodent_select_byfilter_view (struct view_t *view_p, const gchar * select_filter) {
    population_t **population_pp;
    gint items = 0;
    if(!view_p || !view_p->en)
        return items;
    if(!select_filter || !strlen (select_filter))
        return items;
    if(strcmp (select_filter, "*") == 0)
        return rodent_select_all_view (view_p);
    rfm_cursor_wait (view_p->widgets.window);

    gchar *filter = g_strdup (select_filter);
    memset (filter, 0, strlen (select_filter));
    int j = 0,
        i = 0,
        caso;
    for(; i < strlen (select_filter); i++) {
        if(select_filter[i] != '*')
            filter[j++] = select_filter[i];
    }

    if(select_filter[0] == '*' && select_filter[strlen (select_filter) - 1] == '*') {
        caso = TOMATO_IN_SANDWICH_MATCH;
    } else if(select_filter[0] == '*') {
        caso = END_MATCH;
    } else if(select_filter[strlen (select_filter) - 1] == '*') {
        caso = START_MATCH;
    } else if(strchr (select_filter, (int)'*')) {
        caso = BREAD_ON_SANDWICH_MATCH;
    } else {
        caso = EXACT_MATCH;
    }

    rodent_unselect_all_pixbuf (view_p);
    rodent_unsaturate_pixbuf (view_p);
    population_pp = view_p->population_pp;
    for(; population_pp && *population_pp; population_pp++) {
        gchar *f;
        gboolean match = FALSE;
        if((*population_pp)->en == NULL)
            continue;
        if((*population_pp)->en->path == NULL)
            continue;
        if(IS_DUMMY_TYPE ((*population_pp)->en->type))
            continue;
        f = g_path_get_basename ((*population_pp)->en->path);
        gchar *p = strcasestr (f, filter);

        if(p && caso == END_MATCH) {
            gchar *pp;
            do {
                pp = p;
                p = strcasestr (p + 1, filter);
            }
            while(p);
            p = pp;
        }

        if(p) {
            DBG ("p=%s, filter=%s\n", p, filter);
            switch (caso) {
            case TOMATO_IN_SANDWICH_MATCH:
                match = TRUE;
                break;
            case END_MATCH:
                if(p[strlen (filter)] == 0)
                    match = TRUE;
                break;
            case START_MATCH:
                if(p == f)
                    match = TRUE;
                break;
            case EXACT_MATCH:
                if(strlen (p) == strlen (filter))
                    match = TRUE;
                break;
            case BREAD_ON_SANDWICH_MATCH:
                //not implemented. Need to do 2 matches.
                //break;
            default:           //START_END_MATCH
                match = TRUE;
                break;

            }
        }
        if(match) {
            items++;
            rodent_saturate_pixbuf (view_p, *population_pp);
            rodent_select_pixbuf (view_p, *population_pp);
        }
        g_free (f);
    }
    rfm_cursor_reset (view_p->widgets.window);
    gchar *plural_text=g_strdup_printf (ngettext ("%'u item", "%'u items",
                                               items),items);
    gchar *g = g_strdup_printf ("%s: %s", _("Selection"), plural_text);
    g_free(plural_text);

    rfm_status (&(view_p->widgets), "xffm/info", g, NULL);
    g_free (g);
    g_free (filter);

    return items;
}

static void
rodent_unselect_all_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!view_p)
        return;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }
    rodent_unselect_all_view (view_p);
    rfm_population_read_unlock (view_p);
}

static void
rodent_select_all_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    int items = 0;
    if(!view_p)
        return;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }

    rfm_status (&(view_p->widgets), "xffm/info", _("Select _All"), "...", NULL);
    rfm_cursor_wait (view_p->widgets.window);
    gdk_flush ();
    items = rodent_select_all_view (view_p);
    gchar *plural_text=g_strdup_printf (ngettext ("%'u item", "%'u items",
		g_list_length(view_p->selection_list)),g_list_length(view_p->selection_list));
//                                               view_p->selection_count),view_p->selection_count);
    gchar *g = g_strdup_printf ("%s: %s", _("Selection"), plural_text);
    g_free(plural_text);

    rfm_status (&(view_p->widgets), "xffm/info", g, NULL);
    g_free (g);
   
    rfm_cursor_reset (view_p->widgets.window);
    rfm_population_read_unlock (view_p);
}

static void
rodent_touch_activate (GtkMenuItem * menuitem, gpointer user_data) {
    record_entry_t *en = NULL;
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }
    if(view_p && view_p->selection_list && view_p->selection_list->data) {
        en = (record_entry_t *) view_p->selection_list->data;
    }

    if(!en || !g_file_test (en->path, G_FILE_TEST_EXISTS)) {
        rfm_show_text(&(view_p->widgets));
        rfm_diagnostics (&(view_p->widgets), "xffm/error", NULL);
        rfm_diagnostics (&(view_p->widgets), "xffm_tag/stderr", strerror (EINVAL), "\n", NULL);
        rfm_population_read_unlock (view_p);
        return;
    }
    rodent_touch (&(view_p->widgets), en->path);
    rfm_population_read_unlock (view_p);
    return;
}

/***   properties_callbacks  ***/

static void
rodent_on_prop_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    TRACE ("on_prop_activate view_p=0x%lx\n", (unsigned long)view_p);
    if(!view_p){
        return;
    }
    if (!view_p->selection_list || g_list_length (view_p->selection_list) < 1){
        return;
    }
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }
    if(view_p->module_name 
	    && 
	    rfm_rational ("rfm", view_p->module_name, widgets_p, view_p->selection_list, "do_properties")
	    ) {
        // module has done the properties thing here;
    } else {
        PROP_do_prop (widgets_p, view_p->selection_list);
    }
    // memory associated to  select_list_p is freed by properties thread.

    rfm_population_read_unlock (view_p);
}


static void
rodent_select_by_filter (GtkMenuItem * menuitem, gpointer user_data) {
    const gchar *filter_s;
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!view_p || !view_p->en || !view_p->en->path)
        return;
    gchar *b = g_strdup_printf ("%s/", view_p->en->path);
    gchar *label = rfm_utf_string (rfm_chop_excess (b));
    filter_s = rodent_get_response (&(view_p->widgets), _("Select Items Matching..."), label, "*");
    g_free (label);
    g_free (b);

    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }
    if(filter_s && strlen (filter_s)) {
        rodent_select_byfilter_view (view_p, filter_s);
    }
    rfm_population_read_unlock (view_p);
    return;
}

static void
rodent_hidden_toggled (GtkToggleButton * togglebutton, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p;
    view_p = widgets_p->view_p;
    if(!user_data)
        return;
    if(view_p->disable_toggle_refresh)
        return;
    view_p->preferences ^= __SHOW_HIDDEN;
    rodent_save_view_preferences (view_p);
    DBG ("rodent_refresh by rodent_hidden_toggled... %s\n", (view_p->preferences & __SHOW_HIDDEN) ? "ON" : "Off");
    record_entry_t *en = rfm_copy_entry (view_p->en);
    rodent_refresh (widgets_p, en);
}

static void
rodent_preview_toggled (GtkToggleButton * togglebutton, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p;
    if(!user_data)
        return;
    view_p = widgets_p->view_p;
    if(view_p->disable_toggle_refresh)
        return;
    view_p->preferences ^= __SHOW_IMAGES;
    rodent_save_view_preferences (view_p);
    DBG ("rodent_refresh by rodent_preview_toggled... %s\n", (view_p->preferences & __SHOW_IMAGES) ? "ON" : "Off");
    record_entry_t *en = rfm_copy_entry (view_p->en);
    rodent_refresh (widgets_p, en);
}

/* sort type radio: */
static void
rodent_on_radio (GtkCheckMenuItem * checkmenuitem, gpointer user_data) {
    view_t *view_p = (view_t *) g_object_get_data (G_OBJECT (checkmenuitem), "view_p");
    if(!view_p)
        return;
    if (view_p->type == DESKVIEW_TYPE) {
	rodent_erase_deskview_geometries ();
    }
    widgets_t *widgets_p = &(view_p->widgets);
    view_p->sortcolumn = GPOINTER_TO_INT (user_data);
    rodent_save_view_preferences (view_p);
    if(!view_p->disable_toggle_refresh) {
        record_entry_t *en = rfm_copy_entry (view_p->en);
        rodent_refresh (widgets_p, en);
    }
}

static void
on_ascending (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!view_p){
	return;
    }
    if (view_p->type == DESKVIEW_TYPE) {
	rodent_erase_deskview_geometries ();
    }
	
    view_p->preferences |= SORT_ASCENDING;
    rodent_save_view_preferences (view_p);
    if(!view_p->disable_toggle_refresh) {
        record_entry_t *en = rfm_copy_entry (view_p->en);
        rodent_refresh (widgets_p, en);
    }
    return;
}

static void
on_descending (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!view_p)
        return;
    if (view_p->type == DESKVIEW_TYPE) {
	rodent_erase_deskview_geometries ();
    }
    view_p->preferences &= (SORT_ASCENDING ^ 0xffffffff);
    rodent_save_view_preferences (view_p);
    if(!view_p->disable_toggle_refresh) {
        record_entry_t *en = rfm_copy_entry (view_p->en);
        rodent_refresh (widgets_p, en);
    }
    return;
}

static void
private_mount (view_t * view_p, int mounted) {
    record_entry_t *en = NULL;
    gchar *mount_point;
    widgets_t *widgets_p = &(view_p->widgets);
    if(view_p->selected_p && view_p->selected_p->en) {
        en = view_p->selected_p->en;
    }
    DBG ("MOUNT: private_mount %d (0x%lx)\n", mounted, (unsigned long)((long)view_p));
    if(!en || !en->path) {
        g_warning ("no entry to mount");
        return;
    }
    if(mounted){
        SET_MOUNTED_TYPE (en->type);
    } else{
        UNSET_MOUNTED_TYPE (en->type);
    }
    mount_point = (gchar *) FSTAB_fstab_mount (en, &(view_p->widgets));
    rodent_unselect_all_activate (NULL, (gpointer) widgets_p);
    rfm_cursor_reset (view_p->widgets.window);

    DBG ("MOUNT: private_mount done\n");
    return;
}

static void
rodent_mount (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }
    private_mount (view_p, FALSE);
    rfm_population_read_unlock (view_p);
    return;
}

static void
rodent_unmount (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }
    private_mount (view_p, TRUE);
    rfm_population_read_unlock (view_p);
    return;
}

/*   pasteboard callbacks   */

static void
private_paste (widgets_t * widgets_p, view_t * view_p, gboolean symlink) {
    gboolean cut;
    GList *list = NULL;
    record_entry_t *t_en = NULL;
    int i;

    /*if(!rfm_details->pastepath) return;
    TRACE ("TRACE:pastepath=%s\n", rfm_details->pastepath);*/
    
    if((t_en = view_p->en) == NULL)
        return;


    if((i = gui_pasteboard_list (&list)) == 0)
        return;
    if(i == 1)
        cut = TRUE;
    else
        cut = FALSE;
    i = gui_pasteboard_transfer (widgets_p, t_en, list, cut, symlink);

    list = rfm_uri_free_list (list);

    if(i) {
        if(cut)
            XStoreBuffer (GDK_DISPLAY (), "", 1, CUT_BUFFER);
    }
    return;
}

static void
copy_cut_callback (gboolean cut, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!view_p->selection_list)
        return;

    gui_pasteboard_copy_cut (widgets_p, cut, &(view_p->selection_list));
    // unselect stuff if it is cut or copied.
    g_list_free (view_p->selection_list);
    view_p->selection_list = NULL;

    if(rfm_population_try_read_lock (view_p)) {
        population_t **pp;
        for(pp = view_p->population_pp; pp && *pp; pp++) {
            population_t *population_p = *pp;
            population_p->selected = FALSE;
        }
        rfm_population_read_unlock (view_p);
    }
    // force a icon reload for the monitor thread:
    // not really necesary since the change in pasteboard
    // is detected and will cause an icon update without
    // a reload.
    // rfm_soft_reload(view_p);
}

static void
rodent_cut_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }
    copy_cut_callback (TRUE, user_data);
    rfm_population_read_unlock (view_p);
    return;
}

static void
rodent_copy_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }
    copy_cut_callback (FALSE, user_data);
    rfm_population_read_unlock (view_p);
    return;
}

static void
paste_callback (gboolean symlink, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!view_p->en)
        return;
    //rfm_details->pastepath = view_p->en->path;
    private_paste (widgets_p, view_p, symlink);
}

static void
rodent_paste_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }
    paste_callback (FALSE, user_data);
    rfm_population_read_unlock (view_p);
    return;
}

/*static void rodent_pastelink_activate(GtkMenuItem * menuitem, gpointer data)
{
    paste_callback(TRUE, data);
    return;
}*/


static void
rodent_run_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!rfm_population_try_read_lock (view_p)) {
        return;
    }
    execute (menuitem, user_data, FALSE);
    rfm_population_read_unlock (view_p);
}

static void
rodent_pop_view_go_history (view_t * view_p) {
    GList *last;
    if(!view_p->go_list)
        return;
    last = g_list_last (view_p->go_list);
    if(!last) {
        g_list_free (view_p->go_list);
        view_p->go_list = NULL;
        return;
    }
    view_p->go_list = g_list_remove (view_p->go_list, last->data);
    if(!g_list_length (view_p->go_list)) {
        g_list_free (view_p->go_list);
        view_p->go_list = NULL;
        return;
    }
    return;
}

static void
rodent_goup_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!view_p) {
        g_warning ("!view_p");
        return;
    }
    if (view_p->protected) {
        rfm_diagnostics(widgets_p, "xffm/warning",NULL);
        rfm_diagnostics(widgets_p, "xffm_tag/stderr",_("Go Up"),": ",strerror(EBUSY),"\n",NULL);
        return;
    }
    if(!view_p->en || !view_p->en->path)
        return;
    gchar *b = g_path_get_dirname (view_p->en->path);
    if(!g_file_test (b, G_FILE_TEST_IS_DIR))
        return;
    if(view_p->child_constructor) {
        (*view_p->child_constructor) (widgets_p, b);
    } else {
        record_entry_t *en = rfm_stat_entry ((gchar *) b, __ROOT_TYPE | __LOCAL_TYPE);
        rodent_push_view_go_history (view_p);
        view_p->module_name = NULL;
        rodent_refresh (widgets_p, en);
    }
    g_free (b);
}

////////////////////  rodent_popup.i

static void
  rodent_mk_module_popup_menu (const gchar * module_name, view_t * view_p, record_entry_t * en);
static void execute_autotype (GtkMenuItem * menuitem, gpointer user_data);
static void rodent_mount (GtkMenuItem * menuitem, gpointer user_data);
static void rodent_unmount (GtkMenuItem * menuitem, gpointer user_data);

static void
rodent_save_workdir_history (char *p) {
    gchar *f = g_build_filename (g_get_user_data_dir (), WORKDIR_DBH_FILE, NULL);
    COMBOBOX_save_to_history (f, p);
    g_free (f);
}
/***********************************************************************************/

static void
remove_menu_item (GtkWidget * widget, gpointer data) {
    TRACE ("removing 0x%lx\n", (unsigned long)widget);
    gtk_widget_destroy (widget);
}

static void
close_menu_item_activate (GtkMenuItem * menuitem, gpointer data) {
    widgets_t *widgets_p = data;
    view_t *view_p = widgets_p->view_p;
    rodent_destroy_view_item (view_p);
    return;
}

static void
mk_module_menu_items (widgets_t * widgets_p, GtkWidget * parent, GtkAccelGroup * accel_group, GtkWidget * item22_menu) {
    GSList *list,
     *plugin_list = rfm_find_root_plugins ();
    int module_number = 0;
    view_t *view_p = widgets_p->view_p;
    for(list = plugin_list; list; list = list->next) {
        if(!rfm_void ("rfm", (gchar *) (list->data), "is_root_module")) {
            continue;
        }

        if(++module_number > DEEPEST_MODULE_MENU_LEVELS)
            break;
        gchar *name = g_strdup ((gchar *) (list->data));
        const gchar *icon = rfm_natural ("rfm", name, NULL, "module_icon_id");
        if(!icon)
            icon = "xffm/stock_go-forward";
        gchar *label = rfm_natural ("rfm", name, NULL, "module_label");
        if(!label)
            label = g_strdup (name);
        gui_add_menu_item (widgets_p, item22_menu, label, icon, name, module_goto, widgets_p, NULL, 0, 0);
        g_free (label);
        GtkWidget *w = rodent_get_widget_by_name (view_p->widgets.paper, name);
        g_object_set_data (G_OBJECT (w), "module_name", (gpointer) name);
    }
}

static void
mk_directory_menu_items (widgets_t * widgets_p, GtkWidget * parent, GtkAccelGroup * accel_group, GtkWidget * item22_menu) {
    int level;
    for(level = 0; level < DEEPEST_DIR_MENU_LEVELS; level++) {
        char *name = g_strdup_printf ("level-%d", level);
        gui_add_menu_item (widgets_p, item22_menu, name, "xffm/stock_jump-to", name, level_goto, widgets_p, NULL, 0, 0);
    }

}

static void
mk_navigation_menu_items (widgets_t * widgets_p, GtkWidget * parent, GtkAccelGroup * accel_group, GtkWidget * item22_menu) {
    GtkWidget *w;

    gui_add_menu_item (widgets_p,
                       item22_menu,
                       _("Go Up"),
                       "xffm/stock_go-up",
                       "goup_menuitem", (gpointer) rodent_goup_activate, widgets_p, accel_group, GDK_Up, GDK_CONTROL_MASK);

    gui_add_menu_item (widgets_p, item22_menu,
                       _("Go back"),
                       "xffm/stock_go-back", "go_back_menuitem",
                       (gpointer) rodent_back_activate, widgets_p, accel_group, GDK_Left, GDK_CONTROL_MASK);
    gui_add_menu_item (widgets_p, item22_menu,
                       _("Go To"),
                       "xffm/stock_jump-to", "goto2",
                       (gpointer) rodent_jump_to_activate, widgets_p, accel_group, GDK_Right, GDK_CONTROL_MASK);

    GtkWidget *goto_menu;
    goto_menu = rodent_mk_menu (widgets_p, _("_Go"),     /* label */
                                "goto_menu",    /* name */
                                item22_menu,    /* parent */
                                NULL,   /* callback (or NULL) */
                                NULL);  /* iconfile (or NULL) */
    gui_add_menu_item (widgets_p,
                       goto_menu,
                       _("Reload"),
                       "xffm/stock_refresh",
                       "refresh3", (gpointer) rodent_refresh_activate, widgets_p, accel_group, GDK_Down, GDK_CONTROL_MASK);
    gui_add_menu_item (widgets_p, goto_menu,
                       g_get_host_name (),
                       "xffm/stock_system", "host2",
                       (gpointer) rodent_host_activate, widgets_p, accel_group, GDK_End, GDK_CONTROL_MASK);
    gui_add_menu_item (widgets_p, goto_menu,
                       _("Home"),
                       "xffm/stock_home", "home2",
                       (gpointer) rodent_home_activate, widgets_p, accel_group, GDK_Home, GDK_CONTROL_MASK);

    // module elements:
    mk_module_menu_items (widgets_p, parent, accel_group, goto_menu);
    // navigation bar equivalent. Subject to revision on merits...
    mk_directory_menu_items (widgets_p, parent, accel_group, goto_menu);

    /* shift mask: */
    /*add_to_keylist(accel_key, accel_mask,  callback,  callback_data); */
    gui_add_to_keylist (GDK_Up, GDK_CONTROL_MASK, (gpointer) rodent_up_activate);

    /* directory menu-items */

    w = gui_add_menu_separator (item22_menu);
    RODENT_HOOKUP_OBJECT (widgets_p->paper, w, "navigation_separator");
    widgets_p->navigation_popup = goto_menu;

}

static void
mk_view_menu_items (widgets_t * widgets_p, GtkWidget * parent, GtkAccelGroup * accel_group, GtkWidget * item22_menu) {
    GtkWidget *w;
    //view_t *view_p=widgets_p->view_p;
    //

    GtkWidget *sort1_menu;
    sort1_menu = rodent_mk_menu (widgets_p, _("Sort By"),  /* label */
                                 "sort1",       /* name */
                                 item22_menu,   /* parent */
                                 NULL,  /* callback (or NULL) */
                                 "xffm/convert");       /* iconfile (or NULL) */

    gui_add_menu_item (widgets_p, sort1_menu,
                       _("Descending"),
                       "xffm/stock_sort-ascending", "descending1", (gpointer) on_ascending, widgets_p, NULL, 0, 0);
    gui_add_menu_item (widgets_p, sort1_menu,
                       _("Ascending"),
                       "xffm/stock_sort-descending", "ascending1", (gpointer) on_descending, widgets_p, NULL, 0, 0);

    {
        GSList *radiogroup = NULL;
        GtkWidget *unsorted6,
         *name6,
         *size6,
         *date6,
         *owner6,
         *group6,
         *mode6;

        unsorted6 = gui_mk_radio_button (widgets_p, _("Type"), "unsorted6", sort1_menu, &radiogroup);
        name6 = gui_mk_radio_button (widgets_p, _("By Name"), "name6", sort1_menu, &radiogroup);
        size6 = gui_mk_radio_button (widgets_p, _("By Size"), "size6", sort1_menu, &radiogroup);
        date6 = gui_mk_radio_button (widgets_p, _("By Date"), "date6", sort1_menu, &radiogroup);
        owner6 = gui_mk_radio_button (widgets_p, _("By Owner"), "owner6", sort1_menu, &radiogroup);
        group6 = gui_mk_radio_button (widgets_p, _("By Group"), "group6", sort1_menu, &radiogroup);
        mode6 = gui_mk_radio_button (widgets_p, _("By Permissions"), "mode6", sort1_menu, &radiogroup);
        {
            view_t *view_p = widgets_p->view_p;
            if(view_p->sortcolumn == TYPE_SORT)
                gtk_check_menu_item_set_active ((GtkCheckMenuItem *) unsorted6, TRUE);
            else if(view_p->sortcolumn == NAME_SORT)
                gtk_check_menu_item_set_active ((GtkCheckMenuItem *) name6, TRUE);
            else if(view_p->sortcolumn == SIZE_SORT)
                gtk_check_menu_item_set_active ((GtkCheckMenuItem *) size6, TRUE);
            else if(view_p->sortcolumn == DATE_SORT)
                gtk_check_menu_item_set_active ((GtkCheckMenuItem *) date6, TRUE);
            else if(view_p->sortcolumn == OWNER_SORT)
                gtk_check_menu_item_set_active ((GtkCheckMenuItem *) owner6, TRUE);
            else if(view_p->sortcolumn == GROUP_SORT)
                gtk_check_menu_item_set_active ((GtkCheckMenuItem *) group6, TRUE);
            else if(view_p->sortcolumn == MODE_SORT)
                gtk_check_menu_item_set_active ((GtkCheckMenuItem *) mode6, TRUE);

            g_signal_connect ((gpointer) unsorted6, "toggled", G_CALLBACK (rodent_on_radio), GINT_TO_POINTER (TYPE_SORT));
            g_object_set_data (G_OBJECT (unsorted6), "view_p", view_p);

            g_signal_connect ((gpointer) name6, "toggled", G_CALLBACK (rodent_on_radio), GINT_TO_POINTER (NAME_SORT));
            g_object_set_data (G_OBJECT (name6), "view_p", view_p);

            g_signal_connect ((gpointer) size6, "toggled", G_CALLBACK (rodent_on_radio), GINT_TO_POINTER (SIZE_SORT));
            g_object_set_data (G_OBJECT (size6), "view_p", view_p);

            g_signal_connect ((gpointer) date6, "toggled", G_CALLBACK (rodent_on_radio), GINT_TO_POINTER (DATE_SORT));
            g_object_set_data (G_OBJECT (date6), "view_p", view_p);

            g_signal_connect ((gpointer) owner6, "toggled", G_CALLBACK (rodent_on_radio), GINT_TO_POINTER (OWNER_SORT));
            g_object_set_data (G_OBJECT (owner6), "view_p", view_p);

            g_signal_connect ((gpointer) group6, "toggled", G_CALLBACK (rodent_on_radio), GINT_TO_POINTER (GROUP_SORT));
            g_object_set_data (G_OBJECT (group6), "view_p", view_p);

            g_signal_connect ((gpointer) mode6, "toggled", G_CALLBACK (rodent_on_radio), GINT_TO_POINTER (MODE_SORT));
            g_object_set_data (G_OBJECT (mode6), "view_p", view_p);
        }

    }

    /* end of sort menu items */

    w = gtk_check_menu_item_new_with_mnemonic (_("Show hidden files"));
    gtk_widget_show (w);
    gtk_container_add (GTK_CONTAINER (item22_menu), w);
    RODENT_HOOKUP_OBJECT (widgets_p->paper, w, "show_hidden_menuitem");
    g_signal_connect ((gpointer) w, "toggled", G_CALLBACK (rodent_hidden_toggled), widgets_p);

    w = gtk_check_menu_item_new_with_mnemonic (_("Show preview"));
    gtk_container_add (GTK_CONTAINER (item22_menu), w);
    RODENT_HOOKUP_OBJECT (widgets_p->paper, w, "preview_images_menuitem");
    gtk_widget_show (w);
    g_signal_connect ((gpointer) w, "toggled", G_CALLBACK (rodent_preview_toggled), widgets_p);

    w = gui_add_menu_separator (item22_menu);
    RODENT_HOOKUP_OBJECT (widgets_p->paper, w, "view_separator");

}

static void
mk_file_menu_items (widgets_t * widgets_p, GtkWidget * parent, GtkAccelGroup * accel_group, GtkWidget * item22_menu) {
    GtkWidget *w;
    DBG ("POPUPx: mk_file_menu_items...\n");
    //view_t *view_p=widgets_p->view_p;
    {
        gchar *openwith=g_strdup_printf(_("Open with %s"),"...");
        gchar *string[4] = { openwith, "", "", NULL };
        gchar *iconfile[4] = { "xffm/stock_open", "xffm/stock_run", "xffm/stock_run", NULL };
        gchar *id[4] = { "open_with_menuitem", "autotype_R", "autotype_Prun", NULL };
        gpointer callback[4] = { (gpointer) on_open_with_activate, (gpointer) on_autotype_R, (gpointer) on_open_exec, NULL };
        guint akey[4] = { GDK_o, 0, 0, 0 };
        int i;

        for(i = 0; string[i]; i++) {
            gui_add_menu_item (widgets_p, item22_menu,
                               _(string[i]),
                               iconfile[i],
                               id[i], callback[i], widgets_p, ((akey[i]) ? accel_group : NULL), akey[i], GDK_CONTROL_MASK);
        }
    }
    w = gui_add_menu_separator (item22_menu);
    RODENT_HOOKUP_OBJECT (widgets_p->paper, w, "file_separator2");
    {
        gchar *string[] = { N_("Properties"), 
	    N_("Open terminal"),
            N_("Execute Shell Command"), 
	    N_("Search for"), //"Find..." "Search"
            N_("Compare files (diff)"), 
            N_("Mount the volume associated with this folder"), 
            N_("Unmount the volume associated with this folder"),
            N_("Duplicate each selected item"), //N_("D_uplicate"), 
	    N_("Create Symbolic Link"),   //N_("_Link Here"),
            N_("Touch")" (touch)", 
	    N_("Rename"),
            NULL
        };
        gchar *iconfile[] = { "xffm/stock_properties", "xffm/x-xterm",
            "xffm/stock_run", "xffm/stock_search",
            "xffm/x-rodent-diff", "xffm/stock_yes", "xffm/stock_no",
            "xffm/stock_dnd-multiple", "xffm/stock_convert",
            "xffm/stock_index", "xffm/stock_save-as",
            NULL
        };
        gchar *id[] = { "properties2", "terminal2",
            "run2", "find2",
            "differences2", "mountP", "unmountP",
            "duplicate_menuitem", "symlink_menuitem",
            "touch_menuitem", "rename_menuitem",
            NULL
        };
        gpointer callback[] = { (gpointer) rodent_on_prop_activate,
            (gpointer) rodent_open_in_terminal_activate,
            (gpointer) rodent_run_activate,
            (gpointer) rodent_glob_activate,
            (gpointer) rodent_differences_activate, NULL, NULL,
            (gpointer) rodent_duplicate_activate,
            (gpointer) rodent_symlink_activate,
            (gpointer) rodent_touch_activate,
            (gpointer) rodent_rename_activate,
            NULL
        };
        guint akey[] = { GDK_p, GDK_t,
            GDK_e, GDK_f,
            GDK_h, 0, 0,
            GDK_d, GDK_s,
            GDK_t, GDK_r,
            0
        };
        guint amask[] = { GDK_CONTROL_MASK, GDK_CONTROL_MASK,
            GDK_CONTROL_MASK, GDK_CONTROL_MASK,
            GDK_CONTROL_MASK, 0, 0,
            GDK_CONTROL_MASK, GDK_CONTROL_MASK,
            GDK_MOD1_MASK, GDK_CONTROL_MASK,
            0
        };
        int i;

        for(i = 0; string[i]; i++) {
            gui_add_menu_item (widgets_p, item22_menu,
                               _(string[i]),
                               iconfile[i], id[i], callback[i], widgets_p, ((akey[i]) ? accel_group : NULL), akey[i], amask[i]);
        }
    }

    GtkWidget *new_menu;
    new_menu = rodent_mk_menu (widgets_p, _("New"),     /* label */
                               "new_menu",      /* name */
                               item22_menu,     /* parent */
                               NULL,    /* callback (or NULL) */
                               "xffm/convert"); /* iconfile (or NULL) */

    {
        gchar *string[4] = { N_("Create new file in the given directory"), 
	    N_("Create a new empty folder inside this folder"),
            N_("Open in New Window"),
            NULL
        };
        gchar *iconfile[4] = { "xffm/stock_new", "xffm/stock_new-dir",
            "xffm/stock_book-open", NULL
        };
        gchar *id[4] = { "newfile_menuitem", "newdirectory_menuitem",
            "newwindow",
            NULL
        };
        gpointer callback[4] = { (gpointer) on_newfile_activate, (gpointer) on_newdir_activate,
            (gpointer) on_view_activate, NULL
        };
        guint akey[4] = { GDK_Insert, GDK_Insert, GDK_Page_Up, 0 };
        guint amask[5] = { GDK_CONTROL_MASK, GDK_MOD1_MASK,
            GDK_CONTROL_MASK,
            0
        };
        int i;

        for(i = 0; string[i]; i++) {
            gui_add_menu_item (widgets_p, new_menu,
                               _(string[i]),
                               iconfile[i], id[i], callback[i], widgets_p, ((akey[i]) ? accel_group : NULL), akey[i], amask[i]);
        }
    }

    w = gui_add_menu_separator (item22_menu);
    RODENT_HOOKUP_OBJECT (widgets_p->paper, w, "file_separator");

}

static void
mk_edit_menu_items (widgets_t * widgets_p, GtkWidget * parent, GtkAccelGroup * accel_group, GtkWidget * item22_menu) {
    GtkWidget *w;
    //view_t *view_p=widgets_p->view_p;
    {
        gchar *string[6] = { N_("Cut"), 
	    N_("Copy"), 
	    N_("Paste"),
            N_("Remove"),
            NULL
        };
        gchar *iconfile[6] = { "xffm/stock_cut", "xffm/stock_copy", "xffm/stock_paste",
            "xffm/stock_delete", NULL
        };
        gchar *id[6] = { "cut_menuitem", "copy_menuitem", "paste_menuitem",
            "remove_menuitem", NULL
        };
        gpointer callback[6] = { (gpointer) rodent_cut_activate, (gpointer) rodent_copy_activate,
            (gpointer) rodent_paste_activate,
            (gpointer) rodent_remove_activate, NULL
        };
        guint akey[6] = { GDK_x, GDK_c, GDK_v, 0, GDK_Delete, 0 };
        guint amask[6] = { GDK_CONTROL_MASK, GDK_CONTROL_MASK, GDK_CONTROL_MASK,
            0, 0
        };
        int i;

        for(i = 0; string[i]; i++) {
            gui_add_menu_item (widgets_p, item22_menu,
                               _(string[i]),
                               iconfile[i], id[i], callback[i], widgets_p, ((akey[i]) ? accel_group : NULL), akey[i], amask[i]);
        }
    }
    GtkWidget *select_menu;
    select_menu = rodent_mk_menu (widgets_p, _("Select"),       /* label */
                                  "select_menu",        /* name */
                                  item22_menu,  /* parent */
                                  NULL, /* callback (or NULL) */
                                  NULL);        /* iconfile (or NULL) */

    {
        gchar *string[4] = { N_("Select _All"), 
	    N_("Select Items Matching..."), 
	    N_("Unselect"),
            NULL
        };
        gchar *iconfile[4] = { "xffm/stock_add", "xffm/stock_apply", "xffm/stock_remove", NULL };
        gchar *id[4] = { "select_all_menuitem", "select_byfilter_menuitem",
            "unselect_all_menuitem", NULL
        };
        gpointer callback[4] = { (gpointer) rodent_select_all_activate,
            (gpointer) rodent_select_by_filter,
            (gpointer) rodent_unselect_all_activate, NULL
        };
        guint akey[4] = { GDK_a, GDK_s, GDK_u, 0 };
        guint amask[4] = { GDK_MOD1_MASK, GDK_MOD1_MASK, GDK_MOD1_MASK, 0 };

        int i;

        for(i = 0; string[i]; i++) {
            gui_add_menu_item (widgets_p, select_menu,
                               _(string[i]),
                               iconfile[i], id[i], callback[i], widgets_p, ((akey[i]) ? accel_group : NULL), akey[i], amask[i]);
        }
    }

    w = gui_add_menu_separator (item22_menu);
    RODENT_HOOKUP_OBJECT (widgets_p->paper, w, "edit_separator");
}

static GtkWidget *
common_mk_popup_menu (widgets_t * widgets_p, GtkWidget * parent, GtkAccelGroup * accel_group) {
    /*GtkAccelGroup *accel_group; */
    GtkWidget *w;
    GtkWidget *item22_menu;
    //view_t *view_p=widgets_p->view_p;

    if(!accel_group) {
        accel_group = gtk_accel_group_new ();
        gtk_window_add_accel_group (GTK_WINDOW (widgets_p->window), accel_group);
    }

    item22_menu = rodent_mk_menu (widgets_p, "item22",  /* label */
                                  "item22",     /* name */
                                  parent,       /* parent */
                                  NULL, NULL);  /* callback (or NULL) */
    w = rodent_mk_menu (widgets_p, "",  /* label */
                        "module1",      /* name */
                        item22_menu,    /* parent */
                        NULL,   /* callback (or NULL) */
                        NULL);  /* iconfile (or NULL) */

    /* navigation menu-items */
    mk_navigation_menu_items (widgets_p, parent, accel_group, item22_menu);

    /* file menu-items */
    mk_file_menu_items (widgets_p, parent, accel_group, item22_menu);

    /* view menu-items */
    mk_view_menu_items (widgets_p, parent, accel_group, item22_menu);

    /* edit menu items */
    mk_edit_menu_items (widgets_p, parent, accel_group, item22_menu);

    /* general group of menuitems: */

    gui_add_menu_item (widgets_p, item22_menu,
                       _("Personal settings"),
                       "xffm/stock_preferences",
                       "settings2", (gpointer) on_settings_show, widgets_p, accel_group, GDK_p, GDK_MOD1_MASK);

    gui_add_menu_item (widgets_p, item22_menu,
                       _("_About"),
                       "xffm/stock_about",
                       "about2", (gpointer) gui_on_about_activate, widgets_p, accel_group, GDK_i, GDK_MOD1_MASK);

    gui_add_menu_item (widgets_p, item22_menu,
                       _("Close"),
                       "xffm/stock_close",
                       "exit2", (gpointer) close_menu_item_activate, widgets_p, accel_group, GDK_q, GDK_CONTROL_MASK);

    RODENT_HOOKUP_OBJECT (widgets_p->paper, item22_menu, "item22_menu");
    return item22_menu;

}

/////////////////////////////////////////////77
//////  rodent_menu

/*************************** specific commands to menu ***************/

static int
rodent_diff (widgets_t * widgets_p, char *file1, char *file2) {
    gchar *argv[4];
    int i = 0;
    argv[i++] = "rodent-diff";
    if(file1)
        argv[i++] = file1;
    if(file2)
        argv[i++] = file2;
    argv[i] = NULL;

    rfm_thread_run_argv (widgets_p, argv, FALSE);
    return TRUE;
}

/****************** gtk functions for callbacks *********************/
// mount_item: shows the menu mount item if fstab module available
static void
mount_item (view_t * view_p, population_t * population_p) {
    if(g_list_length (view_p->selection_list) != 1)
        return;
    if(rfm_void ("rfm", "fstab", "is_root_module") == NULL)
        return;

    gboolean mounted = FALSE;
    gboolean in_fstab = FALSE;
    gboolean partition = FSTAB_is_partition_type (population_p->en);
    if(partition || g_file_test (population_p->en->path, G_FILE_TEST_IS_DIR)) {
        mounted = FSTAB_is_mounted (population_p->en->path);
        in_fstab = FSTAB_is_in_fstab (population_p->en->path);
    }
    DBG ("POPUPx: %s mounted=%d, in_fstab=%d, partition=%d, IS_MOUNTED(population_p->en->type)= %d\n", population_p->en->path,
         mounted, in_fstab, partition, IS_MOUNTED (population_p->en->type));

    DBG ("POPUPx  %s is mounted=%d\n", population_p->en->path, mounted);
    gboolean do_mount_item = mounted || in_fstab || partition;
    record_entry_t *en = population_p->en; 
    if ((en->mimetype && strstr(en->mimetype, "application/x-iso9660-image")) ||
	(en->mimemagic && strstr(en->mimemagic, "application/x-iso9660-image"))){
	do_mount_item=TRUE;
    }
    if(do_mount_item) {
        if(mounted)
            SHOW_IT (view_p->widgets.paper, "unmountP");
        else
            SHOW_IT (view_p->widgets.paper, "mountP");
    }
}

static void
remove_item (view_t * view_p, population_t * population_p) {
    if(!view_p || !population_p || !population_p->en || !population_p->en->path)
        return;
    /* for local remove: */
    if(IS_BROKEN_LNK (population_p->en->type)
       || g_file_test (population_p->en->path, G_FILE_TEST_EXISTS)
       ) {
        SHOW_IT (view_p->widgets.paper, "remove_menuitem");
    }
}

static void
properties_item (view_t * view_p, population_t * population_p) {
    if(!view_p || !population_p || !population_p->en || !population_p->en->path)
        return;

    /*if (view_p->module_name) {
       if (rfm_natural("rfm", view_p->module_name, population_p->en, "properties2"))  SHOW_IT(view_p->widgets.paper, "properties2");
       return;
       } */

    /* for local remove: */
    if(IS_BROKEN_LNK (population_p->en->type)
       || g_file_test (population_p->en->path, G_FILE_TEST_EXISTS)) {
        SHOW_IT (view_p->widgets.paper, "properties2");
    }
    if(g_file_test (population_p->en->path, G_FILE_TEST_IS_DIR)) {
        SHOW_IT (view_p->widgets.paper, "find2");
    }
}

static void
recursive_dirname (widgets_t * widgets_p, char *path, int level) {
    view_t *view_p = widgets_p->view_p;
    if(!path || strcmp (path, "/") == 0)
        return;
    if(level >= DEEPEST_DIR_MENU_LEVELS)
        return;
    char *b = g_path_get_dirname (path);
    char *name = g_strdup_printf ("level-%d", level);

    GtkWidget *a = rodent_get_widget_by_name (widgets_p->paper, name);
    GtkWidget *label = gtk_bin_get_child (GTK_BIN (a));
    gchar *q = rfm_utf_string (b);
    gtk_label_set_text ((GtkLabel *) label, q);
    g_free (q);
    char *old_b;
    if((old_b = g_object_get_data (G_OBJECT (a), "path")) != NULL) {
        g_object_set_data (G_OBJECT (a), "path", NULL);
        g_free (old_b);
    }
    g_object_set_data (G_OBJECT (a), "path", (gpointer) b);
    SHOW_IT (view_p->widgets.paper, name);

    recursive_dirname (widgets_p, b, level + 1);
    g_free (name);
    return;
}

static void
refresh_item (view_t * view_p, population_t * population_p) {
    SHOW_IT (view_p->widgets.paper, "goto_menu");
    if(view_p->go_list)
        SHOW_IT (view_p->widgets.paper, "go_back_menuitem");
    else
        HIDE_IT (view_p->widgets.paper, "go_back_menuitem");
    SHOW_IT (view_p->widgets.paper, "goto2");
    SHOW_IT (view_p->widgets.paper, "host2");
    SHOW_IT (view_p->widgets.paper, "home2");
    int i;
    for(i = 0; i < DEEPEST_DIR_MENU_LEVELS; i++) {
        char *name = g_strdup_printf ("level-%d", i);
        HIDE_IT (view_p->widgets.paper, name);
        g_free (name);
    }
    /* this shows valid directories to goto */
    recursive_dirname (&(view_p->widgets), view_p->deepest_dir, 0);


    SHOW_IT (view_p->widgets.paper, "navigation_separator");

    gchar **w,
     *widget_name[] = { "show_hidden_menuitem", "preview_images_menuitem", "new_menu", NULL };
    SHOW_IT (view_p->widgets.paper, "refresh3");
    if(view_p->en && view_p->en->path) {
        gchar *d = g_path_get_dirname (view_p->en->path);
        if(g_file_test (d, G_FILE_TEST_IS_DIR)) {
            SHOW_IT (view_p->widgets.paper, "goup_menuitem");
        }
        g_free (d);
    }

    for(w = widget_name; *w; w++) {
        if(view_p->en && view_p->en->path) {
            if(view_p->module_name) {
                if(rfm_void ("rfm", view_p->module_name, *w))
                    SHOW_IT (view_p->widgets.paper, *w);
            } else if(g_file_test (view_p->en->path, G_FILE_TEST_IS_DIR)) {
                SHOW_IT (view_p->widgets.paper, *w);
            }
        }
    }

    SHOW_IT (view_p->widgets.paper, "view_separator");

    SHOW_IT (view_p->widgets.paper, "sort1");

    if(view_p->preferences & SORT_ASCENDING){
        HIDE_IT (view_p->widgets.paper, "descending1");
        SHOW_IT (view_p->widgets.paper, "ascending1");
    } else {
        HIDE_IT (view_p->widgets.paper, "ascending1");
        SHOW_IT (view_p->widgets.paper, "descending1");
    }

    SHOW_IT (view_p->widgets.paper, "run2");
    if(rfm_valid_pasteboard (view_p) && view_p->en && view_p->en->path) {
        SHOW_IT (view_p->widgets.paper, "paste_menuitem");
    }

    if(view_p->constructor) {
        SHOW_IT (view_p->widgets.paper, "newwindow");

    }

    SHOW_IT (view_p->widgets.paper, "terminal2");
    SHOW_IT (view_p->widgets.paper, "find2");
    SHOW_IT (view_p->widgets.paper, "differences2");
    SHOW_IT (view_p->widgets.paper, "file_separator");

    if(view_p->en) {
        SHOW_IT (view_p->widgets.paper, "select_menu");
        SHOW_IT (view_p->widgets.paper, "select_all_menuitem");
        SHOW_IT (view_p->widgets.paper, "select_byfilter_menuitem");
        SHOW_IT (view_p->widgets.paper, "unselect_all_menuitem");
        SHOW_IT (view_p->widgets.paper, "edit_separator");
    }
    SHOW_IT (view_p->widgets.paper, "settings2");
    SHOW_IT (view_p->widgets.paper, "about2");
    //if (view_p->type != DESKVIEW_TYPE) 
    SHOW_IT (view_p->widgets.paper, "exit2");

}

#define MENU_IN_MODULE(x) (view_p->module_name && rfm_void("rfm",view_p->module_name,(x)))

static void
open_item (view_t * view_p, population_t * population_p) {
    DBG ("POPUPx: open_item...\n");
    gchar *symbols[] = {
        "open_with_menuitem", "copy_menuitem", "cut_menuitem",
        "rename_menuitem", "duplicate_menuitem",
        "touch_menuitem", "symlink_menuitem",
        NULL
    };
    gboolean multiple_select[] = { TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, 0 };
    int i;

    if(!view_p || !population_p || !population_p->en || !population_p->en->path)
        return;

    for(i = 0; symbols[i]; i++) {
        HIDE_IT (view_p->widgets.paper, symbols[i]);
        if(g_file_test (population_p->en->path, G_FILE_TEST_EXISTS)
           || MENU_IN_MODULE (symbols[i])) {
            if(g_list_length (view_p->selection_list) > 1 && !multiple_select[i])
                continue;
            SHOW_IT (view_p->widgets.paper, "file_separator2");
            SHOW_IT (view_p->widgets.paper, "file_separator");
            SHOW_IT (view_p->widgets.paper, symbols[i]);
        }
    }

    if(g_file_test (population_p->en->path, G_FILE_TEST_IS_DIR)
       || MENU_IN_MODULE ("paste_menuitem")) {
        if(rfm_valid_pasteboard (view_p) && view_p->en && view_p->en->path) {
            SHOW_IT (view_p->widgets.paper, "paste_menuitem");
        }
    }

    if(g_list_length (view_p->selection_list) > 0 &&
       g_list_length (view_p->selection_list) < 3 &&
       view_p->en && view_p->en->path && g_file_test (view_p->en->path, G_FILE_TEST_IS_DIR)) {
        SHOW_IT (view_p->widgets.paper, "differences2");
    }

}

static void
set_menu_toggles (view_t * view_p) {
    {
        DBG ("setting sort toggles now...\n");
        view_p->disable_toggle_refresh = TRUE;
        if(view_p->sortcolumn == TYPE_SORT)
            gtk_check_menu_item_set_active ((GtkCheckMenuItem *)
                                            rodent_get_widget_by_name (view_p->widgets.paper, "unsorted6"), TRUE);
        else if(view_p->sortcolumn == NAME_SORT)
            gtk_check_menu_item_set_active ((GtkCheckMenuItem *)
                                            rodent_get_widget_by_name (view_p->widgets.paper, "name6"), TRUE);
        else if(view_p->sortcolumn == SIZE_SORT)
            gtk_check_menu_item_set_active ((GtkCheckMenuItem *)
                                            rodent_get_widget_by_name (view_p->widgets.paper, "size6"), TRUE);
        else if(view_p->sortcolumn == DATE_SORT)
            gtk_check_menu_item_set_active ((GtkCheckMenuItem *)
                                            rodent_get_widget_by_name (view_p->widgets.paper, "date6"), TRUE);
        else if(view_p->sortcolumn == OWNER_SORT)
            gtk_check_menu_item_set_active ((GtkCheckMenuItem *)
                                            rodent_get_widget_by_name (view_p->widgets.paper, "owner6"), TRUE);
        else if(view_p->sortcolumn == GROUP_SORT)
            gtk_check_menu_item_set_active ((GtkCheckMenuItem *)
                                            rodent_get_widget_by_name (view_p->widgets.paper, "group6"), TRUE);
        else if(view_p->sortcolumn == MODE_SORT)
            gtk_check_menu_item_set_active ((GtkCheckMenuItem *)
                                            rodent_get_widget_by_name (view_p->widgets.paper, "mode6"), TRUE);

        DBG ("before setting __SHOW_IMAGES toggle now...%s\n", (view_p->preferences & __SHOW_IMAGES) ? "ON" : "Off");
        gtk_check_menu_item_set_active ((GtkCheckMenuItem *)
                                        rodent_get_widget_by_name (view_p->widgets.paper,
                                                                   "preview_images_menuitem"), view_p->preferences & __SHOW_IMAGES);

        DBG ("after setting __SHOW_IMAGES toggle now...%s\n", (view_p->preferences & __SHOW_IMAGES) ? "ON" : "Off");
        gtk_check_menu_item_set_active ((GtkCheckMenuItem *)
                                        rodent_get_widget_by_name (view_p->widgets.paper,
                                                                   "show_hidden_menuitem"), view_p->preferences & __SHOW_HIDDEN);
        DBG ("setting __SHOW_HIDDEN toggle now...%s\n", (view_p->preferences & __SHOW_HIDDEN) ? "ON" : "Off");

        view_p->disable_toggle_refresh = FALSE;

    }
}

////////////////////////////////////////////////
////   menu.i

static GList *key_list = NULL;

/* Use as many different names for as many different
 * autotype menu entries for one file extension type. */
static char *auto_C_name[] = {
    "ejecutar1",
    "ejecutar2",
    "ejecutar3",
    "ejecutar4",
    "ejecutar5",
    "ejecutar6",
    "ejecutar7",
    "ejecutar8",
    "ejecutar9",
    "ejecutar10",
    NULL
};

static GtkWidget *
create_menu_entry (widgets_t * widgets_p, char *name, char *label, GtkWidget * parent, void *callback, void *data, int offset) {
    GtkWidget *w,
     *image;
    w = gtk_image_menu_item_new_with_label (label);
    if(offset < 0)
        gtk_container_add (GTK_CONTAINER (parent), w);
    else {
        gtk_menu_shell_insert (GTK_MENU_SHELL (parent), w, offset);
    }

    /* get the application icon would look nice */
    {
        GdkPixbuf *p = rfm_get_pixbuf ("xffm/stock_run", SIZE_BUTTON);
        if(p) {
            image = gtk_image_new_from_pixbuf (p);
        } else
            image = NULL;
    }
    /*image = gtk_image_new_from_stock ("gtk-execute", GTK_ICON_SIZE_MENU); */
    gtk_widget_show (image);
    gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (w), image);
    g_object_set (GTK_IMAGE_MENU_ITEM (w), "always-show-image", TRUE, NULL);
    g_signal_connect ((gpointer) w, "activate", G_CALLBACK (callback), data);
    TRACE ("created %s at 0x%lx window\n", name, (unsigned long)widgets_p->paper);
    /*gtk_widget_show(w); */
    RODENT_HOOKUP_OBJECT (widgets_p->paper, w, name);
    return w;
}

static void
clean_object_data(GtkWidget *a, const gchar *data_id)
{
    gchar *old_text = g_object_get_data (G_OBJECT (a), data_id);
    if((old_text = g_object_get_data (G_OBJECT (a), data_id)) != NULL) {
        g_object_set_data (G_OBJECT (a), data_id, NULL);
        g_free (old_text);
    }
}

static int
set_auto_command (widgets_t * widgets_p,
                  int j,
                  GtkWidget * popup,
                  const gchar * name,
                  const gchar * alt_label,
                  const gchar * app,
                  const gchar * path
                  ) {
    int jj;

    gchar *dirname = g_path_get_dirname (path);
    gchar *basename = g_path_get_basename (path);

    GtkWidget *a = rodent_get_widget_by_name (widgets_p->paper, name);

    const gchar * output_ext=MIME_command_output_ext(app);
    view_t *view_p=widgets_p->view_p;
    TRACE ("AUTO: set_auto_command() now... %s\n", app);
    if(!MIME_is_valid_command (app)) {
        if (strcmp(app, "rodent-newtab")==0 && view_p->tab_constructor) {
            TRACE (">> set_auto_command: %s name=%s\n", app, name);
        }
        else {
            DBG ("%s is not valid command", app);
            return j;
        }
    }
    GtkWidget *label;
    gchar *command;
    if(strncmp (app, "sudo -A ", strlen ("sudo -A ")) == 0) {
        if(getuid () == 0)
            return j;           /* don't sudoize root */
        if(!strlen (app + strlen ("sudo -A "))
           || !MIME_is_valid_command (app + strlen ("sudo -A ")))
            return j;
    }

    DBG("looking for command icon \"%s\"\n",app);
    const gchar *icon_id=MIME_command_icon (app);
    if (!icon_id) {
        DBG("command icon \"%s\" not found!\n",app);
        icon_id="xffm/stock_run";
    }

    if (icon_id) {
      GdkPixbuf *p = rfm_get_pixbuf (icon_id, SIZE_BUTTON);
      if(p) {
        GtkWidget *image = gtk_image_menu_item_get_image (GTK_IMAGE_MENU_ITEM (a));
        gtk_widget_destroy(image);
        image = gtk_image_new_from_pixbuf (p);
        gtk_widget_show (image);
            DBG("setting icon for menu item \"%s\"\n",app);
        gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (a), image);
      } else {
            DBG("unable to get pixbuf for menu item \"%s\"\n",app);
      }
    }


    label = gtk_bin_get_child (GTK_BIN (a));
    if(alt_label) { 
        gtk_label_set_text ((GtkLabel *) label, rfm_utf_string (alt_label));
    } else {
        /* allow mixed utf8-eucjp  paths */
        gchar *new_label;
        {
            const gchar *text = MIME_command_text (app);
	    const gchar *text2 = MIME_command_text2 (app);
            TRACE("MIME: looking for %s in hash: %s %s\n",app, text, text2);
            gchar *q;
            if (text) {
		gchar *texto;
		if (text2){
		    texto=g_strconcat(_(text), _(text2), NULL);
		} else {
		    texto=g_strdup(_(text));
		}
                TRACE("MIME: found\n");
                if (strstr(texto,"%s")!=NULL) {
                    q = MIME_mk_command_line (texto, basename);
                } else {
                    if (output_ext) 
                        q = g_strdup_printf("%s (%s)", texto, output_ext);
                    else
                        q = g_strdup(texto);
                }
		g_free(texto);
            }
            else {
                TRACE("MIME: not found\n");
                q = MIME_mk_command_line (app, basename);
            }
            new_label = rfm_utf_string (q);
            g_free (q);
        }
        gtk_label_set_text ((GtkLabel *) label, new_label);
        g_free (new_label);
    }

    if(output_ext) {
        DBG ("command is:  %s output_ext is; %s\n", app, output_ext);
        command = g_strdup (app);
    } else {
        /* this is to use relative paths for arguments for all
         * commands except those where the output directory (or
         * view_p->workdir) is requested from the user.
         * */
        DBG ("making command from (%s, %s)\n", app, path);
        command = MIME_mk_command_line (app, path);
    }

    /* check for duplicates */
    {
        gint argcp1;
        gchar **argvp1;
        GError *error = NULL;
        if(!g_shell_parse_argv ((const gchar *)command, &argcp1, &argvp1, &error)) {
            g_free (command);
            g_error_free (error);
            g_strfreev (argvp1);
            TRACE ("!g_shell_parse_argv at gui_autostuff\n");
            return j;
        }
        for(jj = 0; jj < j; jj++) {
            gboolean is_duplicate = TRUE;
            gint ac,
              argcp2;
            gchar **argvp2;
            gchar *set_command = g_object_get_data (G_OBJECT (rodent_get_widget_by_name (widgets_p->paper, auto_C_name[jj])),
                                                    "command");
            DBG ("AUTO: j=%d: %s == %s\n", jj, set_command, command);
            if(!g_shell_parse_argv ((const gchar *)set_command, &argcp2, &argvp2, &error)) {
                g_free (basename);
                g_free (command);
                g_error_free (error);
                g_strfreev (argvp2);
                TRACE ("!g_shell_parse_argv at gui_autostuff\n");
                return j;
            }
            if(argcp2 != argcp1) {
                is_duplicate = FALSE;
            } else
                for(ac = 0; ac < argcp2; ac++) {
                    if(strcmp (argvp2[ac], argvp1[ac])) {
                        is_duplicate = FALSE;
                        break;
                    }
                }
            TRACE ("0x%lx, 0x%lx\n", (unsigned long)argvp1, (unsigned long)argvp2);

            g_strfreev (argvp2);
            argvp2 = NULL;
            /*if (strcmp(set_command,command)==0) */
            if(is_duplicate) {
                g_free (basename);
                g_free (command);
                DBG ("AUTO: is_duplicate at gui_autostuff\n");
                return j;
            }
        }                       /* end for */
        g_strfreev (argvp1);
        argvp1 = NULL;
    }
    /* OK, by now we know the command is not duplicated... */
    // But is it a valid command??
    

    /* now let's free up any left over data from previous
     * generated menus... */
    clean_object_data(a, "command");
    clean_object_data(a, "workdir");
    g_object_set_data (G_OBJECT (a), "command", (gpointer) command);
    g_object_set_data (G_OBJECT (a), "workdir", (gpointer) dirname);
    if(MIME_command_output (app)) {
        g_object_set_data (G_OBJECT (a), "querypath", (gpointer)_("Select output folder"));
    } else {
        g_object_set_data (G_OBJECT (a), "querypath", NULL);
    }

    if(output_ext) {
        g_object_set_data (G_OBJECT (a), "output_arg", g_strdup (path));
    } else {
        clean_object_data(a, "output_arg");
        g_object_set_data (G_OBJECT (a), "output_arg", NULL);
    }
    DBG ("AUTO:command=%s, output_arg=%s\n", command, path);
    g_object_set_data (G_OBJECT (a), "output_ext", (gpointer) output_ext);
    DBG ("AUTO:now showing %s, active widget=0x%lx\n", name, (unsigned long)widgets_p->paper);
    SHOW_IT (widgets_p->paper, name);
    j++;
    g_free (basename);
    return j;
}
#if 0
static gchar *
alt_menu_label (const gchar * label, const gchar * path, const gchar * extension) {     /* chinese friendly translations: */
    gchar *alt_label = NULL;
    if(!label)
        return NULL;
    if(extension) {
        gchar *fffname = g_path_get_basename (path);
        gchar *fname = rfm_utf_string (fffname);
        gchar *utf_label = rfm_utf_string (label);
        gchar *ffname = g_strconcat (fname, extension, NULL);
        g_free (fffname);
        if(strstr (label, "%s")) {
            alt_label = g_strdup_printf (utf_label, ffname);
        } else {
            alt_label = g_strdup_printf ("%s --> %s", utf_label, ffname);
        }
        g_free (fname);
        g_free (ffname);
        g_free (utf_label);
    } else {
        alt_label = rfm_utf_string (label);
    }

    return alt_label;
}
#endif

static void
gui_autostuff (widgets_t * widgets_p,
               GtkWidget * popup, 
	       record_entry_t * en, 
	       gpointer callback, 
	       gpointer on_mount, 
	       gpointer on_unmount) {
    int j = 0;
        TRACE ("gui_autostuff\n");

    if(!rodent_get_widget_by_name (widgets_p->paper, auto_C_name[0])) {
        TRACE ("creating auto menu elements from gui_autostuff\n");
        gui_add_autotype_C_widgets (widgets_p, popup, callback, on_mount, on_unmount);
    }
    if(FSTAB_is_partition_type (en)) {
        // only first item (open in new window...)
        gchar *mount_point = FSTAB_get_mnt_dir (en->path);
        if(mount_point) {
	    gchar *cc=g_strdup("rodent");
            j = set_auto_command (widgets_p, j, popup, auto_C_name[j], 
            _("Open in New Window"), cc, mount_point);
	    g_free(cc);
            g_free (mount_point);
        }
    }
    else if(IS_EXE_TYPE (en->type) && !g_file_test(en->path, G_FILE_TEST_IS_DIR)) {
        gchar *cmd;
        GtkWidget *label,
         *a = rodent_get_widget_by_name (widgets_p->paper, "autotype_Prun");
        label = gtk_bin_get_child (GTK_BIN (a));
        gchar *b = g_path_get_basename (en->path);
        gchar *q = rfm_utf_string (rfm_chop_excess (b));
        cmd = g_strdup_printf ("%s (%s)", _("Execute File?"), q);
//        cmd = g_strdup_printf ("%s: %s", _("Execute File?"), q);
        g_free (b);
        g_free (q);
        q = rfm_utf_string (cmd);
        gtk_label_set_text ((GtkLabel *) label, q);
        g_free (cmd);
        g_free (q);
        SHOW_IT (widgets_p->paper, "autotype_Prun");
    }
    /* autotypes */
#if 0
    // This itemizes the last command used on the particular
    // file being examined...
    //
    // FIXME: DBH lock goes into eternal wait. why?
    gboolean in_term;
    const gchar *p;
    p = rfm_get_from_open_history (en->path, &in_term);
    /* XXX: allow in term to be processed too (later...) */
    if(p && !in_term) {
        j = set_auto_command (widgets_p, j, popup, auto_C_name[j], NULL, p, en->path, NULL, NULL);
    }
#endif
    // these are file type specific auto commands and depend on 
    // the file extension of the path. Commands are static
    // and are for .rar .tar, .tar.gz, .tar.bz2 and .iso
    //
    // 2b deprecated some time in the future.
    //
    // This is the nitty gritty. Last but not least.
    // Suggest an application based on the file's mimetype.
    DBG ("AUTO: menu nitty-gritty now\n");
    gchar *magic_type = MIME_magic (en->path);
    gchar *mime_type = MIME_type (en->path, en->st);
    if(g_file_test(en->path, G_FILE_TEST_EXISTS)) {
        int k;
        DBG ("AUTO: menu nitty-gritty mime-type=%s magic-type=%s\n", mime_type, magic_type);

        // here we use the freedesktop and magic
        // mimetype to get the run program.
        for(k = 0; k < 2; k++) {
            gchar **apps = NULL;
            int i = 0;
            if(k == 0) {
                apps = MIME_apps (mime_type);
                DBG ("AUTO: menu nitty-gritty (%s) apps=0x%lx\n", mime_type, (long)apps);
            } else {
                apps = MIME_apps (magic_type);
                DBG ("AUTO: menu nitty-gritty (%s) apps=0x%lx\n", magic_type, (long)apps);
            }

            if(apps && apps[i])
                for(; apps && apps[i]; i++) {
                    DBG ("AUTO: j=%d, i=%d %s -> %s\n", j, i, auto_C_name[j], apps[i]);

                    if(!auto_C_name[j])
                        continue;
                    j = set_auto_command (widgets_p, j, popup, auto_C_name[j], NULL,
                                          // the command:
                                          apps[i], en->path);
                }
            g_strfreev (apps);
        }

    }
#if 10
    // we magic the file because the freedesktop standard
    // is borked and changes text/ for application/ on a 
    // number of files. Precisely what should not be done.
    gchar **text_editors = NULL;
    DBG ("ESC:magic type=%s\n", magic_type);
    if(magic_type) {
        if(strstr (magic_type, "text/"))
            text_editors = MIME_apps ("text/plain");
    } else if(mime_type) {
        if(strstr (mime_type, "text/"))
            text_editors = MIME_apps ("text/plain");
    }
    if(!text_editors) {
        /* even magic mimetype misses the mark. Take care of that here... */
        if (!en->filetype) en->filetype= MIME_file(en->path);
        if(en->filetype && strstr (en->filetype, "text"))
            text_editors = MIME_apps ("text/plain");
    }
    if(text_editors) {
        /* OK to apply an editor */
        int k;

        if(getenv ("EDITOR_CMD") && strlen (getenv ("EDITOR_CMD"))) {
            gchar *basename=g_path_get_basename(getenv ("EDITOR_CMD"));
            j = set_auto_command (widgets_p, j, popup, auto_C_name[j], NULL,
                                  // the command:
                                  basename, en->path);
            g_free(basename);

        }
        // other editors
        DBG ("text_editors != NULL\n");
        for(k = 0; text_editors[k]; k++) {
            j = set_auto_command (widgets_p, j, popup, auto_C_name[j], NULL,
                                  // the command:
                                  text_editors[k], en->path);
        }
    }
    g_strfreev (text_editors);
#endif

    if(g_file_test (en->path, G_FILE_TEST_EXISTS)) {
        SHOW_IT (widgets_p->paper, "open_with_menuitem");
    }

    g_free (magic_type);
    g_free (mime_type);
    return;
}

typedef struct extra_key_t {
    GtkWidget *check1;
    GtkEntry *entry;
    gboolean flag1;
    gboolean flag2;
    const gchar *response;
    const gchar *flagfile;
    const gchar *history_file;
} extra_key_t;

static void
activate_entry (GtkEntry * entry, gpointer data) {
    gtk_dialog_response ((GtkDialog *) data, GTK_RESPONSE_YES);
}

static void
cancel_entry (GtkEntry * entry, gpointer data) {
    gtk_dialog_response ((GtkDialog *) data, GTK_RESPONSE_CANCEL);
}

static void
recover_flags (extra_key_t * extra_key_p) {
    DBHashTable *runflags;
    GString *gs;
    int *flags;

    if((runflags = dbh_open ((char *)extra_key_p->flagfile)) == NULL) {
        TRACE ("Cannot open %s\n", extra_key_p->flagfile);
        extra_key_p->flag1 = 0;
        extra_key_p->flag2 = 0;
        return;
    }
    gs = g_string_new (extra_key_p->response);
    sprintf ((char *)DBH_KEY (runflags), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);
    /*TRACE("TRACE: bookmarking with key=%s\n",d_path); */
    flags = (int *)runflags->data;
    dbh_load (runflags);
    extra_key_p->flag1 = flags[0];
    extra_key_p->flag2 = flags[1];
    dbh_close (runflags);

    DBG ("flags recovered from dbh file for %s, flag1=%d flag2=%d\n",
         extra_key_p->response, extra_key_p->flag1, extra_key_p->flag2);
}

static void
save_flags (extra_key_t * extra_key_p) {
    DBHashTable *runflags;
    GString *gs;
    int *flags;

    if((runflags = dbh_open (extra_key_p->flagfile)) == NULL) {
        DBG ("Creating cache file: %s", extra_key_p->flagfile);
        if((runflags = dbh_create (extra_key_p->flagfile, 11)) == NULL) {
            unlink (extra_key_p->flagfile);
            if((runflags = dbh_create (extra_key_p->flagfile, 11)) == NULL) {
                g_warning ("Cannot create %s\n", extra_key_p->flagfile);
                return;
            }
        }
    }
    gs = g_string_new (extra_key_p->response);
    sprintf ((char *)DBH_KEY (runflags), "%10u", g_string_hash (gs));
    g_string_free (gs, TRUE);
    /*TRACE("TRACE: bookmarking with key=%s\n",d_path); */
    flags = (int *)runflags->data;
    flags[0] = extra_key_p->flag1;
    flags[1] = extra_key_p->flag2;
    dbh_set_recordsize (runflags, 2 * sizeof (int));

    dbh_update (runflags);
    dbh_close (runflags);
    DBG ("flags saved in dbh file  %s flag1=%d flag2=%d\n", extra_key_p->flagfile, extra_key_p->flag1, extra_key_p->flag2);

}

static int
extra_key_completionR (gpointer data) {
    extra_key_t *extra_key_p = (extra_key_t *) data;

    DBG("extra_key_completionR\n");
    if(!extra_key_p) {
        g_warning ("!extra_key_p");
        return FALSE;
    }
    if(!GTK_IS_ENTRY (extra_key_p->entry))
        return FALSE;
    extra_key_p->response = gtk_entry_get_text (extra_key_p->entry);

    recover_flags (extra_key_p);
    gtk_toggle_button_set_active ((GtkToggleButton *) extra_key_p->check1, extra_key_p->flag1);
    return FALSE;
}

static void
toggle_flag (GtkToggleButton * togglebutton, gpointer user_data) {
    extra_key_t *extra_key_p = (extra_key_t *) user_data;
    if(gtk_toggle_button_get_active (togglebutton))
        extra_key_p->flag1 = TRUE;
    else
        extra_key_p->flag1 = FALSE;
    save_flags (extra_key_p);
}

static gpointer
thread_preload_f (gpointer data) {
    const gchar *folder = (const gchar *)data;
    if(folder && g_file_test (folder, G_FILE_TEST_IS_DIR)) {
        GError *error = NULL;
        const gchar *path;
        GDir *dir = g_dir_open (folder, 0, &error);
        while((path = g_dir_read_name (dir)) != NULL) {
            /* just to avoid compiler optimization: */
            if(strcmp (path, ".") == 0)
                g_warning ("This should never happen: strcmp (path, \".\") == 0");
        }

        g_dir_close (dir);
    }

    return NULL;
}

static
    void
filechooser (GtkButton * button, gpointer user_data) {
    filechooser_t *filechooser_p = (filechooser_t *) user_data;
    const gchar *text;
    if(filechooser_p->filechooser_action == GTK_FILE_CHOOSER_ACTION_OPEN) {
        text = _("File Browser");
    } else if(filechooser_p->filechooser_action == GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER) {
        text = _("Open Folder W_indow");
    } else {
        text = "FIXME";
    }
    GtkWidget *dialog = gtk_file_chooser_dialog_new (text,
                                                     GTK_WINDOW (filechooser_p->parent),
                                                     filechooser_p->filechooser_action,
                                                     GTK_STOCK_CANCEL,
                                                     GTK_RESPONSE_CANCEL,
                                                     GTK_STOCK_OPEN,
                                                     GTK_RESPONSE_ACCEPT,
                                                     NULL);
    gtk_file_chooser_set_use_preview_label(GTK_FILE_CHOOSER (dialog), FALSE);
    gtk_file_chooser_set_preview_widget_active (GTK_FILE_CHOOSER (dialog), FALSE);
    // XXX gtk_file_chooser_set_show_hidden isn't working. Gtk bug?
    gtk_file_chooser_set_show_hidden(GTK_FILE_CHOOSER (dialog), TRUE);
    gtk_file_chooser_set_action (GTK_FILE_CHOOSER (dialog), filechooser_p->filechooser_action);
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), filechooser_p->folder);
    gtk_file_chooser_set_create_folders (GTK_FILE_CHOOSER (dialog),TRUE);

    if(gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
        char *filename;
        void *combo_info = filechooser_p->combo_info;
        filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
        COMBOBOX_set_entry (combo_info, filename);
        DBG ("Got %s\n", filename);
        g_free (filename);
	gtk_widget_destroy (dialog);
	if(filechooser_p->activate_func) {
            (*(filechooser_p->activate_func)) (filechooser_p->entry, filechooser_p->activate_user_data);
    }

    } else {
	gtk_widget_destroy (dialog);
    }

}

static void
preload (const gchar * folder) {
    DBG ("THREAD: g_thread_create(thread_preload_f) at rodent_popup.i\n");
    g_thread_create (thread_preload_f, (gpointer) (folder), FALSE, NULL);
}

static void
gui_add_autotype_C_widgets (widgets_t * widgets_p,
                            GtkWidget * popup, 
			    gpointer callback, 
			    gpointer on_mount, 
			    gpointer on_unmount) {
    int i;
    GtkWidget *w;

    /* check for optional fstab plugin... */
    if(rfm_void ("rfm", "fstab", "module_name")) {
        if(on_mount) {
            w = rodent_get_widget_by_name (widgets_p->paper, "mountP");
            if(w){
                g_signal_connect ((gpointer) w, "activate", G_CALLBACK (on_mount), widgets_p);
	    }
        }
        if(on_unmount) {
            w = rodent_get_widget_by_name (widgets_p->paper, "unmountP");
            if(w){
                g_signal_connect ((gpointer) w, "activate", G_CALLBACK (on_unmount), widgets_p);
	    }
        }
    }

    if(callback)
        for(i = 0; auto_C_name[i]; i++) {
            GtkWidget *w;
            w = create_menu_entry (widgets_p, auto_C_name[i], " ", popup, callback, NULL, i + 6);
            g_object_set_data (G_OBJECT (w), "view_p", widgets_p->view_p);
        }

}

/* gui_autofunction for directories works different: 
 * directories: the path argument is relative, so we must do a chdir
 * 		to the path directory.
 * 		The output specified will carry along the fullpath for 
 * 		output. The workdir will be that where the directory 
 * 		resides. We must do a chdir here.
 * files:	the path argument is absolute. So we just do a chdir to
 * 		the specified workdir before execution.
 *
 * 		*/

/* keybindings... accel group will not work if menus is not around. This is a
 * gtk feature, so we have to do our own keybinding. 
 **/

/* function to add a keybind */

static
    void
gui_add_to_keylist (guint in_key, guint mask, gpointer callback) {
    GList *tmp;
    keybind_t *keybind_p;
    guint key;

    if(in_key < 128) {
        TRACE ("key=toupper(in_key)\n");
        key = toupper (in_key);
    } else
        key = in_key;

    for(tmp = key_list; tmp; tmp = tmp->next) {
        keybind_p = (keybind_t *) tmp->data;
        if(!keybind_p)
            continue;
        if(key == keybind_p->key && mask == keybind_p->mask)
            return;
    }
    TRACE ("adding to keylist=%u callback=0x%lx\n", key, (unsigned long)callback);
    keybind_p = (keybind_t *) malloc (sizeof (keybind_t));
    memset (keybind_p, 0, sizeof (keybind_t));
    keybind_p->key = key;
    keybind_p->mask = mask;
    keybind_p->callback = callback;
    key_list = g_list_append (key_list, keybind_p);
    return;
}

/* function to run callback with keybind values */
typedef void (*keybind_f) (GtkMenuItem *, gpointer);

static
    int
gui_keybind_exec (guint in_key, guint mask, void *data) {
    GList *tmp;
    void (*callback) (GtkMenuItem * menuitem, gpointer data);
    guint key;

    if(in_key < 128)
        key = toupper (in_key);
    else
        key = in_key;

    for(tmp = key_list; tmp; tmp = tmp->next) {
        keybind_t *keybind_p = (keybind_t *) tmp->data;
        if(!keybind_p)
            continue;
        TRACE ("key: %u=%u?\n", (char)keybind_p->key, key);
        if(key == keybind_p->key && mask == keybind_p->mask) {
            callback = (keybind_f) keybind_p->callback;
            TRACE ("callback=0x%lx, data=0x%lx\n", (unsigned long)callback, (unsigned long)data);
            (*callback) (NULL, data);
            return TRUE;
        }
    }
    TRACE ("no keybinding...\n");
    return FALSE;
}

static
GtkWidget *
gui_add_menu_separator (GtkWidget * parent) {
    GtkWidget *w = gtk_menu_item_new ();
    gtk_widget_show (w);
    gtk_container_add (GTK_CONTAINER (parent), w);
    gtk_widget_set_sensitive (w, FALSE);
    return w;
}


/* function valid for only one radio group: */
static
GtkWidget *
gui_mk_radio_button (widgets_t * widgets_p, const gchar * label, const gchar * name, GtkWidget * parent, GSList ** radiogroup) {
    GtkWidget *w = gtk_radio_menu_item_new_with_mnemonic (*radiogroup, label);

    /*w = gtk_radio_menu_item_new_with_mnemonic (*radiogroup,label); */
    *radiogroup = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (w));
    gtk_widget_show (w);
    gtk_container_add (GTK_CONTAINER (parent), w);
    //gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (w), TRUE);
    RODENT_HOOKUP_OBJECT (widgets_p->paper, w, name);
    return w;
}

static
    void
gui_add_menu_item (widgets_t * widgets_p,
                   GtkWidget * parent,
                   const gchar * label,
                   const gchar * icon_id,
                   const gchar * name,
                   gpointer callback, gpointer callback_data, GtkAccelGroup * accel_group, guint accel_key, guint accel_mask) {
    GtkWidget *w = gtk_image_menu_item_new_with_mnemonic (label);
    gtk_image_menu_item_set_always_show_image ((GtkImageMenuItem *) w, TRUE);
    TRACE ("gui_add_menu_item\n");
    if(icon_id)
        rodent_mk_pixmap_menu (widgets_p, icon_id, w, MENU_PIXMAP);
    gtk_widget_show (w);
    gtk_container_add (GTK_CONTAINER (parent), w);
    RODENT_HOOKUP_OBJECT (widgets_p->paper, w, name);
    if(callback)
        g_signal_connect ((gpointer) w, "activate", G_CALLBACK (callback), callback_data);
    if(accel_key) {
        gui_add_to_keylist (accel_key, accel_mask, callback);
    }
    if(accel_group)
        gtk_widget_add_accelerator (w, "activate", accel_group, accel_key, accel_mask, GTK_ACCEL_VISIBLE);
}

static
    void
rodent_mk_module_popup_menu (const gchar * module_name, view_t * view_p, record_entry_t * en) {
    gchar *txt;
    if(!module_name)
        return;
    txt = rfm_natural ("rfm", module_name, en, "module_label");

    /* set the submenu label */
    if(txt) {
        GtkWidget *label,
         *a = rodent_get_widget_by_name (view_p->widgets.paper, "module1");
        label = gtk_bin_get_child (GTK_BIN (a));
        gchar *q = rfm_utf_string (txt);
        gtk_label_set_text ((GtkLabel *) label, q);
        g_free (q);
        g_free (txt);
    }
    /* remove anything left over */
    gtk_container_foreach ((GtkContainer *)
                           rodent_get_widget_by_name (view_p->widgets.paper, "module1_menu"), remove_menu_item, NULL);

    /* add new stuff to submenu */
    rfm_rational ("rfm",
                       module_name,
                       (void
                        *)(rodent_get_widget_by_name (view_p->widgets.paper, "module1_menu")), &(view_p->widgets), "extend_popup");

}


static
GtkWidget *
mk_popup_menu (GtkWidget * parent, GtkAccelGroup * accel_group, view_t * view_p) 
{
    return (common_mk_popup_menu (&(view_p->widgets), parent, accel_group));
}

