
/*****    dialog callbacks   *****/

static gboolean
on_destroy_event (GtkWidget * widget, GdkEvent * event, gpointer data) {
    widgets_t *widgets_p = (widgets_t *) data;

    gtk_widget_destroy (widgets_p->remove);
    widgets_p->remove = GINT_TO_POINTER (RM_CANCEL);
    gtk_main_quit ();
    return TRUE;
}

static void
on_remove_no_activate (GtkButton * button, gpointer data) {
    widgets_t *widgets_p = (widgets_t *) data;
    GtkWidget *check = rodent_get_widget_by_name (widgets_p->remove, "togglebutton1");
    gboolean all = gtk_toggle_button_get_active ((GtkToggleButton *) check);
    gtk_widget_destroy (widgets_p->remove);
    if(all) {
        widgets_p->remove = GINT_TO_POINTER (RM_CANCEL);
    } else {
        widgets_p->remove = GINT_TO_POINTER (RM_NO);
    }
    gtk_main_quit ();
}

static void
on_remove_yes_activate (GtkButton * button, gpointer data) {
    widgets_t *widgets_p = (widgets_t *) data;
    GtkWidget *check = rodent_get_widget_by_name (widgets_p->remove, "togglebutton1");
    DBG ("REMOVE: on_remove_yes_activate\n");
    gboolean all = gtk_toggle_button_get_active ((GtkToggleButton *) check);
    gtk_widget_destroy (widgets_p->remove);
    if(all) {
        widgets_p->remove = GINT_TO_POINTER (RM_YES_ALL);
    } else {
        widgets_p->remove = GINT_TO_POINTER (RM_YES);
    }
    gtk_main_quit ();

    return;
}

#ifdef GNU_SHRED
static void
on_shred_yes_activate (GtkButton * button, gpointer data) {
    widgets_t *widgets_p = (widgets_t *) data;
    GtkWidget *check = rodent_get_widget_by_name (widgets_p->remove, "togglebutton1");
    gboolean all = gtk_toggle_button_get_active ((GtkToggleButton *) check);
    gtk_widget_destroy (widgets_p->remove);

    if(all) {
        widgets_p->remove = GINT_TO_POINTER (SHRED_YES_ALL);
    } else {
        widgets_p->remove = GINT_TO_POINTER (SHRED_YES);
    }
    gtk_main_quit ();

    return;
}
#endif

// specify mode as shred or delete
//shred
//rm
static void
do_the_remove (widgets_t * widgets_p, GList ** rm_list, gboolean unlink_mode) {
    gchar *sources[MAX_COMMAND_ARGS];
    gchar *shred_iterations = NULL;
    int i = 0;
    if(unlink_mode == MODE_SHRED) {
        sources[i++] = "shred";
        if(getenv ("RFM_SHRED_ITERATIONS") && strlen (getenv ("RFM_SHRED_ITERATIONS"))) {
            gboolean ok = TRUE;
            char *p;
            for(p = getenv ("RFM_SHRED_ITERATIONS"); *p; p++) {
                if(!isdigit (*p)) {
                    ok = FALSE;
                    break;
                }
            }
            if(ok) {
                shred_iterations = g_strdup_printf ("--iterations=%s", getenv ("RFM_SHRED_ITERATIONS"));
            }
            sources[i++] = shred_iterations;
        }
        if(getenv ("RFM_SHRED_ZERO") && strlen (getenv ("RFM_SHRED_ZERO"))) {
            sources[i++] = "--zero";
        }
        if(getenv ("RFM_SHRED_REMOVE") && strlen (getenv ("RFM_SHRED_REMOVE"))) {
            sources[i++] = "--remove";
        }
/*      sources[i++] = " -s, --size=N"; */
    } else {
        sources[i++] = "rm";
        sources[i++] = "-r";
        if(getenv ("RFM_RM_ONE_FILE_SYSTEM") && strlen (getenv ("RFM_RM_ONE_FILE_SYSTEM"))) {
            sources[i++] = "--one-file-system";
        }
    }

    // common options:
    sources[i++] = "-f";
    if(unlink_mode == MODE_RM &&
	    getenv ("RFM_RM_VERBOSE") && 
	    strlen (getenv ("RFM_RM_VERBOSE"))) {
        sources[i++] = "-v";
    }
    if(unlink_mode == MODE_SHRED && 
	    getenv ("RFM_SHRED_VERBOSE") && 
	    strlen (getenv ("RFM_SHRED_VERBOSE"))) {
        sources[i++] = "-v";
    }
    //shred:
    gboolean ok = FALSE;
    gboolean excess=FALSE;
    while(*rm_list && (*rm_list)->data) {
        record_entry_t *en = (record_entry_t *) (*rm_list)->data;
        /*if (!en || !en->path || !g_file_test(en->path, G_FILE_TEST_EXISTS)) {
           continue;
           } */
        if(en->path && (g_file_test (en->path, G_FILE_TEST_EXISTS) || g_file_test (en->path, G_FILE_TEST_IS_SYMLINK))) {
            sources[i++] = en->path;
            *rm_list = g_list_remove (*rm_list, (*rm_list)->data);
            if (i == MAX_COMMAND_ARGS - 1  && *rm_list){
                excess=TRUE;
                // we are at the last argv, for NULL
                // remaining items will not be processed.
                while (*rm_list) {*rm_list = g_list_remove (*rm_list, (*rm_list)->data);}
                break;
            }
            ok = TRUE;
        } else {
            *rm_list = g_list_remove (*rm_list, (*rm_list)->data);
        }
    };
    sources[i++] = NULL;

    view_t *view_p=widgets_p->view_p;
    if (view_p->type != DESKVIEW_TYPE) {
	rfm_show_text(widgets_p);
    }

    // note: with run_argv we do not need to escape paths.
    if(ok) {
        rfm_thread_run_argv (widgets_p, sources, FALSE);
    } else {
        DBG ("PASTE CP: ok == FALSE\n");
    }
    g_free (shred_iterations);
    if (excess){
        rfm_diagnostics(widgets_p, "xffm/warning",NULL);
        gchar *max=g_strdup_printf("%d",MAX_COMMAND_ARGS);
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", sources[0], ": ", strerror(E2BIG)," (> ",max,")","\n", NULL);
        g_free(max);
    }


}

static
    void
make_dialog (widgets_t * widgets_p, GList ** rm_list) {
    record_entry_t *en;
    gchar *mess = NULL;


    if(*rm_list == NULL)
        return;
    if(widgets_p->remove) {
        g_warning ("widgets_p->remove != NULL");
        return;
    }

    widgets_p->remove = rodent_create_remove (widgets_p);

    g_signal_connect (rodent_get_widget_by_name (widgets_p->remove, "cancelbutton"),
                      "clicked", G_CALLBACK (on_remove_no_activate), widgets_p);
    g_signal_connect (rodent_get_widget_by_name (widgets_p->remove, "removebutton"),
                      "clicked", G_CALLBACK (on_remove_yes_activate), widgets_p);
#ifdef GNU_SHRED
    g_signal_connect (rodent_get_widget_by_name (widgets_p->remove, "shredbutton"),
                      "clicked", G_CALLBACK (on_shred_yes_activate), widgets_p);
#endif

    g_signal_connect (widgets_p->remove, "delete-event", G_CALLBACK (on_destroy_event), widgets_p);
    g_signal_connect (widgets_p->remove, "destroy-event", G_CALLBACK (on_destroy_event), widgets_p);

    /* dialog specifics */

    if(g_list_length (*rm_list) < 2) {
        HIDE_IT (widgets_p->remove, "togglebutton1");
    }
    HIDE_IT (widgets_p->remove, "warning");

    en = (record_entry_t *) (*rm_list)->data;

    if(!en) {
        g_warning ("critical: en==NULL");
    }

    {
        gchar *b = g_path_get_basename (en->path);
        gchar *q = rfm_utf_string (rfm_chop_excess (b));
        gchar *s1 = rfm_time_to_string (en->st->st_mtime);
        gchar *s2 = rfm_sizetag ((off_t) en->st->st_size, -1);
        mess = g_strconcat (q, "\n", "(", s1, " ", s2, ")", NULL);
        g_free (b);
        g_free (q);
        g_free (s1);
        g_free (s2);
    }

    gtk_label_set_text ((GtkLabel *)
                        rodent_get_widget_by_name (widgets_p->remove, "label16"), _("Remove"));
    gchar *q = rfm_utf_string (mess);
    gtk_label_set_text ((GtkLabel *)
                        rodent_get_widget_by_name (widgets_p->remove, "label20"), q);
    g_free (q);
    g_free (mess);
    mess = NULL;
    gtk_widget_realize (widgets_p->remove);

    gtk_widget_grab_focus (rodent_get_widget_by_name (widgets_p->remove, "removebutton"));
    view_t *view_p=widgets_p->view_p;

    if(view_p->type != DESKVIEW_TYPE) {
        gtk_window_set_transient_for (GTK_WINDOW (widgets_p->remove), GTK_WINDOW (widgets_p->window));
    }
    gdk_window_set_keep_above (widgets_p->remove->window, TRUE);

    gtk_widget_show (widgets_p->remove);
    gtk_main ();
    DBG ("REMOVE: dialog done\n");

    int result = GPOINTER_TO_INT (widgets_p->remove);
    widgets_p->remove = NULL;
    int mode = MODE_RM;
    GList *single_list = NULL;
    switch (result) {
    case SHRED_YES:
        mode = MODE_SHRED;
    case RM_YES:
        DBG ("REMOVE: RM_YES  \\n");
        single_list = g_list_append (single_list, (*rm_list)->data);
        do_the_remove (widgets_p, &single_list, mode);
        g_list_free (single_list);
        en = (record_entry_t *) (*rm_list)->data;
        *rm_list = g_list_remove (*rm_list, (*rm_list)->data);
        rfm_destroy_entry (en);
        break;
    case RM_NO:
        en = (record_entry_t *) (*rm_list)->data;
        *rm_list = g_list_remove (*rm_list, (*rm_list)->data);
        rfm_destroy_entry (en);
        break;
    case SHRED_YES_ALL:
        mode = MODE_SHRED;
    case RM_YES_ALL:
        do_the_remove (widgets_p, rm_list, mode);
    case RM_CANCEL:
    default:
        while(*rm_list && (*rm_list)->data) {
            en = (record_entry_t *) (*rm_list)->data;
            *rm_list = g_list_remove (*rm_list, (*rm_list)->data);
            rfm_destroy_entry (en);
        }
    }

}

/**************/

static int
make_single_rm_dialog (widgets_t * widgets_p, const gchar * nfile, const gchar * ofile) {
    gchar *mess = NULL,
        *mess1 = NULL;
    struct stat st;
    if(widgets_p->remove)
        g_assert_not_reached ();
    if(!nfile)
        g_assert_not_reached ();

    widgets_p->remove = rodent_create_remove (widgets_p);

    g_signal_connect (rodent_get_widget_by_name (widgets_p->remove, "cancelbutton"),
                      "clicked", G_CALLBACK (on_remove_no_activate), widgets_p);
    g_signal_connect (rodent_get_widget_by_name (widgets_p->remove, "removebutton"),
                      "clicked", G_CALLBACK (on_remove_yes_activate), widgets_p);

    g_signal_connect (widgets_p->remove, "delete-event", G_CALLBACK (on_destroy_event), widgets_p);
    g_signal_connect (widgets_p->remove, "destroy-event", G_CALLBACK (on_destroy_event), widgets_p);

    /* dialog specifics */

#ifdef GNU_SHRED
    HIDE_IT (widgets_p->remove, "shredbutton");
#endif

    {
        HIDE_IT (widgets_p->remove, "warning");
        mess1 = g_strdup (" ");
    }

    lstat (nfile, &st);

    if(ofile) {
        gtk_label_set_text ((GtkLabel *)
                            rodent_get_widget_by_name (widgets_p->remove, "label22"), _("Overwrite"));
    } else {
        gtk_label_set_text ((GtkLabel *)
                            rodent_get_widget_by_name (widgets_p->remove, "label22"), _("Remove"));

    }

    gtk_label_set_text ((GtkLabel *)
                        rodent_get_widget_by_name (widgets_p->remove, "label16"), _("Remove"));

    {
        gchar *b = g_path_get_basename (nfile);
        gchar *s1 = rfm_time_to_string (st.st_mtime);
        gchar *s2 = rfm_sizetag ((off_t) st.st_size, -1);
        mess = g_strconcat (mess1, b, "\n", "(", s1, " ", s2, ")", NULL);
        g_free (b);
        g_free (s1);
        g_free (s2);
    }

    gtk_label_set_text ((GtkLabel *)
                        rodent_get_widget_by_name (widgets_p->remove, "label20"), mess);

    g_free (mess);
    mess = NULL;
    g_free (mess1);
    mess1 = NULL;
    gtk_widget_realize (widgets_p->remove);

    view_t *view_p=widgets_p->view_p;
    if(view_p->type != DESKVIEW_TYPE) {
        gtk_window_set_transient_for (GTK_WINDOW (widgets_p->remove), GTK_WINDOW (widgets_p->window));
    }
    gdk_window_set_keep_above (widgets_p->remove->window, TRUE);
    HIDE_IT (widgets_p->remove, "togglebutton1");
    gtk_widget_grab_focus (rodent_get_widget_by_name (widgets_p->remove, "removebutton"));
    gtk_widget_show (widgets_p->remove);
    gtk_main ();
    int result = GPOINTER_TO_INT (widgets_p->remove);
    widgets_p->remove = NULL;
    return result;
}
