/*
 *
 * Edscott Wilson Garcia 2001-2010 edscott@xfce.org
 *
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/stat.h>

#include <errno.h>
#include <limits.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <gtk/gtk.h>
#include <glib.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <magic.h>
#include <gmodule.h>

#include "rodent_libs.h"


#if JUST_TRANSLATE 
// translations from applications.xml:
gchar *other_translations[]={
        N_("List files only"),
        N_("Extract All Files"),
        N_("Open in New Window"), 
        N_("Open in New Tab"),
        N_("Create a compressed archive with the selected objects"),
        N_("Create a new archive"),
        N_("Install"),
        N_("Uninstall"),
        N_("Information about the program"),
        N_("Show Information")
        N_(" [Simulation]")
        N_("Write CD/DVD")
};
#endif
#define USER_APPLICATIONS g_get_user_config_dir(),"rfm","user-applications"
#define CACHE_DIR g_get_user_cache_dir(),"rfm","modules"

/* this should be first 2 lines after headers: */
G_MODULE_EXPORT LIBRFM_MODULE
#include "applications-module.h"
#include "applications-module.i"
/****************************************************************************/
  G_MODULE_EXPORT gchar *
mime_type (void *p, void *q) {
    const gchar *file = p;
    struct stat *st_p = q;
    struct stat st_v;
    if(!st_p) {
        st_p = &st_v;
        stat (file, st_p);
    }
    gchar *type = mimeable_file (st_p);
    if(type) {
        TRACE ("MIME: stat mime_type(%s) -> %s\n", file, type);
        return (void *)type;
    }

    type = locate_mime_t (file);
    if(type) {
        TRACE ("MIME:locate_mime_t(%s) -> %s\n", file, type);
        return (void *)type;
    }

    /* certain back up types? */
    if(file[strlen (file) - 1] == '~') {
        TRACE ("MIME: certain back up types(%s)-> %s\n", file, "application/x-trash");
        return (void *)g_strdup ("application/x-trash");
    }

    TRACE ("MIME: locate_mime_t FAILED\n");

    return NULL;
}

#ifdef HAVE_LIBMAGIC
// thread safe 
static gchar *
lib_magic (const gchar * file, int flags) {
    const char *type;
    gchar *magic_mime_type = NULL;
    g_mutex_lock (magic_mutex);
    {
        magic_setflags (cookie, flags);
        type = magic_file (cookie, file);
        if(type) {
            magic_mime_type = g_strdup (type);
        } else {
            magic_mime_type = g_strdup ("unknown");
        }
    }
    g_mutex_unlock (magic_mutex);
    return magic_mime_type;
}

// see "man libmagic" for explantion of flags
G_MODULE_EXPORT void *
mime_magic (void *p) {
    const gchar *file = p;
    int flags =
        MAGIC_MIME_TYPE |
        MAGIC_NO_CHECK_APPTYPE |
        MAGIC_NO_CHECK_ENCODING | MAGIC_SYMLINK | MAGIC_NO_CHECK_COMPRESS | MAGIC_NO_CHECK_TAR | MAGIC_PRESERVE_ATIME;
    return (void *)(lib_magic (file, flags));
}

G_MODULE_EXPORT void *
mime_encoding (void *p) {
    const gchar *file = p;
    int flags =
        MAGIC_MIME_ENCODING |
        MAGIC_NO_CHECK_APPTYPE | MAGIC_SYMLINK | MAGIC_NO_CHECK_COMPRESS | MAGIC_NO_CHECK_TAR | MAGIC_PRESERVE_ATIME;
    return (void *)(lib_magic (file, flags));
}

G_MODULE_EXPORT void *
mime_file (void *p) {
    const gchar *file = p;
    int flags = MAGIC_NO_CHECK_ENCODING | MAGIC_SYMLINK | MAGIC_NO_CHECK_COMPRESS | MAGIC_NO_CHECK_TAR | MAGIC_PRESERVE_ATIME;
    return (void *)(lib_magic (file, flags));
}

#else
// deprecated forked magic...
# error "Deprecated forked magic no longer supported"
#endif

G_MODULE_EXPORT void *
mime_is_valid_command (void *p) {
    const char *cmd_fmt = p;
    TRACE ("MIME: mime_is_valid_command(%s)\n", cmd_fmt);
    GError *error = NULL;
    int argc;
    gchar *path;
    gchar **argv;
    if(!cmd_fmt)
        return GINT_TO_POINTER (FALSE);
    if(!g_shell_parse_argv (cmd_fmt, &argc, &argv, &error)) {
        gchar *msg = g_strcompress (error->message);
        g_warning ("%s: %s", msg, cmd_fmt);
        g_error_free (error);
        g_free (msg);
        return GINT_TO_POINTER (FALSE);
    }
    path = g_find_program_in_path (argv[0]);
    if(!path) {
        gboolean direct_path = g_file_test (argv[0], G_FILE_TEST_EXISTS) ||
            strncmp (argv[0], "./", strlen ("./")) == 0 || strncmp (argv[0], "../", strlen ("../")) == 0;
        //g_warning("argv[0]=%s",argv[0]);
        if(direct_path) {
            path = g_strdup (argv[0]);
        }
    }
    DBG ("mime_is_valid_command(): g_find_program_in_path(%s)=%s\n", argv[0], path);

    //if (!path || access(path, X_OK) != 0) {
    if(!path) {
        g_strfreev (argv);
        if(!path)
            errno = ENOENT;
        return GINT_TO_POINTER (FALSE);
    }
    // here we test for execution within sudo
    // XXX we could also check for commands executed in a terminal, but not today...
    void *retval=GINT_TO_POINTER(TRUE);
    if (strcmp(argv[0],"sudo")==0) {
        int i=1;
        if (strcmp(argv[i],"-A")==0) i++;
        retval=mime_is_valid_command((gpointer)argv[i]);
    }

    g_strfreev (argv);
    g_free (path);
    return retval;
}

static
gchar *get_hash_key_strstrip (void *p){
    gchar *pp=g_strdup((char *)p);
    g_strstrip(pp);
    gchar *key=get_hash_key ((gchar *)pp);
    g_free(pp);
    return key;
}

G_MODULE_EXPORT void *
mime_command_text (void *p) {
    if (!p) return NULL;
    gchar *key=get_hash_key_strstrip ((gchar *)p);
    void *value=g_hash_table_lookup (application_hash_text, key);
    g_free(key);
    return value;
}

G_MODULE_EXPORT void *
mime_command_text2 (void *p) {
    if (!p) return NULL;
    gchar *key=get_hash_key_strstrip ((gchar *)p);
    void *value=g_hash_table_lookup (application_hash_text2, key);
    g_free(key);
    return value;
}

G_MODULE_EXPORT void *
mime_command_icon (void *p) {
    if (!p) return NULL;
    gchar *key=get_hash_key_strstrip ((gchar *)p);
    void *value=g_hash_table_lookup (application_hash_icon, key);
    g_free(key);
    return value;
}

G_MODULE_EXPORT void *
mime_command_output (void *p) {
    if (!p) return NULL;
    gchar *key=get_hash_key_strstrip ((gchar *)p);
    void *value=g_hash_table_lookup (application_hash_output, key);
    g_free(key);
    return value;
}
G_MODULE_EXPORT void *
mime_command_output_ext (void *p) {
    if (!p) return NULL;
    gchar *key=get_hash_key_strstrip ((gchar *)p);
    void *value=g_hash_table_lookup (application_hash_output_ext, key);
    g_free(key);
    return value;
}

    
G_MODULE_EXPORT void *
mime_command (void *p) {
    const char *type = p;
    DBG ("APPS: mime_command(%s)\n", type);
    gchar **apps;
    int i;
    gchar *cmd_fmt = NULL;
    apps = locate_apps (type);
    if(!apps) {
        DBG ("APPS: --> NULL\n");
        return NULL;
    }
    if(!apps[0]) {
        DBG ("APPS: --> NULL\n");
        g_free (apps);
        return NULL;
    }

    for(i = 0; apps[i]; i++) {
        g_free (cmd_fmt);
        cmd_fmt = g_strcompress (apps[i]);
        if(mime_is_valid_command (cmd_fmt)) {
            g_strfreev (apps);
            DBG ("APPS: --> %s\n", cmd_fmt);
            return (void *)cmd_fmt;
        }
    }
    g_free (cmd_fmt);
    g_strfreev (apps);
    DBG ("APPS: --> NULL\n");
    return NULL;
}

G_MODULE_EXPORT void *
mime_apps (void *p) {
    const gchar *type = p;
    TRACE ("MIME: mime_apps(%s)\n", type);
    gchar **apps;
    apps = locate_apps (type);
    if(!apps)
        return NULL;
    if(!apps[0]) {
        g_free (apps);
        return NULL;
    }
    return (void *)apps;
}

G_MODULE_EXPORT void *
mime_typeinfo (void *p) {
    const gchar *type = p;
    xmlNodePtr node;
    xmlNodePtr subnode;
    xmlDocPtr doc;
    gchar *mimefile = NULL;
    gchar *mimetype = NULL;
    gchar *info = NULL;

    mimefile = g_build_filename (FREEDESKTOP_MIME_FILE, NULL);

    TRACE ("MIME: mimefile=%s\n", mimefile);
    if(access (mimefile, R_OK) != 0) {
        g_free (mimefile);
        mimefile = NULL;
        TRACE ("MIME: access(mimefile,R_OK)!=0\n");
        return NULL;
    }
    xmlKeepBlanksDefault (0);

    if((doc = xmlParseFile (mimefile)) == NULL) {
        TRACE ("MIME: rfm-modules: invalid xml file: %s\n", mimefile);
        g_free (mimefile);
        mimefile = NULL;
        return NULL;
    }

    node = xmlDocGetRootElement (doc);
    if(!xmlStrEqual (node->name, (const xmlChar *)"mime-info")) {
        TRACE ("MIME: rfm-modules: invalid xml file %s\n", mimefile);
        g_free (mimefile);
        mimefile = NULL;
        xmlFreeDoc (doc);
        return NULL;
    }

    /* Now parse the xml tree */
    TRACE ("MIME: TRACE: parsing %s\n", mimefile);
    TRACE ("MIME: ------------looking for %s\n", type);
    for(node = node->children; node; node = node->next) {
        if(xmlStrEqual (node->name, (const xmlChar *)"mime-type")) {
            mimetype = (char *)xmlGetProp (node, (const xmlChar *)"type");
            TRACE ("MIME: %s==%s\n", mimetype, type);
            if(xmlStrEqual ((const xmlChar *)mimetype, (const xmlChar *)type)) {
                for(subnode = node->children; subnode; subnode = subnode->next) {
                    if(xmlStrEqual (subnode->name, (const xmlChar *)"comment")) {
                        char *e = (char *)xmlNodeListGetString (doc, subnode->children,
                                                                1);
                        g_free (mimetype);
                        xmlFreeDoc (doc);
                        g_free (mimefile);
                        info = g_strdup (e);
                        if(e)
                            g_free (e);
                        return info;
                    }
                }

            }
            g_free (mimetype);
        }
    }
    xmlFreeDoc (doc);
    g_free (mimefile);
    return NULL;
}


G_MODULE_EXPORT void *
mime_add (void *p, void *q) {
    gchar *type = p;
    gchar *command = g_strdup((char *)q);
    g_strstrip(command);
    if(!command || !strlen (command)){
	g_free(command);
        return NULL;
    }
//    load_hashes ();

    DBG ("OPEN APPS: adding type %s->%s\n", type, command);
    add_type_to_hashtable(type, command);

    DBG ("THREAD: g_thread_create(gencache) from applications-module.c\n");

    GError *error = NULL;
    // thread will dispose of config_command:
    gchar *config_command=g_strdup_printf("%s:%s", type, command);
    g_free(command);
    g_thread_create (gencache, config_command, FALSE, &error); // non-hack
    if(error) {
          g_warning ("%s", error->message);
          g_error_free (error);
    }

    return NULL;
}

G_MODULE_EXPORT void *
mime_mk_command_line (void *p, void *q) {
    const gchar *command_fmt = p;
    const gchar *path = q;

    DBG ("MIME: mime_mk_command_line(%s)\n", path);
    gchar *command_line = NULL;
    gchar *fmt = NULL;

    if(!command_fmt)
        return NULL;
    if(!path)
        path = "";

    DBG ("MIME: command_fmt=%s\n", command_fmt);

    /* this is to send path as an argument */

    if(strstr (command_fmt, "%s")) {
        fmt = g_strdup (command_fmt);
    } else {
        fmt = g_strconcat (command_fmt, " %s", NULL);
    }
    DBG ("MIME: command_fmt fmt=%s\n", fmt);

    TRACE ("MIME: path=%s\n", path);
    gchar *esc_path = rfm_esc_string (path);
    command_line = g_strdup_printf (fmt, esc_path);
    g_free (esc_path);
    DBG ("MIME2: command_line=%s\n", command_line);

    g_free (fmt);
    return (void *)command_line;
}

G_MODULE_EXPORT void *
mime_mk_terminal_line (void *p) {
    const gchar *command = p;
    DBG ("MIME: mime_mk_command_line(%s)\n", command);
    gchar *command_line = NULL;

    if(!command)
        return NULL;

    gchar *term_path = rfm_what_term ();
    gchar *term;
    if(term_path) {
        term = g_path_get_basename (term_path);
    } else {
        term = g_strdup ("xterm");
    }
    const gchar *exec_flag = "-e";
    if(!mime_is_valid_command (term)) {
        g_warning ("%s == NULL", term);
        g_free (term);
        return NULL;
    }
    if(strstr (term, "gnome-terminal") || strstr (term, "gnome2-terminal") || strstr (term, "Terminal")) {
        exec_flag = "-x";
    }
    command_line = g_strdup_printf ("%s %s %s", term, exec_flag, command);
    g_free (term);

    return command_line;
}

/////////////////////////////////////////////////////////////////

G_MODULE_EXPORT const gchar *
g_module_check_init (GModule * module) {
    TRACE("***************g_module_check_init\n");
    application_hash_sfx = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_type = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_icon = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_text = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_text2 = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_output = g_hash_table_new (g_str_hash, g_str_equal);
    application_hash_output_ext = g_hash_table_new (g_str_hash, g_str_equal);
    if(!load_hashes_from_cache ()) 
    {
            mime_build_hashes ();
            GError *error = NULL;
            DBG ("THREAD: g_thread_create(gencache) from applications-module.i, load_hashes()\n");
            g_thread_create (gencache, NULL, FALSE, &error); // non-hack
            if(error) {
                g_warning ("%s", error->message);
                g_error_free (error);
            }
    }
    if(!magic_mutex) {
        magic_mutex = g_mutex_new ();
        cookie = magic_open (MAGIC_NONE);
        magic_load (cookie, NULL);
    }
    return NULL;
}

G_MODULE_EXPORT rodent_mime_functions *
module_init (void) {
    return GINT_TO_POINTER (1);
}

G_MODULE_EXPORT void
g_module_unload (GModule * module) {
    if(application_hash_sfx) {
        destroy_application_hash_sfx ();
    }
    if(application_hash_type) {
        destroy_application_hash_type ();
    }
}
