/*
 * Copyright (C) 2003-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "rodent_libs.h"

#include "find-module_gui.h"
#include "find-module_callbacks.h"

int result_limit=256;
gboolean default_recursive=TRUE;
gboolean default_recursiveH=FALSE;
gboolean default_xdev=TRUE;
gboolean default_case_sensitive=FALSE;
gboolean default_ext_regexp=FALSE;
gboolean default_no_look_in_binaries=TRUE;
int default_type_index=0;
gboolean default_anywhere=TRUE;
gboolean default_match_words=FALSE;
gboolean default_match_lines=FALSE;
gboolean default_match_no_match=FALSE;
    
gboolean have_grep = FALSE;
gboolean have_gnu_grep = FALSE;


GtkWidget *
fill_string_option_menu (GtkComboBox *om, GList * strings) {
    if (!om || ! strings){
	g_warning("create_string_option_menu: GtkComboBox==NULL || GList==NULL\n");
	return NULL;
    }
    GList *p;
    for (p=g_list_first(strings); p && p->data; p=p->next) {
	gtk_combo_box_append_text (om, (const gchar *)p->data);
      /* Note: The model will keep a copy of the string internally, 
       * so the list may be freed */
    }
    gtk_combo_box_set_active(om, default_type_index);
    return (GtkWidget *)om;
}

GtkWidget *
create_find_dialog (widgets_t * widgets_p) {
    GtkWidget *find_dialog;
    //GtkWidget *vbox7b;
    //GtkWidget *notebook1;
    GtkWidget *vbox6;
    GtkWidget *label42;
    GtkWidget *hbox25;
    GtkWidget *hbox14;
    GtkWidget *vbox7a;
    GtkWidget *hbox15;
    GtkWidget *label26;
    GtkWidget *path_combo;
    GtkWidget *hbox16;
    GtkWidget *label27;
    GtkWidget *filter_combo;
    GtkWidget *togglebutton2;
    GtkWidget *image2;
    GtkWidget *scrolledwindow8;
    GtkWidget *viewport3;
    GtkWidget *filter_help;
    GtkWidget *hbox17;
    GtkWidget *vbox12;
    GtkWidget *recursive;
    GtkWidget *recursiveH;
    GtkWidget *xdev;
    GtkWidget *hbox18;
    GtkObject *spinbutton1_adj;
    GtkWidget *spinbutton1;
    GtkWidget *label30;
    GtkWidget *label29;
    GtkWidget *label45;
    GtkWidget *hbox26;
    GtkWidget *vbox8;
    GtkWidget *hbox19;
    GtkWidget *label33;
    GtkWidget *grep_combo;
    GtkWidget *hbox20;
    GtkWidget *vbox13;
    GtkWidget *case_sensitive;
    GtkWidget *ext_regexp;
    GtkWidget *alignment1;
    GtkWidget *togglebutton3;
    GtkWidget *image1;
    GtkWidget *scrolledwindow9;
    GtkWidget *viewport4;
    GtkWidget *regexp_help;
    GtkWidget *hbox27;
    GtkWidget *vbox10;
    GtkWidget *hbox21;
    GtkWidget *label37;
    GtkWidget *file_type_om;
    GtkWidget *hbox28;
    GtkWidget *vbox11;
    GtkWidget *no_look_in_binaries;
    GtkWidget *hbox24;
    GtkWidget *label40;
    GtkWidget *anywhere;
    GSList *anywhere_group = NULL;
    GtkWidget *match_words;
    GtkWidget *match_lines;
    GtkWidget *match_no_match;
    GtkWidget *label41;
    GtkWidget *hbuttonbox2;
    GtkWidget *apply_button;
    GtkWidget *find_button;
    GtkWidget *close_button;
    GtkAccelGroup *accel_group;
    GdkPixbuf *pb;
    gchar *grep;

    grep = g_find_program_in_path ("grep");
    if(grep) {
        FILE *pipe;
        const gchar *cmd = "grep --version";
        have_grep = TRUE;
        pipe = popen (cmd, "r");
        if(pipe) {
            gchar line[256];
            memset (line, 0, 256);
            if(fgets (line, 255, pipe) == NULL)
                g_warning ("fgets: %s", strerror (errno));
            pclose (pipe);
            if(strstr (line, "GNU"))
                have_gnu_grep = TRUE;
        }
    }
    g_free (grep);

    accel_group = gtk_accel_group_new ();

    find_dialog = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (find_dialog), _("Find"));
    gtk_window_set_position (GTK_WINDOW (find_dialog), GTK_WIN_POS_MOUSE);

    view_t *view_p=widgets_p->view_p;
    if (view_p->type == DESKVIEW_TYPE) {
        gtk_window_set_keep_above (GTK_WINDOW(find_dialog), TRUE);
        gtk_window_stick (GTK_WINDOW(find_dialog));
    } else {
        gtk_window_set_modal (GTK_WINDOW (find_dialog), TRUE);
    }
    
    gtk_window_set_default_size (GTK_WINDOW (find_dialog), 425, -1);
    gtk_window_set_icon ((GtkWindow *) find_dialog, rfm_get_pixbuf("xffm/find_result", SIZE_ICON));
	    

  /*******************************************************************************
   vbox7b = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox7b);
    gtk_container_add (GTK_CONTAINER (find_dialog), vbox7b);

    notebook1 = gtk_notebook_new ();
    gtk_widget_show (notebook1);
    gtk_box_pack_start (GTK_BOX (vbox7b), notebook1, TRUE, TRUE, 0);
 **********************************************************************************/   

    vbox6 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox6);
    gtk_container_add (GTK_CONTAINER (find_dialog), vbox6);
//    gtk_container_add (GTK_CONTAINER (notebook1), vbox6);
//    gtk_notebook_set_tab_label_packing (GTK_NOTEBOOK (notebook1), vbox6, FALSE, FALSE, GTK_PACK_START);
    gtk_container_set_border_width (GTK_CONTAINER (vbox6), 5);

    gchar *t=g_strdup_printf("<b>%s</b>", _("Find"));
    label42 = gtk_label_new (t);
    g_free(t);
    gtk_widget_show (label42);
    gtk_box_pack_start (GTK_BOX (vbox6), label42, FALSE, FALSE, 0);
    gtk_label_set_use_markup (GTK_LABEL (label42), TRUE);
    gtk_misc_set_alignment (GTK_MISC (label42), 0, 0.5);

    hbox25 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox25);
    gtk_box_pack_start (GTK_BOX (vbox6), hbox25, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox25), 2);

    hbox14 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox14);
    gtk_box_pack_start (GTK_BOX (hbox25), hbox14, TRUE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox14), 5);

    vbox7a = gtk_vbox_new (FALSE, 5);
    gtk_widget_show (vbox7a);
    gtk_box_pack_start (GTK_BOX (hbox14), vbox7a, TRUE, TRUE, 0);

    hbox15 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox15);
    gtk_box_pack_start (GTK_BOX (vbox7a), hbox15, FALSE, FALSE, 0);

    t=g_strdup_printf("%s:", _("Path"));
    label26 = gtk_label_new (t);
    g_free(t);
    gtk_widget_show (label26);
    gtk_box_pack_start (GTK_BOX (hbox15), label26, FALSE, FALSE, 0);

    path_combo = gtk_combo_box_entry_new ();
    g_object_set_data (G_OBJECT (path_combo), "GladeParentKey", path_combo);
    gtk_widget_show (path_combo);
    gtk_box_pack_start (GTK_BOX (hbox15), path_combo, TRUE, TRUE, 0);


    GtkWidget *button = gtk_button_new ();
    GtkWidget *image = gtk_image_new_from_stock (GTK_STOCK_DIRECTORY, GTK_ICON_SIZE_BUTTON);
    gtk_button_set_image ((GtkButton *) button, image);
    GtkWidget *vbox = gtk_vbox_new (FALSE, 6);
    gtk_box_pack_start (GTK_BOX (hbox15), vbox, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 0);
    gtk_widget_show (vbox);
    gtk_widget_show (button);
    RODENT_HOOKUP_OBJECT (find_dialog, button, "fileselector");

    hbox16 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox16);
    gtk_box_pack_start (GTK_BOX (vbox7a), hbox16, TRUE, FALSE, 5);

    gchar *text=g_strdup_printf("%s ", _("Filter:"));
    label27 = gtk_label_new (text);
    g_free(text);
    gtk_widget_show (label27);
    gtk_box_pack_start (GTK_BOX (hbox16), label27, FALSE, FALSE, 0);

    filter_combo = gtk_combo_box_entry_new ();
    g_object_set_data (G_OBJECT (filter_combo), "GladeParentKey", filter_combo);
    gtk_widget_show (filter_combo);
    gtk_box_pack_start (GTK_BOX (hbox16), filter_combo, TRUE, TRUE, 0);

    togglebutton2 = gtk_toggle_button_new ();
    gtk_widget_show (togglebutton2);
    gtk_box_pack_start (GTK_BOX (hbox16), togglebutton2, FALSE, FALSE, 0);

    pb = rfm_get_pixbuf ("xffm/question", SIZE_BUTTON);
    image2 = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (image2);
    gtk_container_add (GTK_CONTAINER (togglebutton2), image2);

    scrolledwindow8 = gtk_scrolled_window_new (NULL, NULL);
    gtk_box_pack_start (GTK_BOX (vbox7a), scrolledwindow8, TRUE, TRUE, 0);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow8), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

    viewport3 = gtk_viewport_new (NULL, NULL);
    gtk_widget_show (viewport3);
    gtk_container_add (GTK_CONTAINER (scrolledwindow8), viewport3);

    filter_help = gtk_label_new ("");
    gtk_container_add (GTK_CONTAINER (viewport3), filter_help);

    hbox17 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox17);
    gtk_box_pack_start (GTK_BOX (vbox7a), hbox17, TRUE, FALSE, 0);

    vbox12 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox12);
    gtk_box_pack_start (GTK_BOX (hbox17), vbox12, FALSE, FALSE, 0);

    recursive = gtk_check_button_new_with_mnemonic (_("Recursive"));
    gtk_widget_show (recursive);
    gtk_box_pack_start (GTK_BOX (vbox12), recursive, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (recursive), 3);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (recursive), default_recursive);

    recursiveH = gtk_check_button_new_with_mnemonic (_( "Find hidden files and directories"));
    gtk_widget_show (recursiveH);
    gtk_box_pack_start (GTK_BOX (vbox12), recursiveH, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (recursiveH), 3);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (recursiveH), default_recursiveH);

    xdev = gtk_check_button_new_with_mnemonic (_("Stay on single filesystem"));
    gtk_widget_show (xdev);
    gtk_box_pack_start (GTK_BOX (vbox12), xdev, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (xdev), 3);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (xdev), default_xdev);

    hbox18 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox18);
    gtk_box_pack_start (GTK_BOX (hbox17), hbox18, TRUE, FALSE, 0);

    spinbutton1_adj = gtk_adjustment_new (result_limit, 0, 4096, 16, 64, 0);
    spinbutton1 = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton1_adj), 0, 0);
    gtk_widget_show (spinbutton1);
    gtk_box_pack_start (GTK_BOX (hbox18), spinbutton1, TRUE, TRUE, 0);
    gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (spinbutton1), GTK_UPDATE_IF_VALID);

    label30 = gtk_label_new ("  ");
    gtk_widget_show (label30);
    gtk_box_pack_start (GTK_BOX (hbox18), label30, FALSE, FALSE, 0);

    hbox27 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox27);
    gtk_box_pack_start (GTK_BOX (vbox6), hbox27, FALSE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox27), 2);

    vbox10 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox10);
    gtk_box_pack_start (GTK_BOX (hbox27), vbox10, TRUE, TRUE, 0);

    hbox21 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox21);
    gtk_box_pack_start (GTK_BOX (vbox10), hbox21, TRUE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox21), 5);

    label37 = gtk_label_new (_("File type : "));
    gtk_widget_show (label37);
    gtk_box_pack_start (GTK_BOX (hbox21), label37, FALSE, FALSE, 0);

    file_type_om =  gtk_combo_box_new_text();
    gtk_widget_show (file_type_om);
    gtk_box_pack_start (GTK_BOX (hbox21), file_type_om, TRUE, TRUE, 0);


    label29 = gtk_label_new (_("Result limit"));
    gtk_widget_show (label29);
    gtk_box_pack_start (GTK_BOX (hbox18), label29, TRUE, FALSE, 0);
    
    t=g_strdup_printf("<b>%s</b>", _("Content search"));
    label45 = gtk_label_new (t);
    g_free(t);
    gtk_widget_show (label45);
    gtk_box_pack_start (GTK_BOX (vbox6), label45, FALSE, FALSE, 1);
    gtk_label_set_use_markup (GTK_LABEL (label45), TRUE);
    gtk_misc_set_alignment (GTK_MISC (label45), 0, 0.5);

    hbox26 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox26);
    gtk_box_pack_start (GTK_BOX (vbox6), hbox26, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox26), 2);

    vbox8 = gtk_vbox_new (FALSE, 5);
    gtk_widget_show (vbox8);
    gtk_box_pack_start (GTK_BOX (hbox26), vbox8, TRUE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (vbox8), 5);

    hbox19 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox19);
    gtk_box_pack_start (GTK_BOX (vbox8), hbox19, FALSE, FALSE, 0);

    t=g_strdup_printf("%s: ",_("Contains the _text"));
    label33 = gtk_label_new (t);
    g_free(t);
    gtk_widget_show (label33);
    gtk_box_pack_start (GTK_BOX (hbox19), label33, FALSE, FALSE, 0);

    grep_combo = gtk_combo_box_entry_new ();
    g_object_set_data (G_OBJECT (grep_combo), "GladeParentKey", grep_combo);
    gtk_widget_show (grep_combo);
    gtk_box_pack_start (GTK_BOX (hbox19), grep_combo, TRUE, TRUE, 5);
    gtk_widget_set_sensitive (grep_combo, have_grep);

    hbox20 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox20);
    gtk_box_pack_start (GTK_BOX (vbox8), hbox20, FALSE, FALSE, 0);

    vbox13 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox13);
    gtk_box_pack_start (GTK_BOX (hbox20), vbox13, TRUE, TRUE, 0);

    case_sensitive = gtk_check_button_new_with_mnemonic (_("Case sensitive search"));
    gtk_widget_show (case_sensitive);
    gtk_box_pack_start (GTK_BOX (vbox13), case_sensitive, FALSE, FALSE, 0);
    gtk_widget_set_sensitive (case_sensitive, FALSE);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (case_sensitive), default_case_sensitive);

    ext_regexp = gtk_check_button_new_with_mnemonic (_("Extended regexp"));
    gtk_widget_show (ext_regexp);
    gtk_box_pack_start (GTK_BOX (vbox13), ext_regexp, FALSE, FALSE, 0);
    gtk_widget_set_sensitive (ext_regexp, FALSE);
    //gtk_widget_set_sensitive (ext_regexp, have_gnu_grep);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (ext_regexp), default_ext_regexp);

    alignment1 = gtk_alignment_new (0.5, 0.5, 0.5, 0.5);
    gtk_widget_show (alignment1);
    gtk_box_pack_start (GTK_BOX (hbox20), alignment1, FALSE, FALSE, 0);

    togglebutton3 = gtk_toggle_button_new ();
    gtk_widget_show (togglebutton3);
    gtk_container_add (GTK_CONTAINER (alignment1), togglebutton3);

    pb = rfm_get_pixbuf ("xffm/question", SIZE_BUTTON);
    image1 = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (image1);
    gtk_container_add (GTK_CONTAINER (togglebutton3), image1);
    gtk_widget_set_sensitive (togglebutton3, FALSE);

    scrolledwindow9 = gtk_scrolled_window_new (NULL, NULL);
    gtk_box_pack_start (GTK_BOX (vbox8), scrolledwindow9, TRUE, TRUE, 0);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow9), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

    viewport4 = gtk_viewport_new (NULL, NULL);
    gtk_widget_show (viewport4);
    gtk_container_add (GTK_CONTAINER (scrolledwindow9), viewport4);

    regexp_help = gtk_label_new ("");
    gtk_widget_show (regexp_help);
    gtk_container_add (GTK_CONTAINER (viewport4), regexp_help);

    hbox28 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox28);
    gtk_box_pack_start (GTK_BOX (vbox6), hbox28, TRUE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox28), 2);

    vbox11 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox11);
    gtk_box_pack_start (GTK_BOX (hbox28), vbox11, TRUE, TRUE, 0);

    no_look_in_binaries = gtk_check_button_new_with_mnemonic (_("Do not look into binary files"));
    gtk_widget_show (no_look_in_binaries);
    gtk_box_pack_start (GTK_BOX (vbox11), no_look_in_binaries, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (no_look_in_binaries), 5);
    if (have_gnu_grep) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (no_look_in_binaries), default_no_look_in_binaries);
    }
    gtk_widget_set_sensitive (no_look_in_binaries, FALSE);

    hbox24 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox24);
    gtk_box_pack_start (GTK_BOX (vbox11), hbox24, FALSE, FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox24), 5);

    label40 = gtk_label_new (_("match : "));
    gtk_widget_show (label40);
    gtk_box_pack_start (GTK_BOX (hbox24), label40, FALSE, FALSE, 0);
    gtk_misc_set_padding (GTK_MISC (label40), 3, 0);
    gtk_widget_set_sensitive (label40, FALSE);

    anywhere = gtk_radio_button_new_with_mnemonic (NULL, _("anywhere"));
    gtk_widget_show (anywhere);
    gtk_box_pack_start (GTK_BOX (hbox24), anywhere, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (anywhere), anywhere_group);
    anywhere_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (anywhere));
    gtk_widget_set_sensitive (anywhere, FALSE);
    if (default_anywhere) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (anywhere), default_anywhere);
    }
     

    match_words = gtk_radio_button_new_with_mnemonic (NULL, _("whole words"));
    gtk_widget_show (match_words);
    gtk_box_pack_start (GTK_BOX (hbox24), match_words, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (match_words), anywhere_group);
    anywhere_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (match_words));
    gtk_widget_set_sensitive (match_words, FALSE);
    if (default_match_words) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (match_words), default_match_words);
    }


    match_lines = gtk_radio_button_new_with_mnemonic (NULL, _("whole lines"));
    gtk_widget_show (match_lines);
    gtk_box_pack_start (GTK_BOX (hbox24), match_lines, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (match_lines), anywhere_group);
    anywhere_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (match_lines));
    gtk_widget_set_sensitive (match_lines, FALSE);
    if (default_match_lines) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (match_lines), default_match_lines);
    }

    match_no_match = gtk_radio_button_new_with_mnemonic (NULL, _("no match"));
    gtk_widget_show (match_no_match);
    gtk_box_pack_start (GTK_BOX (hbox24), match_no_match, TRUE, FALSE, 0);
    gtk_radio_button_set_group (GTK_RADIO_BUTTON (match_no_match), anywhere_group);
    anywhere_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (match_no_match));
    gtk_widget_set_sensitive (match_no_match, FALSE);
    if (default_match_no_match) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (match_no_match), default_match_no_match);
    }

    label41 = gtk_label_new ("");
    gtk_widget_show (label41);
    gtk_box_pack_start (GTK_BOX (vbox6), label41, FALSE, FALSE, 0);

    hbuttonbox2 = gtk_hbutton_box_new ();
    gtk_widget_show (hbuttonbox2);
    //gtk_box_pack_start (GTK_BOX (vbox7b), hbuttonbox2, FALSE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (vbox6), hbuttonbox2, FALSE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbuttonbox2), 3);
    gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox2), GTK_BUTTONBOX_EDGE);

    GtkWidget *button_image;
    find_button =  gtk_button_new_with_label(_("Find"));
    button_image = gtk_image_new_from_pixbuf (rfm_get_pixbuf("xffm/stock_search", SIZE_BUTTON));
    gtk_button_set_image ((GtkButton *)find_button, button_image);
    apply_button =  gtk_button_new_with_label(_("Apply"));
    button_image = gtk_image_new_from_pixbuf (rfm_get_pixbuf("xffm/stock_apply", SIZE_BUTTON));
    gtk_button_set_image ((GtkButton *)apply_button, button_image);
    close_button =  gtk_button_new_with_label(_("Close"));
    button_image = gtk_image_new_from_pixbuf (rfm_get_pixbuf("xffm/stock_close", SIZE_BUTTON));
    gtk_button_set_image ((GtkButton *)close_button, button_image);

    //gtk_button_new_from_stock ("gtk-find");
    gtk_widget_show (find_button);
    gtk_container_add (GTK_CONTAINER (hbuttonbox2), find_button);
    GTK_WIDGET_SET_FLAGS (find_button, GTK_CAN_DEFAULT);

    //apply_button = gtk_button_new_from_stock ("gtk-apply");
    gtk_widget_show (apply_button);
    gtk_container_add (GTK_CONTAINER (hbuttonbox2), apply_button);
    GTK_WIDGET_SET_FLAGS (apply_button, GTK_CAN_DEFAULT);

    //close_button = gtk_button_new_from_stock ("gtk-close");
    gtk_widget_show (close_button);
    gtk_container_add (GTK_CONTAINER (hbuttonbox2), close_button);
    gtk_widget_add_accelerator (close_button, "clicked", accel_group, GDK_Escape, 0, GTK_ACCEL_VISIBLE);

    /* Store pointers to all widgets, for use by rodent_get_widget_by_name(). */
    RODENT_HOOKUP_OBJECT (find_dialog, find_dialog, "find_dialog");
    //RODENT_HOOKUP_OBJECT (find_dialog, vbox7b, "vbox7b");
    //RODENT_HOOKUP_OBJECT (find_dialog, notebook1, "notebook1");
    RODENT_HOOKUP_OBJECT (find_dialog, vbox6, "vbox6");
    RODENT_HOOKUP_OBJECT (find_dialog, label42, "label42");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox25, "hbox25");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox14, "hbox14");
    RODENT_HOOKUP_OBJECT (find_dialog, vbox7a, "vbox7a");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox15, "hbox15");
    RODENT_HOOKUP_OBJECT (find_dialog, label26, "label26");
    RODENT_HOOKUP_OBJECT (find_dialog, path_combo, "path_combo");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox16, "hbox16");
    RODENT_HOOKUP_OBJECT (find_dialog, label27, "label27");
    RODENT_HOOKUP_OBJECT (find_dialog, filter_combo, "filter_combo");
    RODENT_HOOKUP_OBJECT (find_dialog, togglebutton2, "togglebutton2");
    RODENT_HOOKUP_OBJECT (find_dialog, image2, "image2");
    RODENT_HOOKUP_OBJECT (togglebutton2, scrolledwindow8, "scrolledwindow8");
    RODENT_HOOKUP_OBJECT (find_dialog, viewport3, "viewport3");
    RODENT_HOOKUP_OBJECT (find_dialog, filter_help, "filter_help");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox17, "hbox17");
    RODENT_HOOKUP_OBJECT (find_dialog, vbox12, "vbox12");
    RODENT_HOOKUP_OBJECT (find_dialog, recursive, "recursive");
    RODENT_HOOKUP_OBJECT (find_dialog, recursiveH, "recursiveH");
    RODENT_HOOKUP_OBJECT (find_dialog, xdev, "xdev");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox18, "hbox18");
    RODENT_HOOKUP_OBJECT (find_dialog, spinbutton1, "spinbutton1");
    RODENT_HOOKUP_OBJECT (find_dialog, label30, "label30");
    RODENT_HOOKUP_OBJECT (find_dialog, label29, "label29");
    RODENT_HOOKUP_OBJECT (find_dialog, label45, "label45");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox26, "hbox26");
    RODENT_HOOKUP_OBJECT (find_dialog, vbox8, "vbox8");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox19, "hbox19");
    RODENT_HOOKUP_OBJECT (find_dialog, label33, "label33");
    RODENT_HOOKUP_OBJECT (find_dialog, grep_combo, "grep_combo");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox20, "hbox20");
    RODENT_HOOKUP_OBJECT (find_dialog, vbox13, "vbox13");
    RODENT_HOOKUP_OBJECT (find_dialog, case_sensitive, "case_sensitive");
    RODENT_HOOKUP_OBJECT (find_dialog, ext_regexp, "ext_regexp");
    RODENT_HOOKUP_OBJECT (find_dialog, alignment1, "alignment1");
    RODENT_HOOKUP_OBJECT (find_dialog, togglebutton3, "togglebutton3");
    RODENT_HOOKUP_OBJECT (find_dialog, image1, "image1");
    RODENT_HOOKUP_OBJECT (togglebutton3, scrolledwindow9, "scrolledwindow9");
    RODENT_HOOKUP_OBJECT (find_dialog, viewport4, "viewport4");
    RODENT_HOOKUP_OBJECT (find_dialog, regexp_help, "regexp_help");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox27, "hbox27");
    RODENT_HOOKUP_OBJECT (find_dialog, vbox10, "vbox10");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox21, "hbox21");
    RODENT_HOOKUP_OBJECT (find_dialog, label37, "label37");
    RODENT_HOOKUP_OBJECT (find_dialog, file_type_om, "file_type_om");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox28, "hbox28");
    RODENT_HOOKUP_OBJECT (find_dialog, vbox11, "vbox11");
    RODENT_HOOKUP_OBJECT (find_dialog, no_look_in_binaries, "no_look_in_binaries");
    RODENT_HOOKUP_OBJECT (find_dialog, hbox24, "hbox24");
    RODENT_HOOKUP_OBJECT (find_dialog, label40, "label40");
    RODENT_HOOKUP_OBJECT (find_dialog, anywhere, "anywhere");
    RODENT_HOOKUP_OBJECT (find_dialog, match_words, "match_words");
    RODENT_HOOKUP_OBJECT (find_dialog, match_lines, "match_lines");
    RODENT_HOOKUP_OBJECT (find_dialog, match_no_match, "match_no_match");
    RODENT_HOOKUP_OBJECT (find_dialog, label41, "label41");
    RODENT_HOOKUP_OBJECT (find_dialog, hbuttonbox2, "hbuttonbox2");
    RODENT_HOOKUP_OBJECT (find_dialog, apply_button, "apply_button");
    RODENT_HOOKUP_OBJECT (find_dialog, find_button, "find_button");
    RODENT_HOOKUP_OBJECT (find_dialog, close_button, "close_button");

    gtk_widget_grab_default (find_button);
    gtk_window_add_accel_group (GTK_WINDOW (find_dialog), accel_group);

    return find_dialog;
}
