// system type test:
//
/* NOTE: linux will fail if HAVE_STATFS is before HAVE_GETMNTENT section */
#if defined(HAVE_GETMNTENT) && defined(HAVE_STRUCT_MNTENT_MNT_DIR)
    // Linux...
# define LINUX_CONFIG
#elif defined(HAVE_STATFS) && defined(HAVE_STRUCT_STATFS_F_MNTONNAME)
# define THIS_IS_BSD
// BSD functions are not thread safe... ergo, disable them for the time being.
#endif

/********/
#define MNTTYPE_PROCFS	"proc"
#define MNTTYPE_SMBFS	"smbfs"
#define MNTTYPE_DEV	"devpts"
#define MNTTYPE_SHM	"tmpfs"
#define	MNTTYPE_CDFS	"iso9660"
#define MNTTYPE_SWAP	"swap"
#define MNTTYPE_NFS	"nfs"

#define MNTTYPE_CACHEFS	"cachefs"       // proc
#define MNTTYPE_HSFS	"hsfs"  // proc

#define MNTTYPE_CODAFS	"coda"  //nfs
#define MNTTYPE_CDFS_AIX "cdrfs"        /* AIX */
#define MNTTYPE_KERNFS	"kernfs"        // proc
#define MNTTYPE_MFS	"mfs"   //proc
static GtkWidget *private_popup_widget = NULL;

static int
get_fstab_type (const char *type) {

    if(strstr (type, MNTTYPE_NFS)) {
        TRACE ("    got %s type for fstab item %s\n", MNTTYPE_NFS, type);
        return __NFS_TYPE;
    } else if(strstr (type, MNTTYPE_SMBFS)) {
        TRACE ("    got %s type for fstab item %s\n", MNTTYPE_SMBFS, type);
        return __SMB_TYPE;
    } else if(strstr (type, MNTTYPE_PROCFS)) {
        TRACE ("    got %s type for fstab item %s\n", MNTTYPE_PROCFS, type);
        return __PROC_TYPE;
    } else if(strstr (type, MNTTYPE_DEV)) {
        TRACE ("    got %s type for fstab item %s\n", MNTTYPE_DEV, type);
        return __PROC_TYPE;
    } else if(strstr (type, MNTTYPE_SHM)) {
        TRACE ("    got %s type for fstab item %s\n", MNTTYPE_SHM, type);
        return __PROC_TYPE;
    } else if(strstr (type, MNTTYPE_CDFS)) {
        TRACE ("    got %s type for fstab item%s\n", MNTTYPE_CDFS, type);
        return __CDFS_TYPE;
    } else if(strstr (type, MNTTYPE_CDFS_AIX))
        return __CDFS_TYPE;
    else if(strstr (type, MNTTYPE_CACHEFS))
        return __PROC_TYPE;
    else if(strstr (type, MNTTYPE_HSFS))
        return __PROC_TYPE;
    else if(strstr (type, MNTTYPE_KERNFS))
        return __PROC_TYPE;
    else if(strstr (type, MNTTYPE_MFS))
        return __PROC_TYPE;
    else if(strstr (type, MNTTYPE_MFS))
        return __NFS_TYPE;

    return 1;
}

G_MODULE_EXPORT void *
is_mounted (void *p) {
    if(!p) {
        g_warning ("fstab.i:is_mounted() mnt_point != NULL not met!");
        return NULL;
    }
    const gchar *mnt_point = (const gchar *)p;
    const gchar *mnt_fsname = (const gchar *)p;

#if defined LINUX_CONFIG
    // Linux...
    struct mntent *m;
    const gchar *mnttab;
    FILE *tab_file;

    // try /proc/mounts first

    if((tab_file = fopen ("/proc/mounts", "r")) != NULL) {
        (void)fclose (tab_file);
        mnttab = "/proc/mounts";
    } else
        mnttab = "/etc/mtab";
    tab_file = setmntent (mnttab, "r");

    if(!tab_file) {
        perror ("setmntent:");
        return GINT_TO_POINTER (-1);
    }
    while((m = getmntent (tab_file)) != NULL) {
        DBG ("FSTAB-TRACE:%s:  %s  or  %s\n", mnt_point, m->mnt_dir, m->mnt_fsname);
        if(strcmp (m->mnt_dir, mnt_point) == 0 || strcmp (m->mnt_fsname, mnt_fsname) == 0) {
            DBG ("FSTAB-TRACE: GOTCHA  %s  %s:%s\n", m->mnt_dir, m->mnt_fsname, mnt_point);
            endmntent (tab_file);
            return GINT_TO_POINTER (1);
        }
    }
    endmntent (tab_file);
    return NULL;
#elif defined THIS_IS_BSD
    // BSD
    struct statfs statfsbuf;
    if(statfs (mnt_point, &statfsbuf) == 0) {
        TRACE ("FSTAB: path=%s mnttoname=%s\n", mnt_point, statfsbuf.f_mntonname);
        if(strcmp (mnt_point, statfsbuf.f_mntonname) == 0) {
            TRACE ("FSTAB: path=%s is MOUNTED\n", mnt_point);
            return GINT_TO_POINTER (1);
        }
    }
    return NULL;
#else
    return GINT_TO_POINTER (-1);
# warning "missing code here. "
#endif
}

#if defined LINUX_CONFIG
# define MTAB "/etc/mtab";

/*
 * HP-UX, Linux, SunOS 4.x
 *
 * Ubuntu linux has both HAVE_GETMNTENT and HAVE_STRUCT_MNTENT_MNT_DIR
 */

static gint
my_strcmp (gconstpointer a, gconstpointer b) {
    TRACE ("FSTAB:     compare %s <=> %s\n", (char *)a, (char *)b);
    return (strcmp ((char *)a, (char *)b));
}

static GList *
destroy_fstab_list (GList ** fstab_list_p) {
    GList *p;
    for(p = *fstab_list_p; p && p->data; p = p->next) {
        g_free (p->data);
    }
    g_list_free (*fstab_list_p);
    return NULL;
}

static gboolean
include_in_xfdir (struct mntent *mnt_struct, GList ** fstab_list_p) {
    /* skip swap */
    if(strstr (mnt_struct->mnt_type, MNTTYPE_SWAP))
        return FALSE;
    /* skip missing mount points */
    if(!g_file_test (mnt_struct->mnt_dir, G_FILE_TEST_IS_DIR))
        return FALSE;
    if(g_list_find_custom (*fstab_list_p, mnt_struct->mnt_dir, my_strcmp) == NULL) {
        *fstab_list_p = g_list_append (*fstab_list_p, g_strdup (mnt_struct->mnt_dir));
        TRACE ("FSTAB: adding %s\n", mnt_struct->mnt_dir);
        return TRUE;
    } else {
        TRACE ("FSTAB: skipping %s\n", mnt_struct->mnt_dir);
        return FALSE;
    }
    return TRUE;
}

static int
count_elements (void
    ) {

    GList *fstab_list = NULL;
    struct mntent *mnt_struct;
    FILE *fstab_fd;
    int i = 0;
//    char *files[]={"/etc/fstab", NULL};
    char *files[] = { "/etc/fstab", "/etc/mtab", NULL };
    char **p;
    for(p = files; *p; p++) {
        if((fstab_fd = setmntent (*p, "r")) == NULL) {
            g_warning ("Unable to open %s", "/etc/fstab");
            return (0);
        }
        for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
            if(!include_in_xfdir (mnt_struct, &fstab_list))
                continue;
            i++;
        }
        (void)endmntent (fstab_fd);
    }
    fstab_list = destroy_fstab_list (&fstab_list);
    return i;
}

static gboolean
is_user_type (const gchar * mnt_point) {
    FILE *fstab_fd;
    if((fstab_fd = setmntent ("/etc/fstab", "r")) == NULL) {
        g_warning ("Unable to open %s", "/etc/fstab");
        return FALSE;
    }
    struct mntent *mnt_struct;
    gboolean user_type = FALSE;
    for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
        if(strstr (mnt_struct->mnt_type, MNTTYPE_SWAP))
            continue;
        if(!g_file_test (mnt_struct->mnt_dir, G_FILE_TEST_IS_DIR))
            continue;
        if(strcmp (mnt_point, mnt_struct->mnt_dir) == 0 && strstr (mnt_struct->mnt_opts, "user")) {
            user_type = TRUE;
            break;
        }
    }
    (void)endmntent (fstab_fd);
    return user_type;
}

static void
get_mnt_items (xfdir_t * xfdir_p) {

    GList *fstab_list = NULL;
    char *files[] = { "/etc/fstab", "/etc/mtab", NULL };
    char **p;
    int i = 1;
    struct mntent *mnt_struct;
    for(p = files; *p; p++) {
        TRACE ("FSTAB:  parsing %s\n", *p);
        FILE *fstab_fd = setmntent (*p, "r");

        for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
            if(!include_in_xfdir (mnt_struct, &fstab_list))
                continue;

            xfdir_p->gl[i].en = rfm_stat_entry (mnt_struct->mnt_dir, 0);
            if(strcmp (*p, "/etc/mtab") == 0) {
                DBG ("MTAB: setting MTAB type for %s\n", xfdir_p->gl[i].en->path);
                SET_MTAB_TYPE (xfdir_p->gl[i].en->type);
            } else {
                xfdir_p->gl[i].en->tag = g_strdup (mnt_struct->mnt_fsname);
            }
            xfdir_p->gl[i].pathv = g_strdup (mnt_struct->mnt_dir);
            TRACE ("FSTAB:  adding item %d %s \n", i, xfdir_p->gl[i].pathv);

            SET_FSTAB_TYPE (xfdir_p->gl[i].en->type);
            if(strstr (mnt_struct->mnt_opts, "user")) {
                SET_USER_TYPE (xfdir_p->gl[i].en->type);
            }
            /* set type */
            TRACE ("FSTAB:   checking %s for fstab type\n", mnt_struct->mnt_type);
            int type = get_fstab_type (mnt_struct->mnt_type);
            switch (type) {
            case __NFS_TYPE:
                SET_NFS_TYPE (xfdir_p->gl[i].en->type);
                break;
            case __CDFS_TYPE:
                SET_CDFS_TYPE (xfdir_p->gl[i].en->type);
                break;
            case __PROC_TYPE:
                SET_PROC_TYPE (xfdir_p->gl[i].en->type);
                break;
            case __SMB_TYPE:
                SET_SMB_TYPE (xfdir_p->gl[i].en->type);
                break;
            default:
                break;
            }
            i++;
        }
        (void)endmntent (fstab_fd);
    }
    fstab_list = destroy_fstab_list (&fstab_list);

}

G_MODULE_EXPORT gchar *
get_mnt_dir (gchar * mnt_fsname) {
    FILE *fstab_fd = setmntent ("/etc/mtab", "r");
    if(!fstab_fd)
        return NULL;
    struct mntent *mnt_struct;
    gchar *mnt_dir = NULL;
    for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
        if(strcmp (mnt_fsname, mnt_struct->mnt_fsname) == 0) {
            // hit: multiple entries use first listed 
            // user types have preference and use last listed 
            if(strstr (mnt_struct->mnt_opts, "user")) {
                g_free (mnt_dir);
                mnt_dir = g_strdup (mnt_struct->mnt_dir);
            }
            if(!mnt_dir) {
                mnt_dir = g_strdup (mnt_struct->mnt_dir);
            }
        }
    }
    (void)endmntent (fstab_fd);
    return mnt_dir;
}

G_MODULE_EXPORT gchar *
get_mnt_fsname (gchar * mnt_dir) {
    FILE *fstab_fd = setmntent ("/etc/mtab", "r");
    if(!fstab_fd)
        return NULL;
    struct mntent *mnt_struct;
    gchar *mnt_fsname = NULL;
    for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
        if(strcmp (mnt_dir, mnt_struct->mnt_dir) == 0) {
            // hit: multiple entries use first listed 
            // user types have preference and use last listed 
            if(strstr (mnt_struct->mnt_opts, "user")) {
                g_free (mnt_fsname);
                mnt_fsname = g_strdup (mnt_struct->mnt_fsname);
            }
            if(!mnt_fsname) {
                mnt_fsname = g_strdup (mnt_struct->mnt_fsname);
            }
        }
    }
    (void)endmntent (fstab_fd);
    return mnt_fsname;
}

static void
set_partition_info (record_entry_t * en) {
    FILE *fstab_fd = setmntent ("/etc/fstab", "r");
    if(!fstab_fd)
        return;
    gchar *partition = en->path;
    struct mntent *mnt_struct;
    for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
        if(strcmp (partition, mnt_struct->mnt_fsname) == 0) {
            DBG ("MOUNT: %s == %s\n", partition, mnt_struct->mnt_fsname);
            // hit: multiple entries use last listed if not user type 
            if(strstr (mnt_struct->mnt_opts, "user")) {
                SET_USER_TYPE (en->type);
                g_free (en->tag);
                en->tag = g_strdup (mnt_struct->mnt_dir);
            } else {
                if(!IS_USER_TYPE (en->type)) {
                    g_free (en->tag);
                    en->tag = g_strdup (mnt_struct->mnt_dir);
                }
            }
        }
    }
    (void)endmntent (fstab_fd);
    return;
}

static int
partition_items (xfdir_t * xfdir_p, int i, gboolean full_read) {
    FILE *partitions = fopen ("/proc/partitions", "r");
    if(partitions) {
        char line[1024];
        memset (line, 0, 1024);
        while(fgets (line, 1023, partitions) && !feof (partitions)) {
            if(strlen (line) < 5)
                continue;
            if(strchr (line, '#'))
                continue;
            DBG ("FSTAB: %s\n", line);
            gchar *p = strrchr (line, ' ');
            if(p) {
                DBG ("FSTAB: add --> %s\n", p);
                g_strstrip (p);
                //if (strlen(p) && isdigit(p[strlen(p)-1])) {
                if(strlen (p)) {
                    if(full_read) {
                        xfdir_p->gl[i].pathv = g_strdup_printf ("/dev/%s", p);
                        xfdir_p->gl[i].en = rfm_stat_entry (xfdir_p->gl[i].pathv, 0);
//                xfdir_p->gl[i].en = rfm_mk_entry(0);
                        xfdir_p->gl[i].en->path = g_strdup_printf ("/dev/%s", p);
                        xfdir_p->gl[i].en->mimetype = g_strdup ("inode/blockdevice");
                        TRACE ("FSTAB:  adding item %d %s \n", i, p);
                        SET_PARTITION_TYPE (xfdir_p->gl[i].en->type);
                        set_partition_info (xfdir_p->gl[i].en);
                    }
                    i++;
                }
            }
            memset (line, 0, 1024);
        }
        fclose (partitions);
    }
    return i;
}

static int
count_partitions (void
    ) {
    return partition_items (NULL, 0, FALSE);
}

static int
get_partition_items (xfdir_t * xfdir_p, int i) {
    return partition_items (xfdir_p, i, TRUE);
}

// Linux
static xfdir_t *
private_get_xfdir (xfdir_t * xfdir_p) {

    int elements = count_elements ();
    int partitions = count_partitions ();

    xfdir_p->pathc = elements;
    xfdir_p->pathc += partitions;
    DBG ("FSTAB:  counted %ld entries for fstab and mtab\n", (long)xfdir_p->pathc);
    /* up item: */
    xfdir_p->pathc++;

    xfdir_p->gl = (dir_t *) malloc (xfdir_p->pathc * sizeof (dir_t));
    memset (xfdir_p->gl, 0, xfdir_p->pathc * sizeof (dir_t));

    // up item:
    xfdir_p->gl[0].en = NULL;
    xfdir_p->gl[0].pathv = g_strdup (g_get_host_name ());
    // fstab and mtab:
    get_mnt_items (xfdir_p);
    // partitions:
    get_partition_items (xfdir_p, elements + 1);

    int i;
    for(i = 1; i < xfdir_p->pathc; i++) {
        xfdir_p->gl[i].en->parent_module = module_name ();
    }

    DBG ("FSTAB: gl is 0x%lx with count=%ld\n", (unsigned long)((long)xfdir_p->gl), (long)xfdir_p->pathc);
    // force a icon reload for the monitor thread:
    xfdir_p->invalidate_icons = TRUE;

    return (xfdir_p);
}

/* 
 * This should return the fstab type, -1 if default or NULL is not in fstab...
 * Fstab types *are* (or should be) supported in constants.h.
 */
G_MODULE_EXPORT void *
is_in_fstab (void *p) {
    struct mntent *mnt_struct;
    FILE *fstab_fd;
    int result = 0;
    const gchar *path = (const gchar *)p;
    if((fstab_fd = setmntent ("/etc/fstab", "r")) == NULL) {
        fprintf (stderr, "Unable to open %s\n", "/etc/fstab");
        return (NULL);
    }

    for(mnt_struct = getmntent (fstab_fd); mnt_struct != NULL; mnt_struct = getmntent (fstab_fd)) {
        if(strstr (mnt_struct->mnt_type, MNTTYPE_SWAP))
            continue;
        if(!g_file_test (mnt_struct->mnt_dir, G_FILE_TEST_IS_DIR))
            continue;

        DBG ("TRACE2:%s --->  %s   or   %s\n", path, mnt_struct->mnt_dir, mnt_struct->mnt_fsname);

        if(strcmp (path, mnt_struct->mnt_dir) == 0 || strcmp (path, mnt_struct->mnt_fsname) == 0) {
            result = get_fstab_type (mnt_struct->mnt_type);
            DBG ("TRACE2:%s ---> %d %s\n", mnt_struct->mnt_fsname, result, mnt_struct->mnt_type);
            break;
        }
    }

    (void)endmntent (fstab_fd);

    return GINT_TO_POINTER (result);
}

#elif defined THIS_IS_BSD

/*
 * AIX, BSD, OSF1, ULTRIX
 *
 * Ubuntu linux also has HAVE_GETFSENT, but has met preprocessor criteria above.
 */

static gboolean
is_user_type (const gchar * mnt_point) {
    return FALSE;
}

static gboolean
include_in_xfdir (struct fstab *fs) {
    if(strcmp (MNTTYPE_SWAP, fs->fs_vfstype) == 0)
        return FALSE;
    if(!g_file_test (fs->fs_file, G_FILE_TEST_IS_DIR))
        return FALSE;
    return TRUE;
}

static int
count_elements (void
    ) {
    if(!setfsent ())
        return (0);
    struct fstab *fs;
    int i;
    for(i = 0, fs = getfsent (); fs != NULL; fs = getfsent ()) {
        if(!include_in_xfdir (fs))
            continue;
        i++;
    }
    DBG ("fstab: count_elements = %d\n", i);
    //endfsent ();
    return i;
}

static int
count_partitions (void
    ) {
    return 0;
}

G_MODULE_EXPORT gchar *
get_mnt_dir (gchar * mnt_fsname) {
    struct fstab *fs;

    if(!setfsent ())
        return (NULL);

    gchar *mnt_dir = NULL;
    for(fs = getfsent (); fs != NULL; fs = getfsent ()) {
        if(!include_in_xfdir (fs))
            continue;

        if(strcmp (mnt_fsname, fs->fs_spec) == 0) {
            //if(strcmp (mnt_fsname, mnt_struct->mnt_fsname) == 0) {
            // hit: multiple entries use first listed 
            // user types have preference and use last listed 
            if(strstr (fs->fs_mntops, "user")) {
                g_free (mnt_dir);
                mnt_dir = g_strdup (fs->fs_file);
            }
            if(!mnt_dir) {
                mnt_dir = g_strdup (fs->fs_file);
            }
        }
    }

    //endfsent ();
    return mnt_dir;
}

G_MODULE_EXPORT gchar *
get_mnt_fsname (gchar * mnt_dir) {
    struct fstab *fs;

    if(!setfsent ())
        return (NULL);

    gchar *mnt_fsname = NULL;

    for(fs = getfsent (); fs != NULL; fs = getfsent ()) {
        if(!include_in_xfdir (fs))
            continue;
        if(strcmp (mnt_dir, fs->fs_file) == 0) {
            //if(strcmp (mnt_dir, mnt_struct->mnt_dir) == 0) {
            // hit: multiple entries use first listed 
            // user types have preference and use last listed 
            if(strstr (fs->fs_mntops, "user")) {
                g_free (mnt_fsname);
                mnt_fsname = g_strdup (fs->fs_spec);
            }
            if(!mnt_fsname) {
                mnt_fsname = g_strdup (fs->fs_spec);
            }
        }
    }
    //endfsent ();
    return mnt_fsname;
}

//BSD
static xfdir_t *
private_get_xfdir (xfdir_t * xfdir_p) {
    struct fstab *fs;

    int i = count_elements ();
    if(!setfsent ())
        return (0);

    DBG ("fstab: -count_elements()=%d\n", i);
    /* up item ( + 1): */
    xfdir_p->pathc = i + 1;

    if((xfdir_p->gl = (dir_t *) malloc (xfdir_p->pathc * sizeof (dir_t))) == NULL) {
        endfsent ();
        g_warning ("malloc(%d) failed at fstab-module.i!", xfdir_p->pathc * sizeof (dir_t));
        return (0);
    }
    memset (xfdir_p->gl, 0, xfdir_p->pathc * sizeof (dir_t));

    // up item:
    xfdir_p->gl[0].en = NULL;
    xfdir_p->gl[0].pathv = g_strdup (g_get_host_name ());

    for(i = 1, fs = getfsent (); fs != NULL; fs = getfsent ()) {
        if(!include_in_xfdir (fs)) {
            DBG ("fstab: skipping --> %s\n", fs->fs_file);
            continue;
        }

        xfdir_p->gl[i].en = rfm_stat_entry (fs->fs_file, 0);
        xfdir_p->gl[i].en->parent_module = module_name ();
        xfdir_p->gl[i].pathv = g_strdup (fs->fs_file);

        SET_FSTAB_TYPE (xfdir_p->gl[i].en->type);
        DBG ("fstab: %d --> %s\n", i, fs->fs_file);

        if(strcmp (MNTTYPE_CDFS, fs->fs_vfstype) == 0)
            SET_CDFS_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_CODAFS, fs->fs_vfstype) == 0)
            SET_NFS_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_KERNFS, fs->fs_vfstype) == 0)
            SET_PROC_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_MFS, fs->fs_vfstype) == 0)
            SET_PROC_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_NFS, fs->fs_vfstype) == 0)
            SET_NFS_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_PROCFS, fs->fs_vfstype) == 0)
            SET_PROC_TYPE (xfdir_p->gl[i].en->type);
        else if(strcmp (MNTTYPE_SMBFS, fs->fs_vfstype) == 0)
            SET_SMB_TYPE (xfdir_p->gl[i].en->type);

        i++;
    }

    //endfsent ();
    // force a icon reload for the monitor thread:
    xfdir_p->invalidate_icons = TRUE;

    return (xfdir_p);
}

void *
is_in_fstab (void *p) {
    int result = 0;
    struct fstab *fs;
    const gchar *path = (const gchar *)p;

    if(!setfsent ())
        return (NULL);

    for(fs = getfsent (); fs != NULL; fs = getfsent ()) {
        if(strcmp (MNTTYPE_SWAP, fs->fs_vfstype) == 0)
            continue;
        if(!g_file_test (fs->fs_file, G_FILE_TEST_IS_DIR))
            continue;

        if(strcmp (path, fs->fs_file) == 0) {
            if(strcmp (MNTTYPE_CDFS, fs->fs_vfstype) == 0)
                result = __CDFS_TYPE;
            else if(strcmp (MNTTYPE_CODAFS, fs->fs_vfstype) == 0)
                result = __NFS_TYPE;
            else if(strcmp (MNTTYPE_KERNFS, fs->fs_vfstype) == 0)
                result = __PROC_TYPE;
            else if(strcmp (MNTTYPE_MFS, fs->fs_vfstype) == 0)
                result = __PROC_TYPE;
            else if(strcmp (MNTTYPE_NFS, fs->fs_vfstype) == 0)
                result = __NFS_TYPE;
            else if(strcmp (MNTTYPE_PROCFS, fs->fs_vfstype) == 0)
                result = __PROC_TYPE;
            else if(strcmp (MNTTYPE_SMBFS, fs->fs_vfstype) == 0)
                result = __SMB_TYPE;
            else
                result = -1;
            break;
        }
    }
    //endfsent ();
    return GINT_TO_POINTER (result);
}

#else
# error "This should not happen: fstab-module.i (Solaris no longer supported)"

#endif

///////////   common code:

static gchar *
df (void
    ) {
    gchar *df_string = NULL;
    int line_count = 0;
    char line[2048];
#if defined(THIS_IS_BSD)
    FILE *pipe;
    memset ((void *)line, 0, 2048);
    gchar *command = g_find_program_in_path ("df");
    pipe = popen (command, "r");
    g_free (command);
    if(!pipe) {
        g_warning ("unable to pipe df");
        return "";
    }
    while(fgets (line, 2047, pipe) && !feof (pipe)) {
        line_count++;
/*	if (df_string) df_string=g_strconcat(df_string,line,NULL);
	else df_string=g_strdup(line);*/
    }
    pclose (pipe);
#elif defined(LINUX_CONFIG)
    // This is for Linux...*/
    gchar *files[] = { "/proc/partitions", "/proc/mounts", "/etc/mtab", NULL };
    gchar **p;
    for(p = files; p && *p; p++) {
        FILE *partitions = fopen (*p, "r");
        if(partitions) {
            while(fgets (line, 2047, partitions) && !feof (partitions)) {
                line_count++;
            }
            fclose (partitions);
        }
    }
#endif
    df_string = g_strdup_printf ("line_count=%d", line_count);
    TRACE ("DF: %s\n", df_string);
    return df_string;
}
