/*  rfm_settings_dialog.c
 *
 *  Copyright (C) 2003-2011 edscott wilson garcia <edscott@xfce.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define SETTINGS_MODULE_C

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <gdk/gdkx.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include "rfm_libs.h"
#include "settings-module.h"

#ifndef RFM_MCS_PLUGIN
# define RFM_MCS_PLUGIN
#endif

#define CHANNEL  "rfm"

#define BORDER 6
#define SKIP BORDER

GtkWidget *hide_book,
 *hide_local,
 *hide_network,
 *hide_find,
 *hide_trash,
 *hide_fstab,
 *hide_recent,
 *hide_frequent;

typedef struct {
    gchar *variable;
    gchar *value;
    gboolean editable;
} Item;

enum {
    COLUMN_VARIABLE,
    COLUMN_VALUE,
    COLUMN_EDITABLE,
    NUM_COLUMNS
};

static GtkWidget *dialog = NULL;

/********************************************************/
static McsManager *mcs_manager = NULL;

void *run_rfm_settings_dialog (McsPlugin *);
#include "settings-module-candy.i"
#include "settings-module.i"
LIBRFM_MODULE
//  
static GtkWidget *
create_tab (GtkNotebook * notebook, char *label, char *frame_label) {
    GtkWidget *tab_label1 = gtk_label_new (label);
    gtk_widget_show (tab_label1);
    GtkWidget *page_box = gtk_vbox_new (FALSE, 6);
    GtkWidget *sw = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw), GTK_SHADOW_ETCHED_IN);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_widget_show (sw);
    if(gtk_notebook_append_page (notebook, sw, tab_label1) < 0) {
        g_error ("Cannot append page to gtk_notebook!");
    }

    gtk_scrolled_window_add_with_viewport ((GtkScrolledWindow *) sw, page_box);
    gtk_widget_show (page_box);

    add_spacer (GTK_BOX (page_box));
    GtkWidget *frame = xfce_framebox_new (frame_label, TRUE);
    gtk_widget_show (frame);
    gtk_box_pack_start (GTK_BOX (page_box), frame, FALSE, FALSE, 0);

    GtkWidget *vbox = gtk_vbox_new (FALSE, 6);
    xfce_framebox_add (XFCE_FRAMEBOX (frame), vbox);
    return vbox;

}

static void 
subtitle(GtkWidget *parent, const gchar *text){
    GtkWidget *label_box=gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start ((GtkBox *) parent, label_box, FALSE, FALSE, 0);
    gchar *tab_text = g_strdup_printf ("<b><i>%s</i></b>", text);
    GtkWidget *label = gtk_label_new (tab_text);
    gtk_label_set_use_markup (GTK_LABEL (label), TRUE);
    //gtk_frame_set_label_widget (GTK_FRAME (framebox), label);
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
    gtk_box_pack_start ((GtkBox *) label_box, label, FALSE, FALSE, 0);
    gtk_widget_show_all (label_box);
    g_free(tab_text);
}

static GtkWidget *
make_exec_combo_box (GtkWidget * vbox, gint id, char **options, void *callback) {
    GtkWidget *box = gtk_combo_box_new_text ();
    int place = 0;
    GtkWidget *hbox = gtk_hbox_new (FALSE, 6);
    GtkWidget *label = gtk_label_new (_(environ_v[id].env_text));
    gtk_box_pack_start ((GtkBox *) hbox, label, FALSE, FALSE, 0);

    if(getenv (environ_v[id].env_var)
       && strlen (getenv (environ_v[id].env_var))) {
        gchar *path = g_path_get_basename (getenv (environ_v[id].env_var));
        gchar *cmd = g_find_program_in_path (path);
        if(cmd)
            gtk_combo_box_insert_text ((GtkComboBox *) box, place++, path);
        g_free (cmd);
        g_free (path);
    }
    char **p;
    for(p = options; *p; p++) {
        gchar *path = g_find_program_in_path (*p);
        if(path) {
            if(place == 0) {
                /* default values. Null values are not functionally acceptable */
                gtk_combo_box_insert_text ((GtkComboBox *) box, place++, *p);
                if(!rfm_options[id].value || !strlen (rfm_options[id].value)) {
                    g_free (rfm_options[id].value);
                    rfm_options[id].value = g_strdup (*p);
                    mcs_manager_set_string (mcs_manager, rfm_options[id].name, CHANNEL, *p);
                    mcs_manager_notify (mcs_manager, CHANNEL);
                }
            } else
                gtk_combo_box_append_text ((GtkComboBox *) box, *p);
            g_free (path);
        }
    }

    gtk_combo_box_set_active ((GtkComboBox *) box, 0);
    gtk_box_pack_start ((GtkBox *) hbox, box, FALSE, FALSE, 0);
    gtk_box_pack_start ((GtkBox *) vbox, hbox, FALSE, FALSE, 0);
    g_signal_connect (box, "changed", G_CALLBACK (callback), NULL);
    return box;
}

static GtkWidget *
make_file_chooser_button (int id, gboolean is_folder, GtkWidget * vbox) {
    GtkWidget *hbox = gtk_hbox_new (FALSE, 6);
    GtkWidget *label = gtk_label_new (_(environ_v[id].env_text));
    gtk_box_pack_start ((GtkBox *) hbox, label, FALSE, FALSE, 0);
    GtkWidget *button;
    if(is_folder) {
        button = gtk_file_chooser_button_new (_("Select a folder"), GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
    } else {
        button = gtk_file_chooser_button_new (_("Select Files..."), GTK_FILE_CHOOSER_ACTION_OPEN);
    }

    if(getenv (environ_v[id].env_var) && strlen (getenv (environ_v[id].env_var))) {
        gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (button), getenv (environ_v[id].env_var));
    } else {
        gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (button), g_get_home_dir ());
        if(is_folder)
            gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (button), _("Desktop"));
    }

    gtk_file_chooser_button_set_title ((GtkFileChooserButton *) (button), (const gchar *)_(environ_v[id].env_text));
    gtk_file_chooser_button_set_width_chars ((GtkFileChooserButton *) button, 15);
    gtk_box_pack_start ((GtkBox *) hbox, button, FALSE, FALSE, 0);
    gtk_box_pack_start ((GtkBox *) vbox, hbox, FALSE, FALSE, 0);
    g_signal_connect (button, "selection-changed", G_CALLBACK (folder_changed), (gpointer) ((long)(id)));
    return button;
}

G_MODULE_EXPORT void *
run_rfm_settings_dialog (McsPlugin * mp) {
    GtkWidget *header,
     *vbox,
     *hbox;
    int i;
    GtkTreeIter iter;
    GtkCellRenderer *renderer;
    GtkWidget *treeview;
    GtkWidget *label;

    /* plugin: */
    dialog = gtk_dialog_new_with_buttons (mp->caption,
                                          NULL, GTK_DIALOG_NO_SEPARATOR,
                                          _("Clear cache"), 2,
					  _("Icons"),
					  1, GTK_STOCK_CLOSE, 0, NULL);

    g_signal_connect (dialog, "response", G_CALLBACK (dialog_delete), NULL);
    g_signal_connect (dialog, "delete_event", G_CALLBACK (dialog_delete), NULL);
    g_signal_connect (dialog, "destroy", G_CALLBACK (dialog_delete), NULL);
    vbox = GTK_DIALOG (dialog)->vbox;
    /* pretty header */
    header = xfce_create_header (mp->icon, mp->caption);
    gtk_box_pack_start (GTK_BOX (vbox), header, FALSE, TRUE, 0);

    GtkNotebook *notebook = (GtkNotebook *) gtk_notebook_new ();
    gtk_notebook_set_scrollable (notebook, TRUE);
    gtk_notebook_popup_enable (notebook);
    gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (notebook), TRUE, TRUE, 0);

    ////////////////// General
    //
    vbox = create_tab (notebook, _("General"), _("General Options"));
    for(i = RFM_DOUBLE_CLICK_NAVIGATION; i <= RFM_DISABLE_TIPS; i++) {
        toggle_button[i] = gtk_check_button_new_with_label (_(environ_v[i].env_text));
        g_signal_connect (toggle_button[i], "toggled", G_CALLBACK (option_toggled), (gpointer) ((long)(i)));
        gtk_box_pack_start ((GtkBox *) vbox, toggle_button[i], FALSE, FALSE, 0);
    }

    char *terminals[] = { "xterm", "gnome-terminal", "konsole", "Terminal", NULL };
    terminal_box = make_exec_combo_box (vbox, TERMINAL_CMD, terminals, (void *)terminal_changed);
    char *editors[] = { "gvim", "xemacs", "gedit", "kate", "mousepad", NULL };
    editor_box = make_exec_combo_box (vbox, EDITOR_CMD, editors, (void *)editor_changed);
    ////
    hbox = gtk_hbox_new (FALSE, 6);
    label = gtk_label_new (_(environ_v[RFM_ICONVIEW_COLOR].env_text));
    gtk_box_pack_start ((GtkBox *) hbox, label, FALSE, FALSE, 0);
    iconviewcolor_button = gtk_color_button_new ();
    if(getenv ("RFM_ICONVIEW_COLOR") && strlen (getenv ("RFM_ICONVIEW_COLOR"))) {
        GdkColor color;
        if(gdk_color_parse (getenv ("RFM_ICONVIEW_COLOR"), &color)) {
            gtk_color_button_set_color ((GtkColorButton *) iconviewcolor_button, &color);
        }
    }
    gtk_color_button_set_title ((GtkColorButton *) (iconviewcolor_button), (const gchar *)
                                _(environ_v[RFM_ICONVIEW_COLOR].env_text));
    gtk_box_pack_start ((GtkBox *) hbox, iconviewcolor_button, FALSE, FALSE, 0);
    gtk_box_pack_start ((GtkBox *) vbox, hbox, FALSE, FALSE, 0);
    g_signal_connect (iconviewcolor_button, "color-set", G_CALLBACK (color_changed), (gpointer) ((long)(RFM_ICONVIEW_COLOR)));
    if(gtk_widget_is_composited (dialog)) {
        hbox = gtk_hbox_new (FALSE, 6);
        label = gtk_label_new (_(environ_v[RFM_TRANSPARENCY].env_text));
        gtk_box_pack_start ((GtkBox *) hbox, label, FALSE, FALSE, 0);
        double transparency;
        if(getenv ("RFM_TRANSPARENCY") && strlen (getenv ("RFM_TRANSPARENCY"))) {
            errno = 0;
            transparency = strtod (getenv ("RFM_TRANSPARENCY"), NULL);
            if(errno != 0 || transparency < 0.0)
                transparency = 0.0;
            else if(transparency > 0.75)
                transparency = 0.75;
        }
        GtkWidget *transparency_slider = gtk_hscale_new_with_range (0.0, 0.75, 0.01);
        gtk_scale_set_digits ((GtkScale *) transparency_slider, 2);
        gtk_range_set_value ((GtkRange *) transparency_slider, transparency);
        gtk_box_pack_start ((GtkBox *) hbox, transparency_slider, TRUE, TRUE, 0);
        gtk_box_pack_start ((GtkBox *) vbox, hbox, FALSE, FALSE, 0);
        g_signal_connect (transparency_slider, "change-value",
                          G_CALLBACK (transparency_changed), (gpointer) ((long)(RFM_TRANSPARENCY)));

    }

    gtk_widget_show_all (vbox);

    ///////////////////////// Desktop options
    gchar *tab_text = g_strdup_printf(" %s:", _("Options"));
    vbox = create_tab (notebook, _("Desktop"),  tab_text);
    g_free(tab_text);
    for(i = RFM_ENABLE_DESKTOP; i <= RFM_NAVIGATE_DESKTOP; i++) {
        toggle_button[i] = gtk_check_button_new_with_label (_(environ_v[i].env_text));
        g_signal_connect (toggle_button[i], "toggled", G_CALLBACK (option_toggled), (gpointer) ((long)(i)));
        gtk_box_pack_start ((GtkBox *) vbox, toggle_button[i], FALSE, FALSE, 0);
    }



    desktopdir_button = make_file_chooser_button (RFM_DESKTOP_DIR, TRUE, vbox);
    desktopimage_button = make_file_chooser_button (RFM_DESKTOP_IMAGE, FALSE, vbox);

    ////
    hbox = gtk_hbox_new (FALSE, 6);
    label = gtk_label_new (_(environ_v[RFM_DESKTOP_COLOR].env_text));
    gtk_box_pack_start ((GtkBox *) hbox, label, FALSE, FALSE, 0);
    desktopcolor_button = gtk_color_button_new ();
    if(getenv ("RFM_DESKTOP_COLOR") && strlen (getenv ("RFM_DESKTOP_COLOR"))) {
        GdkColor color;
        if(gdk_color_parse (getenv ("RFM_DESKTOP_COLOR"), &color)) {
            gtk_color_button_set_color ((GtkColorButton *) desktopcolor_button, &color);
        }
    }
    gtk_color_button_set_title ((GtkColorButton *) (desktopcolor_button), (const gchar *)
                                _(environ_v[RFM_DESKTOP_COLOR].env_text));
    gtk_box_pack_start ((GtkBox *) hbox, desktopcolor_button, FALSE, FALSE, 0);
    gtk_box_pack_start ((GtkBox *) vbox, hbox, FALSE, FALSE, 0);
    g_signal_connect (desktopcolor_button, "color-set", G_CALLBACK (color_changed), (gpointer) ((long)(RFM_DESKTOP_COLOR)));

    //////////////////////// Copy options
    gchar *tab_label=g_strdup_printf("%s/%s", _("Copy"), _("Move"));
    tab_text = g_strdup_printf(" %s:", _("Options"));
    vbox = create_tab (notebook, tab_label, tab_text);
    g_free(tab_label);
    g_free(tab_text);

    subtitle(vbox, _("Copy"));

#ifdef GNU_CP
    for(i = RFM_CP_A; i <= RFM_CP_S; i++)
#else
    for(i = RFM_CP_A; i <= RFM_CP_HH; i++)
#endif
    {
        toggle_button[i] = gtk_check_button_new_with_label (_(environ_v[i].env_text));
        g_signal_connect (toggle_button[i], "toggled", G_CALLBACK (option_toggled), (gpointer) ((long)(i)));
        gtk_box_pack_start ((GtkBox *) vbox, toggle_button[i], FALSE, FALSE, 0);
    }

    tab_text = g_strdup_printf("%s/%s", _("Copy"), _("Move"));
    subtitle(vbox, tab_text);
    g_free(tab_text);
#ifdef GNU_CP
    for(i = RFM_CP_B; i <= RFM_CP_U; i++)
    {
        toggle_button[i] = gtk_check_button_new_with_label (_(environ_v[i].env_text));
        g_signal_connect (toggle_button[i], "toggled", G_CALLBACK (option_toggled), (gpointer) ((long)(i)));
        gtk_box_pack_start ((GtkBox *) vbox, toggle_button[i], FALSE, FALSE, 0);
    }
#endif

    //Version control
    hbox = gtk_hbox_new (FALSE, 6);
    label = gtk_label_new (_(environ_v[i].env_text));
    gtk_box_pack_start ((GtkBox *) hbox, label, FALSE, FALSE, 0);

    combo_box = gtk_combo_box_new_text ();
    gtk_combo_box_insert_text ((GtkComboBox *) combo_box, 0, "simple");
    gtk_combo_box_append_text ((GtkComboBox *) combo_box, "numbered");
    gtk_combo_box_append_text ((GtkComboBox *) combo_box, "existing");
    gtk_combo_box_append_text ((GtkComboBox *) combo_box, "none");
    gtk_box_pack_start ((GtkBox *) hbox, combo_box, FALSE, FALSE, 0);
    gtk_box_pack_start ((GtkBox *) vbox, hbox, FALSE, FALSE, 0);

    g_signal_connect (combo_box, "changed", G_CALLBACK (combo_changed), (gpointer) ((long)(VERSION_CONTROL)));

    subtitle(vbox, _("Symbolic Link"));
    for(i = RFM_LN_S; i <= RFM_LN_S; i++)
    {
        toggle_button[i] = gtk_check_button_new_with_label (_(environ_v[i].env_text));
        g_signal_connect (toggle_button[i], "toggled", G_CALLBACK (option_toggled), (gpointer) ((long)(i)));
        gtk_box_pack_start ((GtkBox *) vbox, toggle_button[i], FALSE, FALSE, 0);
    }

    tab_text = g_strdup_printf("%s/%s/%s", _("Copy"), _("Move"), _("Symbolic Link"));
    subtitle(vbox, tab_text);
    g_free(tab_text);
    for(i = RFM_CP_VERBOSE; i <=RFM_CP_VERBOSE ; i++)
    {
        toggle_button[i] = gtk_check_button_new_with_label (_(environ_v[i].env_text));
        g_signal_connect (toggle_button[i], "toggled", G_CALLBACK (option_toggled), (gpointer) ((long)(i)));
        gtk_box_pack_start ((GtkBox *) vbox, toggle_button[i], FALSE, FALSE, 0);
    }

    gtk_widget_show_all (vbox);

    //////////////////////// remove/shred options
    tab_label=g_strdup_printf("%s/%s", _("Remove"), _("Shred"));
    tab_text = g_strdup_printf(" %s:", _("Options"));

    vbox = create_tab (notebook, tab_label, tab_text);
    g_free(tab_label);
    g_free(tab_text);

    subtitle(vbox, _("Remove"));

#ifdef GNU_RM
    for(i = RFM_RM_VERBOSE; i <= RFM_RM_ONE_FILE_SYSTEM; i++) 
#else
    for(i = RFM_RM_VERBOSE; i <= RFM_RM_VERBOSE; i++) 
#endif
    {
        toggle_button[i] = gtk_check_button_new_with_label (_(environ_v[i].env_text));
        g_signal_connect (toggle_button[i], "toggled", G_CALLBACK (option_toggled), (gpointer) ((long)(i)));
        gtk_box_pack_start ((GtkBox *) vbox, toggle_button[i], FALSE, FALSE, 0);
    }
    subtitle(vbox, _("Shred"));

#ifdef GNU_SHRED
//FIXME: put in a spinbutton here, default 3, maximum 300, step 1, for envar:
//RFM_SHRED_ITERATIONS
//    int end = RFM_SHRED_ITERATIONS;

    for(i = RFM_SHRED_VERBOSE; i <= RFM_SHRED_ZERO; i++) {
        toggle_button[i] = gtk_check_button_new_with_label (_(environ_v[i].env_text));
        g_signal_connect (toggle_button[i], "toggled", G_CALLBACK (option_toggled), (gpointer) ((long)(i)));
        gtk_box_pack_start ((GtkBox *) vbox, toggle_button[i], FALSE, FALSE, 0);
    }
#endif
    gtk_widget_show_all (vbox);

    ///////////////////////////
    ///////////////////////////
     vbox = create_tab (notebook, _("Advanced options"),
               _("Edit the list of environment variables and associated values"));

    GtkWidget *sw = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw), GTK_SHADOW_ETCHED_IN);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_widget_show (sw);
    gtk_box_pack_start (GTK_BOX (vbox), sw, FALSE, TRUE, 0);

    model = gtk_list_store_new (NUM_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN);
    for(i = 0; i < RFM_OPTIONS; i++) {
        gtk_list_store_append (model, &iter);
        gtk_list_store_set (model, &iter,
                            COLUMN_VARIABLE,
                            g_strdup (rfm_options[i].name),
                            COLUMN_VALUE, g_strdup (rfm_options[i].value), COLUMN_EDITABLE, TRUE, -1);
    }
    treeview = gtk_tree_view_new_with_model (GTK_TREE_MODEL (model));
    g_object_unref (G_OBJECT (model));
    gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (treeview), TRUE);
    gtk_tree_selection_set_mode (gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview)), GTK_SELECTION_SINGLE);
    /* variable column */
    renderer = gtk_cell_renderer_text_new ();
    g_signal_connect (G_OBJECT (renderer), "edited", G_CALLBACK (environment_changed), model);
    g_object_set_data (G_OBJECT (renderer), "column", (gint *) COLUMN_VARIABLE);

    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (treeview),
                                                 -1, _("Variable"), renderer, "text", COLUMN_VARIABLE, NULL);
    /* value column */
    renderer = gtk_cell_renderer_text_new ();
    g_signal_connect (G_OBJECT (renderer), "edited", G_CALLBACK (environment_changed), model);
    g_object_set_data (G_OBJECT (renderer), "column", (gint *) COLUMN_VALUE);

    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (treeview),
                                                 -1, _("Value"), renderer, "text", COLUMN_VALUE, "editable", COLUMN_EDITABLE, NULL);

    gtk_widget_set_size_request (treeview, 300, 200);
    gtk_widget_show (treeview);
    gtk_container_add (GTK_CONTAINER (sw), treeview);

    set_option_buttons ();
    g_signal_connect (notebook, "switch-page", G_CALLBACK (switch_page), NULL);

    gtk_widget_set_size_request (dialog, 550, 500);
    gtk_window_set_decorated (GTK_WINDOW (dialog), TRUE);
    gtk_window_set_type_hint (GTK_WINDOW (dialog), GDK_WINDOW_TYPE_HINT_NORMAL);
    gtk_window_set_destroy_with_parent (GTK_WINDOW (dialog), TRUE);
    gtk_window_stick (GTK_WINDOW (dialog));
    gchar *title = g_strdup_printf ("%s %s (%s)",
            PACKAGE_NAME, TAG, _("Personal settings"));
    gtk_window_set_title (GTK_WINDOW (dialog), title);
    g_free (title);
    GdkPixbuf *icon_pixbuf = rfm_get_pixbuf ("xffm/stock_preferences", SIZE_ICON);
    if(icon_pixbuf) {
        gtk_window_set_icon (GTK_WINDOW (dialog), icon_pixbuf);
    }

    gtk_widget_show_all (dialog);
    return (void *)(dialog);
}

//rfm_rational(const gchar *librarydir,const gchar *module_name, void *p,void *q,const gchar *function_id);
//rfm_rational("rfm","settings",(void *)setting_name,(void *)setting_value,"mcs_set_var");
G_MODULE_EXPORT void *
mcs_set_var (const gchar * setting_name, const gchar * setting_value) {
    DBG ("SETTINGS: mcs_set_var: %s -> %s\n", setting_name, setting_value);
    mcs_manager_set_string (mcs_manager, setting_name, CHANNEL, setting_value);
    mcs_manager_notify (mcs_manager, CHANNEL);
    return GINT_TO_POINTER (1);
}

G_MODULE_EXPORT void *
mcs_shm_stop (void
    ) {
    TRACE ("stopping instance of mcs-shm\n");
    if(mp) {
        munmap (mp->m, sizeof (mcs_shm_t));
    }
    shm_settings_serial = -1;
    return GINT_TO_POINTER (1);
}

G_MODULE_EXPORT void *
mcs_shm_start (void
    ) {
    DBG ("SETTINGS: mcs_shm_start\n");
    if(!mp) {
        int i;

        gchar *shm_settings_file = g_strdup_printf ("/%d-%s", (int)getuid (), MCS_FILE_NAME);
        DBG ("SETTINGS: starting instance of mcs-shm\n");
        mp = (McsPlugin *) malloc (sizeof (McsPlugin));
        memset (mp, 0, sizeof (McsPlugin));
        mp->manager = (McsManager *) malloc (sizeof (McsManager));
        memset (mp->manager, 0, sizeof (McsManager));
        mcs_manager = mp->manager;
        mp->plugin_name = g_strdup (CHANNEL);
        mp->caption = g_strdup_printf ("%s\n%s %s %s\n(%s)",
		 _("Personal settings"), PACKAGE_NAME, TAG, BUILD, VERSION);
        mp->run_dialog = run_rfm_settings_dialog;

        mp->icon = rfm_get_pixbuf ("xffm/stock_preferences", 48);

        DBG ("SETTINGS: shm_open %s\n", shm_settings_file);
        mp->shm = shm_open (shm_settings_file, O_RDWR, 0700);
        if(mp->shm < 0) {
            DBG ("SETTINGS: shm_open failed\n");
            mcs_shm_t *mcs_shm_p;
            mp->shm = shm_open (shm_settings_file, O_CREAT | O_RDWR, 0700);
            if(mp->shm < 0) {
                free (mp->manager);
                free (mp);
                mp = NULL;
                return NULL;
            }
            if(ftruncate (mp->shm, sizeof (mcs_shm_t)) < 0)
                g_warning ("ftruncate: %s", strerror (errno));

            mp->m = mmap (0, sizeof (mcs_shm_t), PROT_READ | PROT_WRITE, MAP_SHARED, mp->shm, 0);
            memset (mp->m, 0, sizeof (mcs_shm_t));
            mcs_shm_p = (mcs_shm_t *) mp->m;

            mcs_shm_p->serial = 1;
            if(mcs_shm_fileread ()) {
                mcs_shm_bringforth ();
            } else {
                mcs_shm_init ();
            }

            msync (mp->m, sizeof (mcs_shm_t), MS_SYNC);
            DBG ("SETTINGS: creating new shm block\n");
        } else {
            DBG ("SETTINGS: using preexisting shm block\n");
            mp->m = mmap (0, sizeof (mcs_shm_t), PROT_READ | PROT_WRITE, MAP_SHARED, mp->shm, 0);
            mcs_shm_bringforth ();
        }

        /* read environment... */
        DBG ("SETTINGS:read environment\n");
        for(i = 0; i < RFM_OPTIONS; i++) {
#if 0
            if(getenv (rfm_options[i].name)
               && strlen (getenv (rfm_options[i].name))
               && strlen (getenv (rfm_options[i].name))) {
                if(strlen (rfm_options[i].name) >= SHM_STRING_LENGTH) {
                    g_error
                        ("strlen(%s) is not less than %d, I stubbornly refuse to consider it!",
                         rfm_options[i].name, SHM_STRING_LENGTH);
                }
                if(strlen (getenv (rfm_options[i].name)) >= SHM_VALUE_LENGTH) {
                    g_warning
                        ("strlen($%s) is not less than %d, I stubbornly refuse to consider it!",
                         rfm_options[i].name, (int)strlen (getenv (rfm_options[i].name)));

                } else {
                    g_free (rfm_options[i].value);
                    rfm_options[i].value = g_strdup (getenv (rfm_options[i].name));
                    DBG ("*SETTINGS: got %s->%s\n", rfm_options[i].name, getenv (rfm_options[i].name));
                }
            } else
#endif
            {
                mcs_shm_t *mcs_shm_p = (mcs_shm_t *) mp->m;
                if(mcs_shm_p->data[i].value) {
                    g_free (rfm_options[i].value);
                    DBG ("*ENV: SETTINGS: defaulting %s->%s\n", rfm_options[i].name, mcs_shm_p->data[i].value);

                    rfm_options[i].value = g_strdup (mcs_shm_p->data[i].value);
                    rfm_setenv (rfm_options[i].name, rfm_options[i].value, TRUE);
                } else {
                    DBG ("*ENV: SETTINGS: nullifying %s\n", rfm_options[i].name);

                    rfm_setenv (rfm_options[i].name, NULL, TRUE);
                }
            }
        }

    }
    if(mp) {
        mcs_shm_t *mcs_shm_p = (mcs_shm_t *) mp->m;
        shm_settings_serial = mcs_shm_p->serial;
        DBG ("SETTINGS: mcs_shm_p->serial=%d\n", mcs_shm_p->serial);
        settings_timer = g_timeout_add (SETTINGS_TIMERVAL, settings_monitor, NULL);
    }
    return mp;
}

static gboolean module_initialized = FALSE;
G_MODULE_EXPORT void *
module_init (void
    ) {
    if(!module_initialized) {
        DBG ("SETTINGS: module_init->rfm_options...\n");
        module_initialized = TRUE;
        int i;
        for(i = 0; i < RFM_OPTIONS; i++) {
            DBG ("SETTINGS:%d/%d  %s <-> %s <-> %s\n", i+1, RFM_OPTIONS, environ_v[i].env_var, environ_v[i].env_string, environ_v[i].env_text);

            rfm_options[i].name = environ_v[i].env_var;
            if(environ_v[i].env_string)
                rfm_options[i].value = g_strdup (environ_v[i].env_string);
            else
                rfm_options[i].value = NULL;
        }
    }
    mcs_shm_start ();
    return (void *)mp;
}
